#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Auth;

use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Base;
use ExSite::Misc;
use ExSite::Util;

our @ISA = qw(ExSite::Base);


#========================================================================

=pod

=head1 ExSite::Auth -- authentication and security tools

Authorization is performed before any standard operation on
the database, when the operation is performed by one of ExSite::DB's
descendant classes.  The authorization process has the following
steps:

=over 4

=item 1. identify - determine who the user claims to be

=item 2. lookup - determine whom we understand that username to represent

=item 3. authenticate - determine if the user's claim matches our knowledge

=item 4. authorize - grant a certain level of access to the database

=item 5. approve - approve a specific operation

=back

Each of the steps in the authorization process may be replaced with
a custom handler if the procedure differs from the default ExSite logic.

=head2 identify() - determine who the user claims to be

Determines the user's claimed identity (such as by checking the
identity cookie), and stores the result so that authorization methods
can check it.  This routine does not authenticate the user's identity
claim, and so is usually not called directly.

=head2 lookup() - look up the user to see if we know who they are

Consult our own database to see if we have a record corresponding to
this user's identity (determined from C<identify()>).  Returns the
user record in a hash, if found.  The default logic for looking up a
user record is to look up the login name, which implies that unique
login names are required.  Once this lookup has occurred, the
following identity query functions can be used:

=head2 my_name() - return the login name of the current user

=head2 my_uid() - return the primary key of the current user's record

=head2 my_gid() - return the primary key of the current user's group record

=head2 my_user_record() - return the user's identity hash

=head2 authenticate() - verify the user's identity claim

Returns true (1) if the user's identity matches the information
returned by lookup(), false (0) otherwise.  This serves as a simple
check that the user is logged in, and that their identity is valid.

=head2 authorize() - grant the user a specific level of access

Returns the user's access level.  Access level is a value of 0 or
higher.  0 corresponds to "nobody", or the general public.  Values of
1 or more correspond to authenticated users, with higher values giving
greater permissions to the database.  The maximum value is
C<$config{auth}-E<gt>{root_auth_level}>, which grants unrestricted
"super-user" access to the data.  (This level of access can be
dangerous, as it allows one to mess with sensitive data fields and
referential integrity.)

The default authorize routine assumes that the user's access level
is stored in the user table.  This value is simply returned without
any sanity checks.

=head2 approve() - approve a specific database operation

Usage: C<$db-E<gt>approve( OP, TABLE, RECORD );>

OP is a low-level database operation (one of "read", "update",
"insert", or "delete").  TABLE is the table that this operation is
being performed on. RECORD is the actual record datahash that this
operation is being performed on (only used for "update" and "delete").

The return value is true if the operation is permitted, or false if:

=over 4

=item * the user does not have permission to read or write to the table

=item * the user is referencing records owned by a different user or group.

=back

=cut 

#----------------------------------------------------------------------
# identify : determine whom the user claims to be

sub identify {
    my ($this) = shift;
    my ($name,$pwd,$uid,$gid);
    my $stat = $this->run_handler("identify");
    my $authtype = $config{auth}{type};
    if (defined $stat) {
	# application has its own identify algorithm
	($name,$pwd) = @$stat;
    }
    elsif ($authtype eq "cookie") {
	($name,$pwd) = &get_cookie($config{site}{identity_cookie},":",2);
    }
    elsif ($authtype eq "fixed") {
	$name = $config{auth}->{default_identity}->{name};
    }
    elsif ($authtype eq "session") {
	# User identity claim is implicit in the session ID.
	# Therefore this call is kind of meaningless in the case of sessions.
	# Let's return the session ID, and see what happens.
	my $session_cookie = $config{site}->{id}."_sessionID";
	return &get_cookie($session_cookie);
    }
    elsif ($authtype eq "none") {
	$name = "nobody";
    }
    $share{identity} = { password => $pwd };
    $share{identity}{name} = $name;
    return $name;
}

#----------------------------------------------------------------------
# lookup : look up the user, to see if we know who they are

sub lookup {
    my ($this,$login) = @_;
    if ($share{identity}{lookup}) {
	return $share{identity}{lookup};
    }
    # check for custom lookup algorithm
    my $stat = $this->run_handler("lookup");
    return $stat if (defined $stat);
    #
    return undef if ($this->{authentication} eq "none");
    my $udata = $share{DB}->fetch_match($config{auth}->{user_table},
			    {$config{auth}->{user_login_column}=>$login});
    if (@$udata > 0) {
	$share{identity}{lookup} = $udata->[0];
	$share{identity}{uid} = $udata->[0]{$config{auth}{user_uid_column}};
	$share{identity}{gid} = $udata->[0]{$config{auth}{user_gid_column}};
	$share{identity}{access} = $udata->[0]{$config{auth}{user_access_column}};
	return $udata->[0];
    }
    return undef;
}

#----------------------------------------------------------------------
# authenticate : verify the user's identity claim
#
# Returns true (1) if the user's identity matches the information
# returned by lookup(), false (0) otherwise.

sub authenticate {
    my ($this) = shift;
    if (exists $this->{authenticated}) {
	return $this->{authenticated};
    }
    # check for custom authentication algorithm
    my $stat = $this->run_handler("authenticate");
    return $stat if (defined $stat);
    
    # is this deprectated?
    if (exists $ENV{EXSITE_TEST} || $this->{authentication} eq "fixed") {
	$this->{authenticated} = 1;
	$share{identity} = $config{auth}{default_identity};
	return 1;
    }

    else {
	# implicit approval if authorization is turned off or $EXSITE_TEST is
	# set when testing from the shell
	return 1 if ($this->{authentication} eq "none");

	#
	#  case 1: authentication type = "session"
	#

	if ($this->{authentication} eq "session") {
	    require ExSite::Session;
	    my $sess; 
	    if (exists $share{Session}) {
		$sess = $share{Session};
	    }
	    else {
		$sess = new ExSite::Session;
		if ($sess) {
		    $share{Session} = $sess;
		}
		else {
		    $this->error("error initializing session");
		    return 0;
		}
	    }
	    return $sess->validate;
	}
	
	#
	#  case 2: other authentication types
	#
	
	# check identity
	if (!$share{identity}) { &identify($this); }
	if (!$share{identity}{name}) {
	    # nobody logged in - not necessarily an error
	    return 0;
	}
	# look up the user
	$this->info("authenticating user \"$this->{identity}->{name}\"");
	my $user = &lookup($this,$share{identity}{name});
	if (!$user) {
	    # unknown username
	    $this->warn("no such user \"$this->{identity}->{name}\"");
	    return 0;
	}
	elsif (! $user->{$config{auth}->{user_pwd_column}}) {
	    # password field is blank - deny access
	    $this->warn("blank password for user \"$user->{$config{auth}->{user_login_column}}\"");
	    return 0;
	}
	# compare entered and stored passwords
	my $pass;
	if ($config{auth}->{password_storage_method} eq "cleartext") {
	    # encrypt the db password before comparing with identity
	    my $salt = substr($share{identity}{password},0,2);
	    $pass = ($share{identity}{password} eq
		     crypt($user->{$config{auth}->{user_pwd_column}},$salt));
	}
	else {
	    # encrypted db passwd can be compared directly with identity
	    $pass = ($share{identity}{password} eq $user->{$config{auth}->{user_pwd_column}});
	}
	if ($pass) {
	    # passwords match
	    $this->{authenticated} = 1;
	    # store extra user info in identity
	    $share{identity}{gid} = $user->{$config{auth}->{user_gid_column}};
	    ####?
	    $share{identity}{access} = $user->{$config{auth}->{user_access_column}};
	    return 1;
	}
	else {
	    # bad password
	    $this->warn("incorrect password for user \"$user->{$config{auth}->{user_login_column}}\" - access denied");
	    $this->{authenticated} = 0;
	    return 0;
	}
    }
}

#----------------------------------------------------------------------
# authorize : grant the user a specific level of access
#
# Access level is a value of 0 or higher.  0 corresponds to "nobody",
# or the general public.  Values of 1 or more correspond to
# authenticated users, with higher values giving greater permissions
# to the database.  The maximum value is $config::root_auth_level,
# which grants unrestricted "super-user" access to the data.  (This
# level of access can be dangerous, as it allows one to mess with
# sensitive data fields and referential integrity.)
#
# The default authorize routine assumes that the user's access level
# is stored in the user table.  This value is simply returned without
# any sanity checks.

sub authorize {
    my ($this) = shift;
    my $stat = $this->run_handler("authorize");
    return $stat if (defined $stat);
    if (! exists $this->{authenticated}) {
	&authenticate($this);
    }
    if ($this->{authenticated}) {
	return $share{identity}{access};
    }
    # give public access by default
    return 0;
}

#----------------------------------------------------------------------
# approve : approve a specific database operation
#
# Arguments:
# op - "read", "update", "insert", "delete"
# table - the table being operated on
# [record] - reference to a datahash containing the record being
#            operated on (used for update, delete only)
#
# Returns true if the operation is permitted.
# Returns false if:
#   - the user does not have permission to read or write to the table
#   - the user is referencing records owned by a different user or group.
#   - the user does not have the required level of access for this CGI script

sub approve {
    my ($this,$op,$table,$record) = @_;
    my $stat = $this->run_handler("approve",$op,$table,$record);
    return $stat if (defined $stat);
    my $level = $this->authorize();
    my $tablemap = $share{DB}{map}->get_table($table);
    # script okay, now check table permissions
    if (exists $tablemap->{$op}) {
	if ($op =~ /read|insert|search/) {
	    return ($level >= $tablemap->{$op});
	}
	else {
	    # for updates & deletions, must check row ownership
	    if (!$record) {
		if ($level >= $config{auth}->{root_auth_level} - 1) { 
		    return 1; 
		}
		else { return 0; }
	    }
	    elsif ($share{DB}->user_owns($table,$record)) {
		return ($level >= $tablemap->{$op});
	    }
	    else {
		return 0;
	    }
	}
    }
    elsif ($this->{dbmap}) {
	# dbmap kludge - no tablemap or authentication when editing dbmaps
	return 1;
    }
    else {
	# unknown operation
	return 0;
    }
}

# some identity queries, for convenience:

sub my_name {
    my ($this) = shift;
    if ($this->authenticate) {
	return $share{identity}{name};
    }
    return undef;
}

sub my_uid {
    my ($this) = shift;
    if ($this->authenticate) {
	return $share{identity}{uid};
    }
    return undef;
}

sub my_gid {
    my ($this) = shift;
    if ($this->authenticate) {
	return $share{identity}{gid};
    }
    return undef;
}

sub my_user_record {
    my ($this) = shift;
    my $u;
    if ($this->authenticate) {
	if (exists $share{identity}{lookup}) {
	    $u = $share{identity}{lookup};
	}
	else {
	    $u = [];
	}
    }
    return (wantarray && defined $u) ? %$u : $u;
}

sub is_admin {
    my $this = shift;
    return ($this->authorize >= ($config{auth}{root_auth_level} - 1));
}

# return the effective access level the user is operating at

sub level {
    my $this = shift;
    if (! exists $this->{effective_level}) {
	$this->{effective_level} = $this->real_level;
    }
    return $this->{effective_level};
}

# return the real access level of the user

sub real_level {
    my $this = shift;
    if (! exists $this->{level}) {
	$this->{level} = $this->authorize;
    }
    return $this->{level};
}

# change to a different (lower) effective access level
#
# This is used when publishing, to make pages render as if being viewed
# by the public.

sub set_level {
    my ($this,$level) = @_;
    $this->{effective_level} = ($level <= $this->real_level) ? $level : 0;
    return $this->{effective_level};
}

# return to our real access level.

sub reset_level {
    my $this = shift;
    delete $this->{effective_level};
}

# validate a login/password pair, and set an identity cookie

sub login {
    my ($this,$login,$passwd) = @_;
    if ($config{auth}{type} eq "cookie") {  ########### SESSIONS

	# set a pseudo-cookie

	my $cookie = "$config{site}{identity_cookie}=".
	    join(":",$login,crypt($passwd,"wd"));
	$ENV{HTTP_COOKIE} = $cookie;

	# try out this id/pwd combo

	my $access;
	if ($access = $this->authorize()) {

	    # good login!

	    my $cookie = &set_cookie($config{site}{identity_cookie},
				     join(":",$login,crypt($passwd,"wd")) );
	    my $page = $share{Page};
	    if ($page) {
		$page->header($cookie);
	    }
	    else {
		print $cookie;
	    }
	    return $access;
	}    
    }
    return 0;
}

# unset the identity cookie

sub logout {
    my $this = shift;
    
    # clear existing identity value for remainder of this request

    my $cookie = &edit_cookie($config{site}{identity_cookie}, undef );

    # clear cookie for future requests

    my $cookie_domain = $config{server}{host} ? 
	".$config{server}{domain}" : $config{server}{domain};
    $cookie = &set_cookie($config{site}{identity_cookie},
			     undef,
			     $config{server}{CGIpath},
			     $cookie_domain,
			     "Thu, 01-Jan-1970 00:00:00 GMT",
			     );

    my $page = $share{Page};
    if ($page) {
	$page->header($cookie);
    }
    else {
	print $cookie;
    }
    return;
}

#----------------------------------------------------------------------
# validate password : compare a password with the stored version
#
# Returns true (1) if the passed password matches the user's
#
# #### TODO #### $encrypted should be non-0 if the passed password is encrypted

sub validate_password {
    my ($this,$password,$user,$encrypted) = @_;
    my $udata;
    if (! defined $user) {
        $udata = $this->my_user_record;
    }
    elsif (ref $user eq "HASH") {
	$udata = $user;
    }
    else {
	$udata = $share{DB}->lookup($user);
    }
    if (!$udata) {
	# unknown user
	$this->warn("validate_password: unknown user");
	return 0;
    }
    elsif (! $udata->{$config{auth}->{user_pwd_column}}) {
	# password field is blank - deny access
	$this->warn("blank password for user \"$udata->{$config{auth}->{user_login_column}}\"");
	return 0;
    }
    # compare entered and stored passwords
    my $pass;
    if ($config{auth}->{password_storage_method} eq "cleartext") {
	$pass = ($password eq $udata->{$config{auth}{user_pwd_column}});
    }
    elsif ($config{auth}->{password_storage_method} eq "des") {
	# encrypt before comparing with DB password 
	my $salt = substr($udata->{$config{auth}{user_pwd_column}},0,2);
	$pass = ($udata->{$config{auth}{user_pwd_column}} eq
		 crypt($password,$salt));
    }
    elsif ($config{auth}->{password_storage_method} eq "md5") {
# TODO!
#	my $salt = substr($udata->{$config{auth}{user_pwd_column}},0,2);
#	$pass = ($udata->{$config{auth}{user_pwd_column}} eq
#		 crypt($password,$salt));
    }
    return $pass;
}

# change password

sub change_password {
    my ($this,$old,$new,$new2) = @_;
    if ($new) {
	if (! defined $new2 || ($new eq $new2)) {
	    my $re = $share{DB}{map}->regexp('password');
	    if ($new =~ /^$re$/) {
		if (! defined $old || $this->validate_password($old)) {
		    my $stat = $this->set_password($new);
		    return $stat if ($stat); # error
		    # update the login cookie
		    $this->login($this->my_name,$new);
		    return undef;
		}
		else {
		    return "password update failed - incorrect password";
		}
	    }
	    else {
		return "insecure password - needs more characters";
	    }
	}
	else {
	    return "new passwords do not match";
	}
    }
    else {
	return "cannot set a blank password";
    }
}

sub set_password {
    my ($this,$password,$uid) = @_;
    if (! $uid) {
	$uid = $this->my_uid;
    }
    if ($uid) {
	return $share{DB}->update($config{auth}{user_table},
				      {$config{auth}{user_pwd_column}=>$password},
				      {$config{auth}{user_uid_column}=>$uid});
    }
    return "no such user";
}

# fetch all admin tickets

sub my_sites {
    my ($this) = @_;
    return $share{DB}->fetch_m2m("section","member",
				 $this->my_uid,"member_site_link");
}

# ditto, but do not count website templates

sub my_real_sites {
    my ($this) = @_;
    my @real_sites;
    foreach my $site ($this->my_sites) {
	if ($site->{type} !~ /template/) {
	    push @real_sites, $site;
	}
    }
    return @real_sites;
}

# return the current site data (may require some guesswork)

sub this_site {
    my ($this) = @_;
    if ($share{Page}) {
	# we're on a page, so return the page's section
	return $share{Page}->fetch("section",$share{Page}->get("page","section_id"));
    }
    elsif ($share{Section}) {
	# we're on a section, so return it
	return $share{Section}->get();
    }
    elsif ($share{DB}->my_user_record()->{section_id}) {
	# user has a primary site defined
	return $share{DB}->fetch("section",$share{DB}->my_user_record()->{section_id});
    }
    else {
	my @sites = $this->my_real_sites;
	if (@sites > 0) {
	    # ambiguous result if more than one - return first one
	    my @s = $share{DB}->sort_rows("section",\@sites,"section_id");
	    return $s[0];
	}
    }
    return undef;
}

# test for "ownership" of a site

sub owns {
    my ($this,$site) = @_;
    if (! $site) { $site = $this->this_site()->{section_id}; }
    return 0 if (! $site);
    my @temp = $share{DB}->fetch_match("member_site_link",
				       { member_id => $this->my_uid,
					 section_id => $site } );
    return (@temp > 0);
}

# test for license to use a service on a site

sub license {
    my ($this,$service,$site) = @_;
    if (eval "require Modules::$service") {
	my $s = eval "new Modules::$service";
	if ($s->ioctl("isRestricted")) {
	    # this service requires licenses
	    if (!$site && $share{Page}) {
		$site = $share{Page}->id("section");
	    }
	    if ($site) {
		my $licenses = $share{DB}->fetch_match("service",
						       {section_id => $site,
							name => $service }
						       );
		return (scalar @$licenses > 0);
	    }
	    # hmm, that's strange, we're not on a page
	    $this->warn("$service: unknown website, cannot confirm license");
	    return 0;
	}
	# this modules is not a service - anyone can use it
	return 1;
    }
    # module does not seem to exist
    $this->warn("$service: no such service");
    return 0;
}

# test for license to use a service's control panel
# ie. test that a web app is licensed on any of the user's sites

sub license_control_panel {
    my ($this,$service) = @_;
    if (! $share{licenses}) {
	$share{licenses} = {};
	my $db = $share{DB};
	if ($db->{dbtype} =~ /SQL/i) {
	    # use a join query to fetch all licensed services
	    my $uid = $db->my_uid;
	    my @services = $db->custom_query("select service.name from service,section,member_site_link where service.section_id=section.section_id and service.section_id=member_site_link.section_id and member_site_link.member_id=$uid");
	    $share{licenses} = &keywise("name",\@services);
	}
	else {
	    # brute-force loop over all services/sites
	    my @sites = $db->my_real_sites();
	    foreach my $l ($share{DB}->fetch_all("service")) {
		foreach my $s (@sites) {
		    if ($l->{section_id} == $s->{section_id}) {
			$share{licences}{$l->{name}} = 1;
			next;
		    }
		}
	    }
	}
    }
    return (exists $share{licenses}{$service});
}

1;

