#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
#
# ExSite v3 Base class
#
# contains only the most generic functions needed at all levels of ExSite:
#
# - default constructor
# - get/set attribute accessor functions
# - handler installation and execution
# - error and diagnostic management

package ExSite::Base;
use strict;
use ExSite::Config qw(:DEFAULT &install_handlers);

# generic object constructor

sub new ($%) {
    my ($this,%opt) = @_;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    # copy options into object
    while (my ($key,$val) = each %opt) {
	$obj->{$key} = $val;
    }
    $obj->initialize_object;
    return $obj;
}

# set up attributes common to all ExSite objects
# Use a standalone method, so that overloaded new() methods in higher
# level classes can set themselves up without having to copy the code.

sub initialize_object {
    my $this = shift;

    # handlers
    $this->{handler} = {};
    &install_handlers($this);

    # error accumulators
    $this->{diagnostics} = {};
    $this->{diagnostics}->{error} = [];
    $this->{diagnostics}->{warn} = [];
    $this->{diagnostics}->{info} = [];

    # legacy
    $this->{errors} = [];

}

# generic attribute accessor functions get/set

sub get ($$) {
    my ($this,$attr) = @_;
    if (exists $this->{$attr}) {
	return $this->{$attr};
    }
    return undef;
}

sub set ($$$) {
    my ($this,$attr,$val) = @_;
    $this->{$attr} = $val;
    return $val;
}

# handlers

# install a handler

sub handler ($$$) {
    my ($this,$name,$handler_ptr) = @_;
    if (ref $handler_ptr eq "CODE") {
	if (! $this->{handler}->{$name}) {
	    # 1st handler
	    $this->{handler}->{$name} = $handler_ptr;
	}
	elsif (ref $this->{handler}->{$name} eq "ARRAY") {
	    # multiple handlers already; add to list
	    push @{$this->{handler}->{$name}}, $handler_ptr;
	}
	else {
	    # 2nd handler; convert to array
	    my $old_ptr = $this->{handler}->{$name};
	    $this->{handler}->{$name} = [ $old_ptr, $handler_ptr ];
	}
    }
}

# run a handler 

sub run_handler ($$@) {
    my ($this,$name,@args) = @_;
    if (exists $this->{handler}->{$name}) {
	my @handlers;
	if (ref $this->{handler}->{$name} eq "ARRAY") {
	    @handlers = @{$this->{handler}->{$name}};
	}
	else {
	    @handlers = ( $this->{handler}->{$name} );
	}
	foreach my $handler (@handlers) {
	    my $stat = &{$handler} ($this,@args);
	    return $stat if (defined $stat);
	    # Note that we return the first non-null result.  We do not
	    # execute all handlers unconditionally.
	}
    }
    return undef;
}

# diagnostics and error handling

# diagnostics are recorded in 3 places:
# 1) in the object itself (diagnostics specific to this object)
# 2) in %share (diagnostics specific to this job)
# 3) in the logfiles (all jobs)

# All diagnostics return the message passed to it, but reformatted as HTML.

#### no way to generically manage fatal errors, 
#### since we don't really want to die()

# add errors to the error accumulator
# - errors indicate a fault that prevents an action from completing

sub error ($@) {
    my $this = shift;
    # add the errors to the object's error accumulator
    push @{$this->{diagnostics}->{error}}, @_;
    # determine the category of error
    my $type = $this->general_class;
    my $msg = "Error: ".join("; ",@_); 
    # log the errors, if necessary
    if ($config{log}{$type} >= 1 && @_ > 0) { 
	$this->log($type,@_);
	push @{$share{diagnostics}}, $msg;
    }
    # return a formatted message
    return "<p class=\"error\">$msg</p>";
}

# add warnings to the warn accumulator
# - warnings indicate a possible problem, but it can be worked around

sub warn ($$@) {
    my $this = shift;
    push @{$this->{diagnostics}->{warn}}, @_;
    my $type = $this->general_class;
    my $msg = "Warning: ".join("; ",@_); 
    if ($config{log}{$type} >= 2 && @_ > 0) { 
	$this->log($type,@_);
	push @{$share{diagnostics}}, $msg;
    }
    return "<p class=\"error\">$msg</p>";
}

# add info to the info accumulator
# - info simply records program status.
# - an info call should be added to the start (and possibly end) of any
# significant method calls.

sub info ($@) {
    my $this = shift;
    push @{$this->{diagnostics}->{info}}, @_;
    my $type = $this->general_class;
    my $msg = "Info: ".join("; ",@_); 
    if ($config{log}{$type} >= 3 && @_ > 0) { 
	$this->log($type,@_); 
	push @{$share{diagnostics}}, $msg;
    }
    return "<p class=\"help\">$msg</p>";
}

# errorcheck simply tests for the existence of errors; it can be used
# as a simple verification check after performing some operation.

sub errorcheck {
    my ($this, $type) = @_;
    if (! $type) { $type = "error"; }
    return ( @{$this->{diagnostics}->{$type}} > 0 );
}

# fetch_diagnostics returns whichever diagnostic accumulator is requested.
# It is presumed to be used by a debug report method to format the 
# messages to be viewed by the end user.
#
# It is presumed that such debug reports will have 4 levels of 
# diagnostic verbosity:
#
# 0 - quiet (no diagnostics)
# 1 - show error messages only
# 2 - show errors and warnings
# 3 - show errors, warnings, and other info

sub fetch_diagnostics ($$) {
    my ($this,$type) = @_;
    if ($type) {
	if (exists $this->{diagnostics}->{$type}) {
	    return @{$this->{diagnostics}->{$type}};
	}
    }
    return ();
}

sub show_diagnostics ($$$) {
    my ($this,$type,$format) = @_;
    my @diag = $this->fetch_diagnostics($type);
    if ($format =~ /HTML/i) {
	return "<SPAN CLASS=\"error\">\n".join("<br>\n",@diag)."</SPAN>\n";
    }
    else {
	return join("\n",@diag);
    }
}

# determine which general class/inheritance category we belong to
# db = database handling classes
# cms = content management classes
# general = other

sub general_class { 
    my $this = shift;
    my $class = ref $this;
    if ($class =~ /^ExSite::(Content|Page|Section)/) { return "cms"; }
    elsif ($class =~ /^ExSite::/) { return "db"; }
    else { return "general"; }
}

# event logger

sub log {
    my $this = shift;
    my $log = shift;
    # remaining arguments are the messages to be logged
    return if (@_ == 0);
    if ($config{log}{logfile}{$log}) {
	if (-w $config{log}{logfile}{$log}) {
	    if (open LOG, ">>$config{log}{logfile}{$log}") {
		my ($time,$user,$op);
		$time = localtime;
		# don't use my_name(), as it will cause an infinite loop
		$user = exists $share{identity} ? $share{identity}{name} : "nobody";
		if (!$user) { $user = "unknown"; }
		$op = $ENV{REQUEST_URI};
		foreach my $message (@_) {
		    print LOG "$time\t$user\t$op\t$message\n";
		}
		close LOG;
		### FIXME: maintain an open logfile connection for better perf
	    }
	}
    }
}

# fetch contents of logfile

sub showlog {
    my ($this,$log,$nlines) = @_;
    if (-r $config{log}{logfile}{$log}) {
	if (open LOG, "<$config{log}{logfile}{$log}") {
	    my @loglines = <LOG>;
	    close LOG;
	    if ($nlines) { 
		# return last $nlines of file
		return @loglines[($#loglines-$nlines) .. $#loglines];
	    }
	    else {
		# log file could be large - use with caution
		return @loglines;
	    }
	}
    }
}

1;
