#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
# ExSite::Cache
# a generic cache tool for saving re-useable data
#
# The cache is an in-memory hash of arbitrary data values.
# The data can be indexed based on an arbitrary amount of
# indexing data, so that the cache can store completely unrelated
# data with low risk of collision.
#
# The primary use in ExSite is to remember results of database
# queries, so that we don't need to make a DB server call to
# re-fetch something we've already fetched (but forgotten).
#
# For instance, if you've fetched data record #123 from table XYZ,
# you could save it in cache this way:
#
# my $cache = new ExSite::Cache;
# $cache->save($data,"XYZ","123");
#
# The first value is the data to save;  the remaining values (there 
# can be any number) are used to uniquely index the data.  In this case,
# the table name and record number are sufficient to uniquely specify
# this piece of data.
#
# If later someone asks for the same data, eg:
#
# $db->fetch("XYZ",123);
#
# then you could first check the cache for a value before querying the
# the database:
#
# $data = $cache->get("XYZ",123);
# if (! $data) { # go to database...

package ExSite::Cache;
use strict;
use Digest::MD5 qw(md5_hex);

# Cache constructor - bare bones

sub new ($%) {
    my ($this,%opt) = @_;
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    return $obj;
}

# save some data to the cache;
# first arg is the data to save;
# remaining args are used to index the data.

sub save ($$@) {
    my $this = shift;
    my $data = shift; # this is the value you are saving in the hash
    # remaining args are the values to hash
    my $key = md5_hex(@_);
    $this->{$key} = $data;
}

# fetch data from the cache;
# args are used to determine the index to the data.

sub get ($@) {
    my $this = shift;
    my $key = md5_hex(@_);
    if (exists $this->{$key}) { return $this->{$key}; } 
    return undef;
}

# delete data from the cache;
# args are used to determine the index to the data.

sub delete ($@) {
    my $this = shift;
    my $key = md5_hex(@_);
    if (exists $this->{$key}) { delete $this->{$key}; } 
}
    
# clear the entire cache.

sub clear ($) {
    my $this = shift;
    foreach my $key (keys %$this) {
	delete $this->{$key};
    }
}

1;
