#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Config;
use strict;
use myConfig;

=pod

=head1 ExSite::Config - ExSite configuration file

C<Config.pm> contains general configuration parameters,
site-specific setup data, and global variables used throughout
the ExSite package.

This file is designed so that it never has to be modified.
To set configuration variables, use an "exsite.conf" file, which 
will be automatically loaded, overriding all of the presets below.
Handlers and custom setup routines can be defined in the module
myConfig.pm.

=for html <hr>

=head2 Usage:

=over 4

=item Default Mode (imports C<%config> hash and setup functions)

C<use ExSite::Config;>

=item Legacy Mode (imports legacy configuration variables)

C<use ExSite::Config qw(:OLD);>

C<&set_old_config_vars;>

=item Partial Legacy Mode (imports selected configuration variables of
types :SITE, :SERVER, :AUTH, :FORM, :REPORT, or :GENERAL)

C<use ExSite::Config qw(:DEFAULT :SITE :SERVER);>

C<&set_old_config_vars("SITE SERVER");>

=back

In addition, every program should also include the following line at the top:

C<&exsite_init;>

=head2 Configuration Management

The preferred configuration management system stores all configuration info
in the %config hash, which is exported automatically.

The file "exsite.conf" is read into the %config hash as well,
allowing site configuration to be modified without altering any code.
To do this, call &read_exsite_conf - this is done from &exsite_init
by default.

Legacy scripts may expect hard-coded configuration variables to be exported
from the Config module.  This is ugly, messy, and considered Bad Perl, but 
if you need it, you can change the "use ExSite::Config" line to:

C<use ExSite::Config qw(:OLD);>

and add a call to C<&set_old_config_vars> in your script or in
C<&exsite_init>.  This will throw a ton of crap into your namespace.
You have been warned.

=head2 Configuration Parameters

Specific configuration parameters are described in the ExSite
Configuration documentation.

Additional configuration parameters can be stored in the global
C<%config> hash.  This can be useful as an alternative method of
passing information to widely separated routines.

=head2 C<%share>

The C<%share> hash is an alternative configuration hash that can be used
for sharing data between widely separated modules and routines.  The
primary difference between C<%config> and C<%share> is that the former
is regarded as persistent (values stay the same from request to request), 
while the latter is volatile (values vary from request to request).  

The differences are not important for pure CGI setups, since C<%config>
must be reconfigured on each request anyway.  But in a mod_perl setup,
C<%config> may only be configured once at server startup, whereas
C<%share> will be cleared and rebuilt on every request.  It is good 
practice to follow this convention, to ease the transition to a mod_perl
setup, when it comes.

C<%share> is empty when initialized by default, and is cleared every
time C<&exsite_init()> is called.  C<%share> will be used to store
pointers to the current database and page objects (C<$share{DB}> and
C<$share{Page}>, respectively), plus any other shareable data placed
there by localization routines and content modules.

=cut

#=====================================================================
# Exported names:

BEGIN {
    use Exporter;
    use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS 
		@old_config_vars @site_conf_vars @server_conf_vars
		@auth_conf_vars @form_conf_vars @report_conf_vars
		@general_conf_vars);

    $VERSION = "3.2.0";

    @ISA = qw(Exporter);

    @site_conf_vars = qw( 
       $site_name $site_administrator $site_identity_cookie $site_logo
       &set_old_config_vars
    );

    @server_conf_vars = qw( 
       %db_config $dbmap_dir
       $HTMLroot $CGIroot $HTMLpath $CGIpath $CSSpath $CSSfile
       $imgdir $host $domain
       $protocol $port $server $temppath $templatedir
       &set_old_config_vars
    );

    @auth_conf_vars = qw( 
       $default_authentication $root_auth_level
       $default_name $default_uid $default_gid $default_access_level
       $user_table $user_login_column $user_pwd_column $user_uid_column
       $user_gid_column $user_access_column $user_access_column $group_table
       $password_storage_method $foreign_link_level @login_destination
       &set_old_config_vars
    );

    @form_conf_vars = qw( 
       $form_style $reqd_field_style %labeltag %fieldtag
       $sepchar $max_field_size $max_field_rows
       $trim_whitespace $default_date
       &set_old_config_vars
    );

    @report_conf_vars = qw( 
       $report_page_length %report_tag %report_rowtag %report_columntag 
       %report_labeltag
       $highlight_alternate_lines
       &set_old_config_vars
    );

    @general_conf_vars = qw( 
       $page_buffering $show_images
       %db_ops @edit_links
       &set_old_config_vars
    );

    @old_config_vars = (
       @site_conf_vars, @server_conf_vars, @auth_conf_vars, @form_conf_vars,
       @report_conf_vars, @general_conf_vars,
       qw(&exsite_init &page_header &page_footer)
    );

    @EXPORT      = qw(%config &exsite_init &exsite_close &page_header 
		      &page_footer &set_cookie %share
		      &input_get &input_post &input
		      );

    @EXPORT_OK = (qw(&read_exsite_conf &install_handlers), 
		  @old_config_vars);

    %EXPORT_TAGS = ( 
		     OLD    => \@old_config_vars,
		     SITE   => \@site_conf_vars,
		     SERVER => \@server_conf_vars,
		     AUTH   => \@auth_conf_vars,
		     FORM   => \@form_conf_vars,
		     REPORT => \@report_conf_vars,
		     GENERAL=> \@general_conf_vars,
		   );
}

use vars (@EXPORT, @EXPORT_OK);

#=====================================================================
# Non-Exported names:

use vars qw(
	    %server_config %site_config %auth_config 
	    %form_config %report_config %log_config
	    %webtop_config %crypt_config
	    $page_header_flag $page_footer_flag
	    $config_flag
	    );

##################################################################
################ static configuration            #################
##################################################################

# static configuration parameters are valid for every site setup.
# They may still be overridden in particular scripts, by setting
# the given parameters to a new value.

#======================================================================
# SITE CONFIGURATION:
# parameters that define the website as a whole

%site_config = (

    # site name
    name => "ExSite",

    # identity cookie name
    identity_cookie => "ExSiteID",

    # site logo <IMG ...> tag
    logo => "",

    # sysadm's email
    administrator => "morgan\@exware.com",

);

#======================================================================
# AUTHENTICATION CONFIGURATION
# parameters controlling, login, authorization, etc.

%auth_config = (

    # authentication type (fixed, none, cookie)
    type => "fixed",

    # define the password storage method: "cleartext", "des", or "md5"
    password_storage_method => "cleartext",

    # define the default identity to use in case of fixed auth type, 
    # or for simple shell-based tests with no cookie handling.
    default_identity => {
	       name => "nobody",
	       uid => 1,
	       gid => 1,
	       access => 3,
	   },

    # define the tables and columns where login and authentication
    # information can be found
    user_table => "member",
    group_table => "member_group",
    user_login_column => "login",
    user_pwd_column => "password",
    user_access_column => "access",
    user_uid_column => "member_id",
    user_gid_column => "member_group_id",

    # what authorization level corresponds to unlimited access?
    root_auth_level => 4,

    # what user level is permitted to link to records they don't own?
    foreign_link_level => 1,

    # login targets
    login_destination => [ "/index.html", 
			   "home.cgi",
			   "home.cgi",
			   "home.cgi",
			   "home.cgi",
			   ],

    # CGI program access levels, to be filled in exsite.conf
    cgi => { },
);

#======================================================================
# FORM CONFIGURATION
# parameters controlling form layout

%form_config = (

    # input field separator character
    sepchar => '#',

    # multiple-input separator character
    multi_sep => '; ',

    #---------------------------------
    # form style
    #---------------------------------
    # form_style can be:
    # "list" (labels above inputs)
    # "table" (labels to the left of aligned inputs)
    # "none" (labels to the left of unaligned inputs)
    # additional styles can be defined by adding entries to %fieldtag, below

    style => "table",

    # field tags (1st goes before label, 2nd goes between label and input,
    #             3rd goes after input)

    fieldtag => {
		 list => ["<P>","<br>","</P>"],
		 table => ["<TR><TD>",":</TD><TD>","</TD></TR>"],
		 none => ["",":","<br>"]
		},

    # reqd_field_style can be:
    # "bold" (labels shown in bold type)
    # "star" (labels marked with an asterisk)
    # "css" (will be dealt with by css)
    # "none" (required fields are not distinct)

    reqd_field_style => "bold",

    # label tags (1st goes before field name, 2nd goes after)

    labeltag => {
		 bold => ["<B><SPAN CLASS=\"reqd_label\">","</SPAN></B>"],
		 star => ["<SPAN CLASS=\"reqd_label\">","<SUP>*</SUP></SPAN>"],
		 css => ["<SPAN CLASS=\"reqd_label\">","</SPAN>"],
		 none => ["<SPAN CLASS=\"label\">","</SPAN>"]
	        },

    # maximum input field width
    max_cols => 60,
    # maximum rows in a textarea input
    max_rows => 10,

    # form mode:
    # "cgi" to have forms processed by CGI programs (eg. doform.cgi)
    # "exsite" to have forms processed by WebDB DCD
    mode => "cgi",

    #---------------------------------
    # data handling options
    #---------------------------------

    # trim leading/trailing whitespace from form data before 
    # inserting/updating?

    trim_whitespace => 0,

    # default date settings for undefined dates: "today" or "null"

    default_date => "today",

    # default setting for the html editor

    htmleditor => {
    		cms => 0, # use text editor in cms mode?
		# set default command to turn options on/off in the editor's toolbar
		op_table => 0, # insert table
		op_image => 0, # insert image
		op_link => 0, # insert link
		op_dyncontent => 0, # insert dynamic content
		op_undo => 1, # undo
		op_redo => 1, # redo
		op_bold => 1, # bold
		op_italic => 1, # italic
		op_underline => 1, # underline
		op_textcolor => 1, # text text color
		op_bgcolor => 1, # set background color
		op_justify => 1, # text justify
		op_alignleft => 1, # text align left
		op_aligncenter => 1, # text align center
		op_alignright => 1, # text align right
		op_olist => 1, # ordered list
		op_blist => 1, # bulleted list
		op_outdent => 1, # outdent
		op_indent => 1, # indent
		
		op_image_upload => 1, # enable file upload form in the insert image/document popup
    },

    # file upload limits (in btyes)
    # "file" type is capped by field size (64K, which is reasonable)
    # "bigfile" type allows up to 16MB, which is probably unreasonable
    # for web uploads.  Set the actual bigfile size limit here:

    bigfile_maxsize => 1048576,  # 1 MB

    # display of foreign key selectors
    # set to 1 to include foreign key id in selector
    # set to 0 to use only general description of foreign record
    show_foreign_key_id => 1,

);

#======================================================================
# REPORT CONFIGURATION
# parameters controlling report formatting

%report_config = (

    # currency character
    currency_char => '$',

    # maximum number of records to report at a time
    page_length => 20,

    # for table reports, this flag determines whether each row is classified 
    # as A or B for the purposes of highlighting alternate lines.  Note that 
    # turning this flag on does not by itself highlight lines;  you must also 
    # setup your stylesheet to recognize the A/B classes of <TR>.  This 
    # setting will override $reportrecordtag->{table}->[0].
    highlight_alternate_lines => 0,

    # report link mode:
    # "cgi" to have db_ops linked to CGI programs (eg. form.cgi)
    # "exsite" to have db_ops linked back to the page (ie. to WebDB DCD)
    linkmode => "cgi",

    # email format:
    # "plain" displays the email using a regular mailto: URI
    # "obfuscated" uses javascript to obfuscate the address from spambots

    email => "plain",

    #---------------------------------
    # reporting of text data
    #---------------------------------

    text => {
	
	# allow HTML tags found in normal text

	honor_html => 0,

	# treat newlines as meaningful in normal text

	honor_nl => 0,

    },

    #---------------------------------
    # report styles
    #---------------------------------

    # Report styles are:
    #
    # table: records listed in a tabular format; column names are shown as first 
    #        header row
    # full: records listed in a 2-column column/value format, one table per record
    # dump: records listed as "column=value", one column per line.
    # index: records listed in a paragraph, with values delimited by commas
    # list-numbered: like index, but enumerated
    # list-bulleted: like index, but bulleted
    # address: records listed in a paragraph, with values delimited by linebreaks
    # description: records listed using <DT> for first value, <DD> for remaining 
    #	 values

    # default style
    style => "table",

    # the following tags go before and after the report, respectively
    report_tag => {
	table => [ "<TABLE CLASS=\"report_table\">", "</TABLE>" ],
	full => [ "", "" ],
	dump => [ "<hr>", "" ],
	index => [ "", "" ],
	list_numbered => [ "<OL>", "</OL>" ],
	list_bulleted => [ "<UL>", "</UL>" ],
	address => [ "", ""],
	description => [ "<DL>", "</DL>"],
    },

    # the following tags go before and after each record, respectively
    row_tag => {
	table => [ "<TR>", "</TR>" ],
	full => [ "<TABLE CLASS=\"report_record\">", "</TABLE><P>" ],
	dump => [ "<PRE>", "</PRE><hr>" ],
	index => [ "<P>", "</P>" ],
	list_numbered => [ "<LI><P>", "</P></LI>" ],
	list_bulleted => [ "<LI><P>", "</P></LI>" ],
	address => [ "<P>", "</P>" ],
	description => [ "<DT>", "<P>" ],
    },

    # the following tags go before and after each column value, respectively
    column_tag => {
	table => [ "<TD>", "</TD>" ],
	full => [ "<TR><TD CLASS=\"label\">", "</TD></TR>" ],
	dump => [ "", "\n" ],
	index => [ "", ", " ],
	list_numbered => [ "", ", " ],
	list_bulleted => [ "", ", " ],
	address => [ "", "<br>" ],
	description => [ "", "<DD>" ],
    },

    # report_labeltag: if defined, the column label will be shown along with 
    # each data value, with the given string separating the two.  If not 
    # defined, column labels will not be shown with each value.
    label_tag => {
	table => undef,
	full => "</TD><TD CLASS=\"data\">",
	dump => "=",
	index => undef,
	list_numbered => undef,
	list_bulleted => undef,
	address => undef,
	description => undef,
    },

);

#======================================================================
# SERVER CONFIGURATION
# this stuff varies between sites, so the following defaults are
# mostly just placeholders for configuration data to follow.
# The auto flag determines whether autoconfiguration is done by
# default.  Set it to 1 in exsite.conf if you want autoconfiguration.

%server_config = (
    db => { name=>"exsite", server=>"localhost", 
	    user=>"exsite", type=>"text" },
    map => { name=>"dbmap", type=>"text" },
    HTMLroot => "/var/www/html",
    HTMLpath => "",
    CGIroot => "/var/www/cgi-bin",
    CGIpath => "/cgi-bin",
    CSSpath => "/_ExSite/css",  # deprecated
    CSSfile => "ExSite.css",    # deprecated
    protocol => "http",
    host => "localhost",
    domain => "localdomain",
    port => "",
    server => "http://localhost.localdomain",
    temppath => "/tmp",
    imgdir => "/images",
    auto => 0,
);

#======================================================================
# LOG CONFIGURATION
# parameters controlling event logging

%log_config = (

    # Each parameter refers to the logging level for a particular log file.
    # Log levels go from 0 (no logging) to 3 (log everything).
    # Log files must exist and be writable by the webserver.

    # db - database log file
    # 1: log database errors
    # 2: log database modifications
    # 3: log database connections & selects
	       
    db => 0,

    # cms - content management system audit trail
    # 1: log publicly-visible changes
    # 2: log all website, page, and content updates
    # 3: log all cms operations

    cms => 0,

    # diag - general ExSite diagnostics
    # 1: log errors
    # 2: log warnings
    # 3: log info messages

    diag => 0,

    # the log files themselves... (paths are relative to cgi directory)

    logfile => {
	db => "db.log",
	cms => "cms.log",
	diag => "diag.log",
    },

);

#======================================================================
# WEBTOP CONFIGURATION 

%webtop_config = (

    # number of icons across
    x => 6,

);

#======================================================================
# CRYPTO CONFIGURATION 

%crypt_config = (

    # CRYPTOGRAPHY:
    # ExSite can automatically encrypt and decrypt data going to and from
    # the database, if the column has "crypt" datatype.  Crypt fields are
    # treated same as text fields, but the data is encrypted according
    # to the following options:

    # keys for coding/decoding
    # It is good practice to define different keys for fast and strong
    # encryption.  Otherwise the regular key will be used.

    key_fast => "fast dummy pass-phrase",
    key_strong => "strong dummy pass-phrase",
    key => "dummy pass-phrase",

    # number of mixing rounds (higher is better)

    nround_fast => 4,
    nround_strong => 32,

    # decryption rules:
    # admin => only administrator can view plaintext
    # owner => owner of record can view plaintext
    access => "admin",

);

#======================================================================
# GENERAL CONFIGURATION 

%config = (

    #---------------------------------
    # HTML output options:
    #---------------------------------

    # PAGE BUFFERING
    # set true if using the page buffering module, ExSite::Page.pm, by default.
    # set false if output goes straight to stdout by default.
    # NB: page buffering can be changed from the default for specific scripts 
    # by setting this flag at the top of the script.

    page_buffering => 0,

    # Thumbnail size (max edge length in pixels)

    thumbnail_size => 100,

    # handling of database images
    # set true if ExSite::Report::show_data should display images inline
    # set false if ExSite::Report::show_data should display a link to the image

    show_images => 1,

    # handling of database operations
    # each operation has the following parameters:
    #   approve: the low-level operation that needs to be approved by DB::approve
    #   url: the script that executes the operation
    #   img: the image for the operator icon
    #   img_width, img_height: the dimensions of the image
    #   label: the text string to use if the image is not present
    #   table: 1 means operation can be performed on whole tables
    #          (otherwise, op is restricted to individual records)
    # Each url is passed ?tab=<table>[&id=<record>]

    db_ops => {
	edit => { approve=>"update", url=>"webdb.cgi?action=edit", 
		  img=>"edit12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"edit", table=>0 },
	copy => { approve=>"insert", url=>"webdb.cgi?action=copy", 
		  img=>"copy12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"copy", table=>0 },
	delete => { approve=>"delete", url=>"webdb.cgi?action=delete", 
		  img=>"rm12.gif", level=>3,
	          img_width=>12, img_height=>13, label=>"delete", table=>0 },
	new => { approve=>"insert", url=>"webdb.cgi?action=new", 
		  img=>"new12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"add new", table=>1 },
	view => { approve=>"read", url=>"webdb.cgi?action=view", 
		  img=>"view12.gif", level=>1,
		  img_width=>12, img_height=>13, label=>"view details", table=>0 },
	list => { approve=>"read", url=>"webdb.cgi?action=list", 
		  img=>"list12.gif", level=>1,
		  img_width=>12, img_height=>13, label=>"list", table=>1 },
	report => { approve=>"read", url=>"webdb.cgi?action=report", 
		  img=>"list12.gif", level=>3,
		  img_width=>12, img_height=>13, label=>"list", table=>1 },
	searchquery => { approve=>"read", url=>"webdb.cgi?action=query", 
		  img=>"search12.gif", level=>1,
		  img_width=>12, img_height=>13, label=>"search", table=>1 },
    },

    db_ops_grps => {
	#all => [ "list", "view", "search", "edit", "copy", "new", "delete" ],
	all => [ "list", "view", "searchquery", "edit", "copy", "new", "delete" ],
	all_edit => ["edit", "copy", "delete"],
	all_view => ["list", "view", "search"],
	link => ["view"],  # kludge to support old "link" args
    },

    # preferred order for displaying db operators, when more than one is shown
    # ops not in this list will not be displayed at all.

    #db_ops_order => [ "list", "view", "search", "edit", "copy", "new", "delete" ],
    db_ops_order => [ "list", "view", "searchquery", "edit", "copy", "new", "delete" ],

    # HTML snippets (user-defined)

    snippet => {},

    # include the other configuration hashes here

    site => \%site_config,
    auth => \%auth_config,
    form => \%form_config,
    report => \%report_config,
    log => \%log_config,
    webtop => \%webtop_config,
    crypt => \%crypt_config,
    server => \%server_config,  # conditional

    # CRYPTOGRAPHY:

    # DEPRECATED - use %crypt_config, above

    # ExSite can automatically encrypt and decrypt data going to and from
    # the database, if the column has "crypt" datatype.  Crypt fields are
    # treated same as text fields, but the data is encrypted using the
    # CipherSaber method, with the following options:

    crypt_pass => "dummy pass-phrase",

    # set to 0 or 1 for CS-1 encryption; values > 1 use CS-2
    # crypt_level => 20,

    # TEA encryption uses 32
    crypt_level => 32,

    # decryption rules:
    # admin => only administrator can view plaintext
    # owner => owner of record can view plaintext
    decrypt => "admin",

);

#### end of static configuration


##################################################################
################ setup and configuration routines ################
##################################################################

=pod

=head1 Setup and Configuration Routines

=head2 server_autoconf() - autoconfigure ExSite

This routine attempts to set the various server configuration
parameters (ie. file paths and machine names) automatically, by
inferring them from the server environment.

WARNING: auto-configuration will not necessarily work for CGI scripts
run from the command line for testing/debugging purposes.

=cut

sub server_autoconf {
    $server_config{HTMLroot} = $ENV{DOCUMENT_ROOT};
    $ENV{SCRIPT_NAME} =~ m:^.*/([^/]+)$:;
    my $script_name = $1;
    $ENV{SCRIPT_FILENAME} =~ /^(.*)\/$script_name$/;
    $server_config{CGIroot} = $1;
    $server_config{CGIpath} = ($ENV{SCRIPT_NAME} =~ /^(.*)\/.*\.cgi$/)? $1 : "";
    $server_config{protocol} = exists $ENV{HTTPS} ? "https" : "http";
    $server_config{HTMLpath} = $server_config{CGIpath};
    $server_config{HTMLpath} =~ s/\/cgi(-bin)?\//\//;
    $server_config{HTMLpath} =~ s/\/cgi(-bin)?$/\//;
    if ($server_config{HTMLpath} eq "/") {
        $server_config{HTMLpath} = "";
    }

    # obsolete path parsing methods
    #$server_config{HTMLpath} =~ s/\/cgi(-bin)?\//\//;
    #$server_config{HTMLpath} = (($server_config{CGIpath} =~ /^\/.+?(\/.+)$/)? $1 : ""),

    # obsolete host parsing method - only does 3-part hostnames
    #$ENV{HTTP_HOST} =~ /^(.+?)\.(.*?):?(\d+)?$/;
    #$server_config{host} = $1;
    #$server_config{domain} = $2;
    #$server_config{port} = $3 ? ":$3" : "";

    # new method - handles 2- or 3-part names
    $ENV{HTTP_HOST} =~ /^((.+?)\.)?(.*?\..*?)(:\d+)?$/;
    $server_config{host} = $2;
    $server_config{domain} = $3;
    $server_config{port} = $4;


    # compose full server name
    # -> moved to exsite_init()
    #$server_config{server} = "$server_config{protocol}://$server_config{host}.$server_config{domain}$server_config{port}";

    # login destinations (server-specific)
    # after being authenticated, a user is deposited at URL
    # $config{auth}->{login_destination}->[$access_level]

#    foreach my $dest (@{$config{auth}->{login_destination}}) {
#	# add paths to login destinations, where necessary
#	if ($dest !~ /\//) {
#	    if ($dest =~ /\.html?$/) {
#		$dest = "$server_config{server}/".$dest;
#	    }
#	    elsif ($dest =~ /\.cgi$/) {
#		$dest = "$server_config{server}$server_config{CGIpath}/".$dest;
#	    }
#	}
#    }
}

=pod

=head2 set_old_config_vars() - legacy support

Define old configuration variables for the benefit of legacy scripts
that expect hard-coded configuration parameters.  You must import
symbols of tag :OLD to get these.  This subroutine is NOT called by
default;  you should call it from your script in exceptional cases,
or from &exsite_init in the general case.

=cut

sub set_old_config_vars {
    if (scalar @_ == 0) { $_ = "ALL"; }
    else { $_ = shift; }

    # site

    if (/ALL|SITE/) {
	$site_name            = $config{site}->{name};
	$site_administrator   = $config{site}->{administrator};
	$site_identity_cookie = $config{site}->{identity_cookie};
	$site_logo            = $config{site}->{logo};
    }

    # server

    if (/ALL|SERVER/) {
        %db_config = %{$config{server}->{db}};
	$dbmap_dir = $config{server}->{map}->{name};
	$HTMLroot  = $config{server}->{HTMLroot};
	$CGIroot   = $config{server}->{CGIroot};
	$HTMLpath  = $config{server}->{HTMLpath};
	$CGIpath   = $config{server}->{CGIpath};
	$CSSpath   = $config{server}->{CSSpath};  
	$CSSfile   = $config{server}->{CSSfile};
	$imgdir    = $config{server}->{imgdir};
	$host      = $config{server}->{host};
	$domain    = $config{server}->{domain};
	$protocol  = $config{server}->{protocol};
	$port      = $config{server}->{port};
	$server    = $config{server}->{server};
	$temppath  = $config{server}->{temppath};
    }

    # auth
    
    if (/ALL|AUTH/) {
	$default_authentication = $config{auth}->{type};
	$root_auth_level        = $config{auth}->{root_auth_level};
	$default_name           = $config{auth}->{default_identity}->{name};
	$default_uid            = $config{auth}->{default_identity}->{uid};
	$default_gid            = $config{auth}->{default_identity}->{gid};
	$default_access_level   = $config{auth}->{default_identity}->{access};
	$user_table             = $config{auth}->{user_table};
	$user_login_column      = $config{auth}->{user_login_column};
	$user_pwd_column        = $config{auth}->{user_pwd_column};
	$user_uid_column        = $config{auth}->{user_uid_column};
	$user_gid_column        = $config{auth}->{user_gid_column};
	$user_access_column     = $config{auth}->{user_access_column};
	$group_table            = $config{auth}->{group_table};
	$password_storage_method= $config{auth}->{password_storage_method};
	$foreign_link_level     = $config{auth}->{foreign_link_level};
	@login_destination      = @{$config{auth}->{login_destination}};
    }

    # forms

    if (/ALL|FORM/) {
	$form_style       = $config{form}->{style};
	$reqd_field_style = $config{form}->{reqd_field_style};
	%labeltag         = %{$config{form}->{labeltag}};
	%fieldtag         = %{$config{form}->{fieldtag}};
	$sepchar          = $config{form}->{sepchar};
	$max_field_size   = $config{form}->{max_cols};
	$max_field_rows   = $config{form}->{max_rows};
	$trim_whitespace  = $config{form}->{trim_whitespace};
	$default_date     = $config{form}->{default_date};
    }

    # reports

    if (/ALL|REPORT/) {
	$report_page_length = $config{report}->{page_length};
	%report_tag         = %{$config{report}->{report_tag}};
	%report_rowtag      = %{$config{report}->{row_tag}};
	%report_columntag   = %{$config{report}->{column_tag}};
	%report_labeltag    = %{$config{report}->{label_tag}};
	$highlight_alternate_lines = $config{report}->{highlight_alternate_lines};
    }

    # general

    if (/ALL|GENERAL/) {
	$page_buffering   = $config{page_buffering};
	$show_images      = $config{show_images};
	%db_ops           = %{$config{db_ops}};
#	@edit_links       = @{$config{all_edit_links}};
    }

    return;
}

=pod

=head2 read_exsite_conf() - load config parameters from a file

This routine loads configuration values from files.  By default 
it looks for "exsite.conf" files in the ExSite and . directories.
An alternate filename can be given.  By default configuration values 
are read straight into the root of %config, but a starting key can 
be given, eg. C<&read_exsite_conf("MyModule.conf","MyModule")> will
read the values in MyModule.conf into C<$config{MyModule}{...}>.

Configuration files contain parameters and values, separated by "=" 
(with any amount of whitespace around the "=".  Values extend from the 
first non-whitespace character after the "=", to the end of line.
Simple parameters are placed directly into %config's keys.  
Parameters of the form "a.b.c" are stored as $config{a}->{b}->{c}. 
Config array references can be created using "+=" to set the values
in the array.

Example exsite.conf settings:

C<trim_whitespace = 0>

C<form.style=list>

C<db_ops.new.url = append.cgi>

=cut

sub read_exsite_conf {
    my ($file,$key) = @_;
    if (! $file) { $file = "exsite.conf"; }
    if ($key) { $config{$key} = {}; }
    my @confpath = ("./conf", "ExSite", ".");
    foreach my $dir (@confpath) {
	if (-e "$dir/$file") {
	    if (open CONF,"<$dir/$file") {
		foreach (<CONF>) {
		    next if (/^\#/);
		    next if (/^\s*$/);
		    chomp;
		    my ($param,$value,$append);
		    if (/\+=/) { 
			($param,$value) = split /\s*\+=\s*/,$_,2;
			$append = 1;
		    }
		    else {
			($param,$value) = split /\s*=\s*/,$_,2;
			$append = 0;
		    }
		    my @param = split /\./,$param;
		    my $conf = $key ? $config{$key} : \%config;
		    foreach my $par (@param[0..$#param-1]) {
			if (! exists $conf->{$par}) {
			    $conf->{$par} = {};
			}
			# this will crash if not a hash ref
			$conf = $conf->{$par};
		    }
		    if ($append) {
			# append parameter to list
			if (! exists $conf->{$param[-1]}) {
			    $conf->{$param[-1]} = [$value];
			}
			else {
			    if (ref $conf->{$param[-1]} ne "ARRAY") {
				my $tmplist = [ $conf->{$param[-1]} ];
				$conf->{$param[-1]} = $tmplist;
			    }
			    push @{$conf->{$param[-1]}},$value;
			}
		    }
		    else {
			# simple parameter
			$conf->{$param[-1]} = $value;
		    }
		}
		close CONF;
	    }
	}
    }
}

=pod

=head2 exsite_init() - generic ExSite initialization routine

This routine configures ExSite, and performs any local initializations.
This routine is called automatically at the start of all
generic CGI scripts, and should be called by custom scripts
as well.

Local initialization can be done in C<&myConfig::my_exsite_init>,
which is called automatically from here.  Note that C<my_exsite_init()> 
is called on every initialization, so it should try to avoid 
reconfiguring static configurations (eg. handler registration) on 
every pass.

=cut

sub exsite_init {

    if (! $config_flag) {
	# Only setup the config hash if $config_flag is unset;
	# this prevents unnecessary reloading in mod-perl setups.

	# load any extra configuration from conf files
	&read_exsite_conf;

	# autoconfigure server?
	if ($config{server}->{auto}) {
	    &server_autoconf;
	}
	
	# build the server address from the various config params
	if ($config{server}{host}) {
	    $config{server}{server} = "$config{server}{protocol}://$config{server}{host}.$config{server}{domain}$config{server}{port}";
	}
	else {
	    $config{server}{server} = "$config{server}{protocol}://$config{server}{domain}$config{server}{port}";
	}
	
	$config_flag = 1;
    }

    # reset share hash
    %share = (
	      diagnostics => [],
	      );

    # localized initialization
    &my_exsite_init(\%config,\%share,@_);

}

=pod

=head2 exsite_close() - clean up

=cut

sub exsite_close {
    # localized clean-up;
    &my_exsite_close(@_);
    # close default database
    if ($share{server}->{DB}) {
	$share{server}->{DB}->disconnect;
    }
}

=pod

=head2 page_header($title) - standard top-of-page HTML

Outputs the standard HTML to begin a page.  Optionally, a page title
can be provided.

Invokes C<&myConfig::my_page_header>, which should contain any
localized code for generating the page header.

=cut

sub page_header {
    my $out;
    if ($share{DB}) {
	$out = $share{DB}->run_handler("page_header",@_);
    }
    if (! $out) {
	$out = &my_page_header(@_);
	if (! $out) { 
	    # no custom header - use generic ExSite system wrapper
	    my ($title,$opt) = @_;
	    $out = "<HTML>
<HEAD>
<TITLE>$title</TITLE>
<meta http-equiv=\"Content-Type\" content=\"text/html; charset=ISO-8859-1\">
<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"$config{server}{HTMLpath}/_ExSite/css/ExSite.css\">
<script LANGUAGE=javascript>
function popup_help(url) {
  window.open(url,\"\",\"width=500,height=500,resizable,scrollbars=1,menubar=0\")
}
</SCRIPT>
</HEAD>
<BODY MARGINHEIGHT=\"0\" MARGINWIDTH=\"0\" TOPMARGIN=\"0\" LEFTMARGIN=\"0\" RIGHTMARGIN=\"0\" BGCOLOR=\"#9999CC\">
<table cellspacing=0 cellpadding=0 border=0 bgcolor=#ccccdd width=100%>
<tr><td colspan=3 bgcolor=#eeeeff><img src=$config{server}{HTMLpath}/_ExSite/images/null.gif height=1 width=1></td></tr>
<tr><td bgcolor=#eeeeff><img src=$config{server}{HTMLpath}/_ExSite/images/null.gif height=1 width=1></td>
<td WIDTH=100%>

<!-- begin menubar -->

<table cellspacing=0 cellpadding=4 border=0 bgcolor=#ccccdd width=100%>

<td align=left WIDTH=100%>

<!-- logo -->
<a href=http://www.ExsiteWebware.com><img src=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/ex-logo28.gif height=28 width=28 border=1></a>

</td>
<td><A HREF=$config{server}{CGIpath}/home.cgi><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-home.gif HEIGHT=30 WIDTH=30 BORDER=0 TITLE=\"ExSite Home\"></A></td>
<td><A HREF=\"javascript:popup_help('$config{server}{CGIpath}/exsite-help.cgi')\"><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-help.gif HEIGHT=30 WIDTH=30 BORDER=0 TITLE=\"Help\"></A></td>
<td><A HREF=$config{server}{CGIpath}/logout.cgi><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-exit.gif HEIGHT=30 WIDTH=30 BORDER=0 TITLE=\"Log Out\"></A></td>
</tr>
</table>

<!-- end menubar -->

</td>
<td bgcolor=#333366><img src=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/null.gif height=1 width=1></td>
</tr>
<tr><td colspan=6 bgcolor=#333366><img src=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/null.gif height=1 width=1></td></tr>
</table><br>\n\n";

	    if (! $opt || ref $opt) {
		# opt flags the webtop in the default presentation
		# if !opt, then draw a window
		$out .= "<div align=center>
    <TABLE CELLSPACING=0 CELLPADDING=0 BORDER=0 WIDTH=90% CLASS=framed>
      <TR>
        <TH CLASS=titlebar><IMG SRC=$config{server}->{HTMLpath}/_ExSite$config{server}->{imgdir}/null.gif HEIGHT=1 WIDTH=1>
          <A HREF=$config{server}{CGIpath}/home.cgi><IMG SRC=$config{server}{HTMLpath}/_ExSite$config{server}{imgdir}/button-exit-small.gif HEIGHT=20 WIDTH=20 BORDER=0></A>&nbsp;<B>$title</B></Th>
      </TR>
      <TR>
        <TD BGCOLOR=#FFFFFF><DIV CLASS=htmlbox>\n";
	    }
	}
    }
	
    # do not modify anything below this point
    
    if ($config{page_buffering}) { 
	return $out; 
    }
    else { 
	print "Content-type: text/html\n\n";
	print $out; 
    }
}

=pod

=head2 page_footer() - standard end-of-page HTML

Outputs the standard HTML to end a page.
Invokes C<&myConfig::my_page_footer>, which should contain any
localized code for generating the page footer.

=cut

sub page_footer {
    my $out;
    if ($share{DB}) {
	$out = $share{DB}->run_handler("page_footer",@_);
    }
    if (! $out) {
	$out = &my_page_footer(@_);
	if (! $out) { 
	    # no special page footer - use ExSite system wrapper
	    my ($ref) = @_;
	    if (! $ref || ref $ref) {
		# close the pseudo-window
		$out = "        </DIV></TD>
      </TR>
    </TABLE>
</div>\n";
	    }
	    if ($share{DB}) {
		# there is a database object - check for any diagnostics to report
		if ($share{DB}->errorcheck()) {
		    $out .= "<h3>Errors:</h3>\n";
		    $out .= $share{DB}->show_diagnostics("error","html");
		}
		if ($share{DB}->errorcheck("warn")) {
		    $out .= "<h3>Warnings:</h3>\n";
		    $out .= $share{DB}->show_diagnostics("warn","html");
		}
	    }
	    $out .= "</BODY></HTML>\n";
	}
    }
    if ($config{page_buffering}) { return $out; }
    else { print $out; }
}

##################################################################
################ handlers                         ################
##################################################################

=pod

=head2 install_handlers() - register handlers

Handlers allow plug ins or other subsystems to register special 
business logic that should take precedence over the default logic
in certain cases.

The ExSite base system has to do this in a few cases for its content
management tables, which have their own logic that differs from the
generic database handling.  Otherwise, it is expected that most 
handlers will be site-specific (registered in myConfig::my_handlers() )
or plug-in-specific (registered from the plug-in module in 
ExSite::Modules).

=cut

sub install_handlers {
    my $db = $share{DB};
    if ($db) {
	# standard handlers here
	$db->handler("select_foreign_key",\&CMS_select_foreign_key);
    }
    # additional handlers unique to this installation 
    &my_handlers(@_);
}

# handler for foreign key selection on CMS tables
#
# - for *.template_id, select only valid templates
# - for *.section_id, select only from my_sites()
# - for page.parent_id, select only from other pages in same section
# - for content.page_id, select only from pages in same section
# - for section.parent_id, select only from other non-template sections

sub CMS_select_foreign_key {
    my ($this, $table, $rec, $col, $ftable) = @_;
    # ignore, if not using the CMS database
    if ($this->{name} eq $config{server}{db}{name}) {
	if ($col eq "section_id") {
	    if (! $this->is_admin) {
		my @sites;
		@sites = $this->my_real_sites();
		return \@sites;
	    }
	    return undef;
	}
	if ($table eq "page") {
	    if ($col eq "template_id") {
		# select only from templates
		### more authorization needed here, eg. disallow 3rd-party private templates
		return $this->fetch_match("page",{type=>"template"});
	    }
	    elsif ($col eq "parent_id") {
		my (@possible_parents,$page,@parents);
		# select only from pages in same section
		$page = $this->fetch("page",$rec);
		@parents = $this->fetch_match("page",{type=>"page",section_id=>$page->{section_id}});
		# don't allow user to select a page as its own parent
		foreach my $parent (@parents) {
		    if ($parent->{page_id} ne $page->{page_id}){push @possible_parents, $parent;}
		}
		return \@possible_parents;
	    }
	}
	if ($table eq "content") {
	    if ($col eq "page_id") {
		my ($content,$page);
		$content = $this->fetch("content",$rec);
		$page = $this->fetch("page",$content->{page_id});
		return $this->fetch_match("page",{section_id=>$page->{section_id}});
	    }
	}
	if ($table eq "section") {
	    if ($col eq "parent_id") {
		my $sql = "select * from section where type != 'website template'";
		$sql .= ($rec =~ /^\d*$/) ? " and section_id != $rec" : "";
		return $this->custom_query($sql);
	    }
	}
    }
    return undef;
}

1;

