#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Content;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::ContentData;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::ContentData);

#------------------------------------------------------------------------

=pod

=head1 Content.pm : manage content objects

A content object is an element of a web page, such as text, html, 
stylesheets, scripts, images, or downloadable files.  This module
provides methods to act on generic content objects.

new: defines a new content object

Also inherits from the ContentData and ContentBase classes.

=cut

#------------------------------------------------------------------------

# setup info for Content class

sub setup {
    &setupContent(@_);
}

sub setupContent {
    my ($this,%option) = @_;

    # fetch the object in question
    ###### CHECK PERMISSIONS.
    my $content;
    if ($option{id}) {
	# direct lookup
	$content = $this->fetch("content",$option{id});
    }
    elsif ($option{content}) {
	# record already fetched
	$content = $option{content};
    }
    elsif ($option{name} && $option{page}) {
	########################
	### search templates for matching content for general display of page,
	### but don't search in CMS management mode

	my @content = $this->fetch_match("content",
					 {name=>$option{name},
					  page_id=>$option{page}},
					 "content_id");
	# should be only 1, but in case there are more, we'll take the
	# most recently added version.
	$content = $content[-1];
    }

    if (ref $content eq "HASH") {
	# found something
	$this->info("setting up Content $content->{content_id}");
	$this->{data} = $content;
	$this->set_context("child",$option{version});

	#$this->{name} = $content->{name};
#	my $context = "none";
#	if (! defined $option{context} || $option{context}) {
#	    $context = $option{context} ? $option{context} : "all";
#	}
#	$this->set_context($context,$content);
    }
    else {
	$this->info("setting up null Content");
    }

}

# make_content : insert a new content record
#
# can optionally pass a content data record to seed it with

sub make_content {
    my ($this,$content,$cdata,$opt) = @_;
    if ($content->{content_id}) {
	# invalid content
	$this->error("attempt to insert an already-defined content record");
	return undef;
    }
    if (! $content->{name}) {
	$this->error("cannot insert unnamed content");
	return undef;
    }
    if (! $content->{page_id}) {
	$this->error("new content not attached to any page");
	return undef;
    }
    if (! $content->{type}) {
	$content->{type} = "editorial";
    }
    my $cid = $share{DB}->insert("content",$content);
    if (! $cid) {
	$this->error("failed to install new content $content->{name}");
    }
    else {
	$this->info("installed new content $content->{name}");
	if (ref $cdata eq "HASH") {
	    $cdata->{content_id} = $cid;
	    my $cdid = $this->make_content_data($cdata,$opt);
	    # ignore return code; make_content_data will report its own errs
	}
    }
    return $cid;
}

# roll back to the previous version of the content

sub rollback {
    my $this = shift;
    # remove the latest version
    $this->delete;

    ### this will have unintended consequences if the version in use
    ### is not the latest version!

    # reconfigured the context with the previous version
    $this->{context}{"content_data"} = $this->find_content_data();
}

# count the number of versions on file

sub count {
    my $this = shift;
    return $share{DB}->count("content_data",
			     {content_id=>$this->id("content")});
}

# everything else is inherited :-)

### FIXME:
# - is_published() - determine whether content needs to be published
# - publish_date() - determine when content was last published


1;
