#!/usr/bin/perl
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::ContentBase;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Base;
use ExSite::Input;

use vars qw(@ISA @class_structure %class_index);
@ISA = qw(ExSite::Base);
@class_structure = ("section", "page", "content", "content_data");
%class_index = ("ExSite::Section"=>0, 
		"ExSite::Page"=>1, 
		"ExSite::Content"=>2, 
		"ExSite::ContentData"=>3, 
		"ExSite::ContentData2"=>3);

#------------------------------------------------------------------------

=pod

=head1 ContentBase.pm : generic content-management methods

These are tools that can be used/inherited by all CMS data types (ie.
sites, pages, content definitions, and content data).

=cut

#------------------------------------------------------------------------


# options
# name => content object name
# page => page this content is assigned to
# id => direct lookup of content version
# version => the version to select
# DB => database object (optional)

sub new ($%) {
    my($this, %option) = @_;

    # objectify object
    my $cobj = {};
    my $class = ref($this) || $this;
    bless $cobj, $class;

    # define database

    if (defined $option{DB}) {
	$cobj->{DB} = $option{DB};
    }
    else {
	$cobj->{DB} = $share{DB};
    }

    # setup content cache - in share, so that multiple objects can share it

    if (! exists $share{ccache}) {
	$share{ccache} = new ExSite::Cache;
    }

    # setup Base class features

    $cobj->initialize_object;

    # register CMS-specific handlers

#    $cobj->{DB}->handler("select_foreign_key",&CMS_select_foreign_key);

    # setup context hash

    $cobj->{context} = {};

    # end generic constructor; now run local class setup

    $cobj->setup(%option);
    return $cobj;
}

# set the context : needed to find() related content.
#
# Every data object in the CMS is part of a content stack:
# site -> page -> content-def -> content-data
#
# Setting the context consists of finding and loading the other objects 
# in this stack.  It is assumed that the current data object is already 
# loaded.
#
# mode is none|all|parent|child

sub set_context {
    my ($this,$mode,$version) = @_;
    return if $this->{context_set};
    $mode = $mode || "all";
    if ($mode eq "all") { $this->{context_set} = 1; }
    # start with current object
    my $class = ref $this;
    my $index = $class_index{$class};
    $this->{context}{$class_structure[$index]} = $this->{data};
    return if ($mode eq "none");
    if ($mode =~ /all|parent/) {
	# move up thru the content stack by following foreign key refs
	for (my $i = $index; $i > 0; $i--) {
	    my $table = $class_structure[$i];
	    my $ptable = $class_structure[$i-1];
	    my $pkey = $ptable."_id";
	    my $id = $this->{context}{$table}{$pkey};
	    if ($id) {
		$this->{context}{$ptable} = $this->fetch($ptable,$id);
	    }
	    else {
		$this->warn("no $table found");
		last;
	    }
	}
    }
    if ($mode =~ /all|child/) {
	# move down thru the content stack
	my @cstack = @class_structure;
	while (scalar @cstack && $cstack[0] ne 
	       $class_structure[$class_index{$class}]) { 
	    shift @cstack; 
	};
	foreach my $ctype (@cstack) {
	    next if (ref $this->{context}{$ctype} eq "HASH"); # already loaded
	    if ($ctype eq "page") {
		$this->{context}{$ctype} = $this->find_page("index.%");
	    }
	    elsif ($ctype eq "content") {
		$this->{context}{$ctype} = $this->find("page","data");
	    }
	    elsif ($ctype eq "content_data") {
		$this->{context}{$ctype} = $this->find_content_data($version);
	    }
	}
    }
}

#------------------------------------------------------------------------

=pod

context testing routines

=cut

#------------------------------------------------------------------------

sub is {
    my ($this,$type) = @_;
    # return true if this object is a $type
    return (ref $this =~ /^$type$/i);
}

sub has {
    my ($this,$type) = @_;
    # return true if this object has a $type loaded in its context
    return defined $this->{context}{$type};
}

sub has_content {
    my ($this) = shift;
    # return true if this has content data defined 
    return $this->has("content_data");
}

sub get { 
    my ($this,$type,$attribute) = @_;
    if ($type) {
	# default to ::Base behaviour if $type is not a context item
	return $this->{$type} if (ref $this->{context}{$type} ne "HASH");
	# if $type is a context item, return data for it
	return $attribute ? 
	    $this->{context}{$type}{$attribute} : 
	    $this->{context}{$type};
    }
    else {
	# no args - return this object's data
	return $this->{context}{$class_structure[$class_index{ref $this}]};
    }
}

sub id {
    my ($this,$type) = @_;
    my $type = $type ? $type : $class_structure[$class_index{ref $this}];
    return $this->{context}{$type}{$type."_id"};
}


#------------------------------------------------------------------------

=pod

access control methods

DEPRECATED:  see ExSite::Auth for access control tools

=cut

#------------------------------------------------------------------------


sub access {
    my ($this,$method,$class,$id) = @_;
    my $db = $share{DB};
    my $uid = $db->my_uid;
    my @sites = $db->fetch_m2m("section","member",$uid,"member_site_link");
    $this->set_context;

    # crude access control for now - simply check for a section match

    foreach my $site (@sites) {
	if ($site->{section_id} == $this->id("section")) {
	    return 1;
	}
    }
    return 0;
}

#------------------------------------------------------------------------

=pod

=head2 find() - locate related or referenced content

find logic:
- look in current page's content objects
- look in templates' content objects, from nearest relation to furthest
- look in content libraries for this section
- look in content libraries for ancestral sections
- take first eligible match found

mode is "data" or "object"

=cut

#------------------------------------------------------------------------

sub find {
    my ($this,$cname,$mode) = @_;

    if (! $this->{context_set}) {
    #if (! $this->{context}{page}) {
	# missing our context - can't just check context_set flag, because
	# set_context calls us before the context is fully set.
	$this->set_context;
    }

    my $page = $this->{context}{page};
    return undef if (! $page);

    # search the current page
    my $c = $this->find_content($cname);

    # search the template stack
    while (! $c && $page->{template_id}) {
	$page = $this->fetch("page",$page->{template_id});
	$c = $this->find_content($cname,$page->{page_id});
    }

    # search the libraries, if this is not the body
    if (! $c && $cname ne "body") { 
	# get current section
	my $section = $this->{context}{section};
	######## set the parent context first, to be sure it is defined?
	return undef if (! $section);
	my $sid = $section->{section_id};
	while ($sid) {
	    # get all libraries in this section
	    my @libs = $this->fetch_match("page",{type=>"library",
						  section_id=>$sid});
	    if (@libs > 0) {
		# search each library for a matching content object
		my $lib = shift @libs;
		while ($lib && ! $c) {
		    $c = $this->find_content($cname,$lib->{page_id});
		    $lib = shift @libs;
		}
		last if ($c);
	    }
	    # goto previous section
	    $sid = $section->{parent_id};
	    if ($sid) {
		$section = $this->fetch("section",$sid);
	    }
	}
    }

    return $c if ($mode eq "data");
    require ExSite::Content;
    return new ExSite::Content(content=>$c);
}

# more find tools

sub find_page {
    my ($this,$name,$sid) = @_;
    return undef if (! $this->{context}{section});
    if (! $sid) { $sid = $this->{context}{section}{section_id}; }
    return undef if (! $sid);
    my @page = $this->fetch_match("page",{filename=>$name,section_id=>$sid},
				  "page_id");
    return $page[-1];
}

sub find_content {
    my ($this,$name,$pid) = @_;
    return undef if (! $this->{context}{page});
    if (! $pid) { $pid = $this->{context}{page}{page_id}; }
    return undef if (! $pid);
    my @content = $this->fetch_match("content",{name=>$name,page_id=>$pid},
				     "content_id");
    return $content[-1];
}

sub find_content_data {
    my ($this,$type,$cid) = @_;
    return undef if (! $this->{context}{content});
    if (! $cid) { $cid = $this->{context}{content}{content_id}; }
    return undef if (! $cid);
    my @content_data;

    # 2 cases:
    
    # 1) templates (or pages??) may have precompiled "page" content objects
    #    Although this may look a little messy, it saves us a bunch of 
    #    DB lookups to find auxiliary template content.

#    if (($this->{context}{content}{name} eq "page") && 
#	($this->{context}{content}{page_id} != $this->{context}{page}{page_id})) {
	# 2nd condition selects for page objects found outside the current
	# page, ie. in an external template.  This avoids infinite context 
	# loops, but also prevents us from precompiling any page that holds
	# its own page element.  Therefore a template preview will never use
	# a precompiled version of itself, and you cannot precompile a page
	# that has its own page object.

	# Let's get on with the insanity:  make a new object and context

	# find the webpage/template that owns the "page" content object.
#	my $p = new ExSite::Page(id=>$this->{context}{content}{page_id});
#	if ($p->is_template() && 
#	    $p->get("page","publish_method") eq "static") {
#	    my $path = $p->diskpath();
#	    if (-e "$path/index.html") {
#		# found precompiled template - make a virtual record
#		$content_data[0] = {};
#		$content_data[0]{content_data_id} = -1;  # dummy
#		$content_data[0]{data} = &get_file("$path/index.html",{});
#		$content_data[0]{mime_type} = "text/html";
#	    }
#	}
#    }
    
    # Q: can we skip the DB lookup for other published files?

    # 2) if we have nothing, we'll try looking it up in the database

    if (scalar @content_data == 0) {
	@content_data = $this->fetch_child("content_data",
					      "content",
					      $cid,
					      "content_data_id");
    }
    if ($type =~ /current|active/) {
	# return most recent approved version
	for (my $ic = $#content_data; $ic >= 0; $ic--) {
	    # if status flag is set, this content is waiting for approval
	    if (! $content_data[$ic]->{status}) {
		return $content_data[$ic];
	    }
	}
	return undef;
    }
    else {
	# return most recent version
	return $content_data[-1];
    }
}


#------------------------------------------------------------------------

=pod

=head2 expand(): Content Expansion/Substitution

Expands the content in the current context, by replacing CMS tags with
the URLs or the raw content that they represent.

Options:

content => a hash of content names and values to use preferentially over
           those in the database
method => a combination of data,url,content,page,module, depending on 
          which expansions you want to apply to this content.  All are 
          selected, by default.
expand => template|design
          this flag requests partial expansion of the content.  Only
          template, design, content are expanded;  remaining tags
          are left in place.  DCD tags are left unexpanded in both cases.
template => if true, expand template and design elements only;  leave 
          local editorial elements in place as CMS tags.
cms => if true, include CMS tools in the output

=cut

#------------------------------------------------------------------------

sub substitute {
    return &expand(@_);
}

sub expand { 
    my ($this,%option) = @_;

    return $this->{html} if ($this->{expanded} || $this->{noexpand});
    $this->info("expanding html");
    $this->{html} = $this->get_html($option{mode});

    if (! exists $option{content}) { $option{content} = {}; }
    my $method = $option{method} || "data,url,content,page,module"; 
    my $cmsmenu = 0;

    # build a hash of meta data - combine page and section datahashes
    my $meta = $this->{context}{page};
    foreach my $key (keys %{$this->{context}{section}}) {
	$meta->{"site_".$key} = $this->{context}{section}{$key};
    }

    do {

	# clear substitution counter
	$this->{htmlsub} = 0;

	# 1) meta data substitutions, eg. <!--$name-->

	if ($method =~ /data/) { 
	    $this->{html} =~ s/<!--\$(\w+)-->/$meta->{$1}/gs;
	}

	# 2) content-URL substitutions, eg. SRC=[[content-object-name]]

	if ($method =~ /url/) {
	    $this->{html} =~ s/\[\[([\w\.-]+)\]\]/$this->get_content_url($1,\%option)/ges;
	}

	# 3) page-URL substitutions, eg. HREF={{page-name/id}}

	if ($method =~ /page/) { 
	    $this->{html} =~ s/\{\{([\w\.-]+)\}\}/$this->get_page_url($1,\%option)/ges;
	}

	# 4) HTML content substitutions, eg. <!--&content(content-obj-name)-->

	if ($method =~ /content/) { 
	    # repeat until all substitutions are done
	    while ($this->{html} =~ /<\!--\&?content\((.+?)\)-->/) {
		$this->{html} =~ s/<\!--\&?content\((.+?)\)-->/$this->get_content($1,\%option)/ges;
	    }
	}

	# NOTE:  <!--&content(foo)--> is now deprecated in favor of 
	# <!--content(foo)--> which isn't syntactically ambiguous with the
	# DCD tags, below.
	    
	# 5) Dynamic Content Driver substitutions, eg. <!--&module.method(args)-->
	if ($method =~ /module/) {
	    # add a CMS menu DCD tag in advanced CMS mode
	    # NB: editmode selects a non-default way of handling inline CMS 
	    #     funcs; only add the menu if this is not set
	    if ($option{cms} && (! $config{cms}{editmode}) && ! $cmsmenu ) {
		$this->{html} =~ s/(<body[^>]*>)/$1\n<!--\&CMS\(cmscmd=pagemenu\)-->/i;
		# ensure that cmsmenu is only inserted once
		$cmsmenu = 1;
	    }
	    # insert DCD content
	    do {
		$this->{expand_redo} = 0;
		my $html = $this->{html};
		# direct substitution of dynamic content (same server request)
		$html =~ s/<\!--\&(\w+)\((.*?)\)-->/$this->get_dynamic_content($1,$2,\%option)/ges;
		# indirect substitution of dynamic content (separate request)
		$html =~ s/<\!--\&\&(\w+)\((.*?)\)-->/$this->get_dynamic_content_indirect($1,$2,\%option)/ges;
		if (! $this->{expand_redo}) {
		    $this->{html} = $html;
		}
	    } until (! $this->{expand_redo});
	    # normally only one pass through do...until loop
	    # but some DCDs may set the expand_redo flag to rebuild the page
	}
    } until ($this->{htmlsub} == 0);

    if ($option{expand}) {
	# partial expansion was requested, which means we should now have
	# one or more masked CMS tags in our HTML.  Now we have to unmask them.
	$this->{html} = &restore_cms_tags($this->{html});
    }

    $this->{expanded} = 1;  # flag expansion as done
    return $this->{html};
}

# sub_page_url : return replacement text for a {{...}} CMS tag

sub get_page_url {
    my ($this,$id,$option) = @_;
    my $p;
    if ($id =~ /^\d+$/) {
	# page id
	$p = new ExSite::Page(id=>$id,noload=>1);
    }
    else {
	# page name
	$p = new ExSite::Page(filename=>$id,
			      section_id=>$this->id("section"));
    }
    return $p->get_url;
}

# sub_content_url : return replacement text for a [[...]] CMS tag

sub get_content_url {
    my ($this,$id,$option) = @_;
    # content reference
    my ($subobj, $vfile);
    $subobj = $this->find($id);
    if (! $subobj || ! $subobj->defined) {
	# no matching content - escape the tag chars in case this
	# is not a CMS tag
	$vfile = &html_escape("[[$id]]","[]");
    }
    else {
	$vfile = $subobj->get_url($option->{mode});
    }
    return $vfile;
}

# sub_content : return replacement text for a <!--&content(...)--> CMS tag

sub get_content {
    my ($this,$id,$option) = @_;

    $this->{htmlsub}++;
    my ($start,$end,$cobj,$cobjname,$csub,$newtext,@cobjname);
    @cobjname = split(/,/,$id);
    # first content object in list is substituted directly
    $cobjname = shift @cobjname;

    if ($option->{method} =~ /dummy-content/) {
	# we're not really substituting content, 
	# so use a placeholder graphic instead
	$newtext = $this->placeholder_image($id);
    }
    elsif ($option->{content}->{$cobjname}) {
	# predefined content object
	$newtext = $option->{content}->{$cobjname};
    }
    else {
	# look for matching content object in DB
	$cobj = $this->find($cobjname);
	if ($option->{expand} eq "template" && 
	    $cobj->get("content_data","page_id") == $cobj->get("page","page_id") ) {
	    # defer substitution of local page content
	    # (we are using the placeholder image to "escape" the CMS tag)
	    $newtext = $this->placeholder_image($id);
	}
	elsif ($option->{expand} eq "design" && 
	    $cobj->get("content","type") eq "editorial") {
	    # defer substitution of editorial content 
	    $newtext = $this->placeholder_image($id);
	}
	else {
	    # substitute now
	    my $in = new ExSite::Input;
	    my $q = $in->query;

	    # FIXME:
	    # first two clauses that follow are experiments in inline
	    # content editing for a simplified CMS interface.  This should
	    # be cleaned up when it becomes clear which method is preferred.

	    if ($id eq "body" && $config{cms}{editmode} eq "easy" && 
		$option->{cms} && $q->{cmscmd} =~ /update/) {
		# inline edit - do not output content
		$newtext = "<!-- body -->\n";
	    }
	    elsif ($id eq "body" && $config{cms}{editmode} eq "MySite" && 
		$option->{cms} && $q->{ecmd} eq "update") {
		# inline edit - do not output content
		$newtext = "<!-- body -->\n";
	    }
	    else {
		$newtext = $cobj->get_html;
	    }
	    if ($option->{cms} && 
		$cobj->{context}{content}{page_id} == 
		$this->{context}{page}{page_id}) {
		###### should allow hotlinks to undefined content
		###### objects that are being inherited from template
		# can only manage content tied directly to this page
		# to manage templates, view the template in cms mode
		
		# inline editing links
		if ($share{DB}->owns() && 
		    (($config{cms}{editmode} ne "MySite") || $id eq "body")) {
		    $newtext .= $this->edit_content_link($cobj);
		    # highlight editable content sections
		    $newtext = "<div style=\"border-width:1px; border-color:red; border-style:dashed; padding:0px;\">$newtext</div>\n";
		}
	    }
	}
    }
    
    # remaining content objects in list are content objects to 
    # insert into the first content object
    
    my $icobj = 1;
    foreach $cobjname (@cobjname) {
	if ($cobjname =~ /^\".+\"$/) {
	    # quoted string - plaintext substitution
	    # simple strings only - (no commas!)
	    $cobjname =~ s/^\"//;
	    $cobjname =~ s/\"$//;
	    $newtext =~ s/<!--\&?content\($icobj\)-->/$cobjname/g;
	}
	else {
	    # unquoted string - treat as a content object reference
	    $newtext =~ s/<!--\&?content\($icobj\)-->/<!--content($cobjname)-->/g;
	}
	$icobj++;
    }
    return $newtext;
}

sub get_dynamic_content {
    my ($this,$module,$modargs,$option) = @_;

    my ($endname,$modname,$modmethod,$newtext);
    # save dcd objects for re-use
    if (! $this->{dcd}) { $this->{dcd} = {}; }

    # flag a substitution as having occured, so we can expand these results
    $this->{htmlsub}++;

    ($modname,$modmethod) = split(/\./,$module,2);
    if (!$modmethod) { $modmethod = "write"; }
    
    # run the DCD
    
    if ($option->{method} =~ /dummy-module/ || $option->{expand}) {
	my $icon = "$config{server}{HTMLpath}/_Modules/$modname/icon.gif";
	if (! -e "$config{server}{HTMLroot}$icon") {
	    $icon =~ s/\.gif$/\.png/;
	}
	if (! -e "$config{server}{HTMLroot}$icon") {
	    # no icon; use default ExSite app icon
	    $icon = "$config{server}{HTMLpath}/_Modules/icon.gif";
	}
	$newtext = "<img alt=\"$module($modargs)\" src=$icon title=\"$module($modargs)\" name=\"$module($modargs)\" border=\"3\" style=\"border-style:outset;\">";
    }
    elsif ($modmethod =~ /^(read|write|ioctl)$/) {
	# find the module
	my $stat = eval "require Modules::$modname";
	if ($stat) {
	    if (! $this->{dcd}{$modname}) {
		# create the DCD object
		$this->{dcd}{$modname} = eval "new Modules::$modname";
	    }
	    my $dcd = $this->{dcd}{$modname};
	    if ($dcd) { 
		# check licensing
		my $license_ok = 1;
		if ($dcd->ioctl("isRestricted")) {
		    # this service requires licenses
		    my @licenses = $share{DB}->fetch_match("service",
							   {section_id => $this->id("section"),
							    name => $modname }
							   );
		    $license_ok = (@licenses > 0);
		    if (! $license_ok) {
			# log the problem
			$this->warn("$modname: unlicensed access from page ".$this->id("page"));
		    }
		}
		
		if ($license_ok) {
		    # execute the DCD method
		    $newtext = $this->{dcd}{$modname}->$modmethod($modargs) || $@;
		    # disable substitution for read and ioctl methods
		    if ($modmethod ne "write") {
			$newtext = "<!-- $modname.$modmethod executed -->";
		    }
		}
		else {
		    $newtext = "$modname: permission denied (no license)";
		}
	    }
	    else {
		$newtext = "$modname: $@";
	    }
	}
	else {
	    # $@ is $EVAL_ERROR
	    $newtext = "<H4>Error in $modname dynamic content driver:</H4>\n<P>\n$@\n</P>";
	    $this->error("$modname: $@");
	}
    }
    else {
	$newtext = "<!-- Error: invalid dynamic content driver method: $modmethod -->";
	$this->error("invalid DCD method: $modmethod");
    }
    return $newtext;
}

# get_dynamic_content_indirect : fetches the DCD content using a separate
# server request, instead of inlining it directly.
#
# The main advantage is that you can publish the main page to a static 
# HTML file, yet keep some page elements dynamic.  This is especially 
# useful for index.html pages, which must be static, but may contain 
# dynamic elements (eg. recent news, upcoming events, current specials,
# etc.).  The solution is either continuous republishing of the index
# page (which may still be a better solution for heavily loaded sites)
# or using an indirect dynamic content fetch.
#
# The disadvantage is that the full page is slower (although the base
# page may be much faster), and that JavaScript must be enabled to 
# perform the secondary content fetches.
#
# The CMS syntax for an indirect DCD call is:
# <!--&&Module(...)-->

sub get_dynamic_content_indirect {
    my ($this,$module,$modargs,$option) = @_;
    my $newtext;
    if ($option->{method} =~ /dummy-module/ || $option->{expand}) {
	my $icon = "$config{server}{HTMLpath}/_Modules/$module/icon.gif";
	if (! -e "$config{server}{HTMLroot}$icon") {
	    $icon =~ s/\.gif$/\.png/;
	}
	if (! -e "$config{server}{HTMLroot}$icon") {
	    # no icon; use default ExSite app icon
	    $icon = "$config{server}{HTMLpath}/_Modules/icon.gif";
	}
	$newtext = "<img alt=\"&$module($modargs)\" src=$icon title=\"&$module($modargs)\" name=\"&$module($modargs)\" border=\"3\" style=\"border-style:outset;\">";
    }
    else {
	my $ref = &url_escape($ENV{REQUEST_URI});
	$newtext = "<script langauge=\"JavaScript\" src=\"$config{server}{HTMLpath}/_ExSite/js/httprequest.js\"></script>
<script language=\"JavaScript\">
<!--//
document.write(getRequestData('$config{server}{CGIpath}/dcd.cgi/$module?$modargs&ref=$ref'));
//-->
</script>
<noscript>
(You must have javascript enabled to view $module.)
</noscript>\n";
    }
    return $newtext;
}

sub placeholder_image {
    my ($this,$id) = @_;
    $id =~ s/\"/%22/g;
    $id =~ s/</%3C/g;
    $id =~ s/>/%3E/g;
    return "<img alt=\"content($id)\" src=\"$config{server}{HTMLpath}/_ExSite/images/checker.png\" title=\"$id\" name=\"$id\" HEIGHT=50 WIDTH=100>";
}

# wrappers for the DB data fetching methods, so that we can cache and reuse
# content fetched already

### TODO: additonal fetching/caching performance improvements can be
### realized by prefetching all content in advance, rather than doing
### individual searches for each content object.

sub fetch {
    my $this = shift;
    my $dat = $share{ccache}->get(@_);
    if (! $dat) {
	$dat = $this->{DB}->fetch(@_);
	$share{ccache}->save($dat,@_);
    }
    if (wantarray) { return %$dat; }
    return $dat;
}

sub fetch_child {
    my $this = shift;
    my $dat = $share{ccache}->get(@_);
    if (! $dat) {
	$dat = $this->{DB}->fetch_child(@_);
	# cache the whole query
	$share{ccache}->save($dat,@_);
	# now cache the individual records
	#my $pkey = $this->{DB}->get_key($_[0]);  # calls DB server
	my $pkey = $this->{DB}{map}->get_table_attr($_[0],"key"); # no server call
	foreach my $row (@$dat) {
	    # cache under same index as a simple fetch
	    $share{ccache}->save($row,$_[0],$row->{$pkey});
	}
    }
    if (wantarray) { return @$dat; }
    return $dat;
}

sub fetch_match {
    my ($this,$table,$match,$sort) = @_;
    # expand the match hash for best indexing
    # match can be a scalar or hash ref
    my @match = ref $match ? %$match : ($match);
    my $dat = $share{ccache}->get($table,@match,$sort);
    if (! $dat) {
	$dat = $this->{DB}->fetch_match($table,$match,$sort);
	$share{ccache}->save($dat,$table,@match,$sort);
	# now cache the individual records
	#my $pkey = $this->{DB}->get_key($table); # calls DB server
	my $pkey = $this->{DB}{map}->get_table_attr($table,"key"); # no server call
	foreach my $row (@$dat) {
	    # cache under same index as a simple fetch
	    $share{ccache}->save($row,$table,$row->{$pkey});
	}
    }
    if (wantarray) { return @$dat; }
    return $dat;
}

# get_dcd_all : return an array of installed DCD names 

sub get_dcd_all {
    my $this = shift;
    my @dcd;
    opendir MODDIR, "$config{server}{CGIroot}/Modules";
    my ($mod,$napp);
    @dcd = readdir MODDIR;
    closedir MODDIR;
    @dcd = grep /\.pm$/, @dcd; # only accept Perl modules
    #@dcd = grep !/^BaseDCD\.pm$/, @dcd; # discard BaseDCD
    foreach (@dcd) { s/\.pm$//; }
    return sort @dcd;
}

# get_dcd : return an array of DCD names available to the current site

sub get_dcd {
    my $this = shift;
    my $section = shift;
    if (!$section) {
	$section = $this->id("section");
    }
    my @dcd = $this->get_dcd_all;
    my @dcd_allow;
    foreach my $dcd (@dcd) {
	my ($mod,undef) = split /\./,$dcd;
	if (eval "require Modules::$mod") {
	    my $s = eval "new Modules::$mod";
	    if ($s) {
		if ($s->ioctl("Restricted")) {
		    # this service requires licenses
		    my @licenses = $share{DB}->fetch_match("service",
							   {section_id => $section,
							    name => $mod }
							   );
		    if (@licenses > 0) {
			push @dcd_allow, $dcd;
		    }
		}
		else {
		    push @dcd_allow, $dcd;
		}
	    }
	    else {
		$this->error($@);
	    }
	}
    }
    return @dcd_allow;
}

#------------------------------------------------------------------------
#
# placeholder methods : these are dummies that will be overloaded in
# higher-level CMS classes.  They are provided here to allow ContentBase
# to run stand-alone for testing.

sub get_raw {
    return;
}

sub get_html {
    return;
}

sub get_url {
    return;
}

sub setup {
    return;
}

sub edit_content_link {
    return;
}
    

1;
