package ExSite::Form;

#-----------------------------------------------------------------------
# ExSite::Form
# Author: Morgan Burke
# (c) 2003 Exware Solutions
#-----------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

=pod

=head1 ExSite::Form - ExSite form management tools

The Form class provides a variety of tools for composing HTML forms,
reading and parsing form input, and automatically performing standard
database operations (updates, inserts, searches) on forms that 
correspond to database records.

There are two groups of methods in this class:  (1) form processing 
methods, and (2) form composition methods.

The form processing methods are mostly for processing forms that
represent database records;  if these forms are built using the 
C<input_record()> method (below), then they can be automatically
processed (read, parsed, validated, and submitted to the database)
by Form methods.  In other words, you do not need to write the 
script that accepts the form data, in simple cases.

For instance, you can create a form as follows:

C<my $form = new ExSite::Form;>

C<print $form->make({table=>"my_db_table",...});>

This same form can be processed as follows:

C<$form->do;>

Note that C<make()> will automatically set the method, action, and
encoding type, so that no form processing code need be written.

=cut

#======================================================================

use strict;
use MIME::Base64;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Input;
use ExSite::Report;

# inherit from Report (and therefore DB)
use vars qw(@ISA);
@ISA = qw(ExSite::Report);

#========================================================================
# form processing methods
# (page down this file for form construction methods)
#========================================================================
#   get_data : fetch raw input
#   get_action : determine what to do with the data
#   do : process the data
#   do_update : update records
#   do_insert : add new records
#   do_search : find records
#   parse : parse the input
#   validate : validate the data
#========================================================================

=pod

=head2 get_data() - fetch raw input

This method retrieves the form input and parses it (assuming it comes
from C<input_record>.  It returns the total length of the input data.

=cut 

sub get_data {
  my $this = shift;
  $this->{input} = new ExSite::Input;
  # fetch from POST data preferentially, QUERY_STRING data otherwise
  my $data = $this->{input}->post_or_query("data"); 
  if ($this->{handler}->{form_input}) {
    # special input handler
    &{$this->{handler}->{form_input}}($data);
  }
  if ($data->{submit_action}) {
    $this->{submit_action} = $data->{submit_action};
    delete $data->{submit_action};
  }
  #### this is redundant (and wasteful, too!)
  $this->{data} = $data;
  $this->{form} = &parse($this,%$data);
  return $ENV{CONTENT_LENGTH} || length $ENV{QUERY_STRING};
}

=pod

=head2 get_action() - what are we supposed to do with this data?

This method figures out what to do with the data, by either reading
the action field, or guessing based on the data.  (Forms that specify
primary keys are presumed to be updates;  otherwise they are presumed
to be inserts.)

Valid actions are "insert", "update", "search".

=cut 

sub get_action {
  my ($this) = shift;
  my $action;
  if ($this->{action}) {
    $action = $this->{action};
  }
  else {
    if (exists $this->{data}->{action}) {
      # this form explicitly declares its action
      $action = $this->{data}->{action};
      delete $this->{data}->{action};
    }
    else {
	# deprecated logic
      my @id = grep /:id$/, keys %{$this->{data}}; ###################
      if ($#id == 0 && $this->{data}->{$id[0]} > 0) {  ####################
   # there is a key specified; must be an update
   $action = "update";
      }
      else {
   # there is no key specified; must be an insert
   $action = "insert";
      }
    }
    $this->{action} = $action;
  }
  return $action;
}

=pod 

=head2 do() - perform the requested action

C<do()> executes the action suggested by C(get_action()>.

First, however, it checks the submit_action, to make sure the user 
didn't cancel the form.  It also guesses a reasonable reply page, or
uses the reply argument provided by the form.

For updates and inserts, do performs a validation check on the form data,
by calling C<validate()>.  Any validation problems will prevent the
action from proceeding.

=head2 do_db() - perform the do() tasks that directly touch the database

C<do_db()> skips the control-flow logic of form processing (dealing with
cancels, determining the reply page, reporting results), and directly 
invokes the db interaction methods.  This is handy if a non-generic control
program is handling high-level form processing.

=head2 do_update() - perform an automatic update from a form

In cases where C<do()> decides to update the database, it does so by
calling C<do_update()>.

=head2 do_insert() - perform an automatic insert from a form

In cases where C<do()> decides to insert a record into the database, 
it does so by calling C<do_insert()>.

=head2 do_search() - perform an automatic search of a table

In cases where C<do()> decides to search a table in the database, 
it does so by calling C<do_search()>.  Searches are done by looking 
for records that contain all of the provided values (ie. there is an
impled logical AND of the fields);  records are taken as matches if 
they contain the provided values as substrings.

=cut

sub do {
    my ($this,$action) = @_;
    $this->info("ExSite::Form::do called");
    my $reply;
    my $out;
    my $redirect = 1;
    if (!$this->{data}) {
	if (! &get_data($this)) {
	    $this->error("do form: Nothing to do!");
	}
    }
    if (!$action) { $action = &get_action($this); }
    $this->{action} = $action;
    
    # fetch the destination URL
    
    if (exists $this->{data}->{reply}) {
	$reply = $this->{data}->{reply};
	delete $this->{data}->{reply};
	if ($reply =~ /doform\.cgi/) {
	    # recursion is perilous
	    $reply = "";
	}
    }
    
    # cancel, if requested

    if ($this->{submit_action} eq "Cancel") {
	# bail
	if ($this->{handler}->{cancel_form}) {
	    my $out = &{$this->{handler}->{cancel_form}} ($this);
	    if ($out) {
		return $out;
	    }
	}
	if ($reply) {
	    print "Location: $reply\n\n";
	    return;
	}
	else {
	    return "<H3>Operation cancelled.</H3>\n";
	}
    }

    # do everything that touches the database

    $this->do_db($action);

    # report results

    if ($this->errorcheck) {
	# uh-oh!
	$out = "<H3>Problem:</H3>\n<p>";
	$out .= $this->show_diagnostics("error","html");
	$out .= "</p>\n<P><A HREF=javascript:history.back()>Go back to form.</A></P>\n";
    }
    elsif ($reply) {
	# success; return the user to the original page
	print "Location: $reply\n\n";
    }
    elsif ($this->{output}) {
	# success; return the status report to the user
	return $this->{output};
    }
    else {
	# success; generic confirmation message
	$out = "<H3>Done!</H3>";
    }
    return $out;
}

# do the db steps only - no control flow logic

sub do_db {
    my ($this,$action) = @_;
    $this->info("ExSite::Form::do_db called");
    if (!$this->{data}) {
	# in case we didn't come from do()
	if (! &get_data($this)) {
	    $this->error("do form: Nothing to do!");
	}
    }
    
    if ($action eq "search") {
	$this->error(&do_search($this));
    }
    else {
	my @err;
	$this->error(&validate($this));
	if (! $this->errorcheck) {
	    my $stat;
	    # loop over tables/records in record order
	    foreach my $tr (sort @{$this->{recordlist}}) {
		my ($record,$table) = split /:/,$tr;
		if ($record && $record !~ /^_/) {
		    # a meaningful record id - try to update
		    $stat = $this->update($table,$this->{form}->{$table}->{$record});
		}
		else {
		    # no meaningful record id - try to insert
		    my $rec = $this->{form}->{$table}->{$record};
		    # check for virtual keys
		    my @keys = keys %{$this->{virtual_keys}};
		    if (@keys > 0) {
			# check for virtual key reference in this record
			foreach my $col (keys %$rec) {
			    if ($rec->{$col} =~ /^_/) {
				# looks like a virtual key
				if ($this->{map}->get_column_attr($table,$col,"datatype") =~ /^key[:=]/) {
				    # smells like a virtual key
				    # -> substutite the actual key
				    $rec->{$col} = $this->{virtual_keys}{$rec->{$col}};
				}
			    }
			} 
		    }
		    my $id = $this->insert($table,$rec);
		    # check results
		    if (! $id) {
			$this->error("record not inserted");
		    }
		    else {
			# add new primary key to the data
			my $pkey = $this->get_key($table);
			$rec->{$pkey} = $id;
			$this->{virtual_keys}->{$record} = $id;
		    }
		}
	    }
	}
    }
}

sub do_update {
    my ($this) = shift;
    my @err = ();
    ##### Authorize the user to perform the requested action
    #push @err, &validate($this);
    if (@err > 0) { return @err; }
    my $stat;
    foreach my $table (keys %{$this->{form}}) {
	foreach my $record (keys %{$this->{form}->{$table}}) {
	    $stat = $this->update($table,$this->{form}->{$table}->{$record});
	    # check results
	    if ($stat) {
		push @err, "<b>ERROR:</b> $stat";
	    }
	}
    }
    return @err;
}

sub do_insert {
    my ($this) = shift;
    my @err = ();
    ##### Authorize the user to perform the requested action
    if (@err > 0) { return @err; }
    my $id;
    my @reclist = @{$this->{recordlist}};
    my %insert_id;
    while (@reclist > 0) {
	my $rec = shift @reclist;
	#my $data = shift @reclist;
	my ($record,$table) = split /:/,$rec;
	$id = $this->insert($table,$this->{form}{$table}{$record});
	# check results
	if (! $id) {
	    push @err, "<b>ERROR:</b> record not inserted";
	    push @err, $this->{query_error};
	}
	else {
	    # add new primary key to the data
	    my $pkey = $this->get_key($table);
	    $this->{form}->{$table}->{$record}->{$pkey} = $id;
	    $insert_id{$rec} = $id;
	}
    }
    return @err;
}

sub do_search {
    my ($this) = @_;
    my @err = ();
    # can search multiple tables in one call
    foreach my $table (keys %{$this->{form}}) {
	foreach my $record (values %{$this->{form}->{$table}}) {
	    foreach my $col (keys %$record) {
		# delete null-valued keys; add wildcards to the remainder
		#  foreach my $col (values %$key) {
		if ($record->{$col}) {
		    # escape literals
		    $record->{$col} =~ s/%/\\%/g;
		    $record->{$col} =~ s/_/\\_/g;
		    # add wildcards to non-id columns #############
		    if ($col !~ /id$/) {
			$record->{$col} =
			    "%".$record->{$col}."%";
		    }
		}
		else {
		    delete $record->{$col};
		}
	    }
	    #my %udata = &Authorize; ################??
	    if ($this->{handler}->{search}) {
		# special search handler
		$this->{output} = &{$this->{handler}->{search}}($this,$table,$record);
	    }
	    else {
		# don't print to stdout, since do_form doesn't know yet 
		# whether it is outputting a page or redirecting.  Also, 
		# we can't return the HTML, since the return value is for 
		# error messages.  So instead, we simply save the output 
		# in the object, and let "do" figure it out.
		$this->{output} = "<H3>Search Results</H3>\n";
		$this->{output} .= $this->report({table=>$table,data=>$record,link=>1,sort=>1});
	    }
	}
    }
    return @err;
}

=pod

=head2 parse(%data) - decode form data

Parse does more than the usual URL-decoding of form data.  It also:

=over 4

=item * determines which database table, record, and column each form input corresponds to

=item * groups data by record, so that multiple records can be entered into the same form

=item * combines multi-part input fields (eg. dates) into single data values

=item * mime-encodes file inputs, so they can be managed as text strings

=back

The parsed data is returned in a hash with the following structure:

C<$form{TABLE}->{ROW}->{COLUMN} = VALUE>

where "ROW" is the primary key value (record ID) of the row, or _<n> for 
undefined (ie. new) rows.

There are 2 methods for encoding the table/row information in the form: 
 
Method 1 places hidden, encrypted table/record markers into the form data.
The markers are named "R[n]__mark" where [n] is some integer to identify
the record.  All inputs beginning with "R[n]_[column]" are attached to that 
record.  This method is more secure because it is difficult to spoof invalid 
table and record IDs.  It also works better with Javascript.

Method 2 encodes the table and record directly into each input name:
C<table#record_id#column>.  The C<record_id> can be null, or "_" for a
simple insert, or "_[n]" for multiple distinct inserts.  This method has
some security concerns, since the table and record IDs could be tampered
with by an attacker.  Exsite does attempt to trap such tampering, but
there can be situations where normal security is bypassed or insufficient. 
The main advantage to Method 2 is that forms can easily be hand-coded.

=cut

sub parse {
    my ($this,%data) = @_;
    $this->info("parse: auto-parsing form data");
    my %form;
    my $sc = $config{form}->{sepchar};
    my @keys;
    if ($this->{input}) {
	# use original input order from form
	@keys = @{$this->{input}->post_or_query("keys")};
    }
    else {
	@keys = sort keys %data;
    }
    $this->{recordlist} = [];
    
    # break down inputs into actual column values
    
    foreach my $key (@keys) {  # loops over inputs in order
	my $val = $data{$key};
	#print "$key = $val<br>\n";
	my ($tab, $irec, $col, $part);
	if ($key =~ /^R(\d+)_mark$/) { 
	    # encrypted table/record marker
	    ($tab, $irec) = split(/$sc/,$this->decrypt($val));
	    next; 
	}
	elsif ($key =~ /^R(\d+)(_([a-z]+))?_(.+)$/) { 
	    # column
	    $part = $3;
	    $col = $4;
	}
	elsif ($key =~ /$sc/) { 
	    # cleartext table/rec/column encoding
	    ##### FIXME: test access permission to this method
	    ($tab, $irec, $col, $part) = split(/$sc/,$key);
	}
	else {
	    # skip non-table data
	    next; 
	}
	my %map = $this->{map}->get_column($tab,$col);
	my $rec = $irec;  # $irec is the input record id from form data
	# set dummy records to "_"
	if (! $rec) { $rec = "_"; }
	if (! exists $form{$tab}) {
	    # new table
	    $form{$tab} = {};
	}
	if (! exists $form{$tab}->{$rec}) {
	    # new record
	    $form{$tab}->{$rec} = {};
	    push @{$this->{recordlist}}, "$rec:$tab";
	}
	# pre-process certain inputs
	# 1) encrypted fields
	##### FIXME: this should probably be done at a lower level
	if ($map{datatype} eq "crypt") {
	    $val = $this->encrypt($val);
	}
	# 2) multi-part inputs
	if ($part) {
	    # compound value
	    if ($part =~ /^html$/i) {
		# must unmangle any CMS tags that were masked for the editor
		$form{$tab}->{$rec}->{$col} = &restore_cms_tags($data{$key});
	    }
	    elsif ($part eq "filename" || $part eq "clear") {
		# A file:  for regular URL-encoding, only the filename
		# will be passed, but for multipart encoding, the file contents
		# will also be passed.  Either way, the file is encoded in the
		# database as "filename#content_data".  The content_data, if
		# present, is Mime base-64 encoded, so that it can be managed
		# using regular text insert/update statements.  This will result
		# in a decoding/encoding performance hit, but we'll have to live
		# with that until someone thinks of a better universal SQL
		# binary data management method.
		next if (exists $form{$tab}->{$rec}->{$col});
		if (exists $data{"$tab$sc$irec$sc$col${sc}clear"}) {
		    # clear the field if the delete box is checked
		    $form{$tab}->{$rec}->{$col} = "";
		    next;
		}
		my $fdata = undef;
		my $size = 0;
		my $mimedata = "";
		if ($ENV{CONTENT_TYPE} =~ /multipart\/form-data/) {
		    # read file contents
		    my $continue = 1;
		    while ($continue) {
			$continue = read($val,$fdata,1024,$size);
			$size += $continue;
		    }
		    # MIME-encode the file contents
		    $mimedata = encode_base64($fdata,"");
		}
		# strip path from filename
		my @path = split /[:\\\/]/, $val;
		#print "=$path[-1]=<br>\n";
		my $filename;
		if ($path[-1]) { $filename = &url_escape($path[-1]); }
		# 
		# remove the column from the form if it is blank, or else
		# existing value will be cleared.  However, if the delete
		# box is checked, then explicitly clear the column.
		#
		#if ($data{"$tab$sc$irec$sc$col${sc}clear"} =~ /ON/i) {
		#    # clear the field if the delete box is checked
		#    $form{$tab}->{$rec}->{$col} = "";
		#}
		#elsif ($path[-1] && $mimedata) {
		if ($path[-1] && $mimedata) {
		    # treat the field as absent, if it is empty
		    $form{$tab}->{$rec}->{$col} = "$filename$sc$mimedata";
		}
		#$form{$tab}->{$rec}->{$col} = "$path[-1]$sc$mimedata";
	    }
	    elsif ($part =~ /key/) {
		# a foreign key reference
		# -> only requires special processing if it points to a 
		# virtual primary key
		if ($val =~ /^_/) {
		    # virtual primary key - not determined until preceeding
		    # records are actually inserted
		    if (! $this->{virtual_keys}) { 
			$this->{virtual_keys} = {}; 
		    }
		    # remember the key for later reference
		    $this->{virtual_keys}->{$val} = undef; 
		}
		$form{$tab}->{$rec}->{$col} .= $val;
	    }
	    elsif ($part =~ /_set_/) {
		# this is a set item; append the value to the values 
		# already accumulated for this column
		if ($part eq "_set_0") {
		    # dummy data item - represents no selected items.
		    # If there is a value, and no other set items selected,
		    # then we keep the value as a default.  If there is a 
		    # value, and other set items are checked, then we 
		    # ignore the default.  If there is no value, then we
		    # initialize the column as empty.
		    if ($val && ! exists $data{"$tab$sc$irec$sc$col${sc}_set_1"}) {
			# dummy value exists, and no other set data is present
			$form{$tab}->{$rec}->{$col} = $val;
		    }
		    elsif (! $val && ! defined $form{$tab}->{$rec}->{$col}) {
			# dummy value empty, and no other set data has been read
			$form{$tab}->{$rec}->{$col} = "";
		    }
		}
		else {
		    # set value selected; append it to the current value
		    $val =~ s/&\#34;/\"/g;
		    if ($form{$tab}->{$rec}->{$col}) {
			$form{$tab}->{$rec}->{$col} .= "; ";
		    }
		    $form{$tab}->{$rec}->{$col} .= $val;
		}
	    }
	    else {
		# this is a date or time (or datetime) value
		my ($year, $month, $day, $hour, $min);
		# 1) check that we haven't already processed the overall column
		if (! exists $form{$tab}->{$rec}->{$col}) {
		    my ($date);
		    # dates
		    if (defined $data{"$tab$sc$irec$sc$col${sc}year"}) {
			$year = $data{"$tab$sc$irec$sc$col${sc}year"};
			$month = $data{"$tab$sc$irec$sc$col${sc}month"};
			$day = $data{"$tab$sc$irec$sc$col${sc}day"};
			$date = join("-",$year,$month,$day);
#       if (($day > $month_days[$month]) || ($day < 1) || ($year < 1900)) {
#         push @{$this->{errors}},"$date is not a valid date";
#       }
		    }
		    # times
		    if (defined $data{"$tab$sc$irec$sc$col${sc}hour"}) {
			$hour = $data{"$tab$sc$irec$sc$col${sc}hour"};
			if ($data{"$tab$sc$irec$sc$col${sc}ampm"} eq "PM" && $hour < 12) {
			    $hour += 12;
			}
			elsif ($data{"$tab$sc$irec$sc$col${sc}ampm"} eq "AM" && $hour == 12) {
			    $hour = 0;
			}
			$min = $data{"$tab$sc$irec$sc$col${sc}min"};
			if ($date) { $date .= " "; }
			$date .= join(":",$hour,$min,"00");
		    }
		    $form{$tab}->{$rec}->{$col} = $date;
		}
	    }
	}
	else {
	    # simple input
	    $val =~ s/&\#34;/\"/g;
	    $form{$tab}->{$rec}->{$col} = $val;
	}
    }
    return \%form;
}

=head2 formpush() - push new data onto the end of the form

C<push()> is used to manually add new data items onto the form, 
after the form data has already been parsed.  This is typically used
by handlers and wizards to place extra data into the form after the
form has been submitted.

The new data is only added to the {form} attribute, not to the
raw form data, nor to the input object.  It is always added to the
end of the form.

sub formpush {
    my $this = shift;
    my $key = shift;
    my $val = shift;
    while ($key) {
	my $sc = $config{form}{sepchar};
	my ($tab, $irec, $col, $part) = split(/$sc/,$key);
	if ($part) {
	    # problem - we only push simple form values, no multipart values
	    $this->error("cannot push multipart form values");
	    return;
	}
	else {
	    my %map = $this->{map}->get_column($tab,$col);
	    my $rec = $irec;  # $irec is the input record id from form data
	    # set dummy records to "_"
	    if (! $rec) { $rec = "_"; }
	    if (! exists $this->{form}{$tab}) {
		# new table
		$this->{form}{$tab} = {};
	    }
	    if (! exists $this->{form}{$tab}{$rec}) {
		# new record
		$this->{form}{$tab}{$rec} = {};
		push @{$this->{recordlist}}, "$rec:$tab";
	    }
	    $this->{form}{$tab}{$rec}{$key} = $val;
	}
	$key = shift;
	$val = shift;
    }
    return;
}

=pod 

=head2 validate() - validate form data

C<validate()> returns an array of validation error messages; if this
array is empty, then the form was completely validated.

Validate operates on the output of C<parse()>.

=cut

sub validate_old {
    my ($this) = @_;
    $this->info("validate: auto-validating form data");
    my @errors = ();
    foreach my $table (keys %{$this->{form}}) {
	foreach my $row (keys %{$this->{form}->{$table}}) {
	    my $rowname = $row;
	    if ($rowname eq "_") { $rowname = "(new)"; }

	    # check ownership of this row, if it is an update
	    
	    if ($this->{action} eq "update") {
		if (! $this->user_owns($table,$row)) { 
		    push @errors, "$table $rowname: ownership mismatch - you do not have permission to update this record";
		}
	    }

	    # check patterns and data requirements
	    
	    my $ncol = $this->{map}->get_ncol($table);
	    for (my $icol=0; $icol<$ncol; $icol++) {
		my %column = $this->{map}->get_column($table,$icol);
		next if ($column{validate} eq "no");
		my $col = $column{column};
		if (exists $this->{form}->{$table}->{$row}->{$col}) {
		    # column is present
		    if ($this->{form}->{$table}->{$row}->{$col}) {
			# data is present - trim whitespace
			if ($config{form}->{trim_whitespace}) {
			    $this->{form}->{$table}->{$row}->{$col} =~
				s/^\s+//;
			    $this->{form}->{$table}->{$row}->{$col} =~
				s/\s+$//;
			}
			# check data against legal patterns
			my $datum = $this->{form}->{$table}->{$row}->{$col};
			my $datatype = $column{datatype} =~ /^key/ ?
			    "key" : $column{datatype};
			my $re = $this->{map}->regexp($datatype);
			if ($datatype eq "key") {
			    if ($datum !~ /^($re|_\d*)$/) {
				push @errors, "$table $rowname: \"$datum\" is not a valid value for \"$column{label}\"";
			    }
			}
			elsif ($datatype eq "fixed") {
			    if ($datum ne $column{default}) {
				push @errors, "$table $rowname: \"$datum\" is not a valid value for \"$column{label}\"";
			    }
			}
			elsif ($datum !~ /^$re$/) {
			    # bad pattern
			    push @errors, "$table $rowname: \"$datum\" is not a valid value for \"$column{label}\"";
			}
			# check upload file sizes
			if ($column{datatype} =~ /file/) {
			    if ($column{datatype} eq "file" && length $datum > 65535) {
				push @errors, "$table $rowname: $column{label} is too large (48K max)";
			    }
			    elsif ($column{datatype} eq "bigfile" && length $datum > $config{form}{bigfile_maxsize} * 4 / 3 ) {
				my $fmax = int($config{form}{bigfile_maxsize} / 1024);
				push @errors, "$table $rowname: $column{label} is too large ($fmax kb max)";
			    }
			}
			# check ownership of foreign keys
			if ($column{datatype} =~ /^key[=:]/) {
			    if ($this->authorize < $config{auth}->{foreign_link_level}) {
				my (undef,$ftable) = split /[:=]/,$column{datatype};
				if (! $this->user_owns($ftable,$datum)) { 
				    push @errors, "$table $rowname: ownership mismatch - you do not have permission to link to $ftable $datum";
				}
			    }
			}
		    }
		    else {
			# column left blank
			next if ($column{validate} =~ /yes/);
			### next if ($column{validate} eq "soft" && $admin);
			if ($column{validate} eq "soft") {
			    ### 
			    push @errors,"$table $rowname: you must provide a value for \"$column{label}\"";
			}
			elsif ($column{validate} eq "hard") {
			    # KLUDGE: serial primary keys not required for inserts
			    if ($column{datatype} ne "key\#" || $this->{action} ne "insert") {
				push @errors,"$table $rowname: no $column{label} defined";
			    }
			}
		    }
		}
		else {
		    # column missing entirely from form
		    #if ($this->{action} eq "insert") {
		    if ($this->{action} eq "insert" && $column{datatype} ne "key\#") {
			# missing columns could be fatal on inserts;  we ignore this
			# exception for updates, since the missing data may already
			# be in the table.  Note that automatic keys (type key#)
			# can be ignored on inserts, even with a hard requirement.
			if ($column{validate} eq "hard") {
			    # data is required for inserts
			    push @errors,"FATAL: $table $rowname: form is missing required field \"$column{label}\"";
			}
			#elsif ($column{validate} eq "soft" && !$admin) {
			elsif ($column{validate} eq "soft") {
			    # data is required for inserts in non-admin mode
			    push @errors,"FATAL: $table $rowname: form is missing required field \"$column{label}\"";
			}
		    }
		}
	    }
	    
	    #
	    # custom validation handler
	    #

	    my $stat = $this->run_handler("validate",$this->{action},$table,$row,$this->{form}->{$table}->{$row});
	    if (defined $stat) {
		push @errors, $stat;
	    }
	}
    }
    if (@errors > 0) {
	$this->warn(join("; ",@errors));
    }
    return @errors;
}

sub validate {
    my ($this) = @_;
    $this->info("validate: auto-validating form data");
    my @errors = ();
    foreach my $table (keys %{$this->{form}}) {
	foreach my $row (keys %{$this->{form}->{$table}}) {
	    my $rowname = $row;
	    if ($rowname eq "_") { $rowname = "(new)"; }

	    # check ownership of this row, if it is an update
	    
	    if ($this->{action} eq "update") {
		if (! $this->user_owns($table,$row)) { 
		    push @errors, "$table $rowname: ownership mismatch - you do not have permission to update this record";
		}
	    }

	    # check patterns and data requirements
	    
	    my $ncol = $this->{map}->get_ncol($table);
	    for (my $icol=0; $icol<$ncol; $icol++) {
		my %column = $this->{map}->get_column($table,$icol);
		next if ($column{validate} eq "no");
		my $col = $column{column};
		if (exists $this->{form}->{$table}->{$row}->{$col}) {
		    # column is present
		    if ($this->{form}->{$table}->{$row}->{$col}) {
			# data is present - trim whitespace
			if ($config{form}->{trim_whitespace}) {
			    $this->{form}->{$table}->{$row}->{$col} =~
				s/^\s+//;
			    $this->{form}->{$table}->{$row}->{$col} =~
				s/\s+$//;
			}
			# check data against legal patterns
			my $datum = $this->{form}->{$table}->{$row}->{$col};
			my $datatype = $column{datatype} =~ /^key/ ?
			    "key" : $column{datatype};
			my $re = $this->{map}->regexp($datatype);
			if ($datatype eq "key") {
			    if ($datum !~ /^($re|_\d*)$/) {
				push @errors, "$table $rowname: \"$datum\" is not a valid value for \"$column{label}\"";
			    }
			}
			elsif ($datatype eq "fixed") {
			    if ($datum ne $column{default}) {
				push @errors, "$table $rowname: \"$datum\" is not a valid value for \"$column{label}\"";
			    }
			}
			elsif ($datum !~ /^$re$/) {
			    # bad pattern
			    push @errors, "$table $rowname: \"$datum\" is not a valid value for \"$column{label}\"";
			}
			# check upload file sizes
			if ($column{datatype} =~ /file/) {
			    if ($column{datatype} eq "file" && length $datum > 65535) {
				push @errors, "$table $rowname: $column{label} is too large (48K max)";
			    }
			    elsif ($column{datatype} eq "bigfile" && length $datum > $config{form}{bigfile_maxsize} * 4 / 3 ) {
				my $fmax = int($config{form}{bigfile_maxsize} / 1024);
				push @errors, "$table $rowname: $column{label} is too large ($fmax kb max)";
			    }
			}
			# check ownership of foreign keys
			if ($column{datatype} =~ /^key[=:]/) {
			    if ($this->authorize < $config{auth}->{foreign_link_level}) {
				my (undef,$ftable) = split /[:=]/,$column{datatype};
				if (! $this->user_owns($ftable,$datum)) { 
				    push @errors, "$table $rowname: ownership mismatch - you do not have permission to link to $ftable $datum";
				}
			    }
			}
		    }
		    else {
			# column left blank
			next if ($column{validate} =~ /yes/);
			### next if ($column{validate} eq "soft" && $admin);
			if ($column{validate} eq "soft") {
			    ### 
			    push @errors,"$table $rowname: you must provide a value for \"$column{label}\"";
			}
			elsif ($column{validate} eq "hard") {
			    # KLUDGE: serial primary keys not required for inserts
			    if ($column{datatype} ne "key\#" || $this->{action} ne "insert") {
				push @errors,"$table $rowname: no $column{label} defined";
			    }
			}
		    }
		}
		else {
		    # column missing entirely from form
		    #if ($this->{action} eq "insert") {
		    if ($this->{action} eq "insert" && $column{datatype} ne "key\#") {
			# missing columns could be fatal on inserts;  we ignore this
			# exception for updates, since the missing data may already
			# be in the table.  Note that automatic keys (type key#)
			# can be ignored on inserts, even with a hard requirement.
			if ($column{validate} eq "hard") {
			    # data is required for inserts
			    push @errors,"FATAL: $table $rowname: form is missing required field \"$column{label}\"";
			}
			#elsif ($column{validate} eq "soft" && !$admin) {
			elsif ($column{validate} eq "soft") {
			    # data is required for inserts in non-admin mode
			    push @errors,"FATAL: $table $rowname: form is missing required field \"$column{label}\"";
			}
		    }
		}
	    }
	    
	    #
	    # custom validation handler
	    #

	    my $stat = $this->run_handler("validate",$this->{action},$table,$row,$this->{form}->{$table}->{$row});
	    if (defined $stat) {
		push @errors, $stat;
	    }
	}
    }
    if (@errors > 0) {
	$this->warn(join("; ",@errors));
    }
    return @errors;
}

sub validate_column {
    my ($this,$table,$row,$column,$datum) = @_;
    my $rowname = $row;
    if ($rowname =~ /^_/) { $rowname = "(new)"; }
    my %column = $this->{map}->get_column($table,$column);
    if ($datum) {
	# we have data - check its validity
	my $stat = $this->validate_datatype($column{datatype},$datum);
	if ($stat eq "invalid data/format") {
	    # validate datatype returns an unfriendly error message
	    return "$table $rowname: \"$datum\" is not a valid value for \"$column{label}\"";
	}
	elsif ($stat) {
	    return "$table $rowname: $stat";
	}
	return undef;
    }
    else {
	# no data...
	if ($column{datatype} eq "key\#" && $this->{action} eq "insert") {
	    # serial number primary keys not required on inserts
	    return undef;
	}
	next if ($column{validate} =~ /yes/);  # data is optional
	if ($column{validate} eq "soft") {
	    return "$table $rowname: you must provide a value for \"$column{label}\"";
	}
	elsif ($column{validate} eq "hard") {
	    return "$table $rowname: no $column{label} defined";
	}
    }
}

sub validate_datatype {
    my ($this,$type,$datum) = @_;
    
    my $stat = $this->run_handler("validate_datatype",$type,$datum);
    return $stat if (defined $stat);
	
    # check data against legal patterns
    my $datatype = $type =~ /^key/ ? "key" : $type;
    my $re = $this->{map}->regexp($datatype);
    if ($datatype eq "key") {
	if ($datum !~ /^($re|_\d*)$/) {
	    return "invalid foreign key reference";
	}
    }
    elsif ($datum !~ /^$re$/) {
	return "invalid data/format";
    }

    # check upload file sizes
    if ($datatype =~ /file/) {
	if ($datatype eq "file" && length $datum > 65535) {
	    return "uploaded file is too large (48K max)";
	}
	elsif ($datatype eq "bigfile" && 
	       length $datum > $config{form}{bigfile_maxsize} * 4 / 3 ) {
	    my $fmax = int($config{form}{bigfile_maxsize} / 1024);
	    return "uploaded file is too large ($fmax kb max)";
	}
    }

    # check ownership of foreign keys
    if ($type =~ /^key[=:]/) {
	if ($this->authorize < $config{auth}->{foreign_link_level}) {
	    my (undef,$ftable) = split /[:=]/,$type;
	    if (! $this->user_owns($ftable,$datum)) { 
		return "ownership mismatch - you do not have permission to link to $ftable $datum";
	    }
	}
    }
    return undef;
}

# delete_record : remove a record from the parsed form data so that it
# won't be processed.

sub delete_record {
    my ($this,$table,$rec) = @_;
    if (exists $this->{form}{$table}{$rec}) {
	$this->info("delete_record: removing $table $rec from form data");
	delete $this->{form}{$table}{$rec};
	for (my $irec = 0; $irec < scalar @{$this->{recordlist}}; $irec++) {
	    my $r = $this->{recordlist}[$irec];
	    my ($rname,$tname) = split(/:/,$r);
	    if ($tname eq $table && $rname eq $rec) {
                splice(@{$this->{recordlist}},$irec,1);        
		return 1;
	    }
	}
    }
    return 0;  # nothing removed
}

#========================================================================
# form construction methods
#========================================================================
# form construction:
#   make : generic form construction method
#   edit : make a form to edit an existing record
#   append : make a form to add a new record
#   query : make a form to perform a search
# input tag construction:
#   input_record : creates all inputs for entering 1 record
#   input_data : generic input
#   input_foreign_key : record selector
#   input_date : enter a date
#   input_time : enter a time
#   input_datetime : enter a date-time pair
#   input_list : select one item from a list of options
#   input_set : select multiple items from a set of options
#========================================================================

#------------------------------------------------------------------------
# make : create a new form
#
# arguments:
#   $rtab = reference to a list of tables
#   $rrec = reference to a list of records
#   $rhide = reference to a list of fields to NOT display
#   $rshow = reference to a list of fields to be displayed
#   $rextra = extra values to add to the form (as hidden inputs)
#   $search = if true, this is a search form
#   $action = form action (doform.cgi by default)
#   $method = form method (POST by default)
#   $group = column group(s) to show
#
# A form displays a set of database records for editing.  You need
# to indicate which records are to be edited, using two arrayrefs:
#
# [ table1, table2, table3, ...]
# [ record1, record2, record3, ...]
#
# Form fields will be generated for each such record, and populated
# with the contents of that record.  If the record IDs are FALSE or
# null, the fields will be blank, and the record will be inserted.  If the
# IDs are defined and valid, those records will be updated.  If the
# keys are defined, but invalid, new records with those keys will be
# inserted.
#
# A single submit button will process all the records.  If you prefer
# to have individual submit buttons for each record or set of records,
# then invoke &make separately for each.
#------------------------------------------------------------------------

sub make {
  my ($this) = shift;
  $this->info("ExSite::Form::make: called");
  my ($rtab,$rrec,$rhide,$rshow,$rextra,$search,$action,$method,$group,$rdata);
  my (@tables,@records);
  if (ref $_[0] eq "ARRAY") {
      # legacy method of argument passing
      ($rtab,$rrec,$rhide,$rshow,$rextra,$search,$action,$method) = @_;
      @tables = @$rtab;
      @records = @$rrec;
  }
  else {
      # preferred method of argument passing
      my %opt = @_;
      if (ref $opt{table} eq "ARRAY") { @tables = @{$opt{table}}; }
      else { @tables = ( $opt{table} ); }
      if (ref $opt{record} eq "ARRAY") { @records = @{$opt{record}}; }
      else { @records = ( $opt{record} ); }
      $rhide = $opt{hide};
      $rshow = $opt{show};
      $rextra = $opt{extra};
      $search = $opt{search};
      $action = $opt{action};
      $method = $opt{method};
      #
      # added feb 12 by ACD
      #
      $rrec = $opt{data};
      #
      #
      #
      $group = $opt{group};
      $rdata = $opt{data};
      if ($opt{debug}) {
	  $action = "$config{server}->{CGIpath}/debug.cgi";
      }
  }
  my $out;
  if ($config{form}->{style} eq "table") {
      $out .= "<TABLE CLASS=\"form\">\n";
  }
  else {
      $out .= "<SPAN CLASS=\"form\">\n";
  }
  my $nrec = 0;
  while (@tables > 0) {
    $nrec++;
    my $table = shift @tables;
    my $record = shift @records;
    if (! $record) { $record = "_$nrec"; }
           #
           #  feb 12 by ACD
           #  
           # the changes made below was:
           # i added "data=>$rrec" to the hash in the input_records below
           # in order to pass parent ids for new records
           #
           #

    if ($search) {
      # explicit search form
      $this->{action} = "search";
      #$out .= &input_record($this,$table,undef,$rhide,$rshow);
      $out .= $this->input_record({table=>$table,hide=>$rhide,show=>$rshow,group=>$group,data=>$rdata});
    }
    elsif (! $record || $record =~ /^_/) {
      # no record info; this is a completely blank insert
      $this->{action} = "insert";
      #$out .= &input_record($this,$table,undef,$rhide,$rshow);
      
      $out .= $this->input_record({data=>$rrec,table=>$table,hide=>$rhide,show=>$rshow,group=>$group});
    }
    elsif ($rdata) {
      # record is a reference to a datahash containing defaults and/or
      # foreign key references
      $this->{action} = "insert";
      #$out .= &input_record($this,$table,$record,$rhide,$rshow);
      $out .= $this->input_record({table=>$table,data=>$record,
				   hide=>$rhide,show=>$rshow,group=>$group});
    }
    else {
      # record is a key of an existing record
      my %data = $this->fetch($table,$record);
      $this->{action} = "update";
      #$out .= &input_record($this,$table,\%data,$rhide,$rshow);
      $out .= $this->input_record({table=>$table,data=>\%data,
				   hide=>$rhide,show=>$rshow,group=>$group});
    }
  }
  if ($config{form}->{style} eq "table") {
      $out .= "</TABLE>\n";
  }
  else {
      $out .= "</SPAN>\n";
  }
  foreach my $key (keys %$rextra) {
    $out .= "<INPUT TYPE=hidden NAME=$key VALUE=\"$rextra->{$key}\">\n";
  }
  if (!exists $rextra->{action}){
    $out .= "<INPUT TYPE=HIDDEN NAME=action VALUE=\"$this->{action}\">\n";
  }
  $out .= &make_form_buttons;
  $out .= "</p>\n</FORM>\n";

  # we add the opening form tab at the end, so that we can select the
  # encoding type based on the form content...

  # (NB: there are 2 actions: $action = the <FORM> action;
  # $this->{action} = the WebDB action)
  if (!$action) { 
      if ($config{form}->{mode} eq "cgi") {
	  $action = "$config{server}->{CGIpath}/doform.cgi"; 
	  #$action = "$config{server}->{CGIpath}/debug.cgi"; 
      }
      else { # exsite mode
	  # redundancy here: WebDB action passed in QUERY and in POST
	  my $pid = $share{Page}->id;
	  $action = "$config{server}->{CGIpath}/page.cgi?_id=$pid";
	  #$action = "debug.cgi";
	  # for exsite-mode forms, add the page parameters to the form
	  #$rextra->{_id} = $share{Page}->{page_id};
      }
  }
  if (!$method) { $method = "POST"; }
  # direct the form to an appropriate debug script if the debug option is set
  ##### warning: flag case where METHOD=GET and ENCTYPE=multipart
  my $formtag = "<FORM METHOD=$method";
  if (defined $this->{enctype}) {
      $formtag .= " ENCTYPE=\"$this->{enctype}\"";
  }
  if ($action) {
      $formtag .= " ACTION=$action";
  }
  if ($this->{htmleditor})
  {
      $formtag .= " onSubmit=\"validate()\"";
      $formtag = $this->editor_javascript().$formtag;
  }
  $formtag .= ">\n";
  $formtag .= $this->{htmleditor} ?
  	"<input type=\"hidden\" name=\"$this->{htmleditor}\">\n" : "";
  $out = $formtag.$out;

  # handlers can massage the form HTML if necessary...
  if ($this->{handler}->{form_output}) {
    &{$this->{handler}->{form_output}}(\$out);
  }

  return $out;
}

#------------------------------------------------------------------------
# make_form_buttons : separated here so that it can be used as a 
# convenience function for anyone generating their own forms.
#------------------------------------------------------------------------

#### should Cancel button do a javascript:history.back()?

sub make_form_buttons {
  return "<INPUT TYPE=submit NAME=submit_action VALUE=\"Submit\">&nbsp;&nbsp;\n<INPUT TYPE=submit NAME=submit_action VALUE=\"Cancel\">&nbsp;&nbsp;\n<INPUT TYPE=reset>&nbsp;&nbsp;\n";
}

# simplified wrappers for &make - deprecated

#------------------------------------------------------------------------
# edit : edit a particular record
#------------------------------------------------------------------------

sub edit {
  my ($this,$table,$record,$rhide,$rshow,$extra) = @_;
  return &make($this,[$table],[$record],$rhide,$rshow,$extra);
}

#------------------------------------------------------------------------
# append : adds one new record to a table
#------------------------------------------------------------------------

sub append {
  my ($this,$table,$record,$rhide,$rshow,$extra) = @_;
  if ($this->{handler}->{append_data}) {
      $record = &{$this->{handler}->{append_data}}($record);
  }
  return &make($this,[$table],[$record],$rhide,$rshow,$extra);
}

#------------------------------------------------------------------------
# query : collects a data pattern to search for
#------------------------------------------------------------------------

sub search_query {
  my ($this,$table,$rhide,$rshow,$extra) = @_;
  return &make($this,[$table],[0],$rhide,$rshow,$extra,1);
}

#--------------------------------------------------------------------
# input_record : loop over the columns of a table, generating an
#                appropriate input tag for each
#--------------------------------------------------------------------

sub input_record {
    my $this = shift;
    my ($opt,$table,$record,$rdata,$rhide,$rshow,$group);
    if (@_ == 1) {
	$opt = shift;
	$table = $opt->{table};
	$record= $opt->{record};
	$rdata = $opt->{data};
	$rhide = $opt->{hide};
	$rshow = $opt->{show};
	$group = $opt->{group};
    }
    else {
	# deprecated calling convention
	($table,$rdata,$rhide,$rshow) = @_;
    }
    # check that the user is allowed to edit/insert this record
    if (! $this->approve($this->{action},$table,$rdata)) {
	return "Permission denied.";
    }
    my ($out, $sc, $map, $typemap);
    $sc = $config{form}->{sepchar};
    $map = $this->{map}->get_map($table);
    if (!defined $map) {
	$out = "ERROR: unknown table $table";
	return $out;  ####### log error!
    }
    my $authlevel = $this->authorize;

    my (%data,@hide,@show);
    if (defined $rdata) { %data = %$rdata; }
    if (defined $rhide) { @hide = @$rhide; }
    if (defined $rshow) { @show = @$rshow; }
    
    my $pkey = $this->get_key($table);
    if (! $record) { $record = $data{$pkey}; }
    
    my ($write, %column);
    my $ncols = $this->{map}->get_ncol($table);
    
    for (my $icolumn = 0; $icolumn < $ncols; $icolumn++) {
        %column = $this->{map}->get_column($table,$icolumn);
	if (! $column{label}) { $column{label} = $column{column}; }

	if ($this->{action} eq "search" && $column{search} eq "no"){next;}

	$write = ($column{write} <= $authlevel);

	# is this row displayed by default?
	#$write = $write && ($column{display} ne "no");
	# options to vary default logic

	# select which columns to show based on show and hide lists
	my $showflag;
	my $hideflag;
	if ($#show >= 0) {
	    $showflag = scalar grep(/^$column{column}$/,@show);
	    $write = $write && $showflag;
	}
	if ($#hide >= 0 || $column{datatype} eq "fixed") {
	    $hideflag = scalar grep(/^$column{column}$/,@hide) ||
		$column{datatype} eq "fixed";
	    $write = $write && ! $hideflag;
	}

	# select columns based on column groups
	if ($group && $write) {
	    if ($column{group} !~ /$group/) { $write = 0 }
	}
	
	# never put timestamps in forms
	if ($column{datatype} eq "timestamp") { next; }

	# set data value for fixed fields
	if ($column{datatype} eq "fixed") { 
	    $data{$column{column}} = $column{default};
	}

	if (! $write) {
	    # don't put hidden passwords in forms	    
	    if ($authlevel != $config{auth}->{root_auth_level} &&
		$column{datatype} eq "password") { next; }
	    # show readonly fields...
	    if ($config{form}->{show_readonly} &&  # if requested, and
		$column{read} <= $authlevel &&     # read is permitted, and
		(($#show >= 0 && $showflag) ||     # the field is explicitly
		! $hideflag)) {                    # shown or not hidden
		$column{datatype} = "readonly";
		$write = 1;
	    }
	}

	my $go;  #build input for this field?
	if ($this->{action} eq "search") {
	    $go = $authlevel >= $column{read} && 
		$authlevel >= $this->{map}->get_table_attr($table,'search') &&
		$column{search} ne "no";
		#$column{search} =~ /yes|general/;
	}
	else {
	    # insert/update
	    $go = $write || $authlevel == $config{auth}->{root_auth_level};
	}

	if ($go) {
	    $out .= $config{form}{fieldtag}{$config{form}{style}}[0];
	    # emphasize required fields
	    if ($column{validate} =~ /soft|hard/ && $this->{action} ne "search") {
		$out .= "$config{form}{labeltag}{$config{form}{reqd_field_style}}[0]$column{label}$config{form}{labeltag}{$config{form}{reqd_field_style}}[1]";
	    }
	    else {
		$out .= "<SPAN CLASS=\"label\">$column{label}</SPAN>";
	    }
	    if ($column{help}) {
		# this column has help text
		$out .= "<SPAN CLASS=\"help\"> - $column{help}</SPAN>";
	    }
	    $out .= $config{form}->{fieldtag}->{$config{form}->{style}}->[1];
	    if ($this->{handler}->{input}) {
		my $in =
		    &{$this->{handler}->{input}} ($table,$column{column},
						  $data{column},\%column);
		if ($in) {
		    $out .= $in;
		    next;
		}
	    }
	    $out .= $this->input_column(
					table=>$table, 
					column=>$column{column},
					size=>$column{size},
					record=>$record, 
					datatype=>$column{datatype},
					value=>$data{$column{column}},
					);
	    # help!
            # this should work, but doesn't because CGIroot is not set properly
	    #my $helpfile = "$config{server}->{CGIroot}$config{server}->{CGIpath}/$config{server}->{map}->{name}/help/$table/$column{column}";
	    # this is a kludge that works in simple cases
	    #my $helpfile = $ENV{SCRIPT_FILENAME};
	    #$helpfile =~ s/\/\w+\.cgi$//;
	    #$helpfile .= "/$config{server}->{map}->{name}/help/$table/$column{column}";
	    #if (-e $helpfile) {
	    #my $helptext = &get_file($helpfile);
	    #$helptext =~ s/\"/&quot;/g;
	    #$out .= $this->popup("$config{server}->{CGIpath}/help.cgi/$config{server}->{map}->{name}/help/$table/$column{column}","<IMG SRC=$config{server}->{HTMLpath}$config{server}->{imgdir}/qmark.gif WIDTH=20 HEIGHT=20 BORDER=0 TITLE=\"click for help\" ALIGN=middle>",400,200);
	    #}
	    
	    $out .= $config{form}->{fieldtag}->{$config{form}->{style}}->[2]."\n";
	}
	elsif ($data{$column{column}}) {
	    # Only include hidden columns if they contain meaningful data.
	    # But, do not resubmit existing data on updates, unless it's 
	    # the primary key.
	    if ($this->{action} ne "update" || $column{column} eq $pkey) {
		$data{$column{column}} =~ s/\"/&\#34;/g;
		my $name = "${table}${sc}${record}${sc}$column{column}";
		$out .= $this->input_html(
					  type=>"hidden",
					  name=>$name,
					  value=>$data{$column{column}},
					  );
	    }
	}
    }
    return $out;
}

#--------------------------------------------------------------------
# input_column : generate input tags appropriate to a database column
#
# looks up the column's ExSite datatype in the DBmap, and generates
# an input field name giving the table, record, and column name. 
#
# options:
# table => the table this column comes from
# column => the column name itself
# size => the data size of the column
# record => the id of the record being edited (if editing existing data)
# value => a default value to preset the field at
# datatype => datatype to use, if different from the map
# data => a datahash from which record and value can be inferred
#
# NB: no security checks are made on input_column;  if using this
# routine to generate custom forms, you must decide for yourself
# whether the user is allowed to alter the column in question.
#--------------------------------------------------------------------

sub input_column {
    my ($this,%opt) = @_;

    # fetch the datatype, if required
    if (! $opt{datatype}) {
	$opt{datatype} = $this->{map}->get_column_attr($opt{table},$opt{column},"datatype");
    }
    if (! $opt{size}) {
	$opt{size} = $this->{map}->get_column_attr($opt{table},$opt{column},"size");
    }

    # custom handler
    my $out = $this->run_handler("input_column",%opt);
    return $out if ($out);

    if ($opt{data}) {
	if (! $opt{record}) {
	    $opt{record} = $opt{data}->{$this->get_key($opt{table})};
	}
	if (! $opt{value}) {
	    $opt{value} = $opt{data}->{$opt{column}};
	}
    }

    # escape any quotes in data
    $opt{value} =~ s/\"/&\#34;/g;

    my $sc = $config{form}{sepchar};
    my $name = "$opt{table}${sc}$opt{record}${sc}$opt{column}";

    # generate the actual input field

    my $out = $this->input_exsite(name=>$name, size=>$opt{size}, 
				  datatype=>$opt{datatype}, 
				  value=>$opt{value},
				  fmatch=>$opt{fmatch},
				  table=>$opt{table},
				  column=>$opt{column},
				  record=>$opt{record});

    # help!
    my $helpfile = "$config{server}{CGIroot}/$config{server}{map}{name}/help/$opt{table}/$opt{column}";
    # alternate kludge that works in simple cases
    #my $helpfile = $ENV{SCRIPT_FILENAME};
    #$helpfile =~ s/\/\w+\.cgi$//;
    #$helpfile .= "/$config{server}->{map}->{name}/help/$opt{table}/$opt{column}";
    if (-e $helpfile) {
	$out .= $this->popup("$config{server}->{CGIpath}/help.cgi/$config{server}->{map}->{name}/help/$opt{table}/$opt{column}","<IMG SRC=$config{server}->{HTMLpath}/_ExSite/images/qmark.gif WIDTH=20 HEIGHT=20 BORDER=0 TITLE=\"click for help\" ALIGN=middle>",400,200);
    }
    return $out;
}

#-------------------------------------------------------------------
# use_wysiwyg : check the client's browser for wysiwyg editor compatability
#-------------------------------------------------------------------
sub use_wysiwyg
{
	my $uagent = $ENV{'HTTP_USER_AGENT'};	
	if ($uagent =~ /opera/i){return 0;}
	elsif ($uagent =~ /msie/i)
	{
		$uagent =~ /msie (\d*\.\d*)/i;
		my $ver = $1;
		if ($ver >= 6){return 1;}
	}
	elsif ($uagent =~ /netscape/i)
	{
		$uagent =~ /netscape\/(\d*\.\d*)/i;
		my $ver = $1;
		if ($ver >= 7.1){return 1;}		
	}
	elsif ($uagent =~ /gecko/i)
	{
		$uagent =~ /gecko\/(\d*)/i;
		my $date = $1;
		if ($date >= 20030126){return 1;}
	}	
	return 0;
}

#--------------------------------------------------------------------
# input_exsite : generate an HTML INPUT tag for an ExSite data item
#
# Usage is the same as input_html, except that an ExSite datatype is 
# accepted in place of a tag type.
#
# Note that some ExSite inputs are multi-part;  in those cases,
# this method returns multiple inputs, renamed appropriately.
# The inputs are named as NAME#PART, where NAME is the user-provided
# name for the input, and PART is the specific part of the input
# that particular input accepts (eg. year, month, day for dates).
# The form action is responsible for decoding this.
#--------------------------------------------------------------------
# general input options:
#   datatype => the ExSite datatype of the input
#   name => the name of the input
#   value => the default value of the input
# datatype-specific input options:
# date/datetime:
#   null_date_type => "today" or "null"
# time/datetime:
#   minstep => minutes increment for minutes option menu
# set:
#   dummy => value to use for the dummy set indicator
# foreign keys:
#   fmatch => match hash to select foreign keys to select from
#   table => table this form is for
#   column => the column in this table containing the foreign reference
#   record => the id of the record containing the data being modified
#   ftable => foreign table being referred to
# encrypted fields:
#   table => table this form is for
#   column => the column being encrypted
#   record => the id of the record containing the encrypted
#--------------------------------------------------------------------

sub input_exsite {
    my ($this,%opt) = @_;
    my $out = $this->run_handler("input_exsite",%opt);
    return $out if ($out);

    my $sc = $config{form}{sepchar};
    if ($opt{datatype}) {
	if (ref $opt{datatype}) {
	    $_ = $opt{datatype}->{type};
	}
	elsif ($opt{datatype} =~ /^key[:=]/) {
	    $_ = $opt{datatype};
	    $opt{datatype} = $this->{map}->get_datatype("key");
	}
	else {
	    $_ = $opt{datatype};
	    $opt{datatype} = $this->{map}->get_datatype($_);
	}
    }
    else {
	$this->error("ExSite::Form::input_exsite: no datatype specified");
	return "<SPAN CLASS=error>ERROR: cannot make input: no datatype specified</SPAN>";
    }

    # set &input_html options
    my %htmlopt = %opt;
    my @opt;
    if ((/^list:/ && $opt{datatype}->{tag} eq "select") ||
	(/^set:/ && $this->{action} eq "search")) {
	push @opt, ["", "== select =="];
	foreach my $opt (split /\|/,$opt{datatype}->{regexp}) {
	    push @opt, [ $opt, $opt ];
	}
	$htmlopt{options} = \@opt;
    }
    elsif (/^(list|set):/) {
        my @opt = split /\|/,$opt{datatype}->{regexp};
        $htmlopt{options} = \@opt;
    }
    $htmlopt{type} = $opt{datatype}->{tag};

    # new generate the inputs

    if (/^readonly$/) {
	return $opt{value};
    }
    if (/^html$/ && $this->use_wysiwyg)
    {
	if ($this->{action} !~ /search/){
	    return $this->input_htmleditor(%opt);
	}
    }
    elsif (/^(f?text)$/ || (/^html$/)) {
	if ($htmlopt{size} > $config{form}->{max_cols} ||
	    $this->{action} !~ /search/) {
	    # use textarea for large inputs, unless searching
	    $htmlopt{type} = "textarea";
	}
	else {
	    $htmlopt{type} = "text";
	}
	return $this->input_html(%htmlopt);
    }
    elsif (/^crypt$/) {
	# similar to text, but decrypt the data first
	# abort if we're not allowed to view the data
	if ($htmlopt{value}) {
	    my $data = $this->decrypt_column($htmlopt{value},
					     $htmlopt{table},
					     $htmlopt{record},
					     $htmlopt{column});
	    if ($this->errorcheck("warn")) {
		return "<span class=error>This field contains encrypted data, which you do not have permission to decrypt and edit.</span>";
	    }
	    $htmlopt{value} = $data;
	}
	# use text logic
	if ($htmlopt{size} > $config{form}->{max_cols} ||
	    $this->{action} !~ /search/) {
	    # use textarea for large inputs, unless searching
	    $htmlopt{type} = "textarea";
	}
	else {
	    $htmlopt{type} = "text";
	}
	return $this->input_html(%htmlopt);
    }
    elsif (/^list:/ || (($this->{action} eq "search") && /^set:/)) {
	my $value = $htmlopt{value};
	if ($htmlopt{type} eq "radio") {
	    foreach my $item (@{$htmlopt{options}}) { 
		$htmlopt{value} = $item;
		$htmlopt{selected} = ($item eq $value);
		$out .= $this->input_html(%htmlopt);
	    }
	}
	else {
	    $out .= $this->input_html(%htmlopt);
	}
    }
    elsif (/^set:/) {
	# multipart input
	$opt{type} = "checkbox";
	my @values = split /; /,$opt{value};
	my $n = 0;  # dummy set item differentiator
	# dummy set input - to ensure set is represented in form data,
	# even if all items are deselected
	my $name = $opt{name};
	$out = $this->input_html(type=>"hidden",
				  name=>"$name${sc}_set_$n",
				  value=>"$opt{dummy}");
	foreach my $item (@{$htmlopt{options}}) { 
	    $n++;
	    my %setopt = (type=>"checkbox", name=>"$htmlopt{name}${sc}_set_$n", value=>$item);
	    if (grep(/^$item$/,@values) > 0) { $setopt{selected} = 1; }
	    $out .= $this->input_html(%setopt);
	}
	return $out;
    }
    elsif (/^date$/) {
	if (! $opt{null_date_type}) { 
	    $opt{null_date_type} = $config{form}->{default_date}; 
	}
	# value must be in YYYY-MM-DD format
	my ($year,$month,$day);
	if ($opt{value}) {
	    ($year,$month,$day) = split(/-/,$opt{value});
	    if ($month =~ /^0[1-9]$/) {
		# chomp that leading 0 so we can use string equiv operators
		$month = substr($month,1,1);
	    }
	}
	else {
	    if ($opt{null_date_type} eq "today") {
		(undef,undef,undef,$day,$month,$year,undef,undef,undef) =
		    localtime(time);
		$month += 1;
		$year += 1900;
	    }
	    else {
		# use null date by default
		$year = 0;
		$month = 0;
		$day = 0;
	    }
	}
	my ($imonth, @months);
	$imonth = 0;
	foreach my $mname (@ExSite::Report::month_name) {
	    push @months, [$imonth,$mname];
	    $imonth++;
	}
	$out = $this->input_html(type=>"select",
				  name=>"$opt{name}${sc}month",
				  options=>\@months,
				  value=>$month);
	$out .= $this->input_html(type=>"text",
				  name=>"$opt{name}${sc}day",
				  size=>2,
				  value=>$day);
	$out .= ",&nbsp;";
	$out .= $this->input_html(type=>"text",
				  name=>"$opt{name}${sc}year",
				  size=>4,
				  value=>$year);
	return $out;
    }
    elsif (/^time$/) {
	my ($hour,$min,$ampm);
	if ($opt{minstep} <= 0) { $opt{minstep} = 15; }
	if (defined $opt{value} && $opt{value} !~ /^0?0:00:00$/) {
	    ($hour,$min,undef) = split(/:/,$opt{value});
	    $ampm = ($hour >= 12 ? "PM" : "AM");
	}
	else {
	    ($hour,$min) = ("12","00");
	    $ampm = "AM";
	}
	my $hour2 = $hour > 12 ? $hour - 12 : $hour;
	$hour2 =~ s/^0//;

	my @hours = (1,2,3,4,5,6,7,8,9,10,11,12);
	$out = $this->input_html(type=>"select",
				  name=>"$opt{name}${sc}hour",
				  options=>\@hours,
				  value=>$hour2);
	$out .= ":";
	my @minutes;
	for (my $i = 0; $i < 60; $i += $opt{minstep}) {
	    push @minutes, sprintf "%02d",$i;
	}
	$out .= $this->input_html(type=>"select",
				  name=>"$opt{name}${sc}min",
				  options=>\@minutes,
				  value=>$min);
	$out .= $this->input_html(type=>"select",
				  name=>"$opt{name}${sc}ampm",
				  options=>["AM","PM"],
				  value=> $ampm);
	return $out;
    }
    elsif (/^datetime$/) {
	my ($dval,$tval) = split(/ /,$opt{value});
	$htmlopt{value} = $dval;
	$htmlopt{datatype} = "date";
	$out = &input_exsite($this,%htmlopt);
	$out .= ",&nbsp;&nbsp;&nbsp;";
	$htmlopt{value} = $tval;
	$htmlopt{datatype} = "time";
	$out .= &input_exsite($this,%htmlopt);
	return $out;
    }
    elsif (/^key[:=]/) {
	my (undef,$ftable) = split /[:=]/;
	my $rdata;
	$htmlopt{type} = "select";
	if (scalar keys %{$opt{fmatch}}) {
	    # select only form foreign rows matching the match hash
	    $rdata = $this->fetch_match($ftable,$opt{fmatch});
	}
	else {
	    # check if a handler has been provided to filter appropriate
	    # foreign keys to choose from
	    if (! $opt{table}) {
		($opt{table},$opt{record},$opt{column},undef) = 
		    split /$config{form}{sepchar}/,$opt{name};
	    }
	    $rdata = $this->run_handler("select_foreign_key",$opt{table},$opt{record},$opt{column},$ftable);
	    if (ref $rdata ne "ARRAY") {
	    	# nothing!
	    	# select from all foreign rows - could be dangerous
	    	$rdata = $this->fetch_all($ftable);
	    }
	}
	# sort the foreign keys
	my $sort_key = $this->primary_column($ftable);
	my @data = $this->sort_rows($ftable,$rdata,$sort_key);
	# make the menu
	my @options;
	if (@data == 0) {
	    my $title = $this->{map}->table_title($ftable,1);
	    push @options, [0, "WARNING: no $title defined!"];
	}
	else {
	    my @show = $this->primary_column($ftable);
	    my $key = $this->get_key($ftable);
	    my $title = $this->{map}->table_title($ftable,0);
	    push @options, ["", "== select $title =="];
	    foreach my $dat (@data) {
		my $label;
		if ($this->{handler}->{select_foreign_key_label}) {
		    $label = &{$this->{handler}->{select_foreign_key_label}} 
		    ($this,$ftable,$dat,$key,@show);
		}
		if (! $label) {
		    foreach my $col (@show) {
			$label .= "$dat->{$col} ";
		    }
		    # can optionally add IDs to distinguish between similar records
		    if ($config{form}->{show_foreign_key_id}) {
			$label .= "[$dat->{$key}]";
		    }
		}
		push @options, [$dat->{$key}, $label];
	    }
	}
	$htmlopt{name} .= "#key";
	$htmlopt{options} = \@options;
	$out = $this->input_html(%htmlopt);
	return $out;
    }
    elsif (/^money$/) {
	$out = "\$";
	$htmlopt{size}=8;   ### use config value?
	$out .= $this->input_html(%htmlopt);
	return $out;
    }
    elsif (/file$/) {
	my $name = $htmlopt{name};
	my $value = $htmlopt{value};
	$htmlopt{name} = "$name#filename";
	# not meaningful at this time to pass a value for file fields, and it
	# just clogs the page with Mime-encoded file data.
	$htmlopt{value} = undef;
	if ($htmlopt{size} > 100) { $htmlopt{size} = $config{form}{max_cols}; }
	$out .= $this->input_html(%htmlopt);
	if ($value) {
	    $out .= "&nbsp;&rarr;&nbsp;currently&nbsp;".
		$this->show_data($htmlopt{table},$htmlopt{column},
				 $value,$htmlopt{record}); 
	    my %rmopt = %htmlopt;
	    $rmopt{name} = "$name#clear";
	    $rmopt{type} = "checkbox";
	    $rmopt{value} = "remove this file";
	    #$out .= "&nbsp;&nbsp;Check here to clear this file without uploading a new one:";
	    $out .= "&nbsp;&nbsp;";
	    $out .= $this->input_html(%rmopt);
	}
    }
    elsif (/^fixed$/) {
	$htmlopt{type} = "hidden";
	#$htmlopt{type} = "text";
	#$htmlopt{readonly} = 1;
	$htmlopt{value} = $this->{map}->get_column_attr($htmlopt{table},
							$htmlopt{column},
							"default");
	$out .= $this->input_html(%htmlopt);
    }
    else {
	$out .= $this->input_html(%htmlopt);
    }
    return $out;
}

#--------------------------------------------------------------------
# input_html : generate an HTML input tag
#
# INPUT keys = *name,type,value,size,readonly,rows,cols,options
# OUTPUT: the appropriate input tag
#
# This method could be modified/extended by using CGI's standard HTML 
# generation methods.
#--------------------------------------------------------------------

sub input_html {
    my ($this,%opt) = @_;
    if (! $opt{name}) {
	$this->error("ExSite::Form::input_html: no input field name specified");
	return "<span class=error>(error: unnamed input)</span>";
    }
    my $out = "<";
    my $sc = $config{form}->{sepchar};
    my $type = $opt{type} || "text";

    # set the field size
    my $size = $opt{size} < $config{form}->{max_cols} ?
	$opt{size} : $config{form}->{max_cols};

    #
    # 1) basic input definition
    #
    if ($type =~ /textarea/i) {
	$out .= "textarea wrap=virtual";
	my $cols = $opt{cols} || $size;
	my $nrows = $opt{rows} || int($opt{size}/$config{form}->{max_cols})+1;
	if ($nrows > $config{form}->{max_rows}) { 
	    $nrows = $config{form}->{max_rows}; 
	}
	$out .= " rows=$nrows cols=$cols";
    }
    elsif ($type =~ /select/i) {
	$out .= "select";
    }
    else { 
	$out .= "input type=$type";
	if ($opt{size}) { $out .= " size=$size"; }
    }
    $out .= " name=$opt{name}";
    if ($opt{readonly} && $type =~ /text/) { $out .= " readonly"; }
    if ($opt{selected} && $type =~ /radio|checkbox/) { $out .= " checked"; }
    if ($type =~ /select|textarea/) {
	$out .= ">";
    }
    #
    # 2) now add the input data
    #
    if (exists $opt{value} || $type =~ /select|radio/i) {
	if ($type =~ /textarea/i) {
	    # okay to use quotes in a text area
	    my $data = $opt{value};
	    $data =~ s/&\#34;/\"/g;
	    $out .= "$data";
	}
	elsif ($type =~ /select/i) {
	    $out .= "\n";
	    if (ref $opt{options} eq "ARRAY") {
		# an array of options, in various formats
		foreach my $item (@{$opt{options}}) {
		    my $typ = ref $item;
		    my ($val,$text);
		    if ($typ eq "ARRAY") { 
			# reference to an array of value,text
			$val = $item->[0]; 
			$text = $item->[1];
		    }
		    elsif ($typ eq "HASH") { 
			# hash with value, text 
			$val = $item->{value}; 
			$text = $item->{text}; 
		    }
		    else { 
			# just text, no value
			$val = $text = $item; 
		    }
		    $out .= "<option";
		    #if ($val) { $out .= " value=\"$val\""; }
		    $out .= " value=\"$val\"";
		    if ($val eq $opt{value}) { $out .= " selected"; }
		    $out .= ">$text</option>\n";
		}
	    }
	    elsif (ref $opt{options} eq "HASH") {
		# a hash of value=>descriptive text pairs
		while (my ($val,$text) = each %{$opt{options}}) {
		    $out .= "<option value=\"$val\"";
		    if ($val eq $opt{value}) { $out .= " selected"; }
		    $out .= ">$text</option>\n";
		}
	    }
	    else {
		# assume it is preformatted
		$out .= $opt{options};
	    }
	}
	else { 
	    # Netscape 4.7 adds spaces if value not defined
	    $out .= " value=\"$opt{value}\"";
	}
    }
    #
    # 3) now close the tags
    #
    if ($type =~ /textarea/i) {
	$out .= "</textarea>";
    }
    elsif ($type =~ /select/i) {
	$out .= "</select>";
    }
    else { 
	$out .= ">";
    }
    if ($type =~ /radio|checkbox/i) {
	my $label = $opt{label} ? $opt{label} : $opt{value};
	if ($label) { 
	    $out .= "$label&nbsp;&nbsp;&nbsp; ";
	}
    }
    if ($type =~ /file/i) {
	$this->{enctype} = "multipart/form-data";
    }
    return $out;
}

#------------------------------------------------------------------------
# popup : link to a document in a pop-up dialog box
#
# This is a kludge; need a general Javascript management tool
#------------------------------------------------------------------------

sub popup($$$$$) {
    my ($this, $url, $anchor, $width, $height) = @_;
    if (! $width) { $width = 400; }
    if (! $height) { $height = 400; }
    my $out .= "<A HREF=javascript:popup_link(\"$url\")>$anchor</A>";
    if (! $config{form}->{"javascript-popup"}) {
    	$out .= "<script LANGUAGE=javascript>\nfunction popup_link(url) {\n  window.open(url,\"\",\"width=$width,height=$height,resizable,scrollbars=1,menubar=0\")\n}\n</SCRIPT>\n";
    }
    $config{form}->{"javascript-popup"} = 1;   #### not modperl compatible...
    return $out;
}					

#-------------------------------------------------------------------
# input_htmleditor : generate a wysiwyg editor 
#
# options:
# name => the name of the input field
#
# editable content - one of:
# table, record, column => a field in the database
# file => a path to a file on disk
# url => an external URL
#-------------------------------------------------------------------
sub input_htmleditor{
    my ($this,%opt) = @_;
    my ($out,$sc,$value);
    $sc = $config{form}->{sepchar};
    $value = $opt{value};
    $this->{htmleditor} = $opt{name}."#html";
    my ($table,$record,$column);
    if ($opt{table}) {
    	$this->{htmleditor_content_id} = $opt{record};
	$this->{htmleditor_page_id} = $opt{page};
	$this->{htmleditor_section_id} = $opt{section};
	$table = $opt{table};
	$record = $opt{record};
	$column = $opt{column};
    }
    else {($table,$record,$column) = split(/$sc/,$opt{name});}
    my @icons = (
	{form_id=>"op_undo",title=>"Undo",alt=>"Undo",id=>"undo",class=>"imagebutton",imgclass=>"buttonImg",file=>"undo.gif"},
	{form_id=>"op_redo",title=>"Redo",alt=>"Redo",id=>"redo",class=>"imagebutton",imgclass=>"buttonImg",file=>"redo.gif"},
	{form_id=>"op_link",title=>"Insert Link",alt=>"Insert Link",id=>"createlink",class=>"imagebutton",imgclass=>"buttonImg",file=>"link.gif"},
	{form_id=>"op_image",title=>"Insert Image",alt=>"Insert Image",id=>"insertimage",class=>"imagebutton",imgclass=>"buttonImg",file=>"image.gif"},
	{form_id=>"op_table",title=>"Insert Table",alt=>"Insert Table",id=>"createtable",class=>"imagebutton",imgclass=>"buttonImg",file=>"button_table_down.gif"},
	{form_id=>"op_dyncontent",title=>"Insert Web Application",alt=>"Insert Dynamic Content",id=>"dyncontent",class=>"imagebutton",imgclass=>"buttonImg",file=>"dcd.gif"},
	{form_id=>"op_bold",title=>"Bold",alt=>"Bold",id=>"bold",class=>"imagebutton",imgclass=>"buttonImg",file=>"bold.gif"},
	{form_id=>"op_italic",title=>"Italic",alt=>"Italic",id=>"italic",class=>"imagebutton",imgclass=>"buttonImg",file=>"italic.gif"},
	{form_id=>"op_underline",title=>"Underline",alt=>"Underline",id=>"underline",class=>"imagebutton",imgclass=>"buttonImg",file=>"underline.gif"},
	{form_id=>"op_textcolor",title=>"Text Color",alt=>"Text Color",id=>"forecolor",class=>"imagebutton",imgclass=>"buttonImg",file=>"forecol.gif"},
	{form_id=>"op_bgcolor",title=>"Background Color",alt=>"Background Color",id=>"hilitecolor",class=>"imagebutton",imgclass=>"buttonImg",file=>"bgcol.gif"},
	{form_id=>"op_justify",title=>"Justify",alt=>"Justify",id=>"justifyfull",class=>"imagebutton",imgclass=>"buttonImg",file=>"justify.gif"},
	{form_id=>"op_alignleft",title=>"Align Left",alt=>"Align Left",id=>"justifyleft",class=>"imagebutton",imgclass=>"buttonImg",file=>"left.gif"},
	{form_id=>"op_aligncenter",title=>"Center",alt=>"Center",id=>"justifycenter",class=>"imagebutton",imgclass=>"buttonImg",file=>"center.gif"},
	{form_id=>"op_alignright",title=>"Align Right",alt=>"Align Right",id=>"justifyright",class=>"imagebutton",imgclass=>"buttonImg",file=>"right.gif"},
	{form_id=>"op_olist",title=>"Ordered List",alt=>"Ordered List",id=>"insertorderedlist",class=>"imagebutton",imgclass=>"buttonImg",file=>"ordlist.gif"},
	{form_id=>"op_blist",title=>"Bulleted List",alt=>"Bulleted List",id=>"insertunorderedlist",class=>"imagebutton",imgclass=>"buttonImg",file=>"bullist.gif"},
	{form_id=>"op_outdent",title=>"Outdent",alt=>"Outdent",id=>"outdent",class=>"imagebutton",imgclass=>"buttonImg",file=>"outdent.gif"},
	{form_id=>"op_indent",title=>"Indent",alt=>"Indent",id=>"indent",class=>"imagebutton",imgclass=>"buttonImg",file=>"indent.gif"},
    );
    $out .= "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\">\n<tr>\n<td>\n
    <table cellspacing=\"0\" cellpadding=\"0\" bgcolor=\"#d6d3ce\" width=\"100%\" class=\"editorToolBar\"><tr><td>\n
    <table id=\"toolbar1\" height=\"30\" border=\"0\" cellspacing=\"0\" cellpadding=\"0\" bgcolor=\"#d6d3ce\">\n
	<tr>\n";
    foreach my $i (@icons){
    	if($config{form}{htmleditor}{$i->{form_id}} || $opt{editor}{$i->{form_id}}){
    	    $out .= "	<td class=\"tdClass\" width=\"30\"><div class=\"$i->{class}\" id=\"$i->{id}\"><center><img src=\"";
	    $out .= "$config{server}->{HTMLpath}/_ExSite/images/$i->{file}\" title=\"$i->{title}\" ";
	    $out .= "class=\"$i->{imgclass}\" alt=\"$i->{alt}\"></center></div></td>\n";
	}
    }
    $out .= "	</tr>
    </table>\n</td></tr></table>\n";
    my ($source,$cmstag);
    $cmstag = $config{form}{htmleditor}{op_dyncontent} || $opt{editor}{op_dyncontent} ? 2 : 0;
    if ($record && $record !~ /^_/) {
	$source = $config{server}->{CGIpath}."/echo.cgi?tab=$table&col=$column&id=$record";
	if ($table eq "content_data") {
		$cmstag++;
		$source .= "&cms=$cmstag";
	}
    }
    elsif ($opt{file}){
	$source = $config{server}->{CGIpath}."/echo.cgi$opt{file}";
    }
    elsif ($opt{url}){
	$source = $config{server}->{CGIpath}."/echo.cgi?url=$opt{url}";
    }
    else{
    	$source = $config{server}{CGIpath}."/echo.cgi?cms=$cmstag";
    }
    $out .= "\n<iframe class=\"editorTextArea\" id=\"edit\" style=\"HEIGHT: 512px; WIDTH: 700px\"";
    #$out .= $value ? " src=\"$source\"" : "";
    $out .= " src=\"$source\"";
    $out .= "></iframe>\n";
    $out .= "<iframe width=\"250\" height=\"170\" id=\"colorpalette\" src=\"$config{server}->{HTMLpath}/_ExSite/colors.html\" ";
    $out .= "style=\"visibility:hidden; position: absolute;\"></iframe>\n";
    $out .= "    <table bgcolor=\"#C0C0C0\" id=\"toolbar2\" width=\"100%\" class=\"editorEditOption\">
	<tr>
	<td>
	<select id=\"formatblock\" onchange=\"Select(this.id);\">
	  <option value=\"<p>\">Normal</option>
	  <option value=\"<p>\">Paragraph</option>
	  <option value=\"<h1>\">Heading 1 <h1></option>
	  <option value=\"<h2>\">Heading 2 <h2></option>
	  <option value=\"<h3>\">Heading 3 <h3></option>
	  <option value=\"<h4>\">Heading 4 <h4></option>
	  <option value=\"<h5>\">Heading 5 <h5></option>
	  <option value=\"<h6>\">Heading 6 <h6></option>
	  <option value=\"<address>\">Address <ADDR></option>
	  <option value=\"<pre>\">Formatted <pre></option>
	</select> 
	<select id=\"fontname\" onchange=\"Select(this.id);\">
	  <option value=\"Font\">Font</option>
	  <option value=\"Arial\">Arial</option>
	  <option value=\"Courier\">Courier</option>
	  <option value=\"Times New Roman\">Times New Roman</option>
	</select> 
	<select unselectable=\"on\" id=\"fontsize\" onchange=\"Select(this.id);\">
	  <option value=\"Size\">Size</option>
	  <option value=\"1\">1</option>
	  <option value=\"2\">2</option>
	  <option value=\"3\">3</option>
	  <option value=\"4\">4</option>
	  <option value=\"5\">5</option>
	  <option value=\"6\">6</option>
	  <option value=\"7\">7</option>  
	</select>
	</td>
	</tr>
    </table>
</td>\n</tr>\n</table>\n
    ";
    return $out;
}

#-----------------------------------------------------------------
# editor_javascript : generate necessary javascript code to 
#			enable the wysiwyg editor
#-----------------------------------------------------------------
sub editor_javascript{
	my ($this,$section_id) = @_;
	if (! $section_id) { $section_id = $this->{htmleditor_section_id}; }
	my $out =  "
<style type=\"text/css\">
.imagebutton {
	height: 22; 
	width: 23; 
	border: none; 
	background-color: #D6D3CE;
}
.image {
	position: relative; 
	left: 1; 
	top: 1; 
	height: 20; 
	width: 21; 
	border: none;
}
.toolbar {
	height: 30; 
	background-color: #D6D3CE;
}
.tdClass {
	background-color: #D6D6CE;
}
table.editorToolBar td {
	background-color: #D6D6CE;
}
table.editorEditOption td {
	background-color: #C0C0C0;
}
.editorTextArea {
	background-color: #FFFFFF;
}
</style>
<script language='JavaScript' TYPE='text/javascript' >
<!--
window.onload = Start;
var command = \"\";
var curr_id;
function Start() {
  document.getElementById('edit').contentWindow.document.designMode = \"on\";
  try {
    document.getElementById('edit').contentWindow.document.execCommand(\"undo\", false, null);
  } catch (e) {
    alert(\"This editor is not supported on your browser.\");
  }
  InitToolbarButtons();
}
function doDynContent() {
  var link = '$config{server}{CGIpath}/dlg_dyncontent.cgi?section=$section_id';
  varlnkWin = window.open(link,'','height=160,width=350,status=no,menubar=no,toolbar=no,location=no,scrollbars=yes,resizable=yes');
}
function doImage() {
  var link = '$config{server}->{CGIpath}/dlg_image.cgi?";
  $out .= $this->{htmleditor_page_id} ? "cont_id=$this->{htmleditor_page_id}" : "data_id=$this->{htmleditor_content_id}";
  $out .= "'\nvar lnkWin = window.open(link,'','height=380,width=530,status=no,menubar=no,toolbar=no,location=no,scrollbars=yes,resizable=yes');
}
function doTable() {
  var link = '$config{server}->{CGIpath}/dlg_table.cgi';
  var lnkWin = window.open(link,'','height=400,width=470,status=no,menubar=no,toolbar=no,location=no,scrollbars=yes,resizable=yes');
} 
function getDynContent(name,param,url) {
  if (name) {
    var dyn;
    if (isIE()){
      dyn = '<';
      dyn += 'img alt=\"' + name + '(' + param + ')\" name=\"' + name + '\" ';
      dyn += 'src=\"' + url + '\" ';
      dyn += 'title=\"' + name + '\" ';
      dyn += 'border=\"3\" style=\"border-style:outset;\"';
      dyn += '>';
      var selection = edit.document.selection.createRange();
      selection.pasteHTML(dyn);
    }else{
      dyn = document.getElementById('edit').contentWindow.document.createElement('img');
      dyn.alt = name + '(' + param + ')';
      dyn.name = name;
      dyn.src = url;
      dyn.title = name;
      dyn.border = 3;
      dyn.style.borderStyle = 'outset';
      insertNodeAtSelection(document.getElementById('edit').contentWindow,dyn);
    }
  }
}
function getImage(isImage,name,url,align,width,height) {
  var imageUrl;
  if (isImage > 0){
    imageUrl = '[[';
  }else{
    imageUrl = 'content(';
  }
  imageUrl += name;
  if (isImage > 0){
    imageUrl += ']]';
  }else{
    imageUrl += ')';
  }
  if (name){
    var image;
    if (isIE()){
      image = '<img alt=\"' + imageUrl + '\"';
      image += ' src=\"' + url + '\"';
      image += ' title=\"' + name + '\"';
      image += ' width=\"' + width + '\" height=\"' + height + '\"';
      if (align != ''){
        image += ' align=\"' + align + '\"';
      }
      image += '>';
      var selection = edit.document.selection.createRange();
      selection.pasteHTML(image);
    }else{
      image = document.getElementById('edit').contentWindow.document.createElement('img');
      image.alt = imageUrl;
      image.name = name;
      image.src = url;
      image.title = name;
      image.width = width;
      image.height = height;
      if (align != ''){
        image.align = align;
      }
      insertNodeAtSelection(document.getElementById('edit').contentWindow,image);
    }
  }
}
function doLink() {
  var selection;
  var link = '$config{server}->{CGIpath}/dlg_link.cgi?";
  $out .= $this->{htmleditor_page_id} ? "cont_id=$this->{htmleditor_page_id}" : "data_id=$this->{htmleditor_content_id}";
  $out .= "&section_id=${section_id}&title=';
  if (isIE()){
    selection = edit.document.selection.createRange();
    link += selection.text;
  }else{
    selection = document.getElementById('edit').contentWindow.getSelection();
    link += selection.toString();
  }
  var lnkWin = window.open(link,'','height=190,width=420,status=no,menubar=no,toolbar=no,location=no,scrollbars=yes,resizable=yes');
}
function doColor() {
  var color = '$config{server}->{HTMLpath}/_ExSite/colors.html';
  var colorWin = window.open(color,'','height=150,width=200,status=no,menubar=no,toolbar=no,location=no,scrollbars=no,resizable=no');
}
//-->
</script>
<script language='JavaScript' src='$config{server}{HTMLpath}/_ExSite/js/editor.js'></script>
";
	return $out;
}

1;
