package ExSite::Image;

=pod

ExSite::Image - image manager

Provides a few basic image mamagement tools for the ExSite CMS.
This is just a simple wrapper/API to whatever underlying image processing
tools are going to do the heavy lifting, so it is easy to replace or
extend with more powerful tools if needed.

This version tries to use Image::Info for fetching image attributes, 
and the convert shell command for modifying the image itself.

To start working with a new image:

    my $img = new ExSite::Image($name,$rawdata);   # OR
    my $img = new ExSite::Image($encoded_data);

To get the image attributes:

    my $info = $img->info;  # $info is a hash ref

To get the image size:

    my ($x,$y) = $img->dim;

To change the image size:

    my $stat = $img->scale(500,500,$flag);  

    # $flag > 0   scale up to that size, if smaller
    # $flag == 0  scale to that size
    # $flag < 0   scale down to that size, if larger
    # retain aspect ratio in all cases

To thumbnail an image:

    my $stat = $img->thumb;  # equivalent to $img->scale(100,100,-1);  

To get image raw data:

    my $imgdata = $img->get;

To get image data in exsite binary file encoding format:

    my $imgdata = $img->encode;

=cut

use strict;
use MIME::Base64;
use IO::Handle;
use FileHandle;
use IPC::Open3;
use ExSite::Base;
use ExSite::Config;
use ExSite::Misc;

use vars qw(@ISA);
@ISA = qw(ExSite::Base);

sub new {
    my $this = shift;
    # objectify object
    my $obj = {};
    my $class = ref($this) || $this;
    bless $obj, $class;
    return $obj->setup(@_);
}

sub setup {
    my $this = shift;
    if (@_ == 1) {
	($this->{filename},my $mimedata) = split /\#/,shift,2;
	$this->{filedata} = &decode_base64($mimedata);
    }
    elsif (@_ == 2) {
	my @filepath = split /[:\\\/]/, shift;
	$this->{filename} = @filepath[-1];
	$this->{filedata} = shift;
    }
    else {
	$this->error("invalid image data");
	return undef;
    }
    return $this;
}

sub info {
    my $this = shift;
    if (require Image::Info && require IO::String ) {
	my $fdata = $this->{filedata};
	return &Image::Info::image_info(\$fdata);
    }
    # no info tool, so let's kludge something up using convert
    my $pid = open3(*W, *R, *E, "convert -verbose - -");
    print W $this->{filedata}; close W;
    # we have to read (and discard) stdout data before the stderr output
    # will be given to us
    while (<R>) { 1; } 
    close R;
    # now read stderr, which contains the -verbose data we want
    # expected output is something like:
    # /tmp/ABC123=>- GIF 239x239+0+0 PseudoClass 256c 8-bit 0.0u 0:01
    my $imgdata = <E>; close E; 
    waitpid $pid, 0; # perlipc recommends this to avoid zombies
    my @imgdata = split / /,$imgdata;
    $imgdata[2] =~ /([\d]+)x([\d]+)/;
    my $ext = lc($imgdata[1]);
    return {
	file_ext => $ext,
	file_media_type => &MimeType("dummy.$ext"),
	width => $1,
	height => $2,
    }
}

sub dim {
    my $this = shift;
    my $info = $this->info;
    return ($info->{width},$info->{height});
}

sub scale {
    my ($this,$x,$y,$flag) = @_;
    #
    # A better solution here might be to install Perl::Magick and
    # use its interface to the ImageMagick libraries.  However, that
    # package is HUGE and technically unnecessary, since if the 
    # ImageMagick libraries are already in place then we can just
    # use the "convert" shell command to do the job, without the 
    # hassle of installing and supporting Perl::Magick.
    #
    my $geom = "${x}x${y}";
    if ($flag > 0) { $geom .= '\<'; }
    elsif ($flag < 0) { $geom .= '\>'; }
    my $pid = open3(*W, *R, *E, "convert -geometry $geom - -");
    print W $this->{filedata}; close W;
    my $newdata;
    my $size = 0;
    my $continue = 1;
    while ($continue) {
	$continue = read(R,$newdata,1024,$size);
	$size += $continue;
    }
    $this->{filedata} = $newdata;
    close R;
    #my $imgdata = <E>;
    close E;  # don't need stderr in this case
    waitpid $pid, 0;  # anti-zombie spell
    return $size;
}

# shortcut for thumbnailing

sub thumb {
    my $this = shift;
    my $thumbdim = $config{thumbnail_size};
    my ($x,$y) = $this->dim;
    # don't thumbnail if the image is already small
    return 0 if ($x < $thumbdim && $y < $thumbdim);
    return $this->scale($thumbdim,$thumbdim,-1);
}

# return raw image data to user

sub get () {
    my $this = shift;
    return $this->{filedata};
}

# encode image data in exsite format (filename#mime_data)

sub encode {
    my $this = shift;
    my $fname = &url_escape($this->{filename});
    return $fname.$config{form}{sepchar}.&encode_base64($this->{filedata});
}

1;
