package ExSite::Input;
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

use strict;
use ExSite::Base;
use ExSite::Config;

# inherit from Base
use vars qw(@ISA);
@ISA = qw(ExSite::Base);

#========================================================================

=pod

=head1 ExSite::Input  --  tools for reading http inputs

=head2 query() - read input passed via GET method (ie. QUERY_STRING)

=head2 post() - read input passed via POST method

=head2 query_or_post() - read from query, or post if no query

=head2 post_or_query() - read from post, or query if no post

=head2 combine() - read from both post and query combined

These are replacements for the generic input parsing routines in 
ExSite::Misc::.  The differences between these and the generic routines are:

=over 4

=item * these versions cache the input in C<%share>, so that multiple 
modules can share the same POST data.

=item * caching also prevents the same input from being re-parsed
unnecessarily.

=item * raw input is also saved, in case URL-encoding is not used

=item * user can optionally request raw input, instead of decoded input

=back

Input is saved in the following structure:

C<$share{input}{get|post|combine}{raw|data|keylist}>

where get/post/combine selects the input stream, and raw/data/keylist
selects the format you would like to view the data in.  Raw is the
bare input (or a CGI:: object if using multipart encoding), data is 
a hash of key->value pairs, and keylist is an ordered set of input keys.

=cut

#========================================================================

# my $preset_input = new ExSite::Input({post=>"mypostdata",get=>"myquery"});

sub new ($$) {
    my($this, $preset) = @_;

    # objectify object
    my $cobj = {};
    my $class = ref($this) || $this;
    bless $cobj, $class;
    $cobj->info("ExSite::Input object created");

    # input data is stored in share area, not in object, so that it 
    # can serve multiple components simultaneously.

    if (!$share{input}) {
	$share{input} = {};
	foreach my $action ("post","get","combine") {
	    $share{input}{$action} = {};
	    $share{input}{$action}{keylist} = [];
	    $share{input}{$action}{data} = {};
	}
    }

    # check for preset data: allows programs to manually preset the input
    # (useful for automated testing)

    if (keys %$preset > 0) {
	$share{input}{preset} = $preset;
    }
    elsif (! $share{input}{preset}) {
	$share{input}{preset} = {};
    }

    return $cobj;
}

# query : process QUERY_STRING input, ie. "GET" method

sub query {
    my ($this,$fetch_as) = @_;
    if (! $share{input}{get}{raw}) {
	$share{input}{get}{raw} = 
	    $share{input}{preset}{get} || $ENV{QUERY_STRING};
	$this->decode("get",$share{input}{get}{raw});
    }
    return $this->fetch("get",$fetch_as);
}

# get is synonym for query
sub get { return &query(@_); }

# post : process "POST" method input

sub post {
    my ($this,$fetch_as) = @_;
    if (! $share{input}{post}{raw}) {
	if ($ENV{CONTENT_TYPE} =~ /multipart\/form-data/) {
	    $this->info("decoding multipart form");
	    #$share{input}{post}{data}{debug} = $ENV{CONTENT_TYPE};
#	    if (! require CGI) : or { 
#		$this->error("CGI:: module failed to load: $@");
#		return; 
#	    }
	    require CGI;
	    # CONTENT_TYPE=multipart/form-data
	    # (this form contains file uploads; use CGI module to decode it;
	    # that means that the "raw" data is really a CGI:: handle.)
	    $share{input}{post}{raw} = 
		$share{input}{preset}{post} || CGI::new();

	    # must use CGI:: to decode
	    if ($ENV{CONTENT_TYPE} =~ /multipart\/form-data/) {
		# CONTENT_TYPE=multipart/form-data
		my $in = $share{input}{post}{raw};
		foreach my $par ($in->param) {
		    push @{$share{input}{post}{keylist}}, $par;
		    # fetch multiple values, in case of repeated inputs
		    my @pars = $in->param($par);
		    if (@pars > 1) { 
			# there are multiple values; concatenate them
			$share{input}{post}{data}{$par} = 
			    join($config{form}{multi_sep},@pars);
		    }
		    else {
			# single value; do not concatenate, or filehandles
			# will be wrecked
			$share{input}{post}{data}{$par} = $in->param($par);
		    }
		}
	    }
	}
	else {
	    # CONTENT_TYPE=application/x-www-form-urlencoded
	    $this->info("decoding regular form");
	    my $clen = $ENV{'CONTENT_LENGTH'};
	    if ($share{input}{preset}{post}) {
		$share{input}{post}{raw} = $share{input}{preset}{post};
	    }
	    elsif ($clen) {
		my $in;
		read(STDIN,$in,$ENV{'CONTENT_LENGTH'});
		$share{input}{post}{raw} = $in;
	    }
	    $this->decode("post",$share{input}{post}{raw});
	}
    }
    return $this->fetch("post",$fetch_as);
}

# post_or_query, query_or_post : return one or the other, but not both

sub post_or_query {
    my ($this,$fetch_as) = @_;
    my $data = $this->post($fetch_as);
    return $data if (ref $data eq "HASH" && keys %$data > 0);
    return $data if (ref $data eq "ARRAY" && @$data > 0);
    $data = $this->query($fetch_as);
    return $data;
}

sub query_or_post {
    my ($this,$fetch_as) = @_;
    my $data = $this->query($fetch_as);
    return $data if (ref $data eq "HASH" && keys %$data > 0);
    return $data if (ref $data eq "ARRAY" && @$data > 0);
    $data = $this->post($fetch_as);
    return $data;
}

# combine : combine post and query data together into a single input stream

sub combine {
    my ($this,$fetch_as) = @_;
    # no raw input for combined input
    $share{input}{combine}{raw} = undef;
    # load both input streams
    $this->query;
    $this->post;
    # combine the data hashes and key lists
    my @tmplist = 
	(@{$share{input}{get}{keylist}},@{$share{input}{post}{keylist}});
    $share{input}{combine}{keylist} = \@tmplist;

    my %tmphash = %{$share{input}{post}{data}};
    foreach my $key (keys %{$share{input}{get}{data}}) {
	if (! exists $tmphash{$key}) { 
	    $tmphash{$key} = $share{input}{get}{data}{$key};
	}
    }
    $share{input}{combine}{data} = \%tmphash;
    return $this->fetch("combine",$fetch_as);
}

# decode : break input string down into key, value pairs
# arguments: input method, input string, separator character (optional)

sub decode {
    my $this = shift;
    my $method = shift;
    my $sep = $_[1] || "&";
    my @in = split(/$sep/,$_[0]);
    my (%out, @okeys);
    my $out = $share{input}{$method};
    foreach my $i (0 .. $#in) {
        # Convert pluses to spaces
        $in[$i] =~ s/\+/ /g;

        # Split into key and value.
        (my $key, my $val) = split(/=/,$in[$i],2); # splits on the first =

        # Convert %XX from hex numbers to alphanumeric
        $key =~ s/%(..)/pack("c",hex($1))/ge;
        $val =~ s/%(..)/pack("c",hex($1))/ge;

	# save key order
	if (! exists $out->{data}{$key}) { push @{$out->{keylist}}, $key; }
        # Associate key and value. Multiply-defined inputs are concatenated.
	$out->{data}{$key} .= 
	    $config{form}{multi_sep} if (defined($out->{data}{$key}));
	$out->{data}{$key} .= $val;
    }
    return;
}

# fetch : retrieve data in various formats

sub fetch {
    my $this = shift;
    my $method = shift;
    $_ = shift;
    if (/raw/)     { return $share{input}{$method}{raw}; }
    elsif (/keys/) { return $share{input}{$method}{keylist}; }
    else           { return $share{input}{$method}{data}; }
}

# modify the input buffers

sub set {
    my ($this,$key,$val,$action) = @_;
    foreach my $act ("post","get","combine") {
	if (! $action || $action eq $act) {
	    $share{input}{$act}{data}{$key} = $val;
	}
    }
}

sub delete {
    my ($this,$key) = @_;
    # WARNING: only deletes value from data hash, not raw data or keylist
    foreach my $action ("post","get","combine") {
	delete $share{input}{$action}{data}{$key};
    }
}


1;

