#!/usr/bin/perl
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------

package ExSite::Page;

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Section;
use ExSite::Content;
use MIME::Base64;

use vars qw(@ISA);
@ISA = qw(ExSite::Content);

# new() constructor is inherited from ContentData class.

sub setup {
    # so that we can do setup using a non-overloaded method
    &setupPage(@_);
}

# options:
# id = id of page to look up
# page = prefetched page record
# noload = skip content loading phase (much faster if not rendering page)

sub setupPage { 
    my ($this,%option) = @_;

    my $page;
    if ($option{id}) {
	# lookup page record
	$page = $this->fetch("page",$option{id});
	if (! $page) {
	    $this->error("page $option{id} not found");
	}
    }
    elsif ($option{page}) {
	# page record was passed
	if (ref $option{page} eq "HASH") {
	    $page = $option{page};
	}
	else {
	    $this->error("invalid page parameters passed to setupPage");
	}
    }
    elsif ($option{filename} && $option{section_id}) {
	# lookup page by name & section
	my @pages = $this->fetch_match("page",{filename=>$option{filename},
					   section_id=>$option{section_id}});
	$page = @pages[-1];
	if (! $page) {
	    $this->error("\"$option{filename}\" not found");
	}
    }
    if ($page) {
	$this->info("setting up Page $page->{page_id}");
	$this->{data} = $page;
	$this->set_context("none");
    }
    else {
	$this->info("setting up null Page");
    }

    # publish flag
    if (exists $option{publish}) { 
	$this->{publish} = $option{publish}; 
    }
}

### basic queries

sub title { 
    my $this = shift;
    return $this->{data}{title};
}

sub description { 
    my $this = shift;
    return $this->{data}{description};
}

sub keywords { 
    my $this = shift;
    return $this->{data}{keywords};
}

sub is_template { 
    my $this = shift;
    return ($this->{data}{type} eq "template");
}

sub is_library { 
    my $this = shift;
    return ($this->{data}{type} eq "library");
}

sub is_public { 
    my $this = shift;
    return ($this->{data}{access} eq "public");
}

sub publishing { 
    my $this = shift;
    return $this->{publish};
}

## deprecated?
sub flag { 
    my ($this,$flag) = @_;
    if ($this->{data}{flags} =~ /$flag/) {   ## not a very robust regexp
	return 1;
    }
    return 0;
}

# return the timestamp of the page
# defined as the most recent of the timestamps of:
# - the template
# - the page
# - the content objects in the page

sub timestamp {
    my $this = shift;
    my @times;
    # timestamp of current page
    push @times, $this->get("page","mtime");
    # timestamp of all content objects in this page
    my @cobjs = $this->fetch_all_content_objs;
    foreach my $cobj (@cobjs) {
	push @times, $cobj->timestamp;
    }
    # plus the same, for this page's template
    my $template_id = $this->get("page","template_id");
    if ($template_id) {
	my $template = new ExSite::Page(id=>$template_id);
	push @times, $template->timestamp;
    }
    my @sorted_times = sort @times;
    return pop @sorted_times;
}

### some page display tools

sub show {
    my $this = shift;
    
    # dump headers

    my $content_type = $this->content_type || "text/html";
    print $this->{http_header};
    if ($this->{http_header} !~ /content-type:/i) {
	print "Content-type: $content_type\n";
    }
    print "\n";

    # dump page

    if (! $this->{redirect}) {
	print $this->expand;
    }
}

# some HTTP header functions

sub header {
    my $this = shift;
    my $hdr = shift;
    if ($hdr !~ /\n$/) { $hdr .= "\n"; }
    $this->{http_header} .= $hdr;
    return;
}

sub redirect {
    my ($this,$url) = @_;
    $this->header("Location: $url");
    $this->{redirect} = 1;
    return;
}

sub location {
    &redirect(@_);
    return;
}

sub cookie {
    my ($this,$name,$value,$path,$domain,$expiry) = @_;
    if (! $path) { $path = $config{server}{CGIpath}; }
    if (! $domain) { $domain = $config{server}{domain}; }
    my $cookie = "Set-cookie: $name=$value; path=$path; domain=$domain";
    if ($expiry) { $cookie .= "; Expires=$expiry"; }
    $this->header($cookie);
    return;
}

### FIXME: look for an exception template

sub errorpage {
    my ($this,$html) = @_;
    # replace page with error message
    $this->{html} = $html;
    $this->{noexpand} = 1;
    return;
}

# link - like get_url, but always returns a dynamic link, and you can add 
# your own parameters to it.

sub link {
    my ($this,$args) = @_;
    my $id = $this->id("page");
    my $url = "$config{server}{CGIpath}/page.cgi?_id=$id";
    if ($args) {
	if (ref $args eq "HASH") {
	    $args = &EncodeHash($args);
	}
	$url .= "&$args";
    }
    return $url;
}

# name of page file - overloaded variation on ContentData method

sub filename {
    my $this = shift;
    my $page = $this->get;
    if ($page) {
	my $filename = $page->{filename};
	if ($page->{type} eq "page") {
	    return $filename || "page_".$page->{page_id}.".html";
	}
	else {
	    # directory name for templates and libraries
	    return $filename || $page->{page_id};
	}
    }
    return undef;
}

# httppath: the URL path to the page, minus the server
######### DEPRECATED ############

sub httppath {
    my $this = shift;
    my $page = $this->get;
    if ($page) {
	my $s_httppath;
	if ($page->{section_id}) {
	    my $s = new ExSite::Section(id=>$page->{section_id});
	    $s_httppath = $s->httppath;
	}
	else {
	    $s_httppath = $config{server}{HTMLpath};
	}
	$_ = $page->{type};
	if (/template/) {
	    return "$s_httppath/_Templates/$page->{filename}";
	}
	elsif (/library/) {
	    return "$s_httppath/_Libraries/$page->{filename}";
	}
	else {
	    return $s_httppath;
	}
    }
    return undef;
}

# disk path to file

sub diskpath {
    my $this = shift;
    my $dir = shift;  # return directory only if true
    my $page = $this->get("page");
    if ($page) {
	my $s_diskpath;
	if ($page->{section_id}) {
	    my $s = new ExSite::Section(id=>$page->{section_id});
	    $s_diskpath = $s->diskpath;
	}
	else {
	    $s_diskpath = $config{server}{HTMLroot};
	}
	if ($s_diskpath) {
	    if (! $dir || $page->{type} ne "page") { 
		if ($page->{type} eq "template") { 
		    $s_diskpath .= "/_Template";
		}
		elsif ($page->{type} eq "library") { 
		    $s_diskpath .= "/_Library";
		}
	    }
	    return $s_diskpath if ($dir && $page->{type} eq "page");
	    return "$s_diskpath/".$this->filename;
	}
    }
    return undef;
}

# overloaded get_url: return a URL for the page
#### check for published page
#### else, return a link for dynamic page

sub get_url {
    my $this = shift;
    my $static_req = shift; # explicit request for static (1) or dynamic (0)
    my $static = defined $static_req ? $static_req : 1;  # default to static
    #if ($this->{publish}) { $static = 1; }
    my $page = $this->get;
    return undef if (! $page);
    #return undef if ($page->{type} eq "library");
    if ($page->{section_id}) {
	my $file = $this->filename;
	my $s = new ExSite::Section(id=>$page->{section_id});
	my $s_url = $s->get_url;
	my $s_diskpath = $s->diskpath;
	my ($path_to_page,$url_to_page);
	if ($page->{type} eq "template") { 
	    $path_to_page = $s_diskpath."/_Template/".$file;
	    $url_to_page = $s_url."/_Template/".$file;
	}
	elsif ($page->{type} eq "library") { 
	    $path_to_page = $s_diskpath."/_Library/".$file;
	    $url_to_page = $s_url."/_Library/".$file;
	}
	else {
	    $path_to_page = $s_diskpath."/".$file;
	    $url_to_page = $s_url."/".$file;
	}
	########### fixme:  remote sites need new logic here

	# if CMScookie is set, use dynamic pages only, so that we can stay
	# in edit mode while we navigate
	my $CMScookie = &get_cookie("ExSiteCMS");

	# return the page URL
	if ($page->{type} ne "page") {
	    # always return path for libraries and templates
	    return $url_to_page;
	}
	elsif ($page->{publish_method} eq "static" && -e $path_to_page) {
	    # there is a published version to show
	    # use it, unless we are in a CMS preview mode
	    if (! $CMScookie || $static) {
		return $url_to_page;
	    }
	}
	# cannot show a published version - fall back on dynamic
	return "$config{server}{CGIpath}/page.cgi?_id=$page->{page_id}";
	# don't use full path - admin server may not be the same as the
	# public server
	#return "$config{server}{server}$config{server}{CGIpath}/page.cgi?_id=$page->{page_id}";
    }
    return undef;
}

# fetch_all_content_objs - returns array of content objects for this page
#### only returns ones belonging to this page.
#### Might be useful to have a function to return all content objects used
#### by this page, including templated objects, etc.

sub fetch_all_content_objs {
    my $this = shift;
    my @contents = $this->fetch_child("content","page",$this->id(),"content_id");
    my @cobjs;
    foreach my $content (@contents) {
	push @cobjs, new ExSite::Content(content=>$content,use=>"newest");
    }
    return @cobjs;
}

# create a hotlink to edit content data

sub edit_content_link {
    my ($this,$cobj) = @_;
    return if (! $cobj);

    #return "<a href=\"$config{server}->{CGIpath}/CMS.cgi/$this->{page}->{section_id}/$this->{page_id}/$cobj->{content_id}?action=update\" target=\"update\"><img src=\"$config{server}->{HTMLpath}$config{server}->{imgdir}/edit20.gif\" height=20 width=20 border=0 title=\"update $cobj->{name}\"></a>";

    my $path = "$this->{context}{page}{section_id}/$this->{context}{page}{page_id}/$cobj->{context}{content}{content_id}";

    my $name = "$cobj->{context}{content}{name}";

    my $out;

    #my $url = "$config{server}->{CGIpath}/CMS.cgi/$path?";
    #my $url = "$config{server}->{CGIpath}/ctrl-panel.cgi/CMS?path=/$path";

    if ($config{cms}{editmode} eq "easy") {
	$out .= "<!--&CMS(mode=inline&cmscmd=edit_content_link&name=${name}&path=/$path)-->";
    }
    elsif ($config{cms}{editmode} eq "MySite") {
	$out .= "<!--&MySite(ecmd=edit_content_link)-->";
    }
    else {

	my $url = "$config{server}->{CGIpath}/ctrl-panel.cgi/CMS?path=/$path";

	# update content, if user has editorial/design priviledges

	$out .= "<a href=\"${url}&cmscmd=update\"><img src=\"$config{server}->{HTMLpath}/_ExSite/images/plus16.gif\" height=16 width=16 border=0 title=\"update $name\"></a>";

	# manage content, if user has editorial/design priviledges

	$out .= "<a href=\"$url\"><img src=\"$config{server}->{HTMLpath}/_ExSite/images/box16.gif\" height=16 width=16 border=0 title=\"manage $name\"></a>";
    }

=pod
    # approve content, if user has suprvisory priviledges, and content is new

    $out .= "<a href=\"$config{server}->{CGIpath}/CMS.cgi/$path?action=approve\"><img src=\"$config{server}->{HTMLpath}/_ExSite/images/check16.gif\" height=16 width=16 border=0 title=\"approve $cobj->{name}\"></a>";

    # remove/rollback/unapprove content, if user has sufficient priviledges

    $out .= "<a href=\"$config{server}->{CGIpath}/CMS.cgi/$path?action=rollback\"><img src=\"$config{server}->{HTMLpath}/_ExSite/images/x16.gif\" height=16 width=16 border=0 title=\"restore previous $cobj->{name}\"></a>";
=cut
    return $out;
}

# write the page to a file

sub publish {
    my $this = shift;
    my $out;
    #$share{Page} = $this;  # save page pointer for DCDs
    $this->{publish} = 1;  # flag this page as going to disk
    my $url = $this->get_url(1); # get static url
    my $diskpath = $this->diskpath(1); # directory only
    my $page = $this->get;
    if (! $page) { 
	$this->error("Page::publish: no page to publish");
	return "No page to publish!";
    }
    else {
	return "" if ($page->{status} eq "disabled");
	$this->info("Page::publish: publishing page $page->{page_id}");
	$this->set_context;
	my $mkdir = $this->mkdir($diskpath);
	if (! $mkdir) {
	    return $this->show_diagnostics("error","html");
	}
	$out .= $mkdir;
	# First, publish virtual files and external content objects.
	# This allows the page to find the published files and link to them
	# instead of the dynamic versions.
	#### Might be better to save all content objects actually 
	#### inserted into the page during substitute(), and use those.
	my @contents = $this->fetch_all_content_objs;
	foreach my $content (@contents) {
	    if ($content->publishable) {
		$out .= $content->publish($diskpath);
	    }
	}
        if ($page->{type} eq "page" && 
            $page->{access} eq "public" &&
            $page->{publish_method} eq "static" ) {
	    # publish page
	    my $filename = $this->filename;
	    $filename =~ /^([\w-]+.(html?|php))$/;
	    my $secure_filename = $1;
	    ##### do we want to allow publication anywhere on the server?
	    $diskpath =~ /^((\/[\w-]+)+)$/;
	    my $secure_diskpath = $1;
	    if ($secure_filename) {
		#$this->{publish} = 1;  # flag this page as going to disk
		$out .= "<P>Saving page $this->{page_id} to <a href=$url>$secure_filename</a>... ";
		if (open OUTFILE, ">$secure_diskpath/$secure_filename") {
		    # WARNING: tampering with environment to spoof normal 
		    # page generation context
		    $ENV{SCRIPT_NAME} = "$config{server}{CGIpath}/page.cgi";
		    $ENV{QUERY_STRING} = "_id=$page->{page_id}";
		    print OUTFILE $this->expand;
		    close OUTFILE;
		    $out .= "succeeded.</P>\n";
		}
		else {
		    $this->error("Page::publish: publish $secure_filename failed: $!");
		    $out .= "<SPAN CLASS=error>FAILED: $!</SPAN></P>\n";
		    return $out;
		}
	    }
	    else {
		$this->error("Page::publish: invalid filename to publish to: $filename");
		$out .= "<P class=error>Invalid filename to publish to: $filename.</p>\n";
	    }
	}
	elsif ($page->{type} eq "template" && $page->{publish_method} eq "static" ) {
	    # publish precompiled template
	    my $secure_filename = "index.html";	
	    $diskpath =~ /^((\/[\w-]+)+)$/;
	    my $secure_diskpath = $1;
	    #$this->{publish} = 1;  # flag this page as going to disk
	    $out .= "<P>Precompiling template $this->{page_id}... ";
	    if (open OUTFILE, ">$secure_diskpath/$secure_filename") {
		# WARNING: tampering with environment to spoof normal 
		# page generation context
		$ENV{SCRIPT_NAME} = "$config{server}{CGIpath}/page.cgi";
		$ENV{QUERY_STRING} = "_id=$page->{page_id}";
		print OUTFILE $this->expand(expand=>"template");
		close OUTFILE;
		$out .= "succeeded.</P>\n";
	    }
	    else {
		$this->error("Page::publish: publish $secure_filename failed: $!");
		$out .= "<SPAN CLASS=error>FAILED: $!</SPAN></P>\n";
		return $out;
	    }
	}
    }
    return $out;
}

# utility for creating publication directories

sub mkdir {
    my ($this,$dir) = @_;
    my $out = " ";  # dummy return value
    $dir =~ /(\/[\w-\/]+)/;
    my $secure_dir = $1;
    if (! $secure_dir) { 
	$this->error("mkdir: invalid directory: $dir");
	return undef;
    }
    my @path = split(/\//,$secure_dir);
    shift @path; # first one is a dummy
    my $path;
    foreach my $dir (@path) {
	$path .= "/$dir";
	next if (-e "$path");
	$out .= "<P>Creating site directory $dir... ";
	if (! mkdir "$path") {
	    $this->error("mkdir $dir: failed: $!");
	    #$this->error("mkdir $path: failed: $!");
	    return undef;
	}
	else {
	    $this->info("created site directory $dir");
	    $out .= "succeeded.</P>\n";
	}
    }
    return $out;
}

1;


