#------------------------------------------------------------------------
# ExSite::Report
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#----------------------------------------------------------------------------
#
# ExSite reporting package.
#
# Methods:
#   new - constructor
#   primary_column - returns the primary report column
#   title - titles a particular table and/or record
#   show_data - outputs a datum, appropriately formatted
#   report_row - reports the contents of a single record
#   report - reports numerous records at once
#   export_text - exports data in tab-delimited-text format
#
# Queries
#   do_query - fetches the records specified by a query
#   encode_query - converts a query to a printable string
#   decode_query - creates a query from an encoded string
#------------------------------------------------------------------------

package ExSite::Report;
use strict;
use ExSite::Config;
use ExSite::DB;
use ExSite::Misc;
use MIME::Base64;

# globals:
# @ISA : inheritor
# %use_button : db ops image buttons
# $total_records_in_query : number of records found by most recent query
# @month_name, @month_days : basic calendar definitions

use vars qw(@ISA %use_button $total_records_in_query @row_class
	    @month_name @month_days);

@ISA = qw(ExSite::DB);

@row_class = ( "A", "B" );

@month_name = ( "(month)", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" );

@month_days = ( 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );

#------------------------------------------------------------------------
# primary_column : selects the primary display column(s), which is used
#                  for titles & hotlinks.  (Not the same as the
#                  primary key, unless no primary display column has
#                  been defined, in which case the primary key is used.)
#------------------------------------------------------------------------

sub primary_column {
    my ($this,$table) = @_;
    my @primary = ();
    my $choice2 = "";
    my $map = $this->{map}->get_map($table);
    # select column to display - use key column by default
    my $ncol = $this->{map}->get_ncol($table);
    for (my $i = 0; $i < $ncol; $i++) {
	if ($map->{display}->[$i] eq "key") {
	    # this is marked as a key column
	    push @primary, $map->{column}->[$i];
	}
    }
    if (@primary > 0) {
	#
	# If called in array context, return all primary display columns.
	# Otherwise, return only the last.
	#
	if (wantarray) { return @primary; }
	else { return $primary[-1]; }
    }
    # nothing suitable, use the primary key
    return $this->get_key($table);
}

#--------------------------------------------------------------------
# title : returns a title for a particular record
#
# INPUT:
#    - table name
#    - record key
#    - data hash (optional) - use this instead of making a DB query
#--------------------------------------------------------------------

sub title {
    my ($this,$table,$record,$rdata) = @_;
    my $table_title = $this->run_handler("report_title",$table,$record,$rdata);
    return $table_title if (defined $table_title);
    my %data;
    $table_title = $this->{map}->table_title($table);
    return $table_title if (! $record);
    my $pcol = &primary_column($this,$table);
    if (defined $rdata) { %data = %$rdata; }
    else { %data = $this->fetch($table,$record); }
    if (defined $data{$pcol}) {
      return "$table_title \"$data{$pcol}\"";
    }
    else {
      return "$table_title &lt;unknown&gt;";
    }
}

#--------------------------------------------------------------------
# show_data : displays a datum appropriately
#
# INPUT:
#    - table name
#    - column name (or number)
#    - data value
#    - record id (optional, only needed for file datatypes)
#
# OUTPUT:
#    returns HTML formatted data value
#--------------------------------------------------------------------

sub show_data {
    my ($this,$table,$column,$data,$id) = @_;

    # check permissions
    my $access = $this->authorize;
    my $read = $this->{map}->get($table,$column,"read");
    if ($read > $access) {
      return "(no permission to view this data)";
    }

    # use handler, if one is defined
    if ($this->{handler}->{show_data}) {
      my $out = $this->run_handler("show_data",$table,$column,$data,$id);
      if (defined $out) { return $out; }
    }

    return $this->show_data_noauth($table,$column,$data,$id);
}

# show_data_noauth can be called directly to ignore DBmap security

sub show_data_noauth {
    my ($this,$table,$column,$data,$id) = @_;
    return "" if (! $data);

    # get datatype
    $_ = $this->{map}->get($table,$column,"datatype");

    # Prepare data for browser display.
    $data =~ s/&quot;/\"/g; # quote conversion
    if (/^html$/) {
	# neuter script and CMS tags; allow everything else
	$data =~ s/<\/?script>/&lt;\/?script&gt;/g;
	$data =~ s/<!--([$&])/<!-- $1/g;
    }
    elsif (! $config{report}{text}{honor_html}) {
	# neuter all HTML tags
	$data =~ s/</&lt;/g;
	$data =~ s/</&gt;/g;
    }

    if (/^url$/) {
	$data =~ s/^\s+//;
	my $ldata = $data;
	if ($data !~ /^http:\/\//) { $ldata = "http://".$data; }
	return "<A HREF=\"$ldata\" TARGET=\"other\">$data</A>";
    }
    elsif (/^email$/) {
	if ($config{report}{email} eq "obfuscate") {
	    # attempt to hide the email address from spambots
	    my ($user, $domain) = split(/\@/,$data,2);
	    return "\n<SCRIPT LANGUAGE=\"JavaScript\">\n<!-- //\nuser = \"$user\";\ndomain = \"$domain\";\naddr = user + '\@' + domain;\ndocument.write('<a href=\"mailto:' + addr + '\">' + addr + '</a>');\n// -->\n</SCRIPT>\n";
	}
	else {
	    # show the email address in plain sight
	    return "<A HREF=mailto:$data>$data</A>";
	}
    }
    elsif (/^money$/) {
	return "$config{report}{currency_char}$data";
    }
    elsif (/^key[:=]/) {
        # foreign key
	if ($data > 0) {
	  # dereference foreign keys
	  my (undef,$ftable) = split /[:=]/;
	  # look up foreign row
	  my %fdata = $this->fetch($ftable,$data);
	  my $fkey = $this->get_key($ftable);
	  # print primary column of foreign record
	  if ($this->approve("read",$ftable,$fkey)) {
	      # user can link to foreign record
	      my $url = "$config{server}->{CGIpath}/$config{db_ops}{view}{url}";
	      $url .= ($url =~ /\?/) ? "&" : "?";
	      $url .= "tab=$ftable&id=$fdata{$fkey}";
	      return "<A HREF=\"$url\">$fdata{&primary_column($this,$ftable)}</A>";
	  }
	  else {
	      # user can see which foreign record is linked to, but not
	      # the record itself
	      return "$fdata{&primary_column($this,$ftable)}";
	  }
	}
	else {
	  return "n/a";
	}
    }
    elsif (/^date$/) {
        return "" if (!$data);
	my ($year,$mon,$day) = split /-/,$data;
	return "$month_name[$mon] $day, $year";
    }
    elsif (/^time$/) {
        return "" if (!$data);
	my $am;
	my ($hour,$min,undef) = split /:/,$data;
	if ($hour < 12) { $am = 1; }
	elsif ($hour > 12) { $hour -= 12; }
	my $out = "$hour:$min ";
	$out .= $am ? "am" : "pm";
	return $out;
    }
    elsif (/^datetime$/) {
        return "" if (!$data);
	my $am;
	my ($date,$time) = split / /,$data;
	my ($hour,$min,undef) = split /:/,$time;
	if ($hour < 12) { $am = 1; }
	elsif ($hour > 12) { $hour -= 12; }
	my $out = "$hour:$min ";
	$out .= $am ? "am" : "pm";
	my ($year,$mon,$day) = split /-/,$date;
	if ($year > 0) {
	    return "$out, $month_name[$mon] $day, $year";
	}
	else {
	    return "undefined";
	}
    }
    elsif (/^crypt$/) {
	return $this->decrypt_column($data,$table,$id,$column);
    }
    elsif (/^text$/) {
	if ($config{report}{text}{honor_nl}) {
	    $data =~ s/\n/<br>\n/g;       # convert NL to <br>
	}
	else {
	    # insert paragraph tags at any double-linebreaks
	    $data =~ s/\r\n\r\n/<P>\n/g;  # handles CR-LF linebreaks
	    $data =~ s/\n\n/<P>\n/g;      # handles LF linebreaks
	}
	return $data;
    }
    elsif (/^ftext$/) {
	# formatted text
	return "<PRE>\n$data\n</PRE>\n";
    }
    elsif (/^html$/) {
	# no conversion of data
	return $data;
    }
    elsif (/^timestamp$/) {
	if (length $data == 14) {
	    # mysql pseudo-timestamp
	    my $year = substr($data,0,4);
	    my $nmonth = substr($data,4,2);
	    my $month = $nmonth ne "00" ? $month_name[$nmonth] : "00";
	    my $day = substr($data,6,2);
	    my $hour = substr($data,8,2);
	    my $min = substr($data,10,2);
	    return "$year-$month-$day $hour:$min";
	}
	else {
	    # assume a regular Unix timestamp
	    return scalar localtime($data);
	}
    }
    elsif (/^(big)?file$/) {
        # The data value for file fields is the record id.
        # fetch the datum:
        #my %fdata = $this->fetch($table,$data);
	if (!$data) { return ""; }
	my ($name,$content) = split /$config{form}->{sepchar}/,$data,2;
	if (!$name) { $name = "download"; }
	$name =~ s/ /_/g;  # spaces mangle the URL
	# guess the mime-type
	my $mimetype = &MimeType($name);
	# estimate the size
	# (multiply by 6/8 to account for Mime base64 encoding overhead)
	my $size = int(length($content)*6/8/100)/10;

	# show the file itself
	if ($config{show_images} && $mimetype =~ /^image\/(gif|png|jpeg)$/) {
	    return "<IMG SRC=$config{server}->{CGIpath}/getdata.cgi/$name?tab=$table&id=$id&col=$column>";
	    ### no dimensions to image in general case...
	}
	else {
	    # Display a link to download the data.
	    if ($size) {
		return "<A HREF=\"$config{server}->{CGIpath}/getdata.cgi/$name?tab=$table&id=$id&col=$column\">$name</A>&nbsp;($mimetype,&nbsp;$size&nbsp;K)";
	    }
	    else {
		$this->warn("0-length file in $table:$id:$column");
		return "$name";
	    }
	}
    }
    elsif (/^filepath$/) {
        # filepath is the path to a file on the local server
	if (!$data) { return ""; }
	if (-e $data) {
	    my @path = split /\//, $data;
	    # guess the mime-type
	    my $mimetype = &MimeType($path[-1]);
	    my $size = int ((-s $data)/100)/10;
	    # show the file path, type, and size
	    if ($this->authorize >= $config{auth}->{root_auth_level} - 1) {
		return "$data ($mimetype, $size K)";
	    }
	    else {
		return "$path[-1] ($mimetype, $size K)";
	    }
	}
	else {
	    $this->warn("$data: file not found");
	    return "<SPAN CLASS=error>file not found</SPAN>";
	}
    }
    else {
	return $data;
    }
}

#--------------------------------------------------------------------
# report : generic report method
#
# INPUT:
# all input is passed in a control hash
#       table => reference to a list of tables to report on
#       data => reference to a list of datahashes, or to a single matchhash
#       query => a query hash
#       sort => [ sort column(s) ]
#       columns => "brief", "full", or [ RE of column types to display ]
#       max => max number of records to display (default all)
#       start => starting record (default 0)
#       type => report type (from Config.pm report styles)
#       special => { column => tag, ... } column-specific formatting
#
# you can also specify the db ops for the report:
#       all_edit => default edit, copy, delete, and new links
#       link => link each record to dbadm.cgi
#       view => same as link
#       edit => link each record to form.cgi
#       copy => link each record to cp.cgi
#       delete => link each record to rm.cgi
#       new => include a link to append new data
# if set to 1, the default links for the above will be included.
# if set to an URL, that URL will be used in place of the default
#
# OUTPUT:
#    HTML for report
#--------------------------------------------------------------------

sub report {
    my ($this,$opt) = @_;
    my %opt = %$opt;
    my $out;

    $this->info("ExSite::Report::report called");

    # use handler to build the report, if one is defined
    if ($this->{handler}->{report}) {
      $out = &{$this->{handler}->{report}} ($this,$opt);
      if ($out) { return $out; }
    }

    # determine tables to report on
    my ($table,$rlist);
    if (exists $opt{table}) {
	if (! ref $opt{table}) {
	    $table = $opt{table};
	    # reformat for compatibility with queries
	    $opt{table} = [ $table ];
	}
	elsif (ref $opt{table} eq "ARRAY") {
	    # rev.1: only report on first table (ie. no joins)
	    $table = $opt{table}->[0];
	}
	else { 
	    $this->error("report: no table specified");
	    return "<P CLASS=\"error\">Error: no table specified to report on.</P>\n";
	}
    }
    elsif (exists $opt{query}) {
	if (exists $opt{query}->{table}) {
	    $table = $opt{query}->{table}->[0];
	}
	else { 
	    $this->error("report: no table specified");
	    return "<P CLASS=\"error\">Error: no table specified to report on.</P>\n";
	}
    }
    else {
	$this->error("report: no table specified");
	return "<P CLASS=\"error\">Error: no table specified to report on.</P>\n";
    }

    # determine the data source
    if (ref $opt{data} eq "ARRAY") {
	# data has been directly specified; report all records with no paging
	$rlist = $opt{data};
	# sort
	if (exists $opt{sort}) {
	    if (ref $opt{sort} eq "ARRAY") {
		# sort gives a list of columns to sort on
		my @list = $this->sort_rows($table,$rlist,@{$opt{sort}});
		$rlist = \@list;
	    }
	    else {
		# no sort keys; sort on main column
		my @list = $this->sort_rows($table,$rlist,
					    $this->primary_column($table));
		$rlist = \@list;
	    }
	}
    }
    else {
	# use a query to select and page through the data
	if (! exists $opt{query}) {
	    my $page_length = $opt{max} ? $opt{max} : $config{report}->{page_length};
	    $opt{query} = { table=>[$table], start=>0, num=>$page_length, 
			    link=>$opt{link}, edit=>$opt{edit} };
	    if (ref $opt{data} eq "HASH") { $opt{query}->{match} = $opt{data}; }
	    if (ref $opt{sort} eq "ARRAY") {
		$opt{query}->{sort} = $opt{sort};
	    }
	}
	my @data = $this->do_query($opt{query});
	$rlist = \@data;
    }

    if (@$rlist == 0) {
      my $title = $this->{map}->table_title($table,1);
      $out = "<P CLASS=error>No $title found.</P>\n";
      goto ADD_NEW;
    }

    $opt{key} = $this->get_key($table);

    # fetch the columns to appear in this report
    if (! exists $opt{columns}) { $opt{columns} = "brief"; }
    my @colmap;
    if (ref $opt{columns} eq "LIST") {
      # columns to report are provided by the calling agent
      foreach my $col (@{$opt{columns}}) {
	my %colmap = $this->{map}->get_column($table,$col);
	push @colmap, \%colmap;
      }
    }
    else {
      # select all columns of type $opt{columns}
      my $ncol = $this->{map}->get_ncol($table);
      for (my $icol = 0; $icol < $ncol; $icol++) {
	my %colmap = $this->{map}->get_column($table,$icol);
	# "key" columns are _ALWAYS_ included
	if (($colmap{display} eq "key") || 
	    ($colmap{display} =~ /$opt{columns}/)) {
	  push @colmap, \%colmap;
	}
	elsif ($opt{type} eq "full" && $colmap{display} eq "full") {
	  push @colmap, \%colmap;
        }
      }
      if (@colmap == 0) {
	  $this->error("report: no columns in $table can be displayed");
	  return "<P CLASS=error>No columns to display!</P>";
      }
    }

    # report type
    if (! $opt{type}) {
      # check table map for default
      $opt{type} = $this->{map}->table($table,"report");
      if (! $opt{type}) { $opt{type} = $config{report}->{style}; }
    }

    # save default sort key
    my ($default_sort_key, $desc);
    if (!$opt{query}->{sort}){$opt{query}->{sort} = [ $this->primary_column($table) ];}
    $default_sort_key = $opt{query}->{sort};
    if (substr($default_sort_key->[0],0,1) ne "-"){$desc=1;}else{$desc=0;}
    
    # report header

    my $reporturl = $config{db_ops}{report}{url};
    $reporturl .= $config{db_ops}{report}{url} =~ /\?/ ? "&" : "?";
    if ($reporturl !~ /^(http|\/)/) {
	$reporturl = "$config{server}{CGIpath}/$reporturl";
    }
    my $re_sort_columns = exists $opt{re_sort} ? $opt{re_sort} : 1;
    $out .= $config{report}->{report_tag}->{$opt{type}}->[0];
    if ($opt{type} eq "table") {
	# Table reports have special requirements - in particular, we need to 
	# output a header row.
	my $stat = $this->run_handler("report_table_header",$opt);
	if ($stat) { 
	    $out .= $stat;
	}
	else {
	    $out .= "\n<TR>";
	    # control options
	    if ($config{report}->{link_position} eq "before") {
		if ($opt{link}) {
		    $out .= "<TH>Options</TH>";
		}
	    }
	    foreach my $column (@colmap) {
		if (exists $opt{query} && $re_sort_columns) {
		    # re-sort option only available on queries
		    my $newquery = $opt{query};
		    my $colname = ($desc && $default_sort_key->[0] eq $column->{column}) ? "-".$column->{column} : $column->{column};
		    $newquery->{sort} = [ $colname ];
		    my $encoded_query = $this->encode_query($newquery);

		    $out .= "<TH><A HREF=\"${reporturl}query=$encoded_query&link=$opt{link}&all_edit=$opt{all_edit}\">$column->{label} ";
		    if ($desc && ($default_sort_key->[0] eq $column->{column})){$out .= " +";}
		    elsif ($default_sort_key->[0] eq "-".$column->{column}){$out .= " -";}
		    $out .= "</A></TH>";
		}
		else {
		    # fixed report format
		    $out .= "<TH>$column->{label}</TH>";
		}
	    }
	    # reset sort key
	    if (exists $opt{query}) {
		if ($default_sort_key) {
		    $opt{query}->{sort} = $default_sort_key;
		}
		else {
		    delete $opt{query}->{sort};
		}
	    }
	    # control options
	    if ($config{report}->{link_position} ne "before") {
		if ($opt{link}) {
		    $out .= "<TH>Options</TH>";
		}
	    }
	    $out .= "</TR>\n";
	}
    }

    # report data

    my $key = $this->get_key($table);
    my $row_handler = "report_$opt{type}_format";
    my $nrow = 0;
    foreach my $rdata (@$rlist) {
	# use handler to format the row, if one is defined
	my $hout = $this->run_handler($row_handler,$table,$rdata,%opt);
	if ($hout) {
	    $out .= $hout;
	}
	# use a template to format the row, if one is defined
	elsif (my $template = $this->get_template($table)) {
	    $out .= &substitute($template,$rdata);
	}
	# otherwise, use generic formatting rules
	else {
	    $nrow++;
	    if ($opt{type} eq "table" && $config{highlight_alternate_lines}) {
		# special case of table row tags
		my $class = $nrow % 2;
		$out .= "<TR CLASS=\"$row_class[$class]\">";
	    }
	    else {
		$out .= $config{report}->{row_tag}->{$opt{type}}->[0];
	    }
	    # extra functions - display as if another column
	    if ($config{report}->{link_position} eq "before") {
		if ($opt{link} || $opt{all_edit} || $opt{edit} ||
		    $opt{copy} || $opt{delete} || $opt{view}) {
		    $out .= $config{report}->{column_tag}->{$opt{type}}->[0];
		    $out .= &report_links($this,$table,$rdata->{$key},%opt);
		    $out .= $config{report}->{column_tag}->{$opt{type}}->[1];
		}
	    }
	    foreach my $col (@colmap) {
		$out .= $config{report}->{column_tag}->{$opt{type}}->[0];
		if ($config{report}->{label_tag}->{$opt{type}}) {
		    $out .= "$col->{label}$config{report}->{label_tag}->{$opt{type}}";
		}
		if ($opt{special}->{$col}) { $out .= "<$opt{special}->{$col}>"; }
		$out .= show_data($this,$table,$col->{column},
				  $rdata->{$col->{column}},$rdata->{$key});
		if ($opt{special}->{$col}) { $out .= "</$opt{special}->{$col}>"; }
		$out .= $config{report}->{column_tag}->{$opt{type}}->[1];
	    }
	    # extra functions - display as if another column
	    if ($config{report}->{link_position} ne "before") {
		if ($opt{link} || $opt{all_edit} || $opt{edit} ||
		    $opt{copy} || $opt{delete} || $opt{view}) {
		    $out .= $config{report}->{column_tag}->{$opt{type}}->[0];
		    $out .= &report_links($this,$table,$rdata->{$key},%opt);
		    $out .= $config{report}->{column_tag}->{$opt{type}}->[1];
		}
	    }
	    $out .= $config{report}->{row_tag}->{$opt{type}}->[1];
	    $out .= "\n";
	}
    }

    # report footer

    $out .= $config{report}->{report_tag}->{$opt{type}}->[1];

    # now output paging links

    my $linkargs = $this->make_link_args(%opt);
    if (exists $opt{query}) {
	my $qout;
	my $query = $opt{query};
	my $start = $query->{start} + 1;
	my $end = $query->{start}+$query->{num} < $total_records_in_query ?
	    $query->{start}+$query->{num} : $total_records_in_query;
	if ($start != 1 || $end != $total_records_in_query) {
	    $qout .= "Records $start-$end of $total_records_in_query.<br>\nPage: ";
	    for (my $irec = 0; $irec <= $total_records_in_query; $irec += $config{report}->{page_length}) {
		my $ipage = int($irec / $config{report}->{page_length}) + 1;
		if ($irec == $start - 1) {
		    $qout .= "<FONT COLOR=#999999>$ipage</FONT> ";
		}
		else {
		    $query->{start} = $irec;
		    my $q = $this->encode_query($query);
		    $qout .= "<A HREF=\"${reporturl}query=$q$linkargs\">$ipage</A> ";
		}
	    }
	}
	# reset query to original state
	$query->{start} = $start - 1;

	if ($query->{start} > 0) {
	    # previous link
	    $query->{start} -= $query->{num};
	    if ($query->{start} < 0) { $query->{start} = 0; } 
	    my $q = $this->encode_query($query);
	    $qout .= "[<A HREF=\"${reporturl}query=$q$linkargs\">previous $query->{num}</A>]\n";
	    $query->{start} += $query->{num};
	}
	if ($total_records_in_query > $end) {
	    # might be more data
	    ### FIXME: on last page, "next N" should give number of remaining records
	    $query->{start} += $query->{num};
	    my $q = $this->encode_query($query);
	    $qout .= "[<A HREF=\"${reporturl}query=$q$linkargs\">next $query->{num}</A>] ";
	    $query->{start} -= $query->{num};
	}
	if ($qout) { $out .= "$qout<br>\n"; }
    }

    # finally, the optional insert link

  ADD_NEW:
    $out .= $this->report_links($table,undef,%opt);
    return $out;

    # done!

    return $out;
}

#--------------------------------------------------------------------
# report_row : display a single row in verbose format
#
# INPUT:
#    1) table name
#    2) a record id, or a reference to a datahash
#    3) option hash:
#       edit => if true, include editing links
#
# OUTPUT:
#    HTML for report
#--------------------------------------------------------------------

sub report_row {
    my ($this,$table,$row,%opt) = @_;

    $this->info("ExSite::Report::report_row called");

    # fetch data for this row
    my %data;
    if (ref $row eq "HASH") {
      # the data has been passed directly
      %data = %$row;
    }
    else {
      # assume the data has been passed as a record id
      %data = $this->fetch($table,$row);
    }
    #my $key = $this->get_key($table);

    # no "add new" link for report_row
    $opt{new} = 0;

    my $out = $this->report({ table=>$table, data=>[\%data], columns=>"full|brief",
			   type=>"full", %opt });

    #$out .= &report_links($this,$table,$data{$key},0,(all_edit=>$opt{all_edit}));
    return $out;

}

#--------------------------------------------------------------------
# report_db : display top-level view of database
#
# Displays the database tables, and the allowed operations on each.
#
# INPUT:
#    1) option hash:
#       class=> select sub-category of tables to show
#       count=> display record counts, if true
#       title=> include title/intro text
#       [op] => db_ops settings
#
# OUTPUT:
#    HTML for report
#--------------------------------------------------------------------

sub report_db {
    my ($this,$opt) = @_;

    $this->info("ExSite::Report::report_db called");

    my $title;
    my @table = $this->{map}->get_mapped_tables;

    my $access = $this->authorize;
    my $out;
    if ($opt->{title} == 1) {
	$out = "<H3>Database Administration Page</H3>\n<P>Select a table to operate on:</P>\n";
    }
    elsif ($opt->{title}) {
	$out = $opt->{title};
    }
    $out .= "<TABLE CLASS=report>\n<TR>";
    if ($config{report}->{link_position} eq "before") {
	$out .= "<TH>Operations</TH>";
    }
    $out .= "<TH>Table</TH>";
    if ($opt->{count}) {
	$out .= "<TH>Record Count</TH>";
    }
    if ($config{report}->{link_position} ne "before") {
	$out .= "<TH>Operations</TH>";
    }
    $out .= "</TR>\n";
    foreach my $tab (sort @table) {
	my $table_par = $this->{map}->get_table($tab);
	if ($access >= $table_par->{read}) {
	    if (! $opt->{class} || $table_par->{class} =~/$opt->{class}/) {
		$title = $this->{map}->table_title($tab,1);
		$out .= "<TR>";
		if ($config{report}->{link_position} eq "before") {
		    $out .= "<TD>";
		    $out .= $this->report_links($tab,undef,%$opt);
		    $out .= "</TD>";
		}
		my $tablename = $table_par->{alias} ne $tab ? "alias->$table_par->{alias}" : $tab;
		$out .= "<TD><DL CLASS=report_db><DT><B>$title:</b> <I>$tablename</I></DT>\n";
		if (exists $table_par->{description}) {
		    $out .= "<DD>$table_par->{description}</DD>\n";
		}
		$out .= "</DL></TD>";
		if ($opt->{count}) {
		    # not implemented yet
		    $out .= "<TD>n/a</TD>";
		}
		if ($config{report}->{link_position} ne "before") {
		    $out .= "<TD>";
		    $out .= $this->report_links($tab,undef,%$opt);
		    $out .= "</TD>";
		}
		$out .= "</TR>\n";
	    }
	}
    }
    $out .= "</TABLE>\n";
    return $out;
}

#--------------------------------------------------------------------
# report_relational : displays rows and their relations
#
# INPUT:
#    1) option hash:
#       tab,table=>the primary table to report on
#           - if none specified, a 
#       id=>the key of a particular record in the above table;
#           - without this parameter, members of the table are listed
#           for selection.
#           - with this parameter, the associated members of the
#           subtables are listed, with options to edit, delete, or add
#           to the table.
#       stab=>the subtable(s) to manage;
#           selects which subtables to list;  by default all are shown.
#       ptab=>the parent table  } must both be specified
#       pid=>the parent id      } to have a valid meaning
#           - selects which members of the table are listed;  by default
#           all members belonging to the account are listed.
#       all remaining options are passed to the other report() methods.
#
# OUTPUT:
#    HTML for report
#--------------------------------------------------------------------

sub report_relational {
    my ($this,$opt) = @_;

    $this->info("ExSite::Report::report_relational called");

    my ($out,$alias);
    if (! exists $opt->{tab} && exists $opt->{table}) { 
	$opt->{tab} = $opt->{table}; 
    }
    $alias = $this->{map}->get_table_attr($opt->{tab},"alias");
    if (! $alias) { $alias = $opt->{tab}; }
    if (!$opt->{tab}) {
	return $this->report_db($opt);
    }
    elsif (! $this->approve("read",$opt->{tab})) {
	return "<H3 CLASS=error>Problem</H3>You do not have permission for this operation.<P>\n";
    }
    elsif (defined $opt->{id}) {
	# a specific record has been selected; show children of this record
	if ($opt->{title} == 1) {
	    $out .= "<H2>".$this->title($opt->{tab},$opt->{id})."</H2>";
	}
	elsif ($opt->{title}) {
	    $out .= $opt->{title};
	}

	my $option={"all_edit"=>1};
	if ($opt->{reply}) {
	    $option= {%$option, "reply"=>$opt->{reply}};
	}
	$out .= $this->report_row($opt->{tab},$opt->{id},%$option);
	$out .= "<P>\n";
	
	# get related tables to report on
	#my @children = $this->{map}->get_children($opt->{tab});
	my @children = $this->{map}->get_children($alias);
	if (defined $opt->{stab}) {
	    # a specific sub table (child table) has been requested
	    @children = grep(/^$opt->{stab}$/,@children);
	}
	elsif (exists $opt->{stab}) {
	    # do not show any subtables
	    @children = ();
	}
	
	foreach my $child (@children) {
	    #y $refcol = $this->{map}->get_reference_column($child,$opt->{tab});
	    my $refcol = $this->{map}->get_reference_column($child,$alias);
	    my %match;
	    $match{$refcol} = $opt->{id};
	    
	    $out .= "<h3>".$this->{map}->table_title($child,1)." for ".
	    $this->title($opt->{tab},$opt->{id})."</h3>\n";
	    my $newurl;  # = "$config{server}->{CGIpath}/$config{db_ops}->{new}->{url}?tab=$child&ptab=$opt->{tab}&pid=$opt->{id}";
	    if ($config{report}->{linkmode} eq "exsite") {
		#$newurl = "page.cgi?_id=".$share{Page}->id."&action=new&tab=$child&ptab=$opt->{tab}&pid=$opt->{id}";
		$newurl = &relink(action=>"new",tab=>$child,ptab=>$opt->{tab},pid=>$opt->{id});
	    }
	    elsif ($config{report}->{linkmode} eq "cgi") {
		$newurl = "$config{server}->{CGIpath}/$config{db_ops}->{new}->{url}";
		$newurl .= (($newurl =~ /\?/) ? "&" : "?");
		$newurl .= "tab=$child&ptab=$opt->{tab}&pid=$opt->{id}";
	    }
	    $out .= $this->report({table=>$child, data=>\%match, link=>1,all_edit=>1,new=>$newurl,sort=>1});
	    
	}
    }
    else {
	# no specific record has been selected; list valid records
	my $title = $this->{map}->table_title($opt->{tab},1);
	my $newurl;
	if ($config{report}->{linkmode} eq "cgi") {
	    $newurl = "$config{server}{CGIpath}/$config{db_ops}{new}{url}";
	}
	else {
	    #$newurl = "$config{server}->{CGIpath}/page.cgi?_id=".$share{Page}->id."&action=new";
	    $newurl = &relink(action=>"new");
	}
	$newurl .= (($newurl =~ /\?/) ? "&" : "?");
	$newurl .= "tab=$opt->{tab}&ptab=$opt->{ptab}&pid=$opt->{pid}";
	if (! exists $opt->{pid}) {
	    # show all rows
	    $out .= "<h2>$title</h2>\n";
	    $out .= $this->report({table=>$opt->{tab},link=>1,all_edit=>1,sort=>1,new=>$newurl});
	}
	else {
	    # show only those rows belonging to the parent
	    my %data = $this->fetch($opt->{ptab},$opt->{pid});
	    my $keycol = $this->primary_column($opt->{ptab});
	    $out .= "<h2>$title for \"$data{$keycol}\"</h1>\n";
	    my $refcol = $this->{map}->get_reference_column($opt->{tab},$opt->{ptab});
	    my %match = ( $refcol => $opt->{pid} );
	    $out .= $this->report({table=>$opt->{tab}, data=>\%match, link=>1,all_edit=>1,sort=>1,new=>$newurl});
	}
    }

    return $out;
}

#--------------------------------------------------------------------
# report_links : display data links in a report
#
# This method can be invoked on specific records or entire tables;
# not all ops are valid in both cases (only view and new are valid
# for entire tables).
#
# %opt contains the following flags:
#       all_edit => default edit, copy, and delete links
#       link => link each record to report_row
#       view => same as link
#       edit => link each record to edit
#       copy => link each record to cp.cgi
#       delete => link each record to rm.cgi
#       new => include a link to append new data
#       list => include a link to append new data
#       search => include a link to append new data
#       linktype => text/img
# if set to 1, the default links for the above will be included.
# if set to an URL, that URL will be used in place of the default
#
# table, id = the table and record id to link to
# link? = link to report_row for verbose record description
# edit? = display editing links (edit, copy, delete) in addition to
#        a view link
#--------------------------------------------------------------------

sub report_links {
    my ($this,$table,$id,%opt) = @_;

    # check for handler to manage all links
    my $out = $this->run_handler("report_links",$table,$id,%opt);
    return $out if (defined $out);
    
    my %opflag;

    while (my ($key,$val) = each %opt) {
	if (grep /$key/, keys %{$config{db_ops_grps}}) {
	    foreach my $op (@{$config{db_ops_grps}->{$key}}) {
		if (! defined $opt{$op}) { $opt{$op} = $val; }
	    }
	}
    }

    $this->check_for_buttons;
    my $query = "tab=$table";
    if ($id) { $query .= "&id=$id"; }
    my %db_ops = %{$config{db_ops}};

    # std ops
    foreach my $op (@{$config{db_ops_order}}) {
	# check if this op was requested
	if ($opt{$op}) {
	    # check if this op is valid
	    if ($id xor $config{db_ops}->{$op}->{table} == 1) {
		# check if this op is permitted
		if ($this->approve($db_ops{$op}->{approve},$table,$id)) {
		    my $linkurl;
		    my $cgiprog = undef;
		    if ($opt{$op} && $opt{$op} != 1) { 
			$cgiprog = $opt{$op}; 
			if ($cgiprog !~ /\?.+/) { 
			    # in case custom URL has predefined query
			    $query = undef; 
			}
		    }
		    $out .= $this->link($op,$query,$cgiprog);
		}
	    }
	}
    }

    # check for handler to add extra links
    $out .= $this->run_handler("report_links_extra",$table,$id,%opt);
    
    return $out;
}

#--------------------------------------------------------------------
# make_link_args : encode the dbops parameters as an URL argument
#--------------------------------------------------------------------

sub make_link_args {
  my ($this, %opt ) = @_;
  my @args = ( "" );
  if ($opt{link}) { push @args, "link=$opt{link}"; }
  if ($opt{view}) { push @args, "view=$opt{view}"; }
  if ($opt{all_edit}) { push @args, "all_edit=$opt{all_edit}"; }
  if ($opt{edit}) { push @args, "edit=$opt{edit}"; }
  if ($opt{copy}) { push @args, "copy=$opt{copy}"; }
  if ($opt{delete}) { push @args, "delete=$opt{delete}"; }
  if ($opt{new}) { push @args, "new=$opt{new}"; }
  foreach my $arg (@args) {
      # escape argument separators
      $arg =~ s/\&/\%26/g;
  }
  return join("&",@args);
}

# look for graphic buttons

sub check_for_buttons {
    my $this = shift;
    return if scalar(keys %use_button) > 0;  # already done
    my %db_ops = %{$config{db_ops}};
    my $imgpath = "$config{server}->{HTMLroot}$config{server}->{HTMLpath}/_Modules/WebDB";
    foreach my $op (keys %db_ops) {
	$use_button{$op} = (-e "$imgpath/$db_ops{$op}->{img}");
    }
    return;
}

#--------------------------------------------------------------------
# link : return an HTML link for a db operation
#
# op = db operation
# arg = explicit query string, or hash ref to arg list
# script = optional script to use instead of default db_ops url
#
# Links can be returned in 2 formats:
# "cgi" => link directly to CGI programs to perform the db ops
# "exsite" => link to WebDB ExSite dynamic content driver module
#--------------------------------------------------------------------

sub link {
    my ($this,$op,$arg,$script) = @_;
    my $out;
    my $query = ref $arg eq "HASH" ? &EncodeHash(%$arg) : $arg;
    my $default_script;
    if ($config{report}->{linkmode} eq "exsite") {
	#$default_script = "page.cgi?_id=".$share{Page}->id."&action=$op&$query";
	$default_script = &relink(action=>$op);
    }
    elsif ($config{report}->{linkmode} eq "cgi") {
	$default_script = $config{db_ops}->{$op}->{url};
    }

    if (! $script) { $script = $default_script; }
    if ($script !~ /^\//) { $script = "$config{server}->{CGIpath}/$script"; }
    my $url = $script;
    if ($url !~ /\?/) { 
	# add the query if no query is defined
	$url .= "?$query";
    }
    elsif ($url !~ /[&?]$query&?/) { 
	# add the query if no matching query parameter is found
	$url .= "&$query";
    }
    $this->check_for_buttons;
    my %db_ops = %{$config{db_ops}};
    
    # check for graphic buttons

    if ($use_button{$op}) {
	# use graphic button
	$out = "<A HREF=$url><IMG SRC=$config{server}->{HTMLpath}/_Modules/WebDB/$db_ops{$op}->{img} HEIGHT=$db_ops{$op}->{img_height} WIDTH=$db_ops{$op}->{img_width} BORDER=0 ALT=\"$db_ops{$op}->{label}\" TITLE=\"$db_ops{$op}->{label}\"></A>&nbsp;";
    }
    else {
	# use text link
	$out = "[&nbsp;<A HREF=$url>$db_ops{$op}->{label}</A>&nbsp;]\n";
    }
    return $out;
}

#--------------------------------------------------------------------
# export : exports data in various formats
#
# INPUT:  a hash reference, with the following keys:
#   table=>the table to export
#   column=>ref to a list of columns to export
#   format=>"text" or "excel" (default is text)
#   outfile=>output filename or "-" for stdout (default is stdout)
###   convert_keys=>if true, convert foreign keys to labels (unsupported)
#
# by default, all records are exported, unless one of the following
# record selection options is given:
#
#   data=>ref to a list of datahashes of preselected records
#   match=>ref to a match hash to select records
#   owner=>export all records owned by this UID
#   group=>export all records owned by this GID
#--------------------------------------------------------------------

sub export {
    my ($this,$arg) = @_;
    my $table = $arg->{table};
    return "" if (! $table);

    my $out;
    my $access = $this->authorize;
    if (! $access) {
	$this->warn("export: permission denied");
	return "";
    }

    # default to text format
    if (!exists $arg->{format}) { $arg->{format} = "text"; }

    # default to stdout
    if (!exists $arg->{outfile}) { $arg->{outfile} = "-"; }

    # fetch records to appear in report
    my $rlist;
    if ($arg->{match}) {
	# select matching records
	$rlist = $this->fetch_match($table,$arg->{match});
    }
    elsif ($arg->{data}) {
	# preselected data
	$rlist = $arg->{data};
    }
    elsif ($arg->{owner}) {
	# select records by owner
	my $all = $this->fetch_all($table);
	foreach my $rec (@$all) {
	    if ($this->find_owner($table,$rec,"user") == $arg->{owner}) {
		push @$rlist,$rec;
	    }
	}
    }
    elsif ($arg->{group}) {
	# select records by group
	my $all = $this->fetch_all($table);
	foreach my $rec (@$all) {
	    if ($this->find_owner($table,$rec,"group") == $arg->{group}) {
		push @$rlist,$rec;
	    }
	}
    }
    else {
	# select everything
	$rlist = $this->fetch_all($table);
    }

    # report columns or labels
    my $header = $arg->{header}? $arg->{header} : "column";

    # fetch columns to appear in report
    my (@column,@header);
    if (! defined $arg->{column}) {
        # report all readable columns by default
	my $ncol = $this->{map}->get_ncol($table);
	for (my $i = 0; $i <= $ncol; $i++) {
	    if ($this->{map}->get($table,$i,"read") <= $access) {
	        push @column,$this->{map}->get($table,$i,"column");
	        push @header,$this->{map}->get($table,$i,$header);
	    }
	}
    }
    else {
	# report only the requested columns
	@column = @{$arg->{column}};
	foreach (@column) {
	    push @header, $this->{map}->get($table,$_,$header);
	}
    }

    # export the data

    if ($arg->{format} eq "text") {
	# print column headings
	my $out = join("\t",@header)."\r\n";
	# print data
	my @data;
	foreach my $dhash (@$rlist) {
	    foreach my $col (@column) {
		# escape control characters
		$dhash->{$col} =~ s/\%/\%37/g;  # %
		$dhash->{$col} =~ s/\n/\%0A/g;  # newline
		$dhash->{$col} =~ s/\r/\%0D/g;  # CR
		$dhash->{$col} =~ s/\t/\%09/g;  # tab
		push @data,$dhash->{$col};
	    }
	    $out .= join("\t",@data)."\r\n";
	    @data = ();
	}
	if ($arg->{outfile} ne "-") {
	    open F,">$arg->{outfile}";
	    print F $out;
	    close F;
	}
	else {
	    print $out;
	}
    }
    elsif ($arg->{format} eq "excel") {
	my @Colarray;
	foreach my $column (@column) {
	    my $head = shift @header;
	    push @Colarray, "$head;$column";
	}
	require ExSite::Excel;
	&ExSite::Excel::ExcelExport($rlist,$arg->{outfile},\@Colarray);
    }
}

#--------------------------------------------------------------------
# export_text : exports data in tab-delimited text format
# (for backwards compatibility only)
#
# INPUT:
#    1) table name
#    2) match hash
#    3) reference to list of columns to report
#       (if null, display all reportable columns)
#
# OUTPUT:
#    returns the text report in a string
#--------------------------------------------------------------------

sub export_text {
    my ($this,$table,$rmatch,$rshow) = @_;
    return $this->export({table=>$table,format=>"text",column=>$rshow,match=>$rmatch});
}

#--------------------------------------------------------------------
# import : imports data in delimited text format
#
# INPUT:  a hash reference, with the following keys:
#   table=>the table to import into
#   delim=>the field delimiter
#   file=>input file name
#   method=>insert|update
#--------------------------------------------------------------------

sub import {
    my ($this,$arg) = @_;
    my $table = $arg->{table};
    return 0 if (! $table);
    my $access = $this->authorize;
    return 0 if (! $access);
    my $file = $arg->{file};
    return 0 if (! -e $file);
    my $delim = $arg->{delim} || "\t";
    my $method = $arg->{method};

    my $out;
    open F, "<$file" or return 0;
    my $hdr = <F>;
    chomp $hdr;
    my @hdr = split(/$delim/,$hdr);
    my %data;
    while (<F>) {
	chomp;
	my @data = split /\t/;
	foreach my $key (@hdr) {
	    $data{$key} = shift @data;
	}
	my $stat = $this->run_handler("import_record",\%data);
	# return value can be ignored;  \%data is unchanged if no
	# handler is present
	if ($method eq "update") {
	    $this->update($table,\%data);
	}
	else {
	    $this->insert($table,\%data);
	}
    }
    close F;
    return 1;
}		

#------------------------------------------------------------------------
# Queries
#
# A query is a reference to a hash specifying a set of records:
#
# table=>the table the records come from
# match=>ref to a match hash to select records from the table
# sort=>ref to a list of columns to sort on
# start=>first record to display
# num=>number of records to display
#
#   do_query - fetches the records specified by a query
#   sizeof_query - return total number of records in last query
#   encode_query - converts a query to a printable string
#   decode_query - creates a query from an encoded string
#------------------------------------------------------------------------

sub do_query {
    my ($this,$query) = @_;
    if (! $query->{table}) { return (); };
    if (! $query->{start}) { $query->{start} = 0; }
    if (! $query->{num}) { $query->{num} = $config{report}->{page_length}; }
    my $data;
    if (exists $query->{match}) {
	# select matching data
        $data = $this->fetch_match($query->{table}->[0],$query->{match});
    }
    else {
	# no match condition; fetch everything!!!  (use with caution)
	$data = $this->fetch_all($query->{table}->[0]);
    }
    my @data;
    if (ref $query->{sort} eq "ARRAY") {
	# sort requested
	@data = $this->sort_rows($query->{table}->[0],$data,@{$query->{sort}});
    }
    else {
	# no sort
	#@data = @$data;
	# sort on default key - primary display column
	@data = $this->sort_rows($query->{table}->[0],$data,
				 &primary_column($this,$query->{table}->[0]));
    }
    # save total number of records as a package global for later reference
    $total_records_in_query = scalar @data;
    my $start = $query->{start};
    my $end = $query->{start}+$query->{num}-1 < $total_records_in_query - 1 ?
	$query->{start}+$query->{num}-1 : $total_records_in_query - 1;

    return @data[$start..$end];
}

sub sizeof_query {
    return $total_records_in_query;
}

sub encode_query {
    my ($this,$query) = @_;
    my %qhash = %$query;
    if (exists $query->{match}) {
	$qhash{match} = encode_base64(&EncodeHash(%{$query->{match}}),"");
    }
    if (exists $query->{sort}) {
	$qhash{sort} = join ",",@{$query->{sort}};
    }
    my $qurl = &EncodeHash(%qhash);
    return encode_base64($qurl,"");
}

sub decode_query {
    my ($this,$qstr) = @_;
    my $qurl = decode_base64($qstr);
    my %qhash = &DecodeString($qurl);
    if (exists $qhash{match}) {
	my $match_url = decode_base64($qhash{match});
	my %match = &DecodeString($match_url);
	$qhash{match} = \%match;
    }
    if (exists $qhash{sort}) {
	my @sort = split /,/, $qhash{sort};
	$qhash{sort} = \@sort;
    }
    if (exists $qhash{table}) {
	my @tables = split /,/, $qhash{table};
	$qhash{table} = \@tables;
    }
    return \%qhash;
}

#------------------------------------------------------------------------
# Templates
#
# A report template is a string with marker tags to accept data from
# an arbitrary datahash.
#------------------------------------------------------------------------

sub set_template {
    my ($this,$table,$template) = @_;
    if (! $this->{report_template}) {
	$this->{report_template} = {};
    }
    $this->{report_template}{$table} = $template;
    return;
}

sub get_template {
    my ($this,$table) = @_;
    if (ref $this->{report_template} eq "HASH") {
	return $this->{report_template}{$table};
    }
    return undef;
}

1;
