package Modules::CMS;
#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------

# CMS : content-management toolkit
#
# The CMS package provides tools for creating, configuring, updating, 
# and deleting websites, pages, templates, libraries, content objects,
# and content versions.


use strict;
use ExSite::Config;
use ExSite::Form;
use ExSite::Page;
use ExSite::Content;
use ExSite::Misc;
use ExSite::Input;
use ExSite::Image;
use ExSite::Wizard;
use MIME::Base64;
use Modules::BaseDCD;

use vars qw(@ISA);
@ISA = qw(Modules::BaseDCD);

use vars qw($db %query @pathtable @pathcol $section $page $content $cdata 
	    $nonactions $cms);

@pathtable = ( "section", "page", "content", "content_data" );
@pathcol = ( "title", "title", "name", "content_data_id" );

# actions that are implicitly processed, instead of being passed to DoAction
$nonactions = "map";

# new : called implicitly once per page when a DCM is invoked
#
# mode=0 : run as a DCD for inlining into pages (generates pure HTML snippet)
# mode=1 : run as a stand-alone control panel (generates complete HTML page
#          with CBE javascript menus)

sub new {
    my $this = shift;
    my $obj = {};
    $obj->{mode} = shift;
    my $class = ref($this) || $this;
    bless $obj, $class;
    $obj->read_conf;
    $obj->read;
    $cms = $obj;      # kludge for wizards to access the CMS object
    
    # register some CMS-specific handlers

    $share{DB}->handler("select_foreign_key",\&select_foreign_key_handler);

    return $obj;
}

sub read {
    my ($this,$opt) = @_;
    # fetch all inputs
    $this->{Input} = new ExSite::Input;
    $this->{input} = $this->{Input}->combine;         # combined input
    $this->{query} = $this->{Input}->query("data");   # query string
    $this->{formdata} = $this->{Input}->post;         # posted form data
    # path info
    my @path = split /\//, $this->{input}{path};
    $this->{path} = \@path;
    $this->{section} = $this->{path}[1];
    $this->{page} = $this->{path}[2];
    $this->{content} = $this->{path}[3];
    $this->{cdata} = $this->{path}[4];
}

sub write {
    my ($this,$options) = @_;

    my $db = $share{DB};
    my $out;

    my $min_access = $config{CMS}{access} || 2;
    my $access = $this->authorize;
    my %option; 
    if ($options) { %option = &DecodeString($options); }
    if ($option{mode} eq "inline" && $config{CMS}{allow_inline}) {
	$out .= $this->inline_write($options);
    }
    elsif ($access >= $min_access) {
	my $action = $option{cmscmd} || $this->{query}{cmscmd};
	$this->{topmenu} = 1;
	if ($action && $action !~ /^$nonactions$/) {
	    $out .= $this->DoAction($action);
	}
	elsif ($this->{cdata}) {
	    $out .= $this->ManageContentData($this->{cdata});
	}
	elsif ($this->{content}) {
	    $out .= $this->ManageContent($this->{content});
	}
	elsif ($this->{page}) {
	    $out .= $this->ManagePage($this->{page});
	}
	elsif (defined $this->{section}) {  # $section might be 0
	    $out .= $this->ManageSection($this->{section});
	}
	else {
	    $out .= $this->ManageAllSections;
	}
	if ($this->{topmenu}) {
	    $out = $this->TopMenu.$out;
	}
    }
    else {
	my $err = $this->show_diagnostics("error","html");
	$out .= $err || "<P CLASS=error>Permission denied.</P>\n";
    }
    
    if ($db->{sql_error}) {
	$out .= "<hr>\n".$db->{sql_error};
    }

    #$out .= "</div>\n</td></tr></table>\n";

    return $out;

}

sub authorize {
    my $this = shift;
    my $db = $share{DB};
    my $access = $db->authorize;

    if ($config{CMS}{strict}) {

	# Strict security checks have been requested, so we validate
	# ownership of every item in the content stack.

	my ($s,$p,$c,$v,$content_handle);
	if ($this->{section}) {
	    # check for legal site
	    my @sites = $db->my_sites;
	    while ($s = pop @sites) {
		last if ($this->{section} == $s->{section_id});
	    }
	    # section mismatch ?
	    if ($this->{section} != $s->{section_id}) {
		$this->error("Permission denied: invalid site");
		return 0;
	    }
	}
	
	if ($this->{page}) {
	    $p = $db->fetch("page",$this->{page});
	    if ($p->{section_id} != $this->{section}) {
		$this->error("Permission denied: invalid page");
		return 0;
	    }
	}

	if ($this->{content}) {
	    if ($p) {
		$c = $db->fetch("content",$this->{content});
		if ($c->{page_id} != $this->{page}) {
		    $this->error("Permission denied: invalid content");
		    return 0;
		}
	    }
	    else {
		$this->error("Permission denied: invalid context");
		return 0;
	    }
	}

	if ($this->{cdata}) {
	    if ($p && $c) {
		$v = $db->fetch("content_data",$this->{cdata});
		if ($v->{content_id} != $this->{content}) {
		    $this->error("Permission denied: invalid version");
		    return 0;
		}
	    }
	    else {
		$this->error("Permission denied: invalid context");
		return 0;
	    }
	}
    }
    
    return $access;
}

# Inline_write is a more secure form of write(), for the case when 
# CMS functions are inlined directly into the page.  The set of 
# callable content managment functions is reduced over what write allows,
# but the functions are dynamically set, rather than hard-coded in the
# CMS tag.

sub inline_write {
    my ($this,$options) = @_;
    $this->{topmenu} = 0;
    my $db = $share{DB};
    my $out;
    my %option; 
    if ($options) { %option = &DecodeString($options); }
    my $action = $this->{query}{cmscmd} || $option{cmscmd};
    if ($action =~ /edit_content_link/) {
	$out .= $this->edit_content_link($option{path});
    }
    elsif ($action =~ /donewpage1/) {
	$out .= $this->do_easy_newpage;
    }
    elsif ($action =~ /newpage1/) {
	$out .= $this->easy_newpage;
    }
    elsif ($action =~ /doupdate1/) {
	$out .= $this->do_easy_content_update;
    }
    elsif ($action =~ /update1/ && $option{name} eq "body") {
	$out .= $this->easy_content_update;
    }
    elsif ($action =~ /newsection1/) {
	$out .= $this->wizard("easysection","New Website");
    }
    
    if ($db->{sql_error}) {
	$out .= "<hr>\n".$db->{sql_error};
    }

    return $out;
}

# ioctl : all operations that don't qualify as reads or writes
#         invoked as <!--BaseDCD.ioctl(request)-->

sub ioctl {
    my $this = shift;
    $_ = shift;   # $_ is the ioctl request
    if (/ModuleName/) {
	return "Website Editor";
    }
    elsif (/ControlPanel/) {
	return \&ctrl_panel;
    }
    elsif (/isRestricted/) {
	return $config{CMS}{restricted};
    }
    elsif (/ModuleInfo/) {
	return "The Website Editor is an advanced content management system 
for ExSite.  It supports managing multiple websites under a single webmaster 
account, dynamic page management, member-only extranets, advanced templating 
functions, versioning, libraries, and many other features.";
    }
}

# control panel

sub ctrl_panel {
    my $this = shift;
    if ($share{PageBuf}) {

	# CMS control panel is more complex than a generic control panel,
	# so we have to modify the page structure

	my $p = $share{PageBuf};

	# set control panel display mode

	$this->{mode} = 1;

	# rebuild the header

	$p->{headtype} = 1;
	$p->header("ExSite CMS - $config{site}{name}");

	# add CBE stylesheets for menus

	$p->print("<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"$config{server}{HTMLpath}/_ExSite/js/menu.css\">\n");

	# modify the body window

	$p->print("<div align=center>\n<table width=90%><tr><td>\n");
    }
    return $this->write;
}

# internal methods

sub TopMenu {
    my $this = shift;
    my @menupath;
    my $url = &relink(cmscmd=>undef,path=>undef);
    push @menupath, "<a href=$url>Top</a>";
    my $ipath = 0;
    my @idpath = @{$this->{path}};
    shift @idpath;
    my $path_info;
    foreach my $id (@idpath) {
	$path_info .= "/$id";
	if ($id) {
	    my $db = $share{DB};
	    my $data = $db->fetch($pathtable[$ipath],$id);
	    $url = &relink(path=>$path_info,cmscmd=>undef);
	    push @menupath, "<a href=$url>$data->{$pathcol[$ipath]}</a>";
	}
	else {
	    push @menupath, "[n/a]";
	}
	$ipath++;
    }
    my $text = join(" > ",@menupath);
    return "<table><tr><td><img src=$config{server}{HTMLpath}/_Modules/CMS/icon.gif height=64 width=64 border=0></td><td>
<table cellspacing=0 cellpadding=0 border=0><tr>
<td><img src=$config{server}{HTMLpath}/_Modules/CMS/textwindow-left.gif height=25 width=12></td>
<td background=$config{server}{HTMLpath}/_Modules/CMS/textwindow-mid.gif>$text</td>
<td><img src=$config{server}{HTMLpath}/_Modules/CMS/textwindow-right.gif height=25 width=12></td>
</tr></table></td></tr></table>\n";
}

sub ManageAllSections {
    my $this = shift;
    # list all sections to be managed
    my $title = "Websites and Website Sections";
    my $db = $share{DB};
    my $uid = $db->my_uid;
    my @sdata = $db->fetch_m2m("section","member",$uid,"member_site_link");
    my $help = "<p>Your websites and site templates are contained in folders at right.  Click on the website you want to open.</p>\n";
    my @tables = $db->{map}->get_mapped_tables;
    my $text;
    if (grep(/^section$/, @tables) == 0) {
	$text = "<h3>Error</h3><p>The Website Editor has not been fully installed on your website.  Please contact your website administrator.</p>\n";
    }
    else {
	if (@sdata == 1 && $config{CMS}{skip_topview} && $share{PageBuf}) {
	    # only one site; might as well skip ahead to the site view
	    $share{PageBuf}->location(&relink(path=>"/$sdata[0]{section_id}"));
	}
	else {
	    $text = $this->ShowObjects("icon",@sdata);
	}
    }
    return $this->htmlbox($title,$this->MenuBar,$help,$text); 
}

sub ManageSection {
    my $this = shift;
    my $sid = shift;
    my ($section,@pdata);
    my $db = $share{DB};
    $section = {};
    if ($sid > 0) {
	$section = $db->fetch("section",$sid);
	@pdata = $db->fetch_child("page","section",$sid);
    }
    my ($text,$help);
    my $format = $this->{query}{format} || $config{CMS}{pageview};
    if ($format =~ /sitemap/) {
	$text = "<H3>Site Map</h3>\n";
	my $section = new ExSite::Section(id=>$sid);
	$text .= $section->map({show_hidden=>1,link_to=>"cms",show_filename=>1},{});
	$help .= "<p>This site's pages are shown at right.  Click on the one you want to open.</p>\n";
    }
    else {
	$text = $this->ShowObjects("icon",@pdata),
	$help .= "<p>This site's pages, templates, and libraries are shown at right.  Click on the one you want to open.</p>\n";
    }
    $help .= "<P>Details of this website:</P>\n".$db->report_row("section",$section)."\n";
    return $this->htmlbox("Website: $section->{title}",
		    $this->MenuBar(\@pdata,
			     {website_template=>
				  ($section->{type} eq "website template")}),
		    $help,
		    $text,
		    $this->goto());
}

sub ManagePage {
    my $this = shift;
    my $pid = shift;
    my ($page,@cdata);
    my $db = $share{DB};
    $page = $db->fetch("page",$pid);
    @cdata = $db->fetch_child("content","page",$pid);
    my $type = $page->{type};
    my $help = "\n<p>The editable content items in this $type are shown at right.  Click on one to open it.</p>\n<p>To preview the whole page, see the View menu.</p>\n";
    $help .= "<P>Page details:</P>\n".$db->report_row("page",$page);
    return $this->htmlbox("$type: \"$page->{title}\"",
		    $this->MenuBar(\@cdata),
		    $help,
		    $this->ShowObjects("icon",@cdata),
		    $this->goto($this->{path}[1]));
}

sub ManageContent {
    my $this = shift;
    my $id = shift;
    my ($page,$cobj,@data);
    my $db = $share{DB};
    $cobj = $db->fetch("content",$id);
    $page = $db->fetch("page",$cobj->{page_id});
    @data = $db->fetch_child("content_data","content",$id,"content_data_id");
    my $sidebar = "\n<p>Versions are shown at right.  Most recent versions are shown at top.</p>
<p>Click on the version to open it.</p>\n";
    $sidebar .= "<p>Details of \"$cobj->{name}\":</p>\n".$db->report_row("content",$cobj);
    return $this->htmlbox("$cobj->{name} content for $page->{filename} (\"$page->{title}\")",
	    $this->MenuBar(\@data),
		    $sidebar,
		    $this->ShowObjects("list",@data),
		    $this->goto(@{$this->{path}}[1..2]));
}

sub ManageContentData {
    my $this = shift;
    my $id = shift;
    my $ContentData = new ExSite::ContentData(id=>$id);
    my $help = "<p>This content version is previewed at right.  It may display differently when inserted into a page with its own style rules.<p>
<P>References to unresolved content items are displayed as a checker graphic:
<img src=$config{server}{HTMLpath}/_ExSite/images/checker.png height=10 width=20 border=1>.
These can only be substituted in when this content item is displayed as part 
of a page.</P>\n";
# following only works for page.type="page"
#    $help .= "<p>(The entire page can be viewed <a href=$config{server}{CGIpath}/page.cgi?_id=$page target=preview>here</a>.)</p>\n";
    
    return $this->htmlbox("Version $ContentData->{cdata}{content_data_id}-$ContentData->{cdata}{ctime}",
		    $this->MenuBar,
		    $help,
		    #$ContentData->get_html,
		    #$ContentData->expand(mode=>"dynamic",method=>"data,url,content,page"),
		    $ContentData->preview,
		    $this->goto(@{$this->{path}}[1..3]));
}

sub DoAction {
    my $this = shift;
    $_ = shift;
    my $db = $share{DB};
    my ($title,$text,$url);
    if (/newcontent|newdesign/) {
	$title = "New Content";
	$url = $this->goto(@{$this->{path}}[1..2]);
	$text .= $this->wizard("content","New Content",{content_id=>undef,reply=>$url});
    }
    elsif (/^(change|update)template(all)?$/) {
	my ($sel,$option,$tcheck,$type);
	$option = $1;
	$type = $2;
    	$title = "Change Template";
	$url = $type ? $this->goto($this->{path}[1]) : $this->goto(@{$this->{path}}[1..2]);
	$sel = $type ? $db->fetch('section',$this->{section}) : $db->fetch('page',$this->{page});
	if ($type){$tcheck = ($sel->{type} !~ /template/i) ? 1 : 0;}
	else{$tcheck = ($sel->{type} =! /^page$/i) ? 1 : 0;}
	if ($tcheck){$text = ($option =~ /change/i) ? $this->changetemplate($type) : $this->updatetemplate($type);}
	else
	{
		$text = "<p>You can not change the template ";
		$text .= $type ? "for all pages" : "of a template page or a library page";
		$text .= ". :(</p>";
	}
    }
    elsif (/(change|update)body/) {
        my ($thepage,$option);
	$option = $1;
    	$title = "Change Body";
	$url = $this->goto(@{$this->{path}}[1..2]);
	$thepage = $db->fetch('page',$this->{page});
	if ($thepage->{type} =~ /^page$/i){$text = ($option =~ /change/i) ? $this->changebody : $this->updatebody;}
	else{$text = "<p>You can not change the body of a template page or a library page. :(</p>";}
    }
    elsif (/editsection/) {
	$title = "Configure Website";
	$url = $this->goto();
	$text .= $db->make(["section"],[$this->{section}],undef,undef,{reply=>$url});
    }
    elsif (/map2page/) {
	$title = "Site Map";
	$url = $this->goto();
	my %s = $db->fetch("section",$this->{section});
	$text .= "<h3>Pages for $s{title}:</h3>\n";
	my $s = new ExSite::Section(section=>\%s);
	$text .= $s->map({show_hidden=>1,link_to=>"dynamic",show_filename=>1},{});
	#$text .= $this->MapSection("page");
    }
    elsif (/map2cms/) {
	$title = "Site Map";
	$url = $this->goto();
	my %s = $db->fetch("section",$this->{section});
	$text .= "<h3>Pages for $s{title}:</h3>\n";
	my $s = new ExSite::Section(section=>\%s);
	$text .= $s->map({show_hidden=>1,link_to=>"cms",show_filename=>1},{});
	#$text .= $this->MapSection("cms");
    }
    elsif (/newpage/) {
	$title = "New Page";
	$url = $this->goto(@{$this->{path}}[1]);
	$text .= $this->wizard("page","New Page",{reply=>$url});
    }
    elsif (/newtemplatepage/) {
	$title = "New Page";
	$url = $this->goto(@{$this->{path}}[1]);
	$text .= $this->wizard("website_template_page","New Page",{reply=>$url});
    }
    elsif (/newtemplate/) {
	$title = "New Template";
	$url = $this->goto(@{$this->{path}}[1]);
	$text .= $this->wizard("template","New Template",{reply=>$url});
    }
    elsif (/importtemplate/) {
	$title = "Import Template";
	$url = $this->goto(@{$this->{path}}[1]);
	$text .= $this->wizard("import_template","Import Template",{reply=>$url});
    }
    elsif (/newlibrary/) {
    	my $tbname = "page#_1#type";
	$title = "New Library";
	$url = $this->goto(@{$this->{path}}[1]);
	$text .= $this->wizard("library","New Library",{reply=>$url});
    }
    elsif (/addchildpage/) {
	# same as new page, but we pass in the parent id
	$title = "Add Child Page";
	$url = $this->goto(@{$this->{path}}[1]);
	$text .= $this->wizard("page","Add Child Page",{parent_id=>$this->{page},reply=>$url});
    }
    elsif (/editpage/) {
	$title = "Configure page";
	$url = $this->goto(@{$this->{path}}[1..2]);
	$text .= $db->make(["page"],[$this->{page}],undef,undef,{reply=>$url});
    }
    elsif (/editcontent/) {
	$title = "Configure content";
	$url = $this->goto(@{$this->{path}}[1..2]);
	$text .= $db->make(["content"],[$this->{content}],undef,undef,{reply=>$url});
    }
    elsif (/editversion/) {
	$title = "Configure version";
	$url = $this->goto(@{$this->{path}}[1..3]);
	$text .= $db->make(["content_data"],[$this->{cdata}],undef,undef,{reply=>$url});
    }
    elsif (/content-wizard/) {
	$title = "Define Content";
	$url = $this->goto(@{$this->{path}}[1..3]);
	$text .= $this->wizard("content","Content Wizard",{content_id=>$this->{path}[3]});
    }
    elsif (/contentdata-wizard/) {
	$title = "Update Content";
	$url = $this->goto(@{$this->{path}}[1..3]);
	$text .= $this->wizard("content_data","Update Content Wizard",{content_id=>$this->{path}[3]});
    }
    elsif (/page-wizard/) {
	$title = "Page Wizard";
	$url = $this->goto(@{$this->{path}}[1..3]);
	$text .= $this->wizard("page","Page Wizard",{page_id=>$this->{path}[2]});
    }
    elsif (/section-wizard|newsection/) {
	$title = "Website Wizard";
	$url = $this->goto(@{$this->{path}}[1..3]);
	$text .= $this->wizard("section","Website Wizard",{page_id=>$this->{path}[1]});
    }
    elsif (/purge/) {
	$title = "Housecleaning";
	$url = $this->goto(@{$this->{path}}[1..3]);
	if ($this->{path}[3]) {
	    $text .= $this->Purge;
	}
	else {
	    $text .= "No content to clean up!";
	}
    }
    elsif (/restore/) {
	$title = "Restore Version";
	$url = $this->goto(@{$this->{path}}[1..3]);
	if ($this->{path}[3]) {
	    $text .= $this->Restore;
	}
	else {
	    $text .= "No content to promote!";
	}
    }
    elsif (/pagemenu/) {
	# not part of normal page body
	$this->{topmenu} = 0;
	return $this->show_top_menu;
    }
    elsif (/update1/) {
	$this->{topmenu} = 0;
	$title = undef;
	$url = undef;
	$text .= $this->easy_content_update;
    }
    elsif (/update/) {
	$title = "Update Content";
	$url = $this->goto(@{$this->{path}}[1..3]);
	$text .= $this->wizard("content_data","Update Content Wizard",{content_id=>$this->{path}[3]});
    }
    elsif (/bulkimport/) {
	$title = "Bulk Import";
	$url = $this->goto(@{$this->{path}}[1..2]);
	$text .= $this->bulk_import;
    }
    elsif (/do_import/) {
	$title = "Bulk Import";
	$url = $this->goto(@{$this->{path}}[1..2]);
	$text .= $this->do_import;
    }
    else {
	$text = "<P>Unknown command.  Sorry :-(</P>\n";
    }
    return $this->htmlbox($title,undef,undef,$text,$url);

}

sub MapSection {
    my $this = shift;
    my $opt = shift;
    my $out;
    my $db = $share{DB};
    my @pages = $db->fetch_match("page",{section_id=>$this->{section},
					 type=>"page",
					 parent_id=>0,
					 visibility=>"visible"});
    $out .= "<P><UL CLASS=\"sitemap\">\n";
    foreach my $page (@pages) {
	$out .= $this->MapSection_r($page,$opt);
    }
    $out .= "</UL></P>\n";
    $out .= "<P>* Restricted-access pages.</P>\n";
    $out .= "<P>NOTE: hidden pages do not appear in the site map.</P>\n";
}

sub MapSection_r {
    my $this = shift;
    # recursive mapping routine
    my $db = $share{DB};
    my ($page,$opt) = @_;
    my $pid = $page->{page_id};
    my $url;
    if ($opt eq "page") {
	# link to page
	$url = "page.cgi?_id=$pid";
    }
    else {
	# link to CMS
	$url = &relink(path=>"/$this->{section}/$pid",cmscmd=>undef);
    }
    my $out = "<LI><A HREF=$url>$page->{title}</A>";
    if ($page->{access} eq "public") { $out .= " ($page->{filename})"; }
    else { $out .= "*"; }
    $out .= "</LI>\n";
    my $subpage = $db->fetch_match("page",{parent_id=>$pid,
					   type=>"page",
					   visibility=>"visible"});
    if (@$subpage > 0) {
	$out .= "<UL CLASS=\"sitemap\">\n";
	foreach my $page (@$subpage) { $out .= $this->MapSection_r($page,$opt); }
	$out .= "</UL>\n";
    }
    return $out;
}

sub goto {
    my $this = shift;
    my $path;
    foreach my $id (@_) {
	$path .= "/$id";
	next if (! $id);
    }
    return &relink(cmscmd=>undef,path=>$path);
}

sub WhatAmI {
    my $this = shift;
    my $db = $share{DB};
    my $uid = $db->my_uid;
    my $site = $this->{section};
    if ($site) {
	my @data = $db->fetch_match("member_site_link",{section_id=>$site,
							member_id=>$uid});
	if (@data > 0) {
	    return $data[-1]->{priveleges};
	}
	else {
	    return undef;
	}
    }
    else {
	return undef;
    }
}

sub htmlbox {
    my $this = shift;
    my ($title,$menubar,$sidebar,$content,$url) = @_;
    my $out; 
    my $width = ($menubar || $sidebar) ?  '80%' : '100%';
    if (! $this->{mode}) {
	if ($title) {
	    $out = "<h1>$title</h1>\n";
	}
	$out .= "
    <TABLE CELLSPACING=8 CELLPADDING=0 BORDER=0 WIDTH=100%>

<!--menubar-->

      <tr><td valign=\"top\">$menubar<br><br>

<!--sidebar-->

      $sidebar</td>

<!--webtop-->

      <td valign=\"top\" width=\"$width\"><div class=htmlbox>
$content
      </div></td>
      </tr>
    </TABLE>\n";
    }
    else {
	$out = "    <TABLE CELLSPACING=0 CELLPADDING=0 BORDER=0 WIDTH=100% CLASS=framed>
      <TR>
        <TH CLASS=titlebar colspan=3><IMG SRC=$config{server}{HTMLpath}/_ExSite/images/null.gif HEIGHT=1 WIDTH=1>";
    
#    if ($url) {
#        $out .= "<A HREF=$url><IMG SRC=$config{server}{HTMLpath}/_ExSite/images/button-exit-small.gif HEIGHT=20 WIDTH=20 BORDER=0></A>";
#    }
#    else {
        $out .= "<IMG SRC=$config{server}{HTMLpath}/_ExSite/images/button-null-small.gif HEIGHT=20 WIDTH=20 BORDER=0>";
#    }

    $out .= "&nbsp;<B>$title</B></Th>
      </TR>

<!--hrule-->

      <tr><td bgcolor=#000000 colspan=3><img src=$config{server}{HTMLpath}/_ExSite/images/null.gif></td></tr>

<!--menubar-->

      <tr><td class=menubar colspan=3>$menubar</td></tr>

<!--hrule-->

      <tr><td bgcolor=#000000 colspan=3><img src=$config{server}{HTMLpath}/_ExSite/images/null.gif></td></tr>

<!--sidebar-->

      <tr><td class=sidebar WIDTH=20% valign=top>$sidebar</td>

<!--vrule-->

      <td bgcolor=#000000><img src=$config{server}{HTMLpath}/_ExSite/images/null.gif></td>

<!--webtop-->

      <td width=80% bgcolor=#FFFFFF valign=top><div class=htmlbox>
$content
      </div></td>
      </tr>
    </TABLE>\n";
    }

    return $out;
}

sub MenuBar {
    my $this = shift;
    my ($objects,$opt) = @_;
    # objects is a pointer to the iconified objects in the current screen
    # opt is a pointer to a datahash containing extra formatting options
    my $JS = "$config{server}{HTMLpath}/_ExSite/js";
    my $CGI = $config{server}{CGIpath};
    my $HTML = $config{server}{HTMLpath};
    my $db = $share{DB};
    my @menu;
    
    # @menu is a list of menubar-item, menu, menubar-item, menu, etc.
    # menubar-item is just a text label.  menu is a reference to an array
    # of hashes containing the menu items.  Each hash can have keys for 
    # the menu item label, the url, and (optional) the target.
    # eg. @menu=("File",[ {label=>"Exit",url=>"exit.cgi"} ] );

    #my $reply = &relink(cmscmd=>undef);
    if ($this->{cdata}) {
	my $reply  = &url_escape(&relink(cmscmd=>undef,path=>"/$this->{section}/$this->{page}/$this->{content}"));
	# content data menu
	@menu = ( "Version",
		  [
		   { label=>"Configure Version", url=>$this->goto(@{$this->{path}}[1..4])."&cmscmd=editversion", perm=>"administrator", level => 3, },
		   { label=>"Delete Version", url=>"$CGI/webdb.cgi?action=delete&tab=content_data&id=$this->{cdata}&reply=$reply", perm => "administrator", level => 2, },
		   { label=>"Restore Version", url=>$this->goto(@{$this->{path}}[1..4])."&cmscmd=restore", level => 2, },
		   { label=>"Close", url=>$this->goto(@{$this->{path}}[1..3]), level => 2,  },
		   { label=>"Exit", url=>"$CGI/home.cgi", level => 2, },
		   ],
		  "Update",
		  [
		   { label=>"Update", url=>$this->goto(@{$this->{path}}[1..4])."&cmscmd=update", level => 2, },
#		   { label=>"Approve", url=>"javascript:alert('Workflow tools not available')", perm=>"administrator" },
		   # rollback == delete for now
		   { label=>"Rollback", url=>"$CGI/webdb.cgi?action=delete&tab=content_data&id=$this->{cdata}&reply=$reply", perm => "administrator", level => 2,  },
		   ],
		  );
    }
    elsif ($this->{content}) {
	my $reply  = &url_escape(&relink(cmscmd=>undef,path=>"/$this->{section}/$this->{page}"));
	# content menu
	my $rollbackurl = scalar @$objects > 0 ? "$CGI/webdb.cgi?action=delete&tab=content_data&id=$objects->[-1]->{content_data_id}&reply=$reply" : "javascript:alert('nothing to roll back!')";
	@menu = ( 
		  "Content",
		  [
		   { label=>"Update", url=>$this->goto(@{$this->{path}}[1..3])."&cmscmd=update", level => 2,  },
		   { label=>"Configure Content", url=>$this->goto(@{$this->{path}}[1..3])."&cmscmd=editcontent", perm=>"administrator", level => 2, },
		   { label=>"Delete Content", url=>"$CGI/webdb.cgi?action=delete&tab=content&id=$this->{content}&reply=$reply", perm=>"administrator", level => 2, },
		   { label=>"Close", url=>$this->goto(@{$this->{path}}[1..2]), level => 2, },
		   { label=>"Exit", url=>"$CGI/home.cgi", level => 2, },
		   ],
		  "View",
		  [
		   { label=>"View Latest", url=>$this->goto(@{$this->{path}}[1..3]), level => 2, },
		   { label=>"View Current", url=>$this->goto(@{$this->{path}}[1..3]), level => 2, },
		   ],
		  "Update",
		  [
		   { label=>"Update", url=>$this->goto(@{$this->{path}}[1..3])."&cmscmd=update", level => 2, },
		   { label=>"Rollback", url=>$rollbackurl, perm=>"administrator", level => 2, },
		   ],
		  "Versions",
		  [
		   { label=>"Clean Up", url=>$this->goto(@{$this->{path}}[1..3])."&cmscmd=purge", perm=>"administrator", level => 2, },
#		   { label=>"Remove Old Versions", url=>"javascript:alert('Version Clean Up not implemented')", perm=>"administrator" },
		   ],
		  );
    }
    elsif ($this->{page}) {
	# page menu
	my %page = $db->fetch("page",$this->{page});
	my $type = ucfirst($page{type});
	my $reply  = &url_escape(&relink(cmscmd=>undef,path=>"/$this->{section}"));
	# template kludge
	if ($type eq "Template") { $type = "Page"; }
	@menu = ( "$type",
		  [
		   { label=>"New Content Item", url=>$this->goto(@{$this->{path}}[1..2])."&cmscmd=newcontent", level => 2, },
		   { label=>"Bulk Import", url=>$this->goto(@{$this->{path}}[1..2])."&cmscmd=bulkimport", level => 2, },
		   { label=>"New Child Page", url=>$this->goto(@{$this->{path}}[1..2])."&cmscmd=addchildpage", perm=>"administrator|editor", level => 2, },
		   { label=>"Configure $type", url=>$this->goto(@{$this->{path}}[1..2])."&cmscmd=editpage", perm=>"administrator", level => 2, },
		   { label=>"Publish Page", url=>"$CGI/publish.cgi?page=$this->{page}", target=>"publishwindow", perm=>"administrator|editor", level => 2, },
		   { label=>"Delete $type", url=>"$CGI/webdb.cgi?action=delete&tab=page&id=$this->{page}&reply=$reply", perm=>"administrator", level => 2, },
		   { label=>"Close", url=>$this->goto(@{$this->{path}}[1]), level => 2, },
		   { label=>"Exit", url=>"$CGI/home.cgi", level => 2, },
		   ],
		  "Template",
		  [
		   { label=>"Go to Template", url=>$this->goto(@{$this->{path}}[1],$page{template_id}), level => 2, },
		   { label=>"Change Template",url=>$this->goto(@{$this->{path}}[1..2])."&cmscmd=changetemplate",perm=>"adminiaddchildpagestrator|editor|designer", level => 2, },
		   { label=>"Change Body", url=>$this->goto(@{$this->{path}}[1..2])."&cmscmd=changebody", level => 2, },
		   ],
		  "View",
		  [
		   { label=>"Preview", url=>"$CGI/page.cgi?_id=$this->{page}", target=>"preview", level => 2, },
		   { label=>"Editable Preview", url=>"$CGI/page.cgi?_id=$this->{page}&_cms=1", target=>"preview", level => 2, },
		   { label=>"Non-editable Preview", url=>"$CGI/page.cgi?_id=$this->{page}&_cms=0", target=>"preview", level => 2, },
		   { label=>"Edit Mode On", url=>"$CGI/page.cgi?_id=$this->{page}&_cmsmode=1", target=>"preview", level => 2, },
		   { label=>"Edit Mode Off", url=>"$CGI/page.cgi?_id=$this->{page}&_cmsmode=0", target=>"preview", level => 2, },
		   #{ label=>"Published Version", url=>"$HTML/XXX"},
		   ],
		  );
	my $update_menu = [];
	if (@$objects > 0) {
	    foreach my $c (@$objects) {
		push @$update_menu, { label=>"Update $c->{name}", url=>($this->goto(@{$this->{path}}[1..2],$c->{content_id})."&cmscmd=update"), level => 2,  }
	    }
	    push @menu, ("Update", $update_menu);
	}
    }
    elsif ($this->{section}) {
	my $reply  = &url_escape(&relink(cmscmd=>undef,path=>undef));
	# site menu
	my %section = $db->fetch("section",$this->{section});
	@menu = ( "Website",
		  [
		   { label=>"New Sub-Site", url=>$this->goto(@{$this->{path}}[1])."&cmscmd=newsection", perm=>"administrator", level => 3, },
#		   { label=>"New Site Template", url=>"javascript:alert('New Site Template not implemented')", perm=>"administrator" },
		   ##### should be newtemplatepage for website templates
		   { label=>"New Page", url=>$this->goto(@{$this->{path}}[1])."&cmscmd=".($opt->{website_template} ? "newtemplatepage" : "newpage"), perm=>"administrator", level => 2, },
		   { label=>"New Library", url=>$this->goto(@{$this->{path}}[1])."&cmscmd=newlibrary", perm=>"administrator", level => 2, },
		   { label=>"New Template", url=>$this->goto(@{$this->{path}}[1])."&cmscmd=newtemplate", perm=>"administrator", level => 2, },
		   { label=>"Configure Site", url=>$this->goto(@{$this->{path}}[1])."&cmscmd=editsection", perm=>"administrator", level => 3, },
		   { label=>"Delete Site", url=>"$CGI/webdb.cgi?action=delete&tab=section&id=$this->{section}&reply=$reply", perm=>"administrator", level => 3, },
		   { label=>"Close", url=>$this->goto(), level => 2, },
		   { label=>"Exit", url=>"$CGI/home.cgi", level => 2, },
		   ],
		  "Template",
		  [
		   { label=>"New Template", url=>$this->goto(@{$this->{path}}[1])."&cmscmd=newtemplate", perm=>"administrator", level => 2, },
		   { label=>"Change For All Pages", url=>$this->goto(@{$this->{path}}[1])."&cmscmd=changetemplateall",perm=>"administrator|designer", level => 2, },
		   ],
		  "Publish",
		  [
		   { label=>"Publish Site", url=>"$CGI/publish.cgi?section=$this->{section}", target=>"publishwindow", perm=>"administrator|editor", level => 2,  },
#		   { label=>"Un-publish Site", url=>"javascript:alert('Un-publish not implemented')", perm=>"administrator"},
		   ],
		  "View",
		  [
		   #{ label=>"Home Page", url=>"$CGI/page.cgi?_id=XXX"},
		   { label=>"Site Contents", url=>&relink(path=>"/$this->{section}",format=>"contents"), level => 2, },
		   { label=>"Site Map", url=>&relink(path=>"/$this->{section}",format=>"sitemap"), level => 2, },
		   ],
		  );
    }
    else {
	# top menu
	@menu = ( "Websites",
		  [{ label=>"New Site", url=>$this->goto()."&cmscmd=newsection", perm=>"administrator", level => 3, },
#		   { label=>"New Site Template", url=>"javascript:alert('New Site Template not implemented')", perm=>"administrator" },
		   { label=>"Exit", url=>"$CGI/home.cgi", level => 2, }],
		  );
    }

    my $priv = $this->WhatAmI();
    my $level = $db->authorize;
    my $out;

    if (! $this->{mode}) {
	# simple HTML menubar

	$out .= "<TABLE WIDTH=100% CELLSPACING=0 CELLPADDING=0 BORDER=0>\n";

	# now create the menus...

	my $imenu = 0;
	while (@menu) {
	    $imenu++;
	    my $menutitle = shift @menu;
	    my $menu = shift @menu;
	    
	    $out .= "<tr><th>$menutitle</th></tr>\n<tr><td>\n";
	    foreach my $item (@$menu) {
		next if (exists $item->{perm} && $priv !~ /$item->{perm}/);
		next if (exists $item->{level} && $level < $item->{level});
		$out .= "<a href=\"$item->{url}\"";
		if (exists $item->{target}) { $out .= " target=\"$item->{target}\""; }
		$out .= ">$item->{label}</a><br>\n";
	    }
	    $out .= "</td></tr>\n";
	}
	$out .= "</table>\n";
    }

    else {

	# add JavaScript libraries
	$out .= "<script type='text/javascript' src='$JS/cbe_core.js'></script>
<script type='text/javascript' src='$JS/cbe_event.js'></script>
<script type='text/javascript' src='$JS/cbe_util.js'></script><!-- for debugging -->
<script type='text/javascript' src='$JS/cbe_debug.js'></script><!-- for debugging -->
<script type='text/javascript' src='$JS/menu_exsite.js'></script>

<script type='text/javascript'><!--
var mark1, menu1, menu2, menu3;
function windowOnload() {
  mark1 = cbeGetElementById('menuMarker1').cbe;
  menu1 = new cbeMenu(
    true,                       // horizontal,
//    (document.cbe.width()-(3*75))/2, mark1.pageY(), 3, // centered
    mark1.pageX(), mark1.pageY(), 3, // menuX, menuY, menuZ (left-justified)
    [75,75,75], 20, 8,          // lblWidthsArray, lblHeight, lblspacing
    [140,170,155],              // boxWidthsArray,
    '#000099', '#000000',       // activeColor, inactiveColor,
    '#FFFFFF', '#9999AA',       // activeBkgnd, inactiveBkgnd
    '#9999AA'                   // boxBkgnd
  );
  window.cbe.addEventListener(\"resize\", resizeListener, false);
}

--></script>\n";

	# now build the HTML for the menubar

	$out .= "<a name=\"topofpage\"></a>
 <TABLE WIDTH=100% CELLSPACING=0 CELLPADDING=0 BORDER=0> 
  <TR>
   <TD COLSPAN=4><IMG SRC=$config{server}{HTMLpath}/_ExSite/images/null.gif HEIGHT=1 WIDTH=1></TD>
  </TR>
   <TR>
   <TD><IMG SRC=$config{server}{HTMLpath}/_ExSite/images/null.gif HEIGHT=1 WIDTH=1></TD>
   <TD>
     &nbsp;&nbsp;
   </TD>
   <TD WIDTH=100%>
    <div id='menuMarker1' class='menuMarker'>&nbsp;</div>
   </TD>
   <TD><IMG SRC=$config{server}{HTMLpath}/_ExSite/images/null.gif HEIGHT=1 WIDTH=1></TD>
  </TR>
  <TR>
   <TD COLSPAN=4><IMG SRC=$config{server}{HTMLpath}/_ExSite/images/null.gif HEIGHT=1 WIDTH=1></TD>
  </TR>
 </TABLE>\n";

	# now create the menus...

	my $imenu = 0;
	while (@menu) {
	    $imenu++;
	    my $menutitle = shift @menu;
	    my $menu = shift @menu;
	    
	    $out .= "<div id='label1$imenu' class='mLabel'>$menutitle</div>
<div id='box1$imenu' class='mBox'>\n";
	    foreach my $item (@$menu) {
		next if (exists $item->{priv} && $priv !~ /$item->{priv}/);
		next if (exists $item->{level} && $level < $item->{level});
		$out .= "<a class=\"m\" href=\"$item->{url}\"";
		if (exists $item->{target}) { $out .= " target=\"$item->{target}\""; }
		$out .= ">$item->{label}</a>\n";
	    }
	    $out .= "</div>\n";
	}
    }

    return $out;

}

#=======================================================================
# clean out old content
#=======================================================================

sub Purge {
    my $this = shift;
    my $out;
    my $first = 1;
    my $db = $share{DB};
    foreach my $key (keys %{$this->{formdata}}) {
	if ($key =~ /^\d+$/) {
	    $first = 0;
	    $db->delete_key("content_data",$key);
	}
    }
    my @cdata = $db->fetch_child("content_data","content",$this->{content},"content_data_id");
    my $url = $this->goto(@{$this->{path}}[1..3]);
    $out .= "<H1>Housecleaning</h1>\n<p>This form allows you to remove old content versions that no longer need to be archived.  Select which versions to delete.</p>\n<p><b>Delete:</b></p>
<p><form action=\"${url}&cmscmd=purge\" method=\"post\">\n";
    for (my $ic = $#cdata; $ic >= 0; $ic--) {
	$out .= "<input type=checkbox name=\"$cdata[$ic]->{content_data_id}\"";
	if ($first && $ic != $#cdata) {
	    # select all non-current versions by default
	    $out .= " checked";
	}
	##### should link the version to a preview
	$out .= "> $cdata[$ic]->{note}<br>\n";
    }
    $out .= "</p>\n<input type=\"submit\" value=\"Delete\"> <input type=reset>\n";
    return $out;
}

#=======================================================================
# restore old content
#=======================================================================

sub Restore {
    my $this = shift;
    my $db = $share{DB};
    my %cdata = $db->fetch("content_data",$this->{cdata});
    my $message = "Restored content_data $this->{cdata}";
    $cdata{note} = $message;
    #delete $cdata{content_data_id};
    $db->insert("content_data",\%cdata);
    return "<h1>$message</h1>";
}

#=======================================================================
# webtop and icon display
#
#=======================================================================

sub ShowObjects {
    my $this = shift;
    my $format = shift;  # "icon" or "list"
    # remaining args are the data objects
    my $icon = ($format eq "icon");

    my $out = "<table cellspacing=20><tr>\n";
    my $nicon;
    if ($icon) {
	# sort the objects by their name
	my @objs;
	foreach my $obj (@_) {
	    my @iconspec = $this->objtype($obj);
	    push @objs, \@iconspec;
	}
	my @newobjs = sort {$a->[1] cmp $b->[1]} @objs;
	foreach my $obj (@newobjs) {
	    $nicon++;
	    my ($iconname,$label,$description,$url) = @$obj;
	    $out .= "<td class=icon align=center><a href=\"$url\" class=icon title=\"$description\"><img src=$config{server}{HTMLpath}/_Modules/CMS/${iconname}.gif height=48 width=48 border=0><br>$label</a></td>\n"; 
	    if ($nicon % $config{webtop}{x} == 0) { $out .= "</tr><tr>\n"; }
	}
	if (@{$this->{path}} > 1) {
	    $nicon++;
	    my $url = $this->goto(@{$this->{path}}[1..($#{$this->{path}}-1)]);
	    $out .= "<td class=icon align=center><a href=\"$url\" class=icon title=\"back\"><img src=$config{server}{HTMLpath}/_Modules/CMS/back.gif height=48 width=48 border=0><br>[ up ]</a></td>\n"; 
	}
    }
    else {
	# list in inverse order
	while (@_) {
	    my $obj = pop @_;
	    my ($iconname,$label,$description,$url) = $this->objtype($obj);
	    $out .= "<td class=icon align=center><a href=\"$url\" class=icon title=\"$description\"><img src=$config{server}{HTMLpath}/_Modules/CMS/${iconname}.gif height=48 width=48 border=0></a></td><td><a href=\"$url\" class=icon title=\"$description\">$label</a></td>\n</tr></tr>\n"; 
	}
	if (@{$this->{path}} > 1) {
	    $nicon++;
	    my $url = $this->goto(@{$this->{path}}[1..$#{$this->{path}}-1]);
	    $out .= "<td class=icon align=center><a href=\"$url\" class=icon title=\"back\"><img src=$config{server}{HTMLpath}/_Modules/CMS/back.gif height=48 width=48 border=0></td><td><a href=\"$url\" class=icon title=\"back\">[back]</a></td>\n</tr><tr>\n"; 
	}
    }
    $out .= "</tr></table>\n";
    return $out;
}

# return the object type
# site, sitetemplate, page, template, library, econtent, dcontent, data
sub objtype {
    my $this = shift;
    my $obj = shift;
    if (exists $obj->{section_id}) {
	if (exists $obj->{publish_to}) {
	    # a site/section
	    return (($obj->{type} =~ /template/ ? "sitetemplate" : "site"),
		    $obj->{title},
		    $obj->{description},
		    $this->goto($obj->{section_id}));
	}
	else {
	    # a page
	    return ($obj->{type},
		    $obj->{filename},
		    $obj->{description},
		    $this->goto($obj->{section_id}, $obj->{page_id}));
	}
    }
    elsif (exists $obj->{page_id}) {
	# content object
	return (($obj->{type} eq "editorial" ? "econtent" : "dcontent"),
		$obj->{name},
		"$obj->{allow_mime}; $obj->{description}",
		$this->goto(${$this->{path}}[1],$obj->{page_id},$obj->{content_id}));
    }
    else {
	# content data
	my $db = $share{DB};
	return ("data",
		"Updated ".$db->show_data("content_data","ctime",$obj->{ctime})."; $obj->{note}",
		$obj->{mime_type},
		$this->goto($this->{path}[1],$this->{path}[2],$obj->{content_id},$obj->{content_data_id}));
    }
}

#==========================================================================
# EASY update methods - simplified interface to CMS
#
# easy_content_update : 1-step content update
# easy_newpage : 1-step page creation
#==========================================================================

### easy_newpage functions superceded by MySite easy-CMS module

sub easy_newpage {
    my $this = shift;
    my %page;
    my $db = $share{DB};
    my $out = "<h1>Make a new web page</h1>\n";

    my $action = &relink(cmscmd=>"donewpage1");
    
    $out .= "<p>Your new web page will appear in site menus under a
short but descriptive name.  Please specify the menu label to use for
this new page:</p>\n";
    
    $out .= $db->make(table=>"page",
    	    	      show=>["label"],
		      data=>{label=>"New Page"},
		      action=>$action);

    return $out;
}

sub do_easy_newpage {
    my $this = shift;
    my $in = new ExSite::Input;
    my $data = $in->post;
    my $db = $share{DB};
    my $label = $data->{'page##label'};
    my @err;
    my $pid;

    if ($label) {

	my @sites = $db->my_real_sites;
	my $site = $sites[0];

	my %page = (
		    label=>$label,
		    title=>$label,
		    type=>"page",
		    status=>"active",
		    access=>"public",
		    publish_method=>"static",
		    visibility=>"visible",
		    rank=>1,
		    parent_id=>0,
		    section_id=>$site->{section_id},
		    template_id=>$site->{template_id},
		    );
    
	my $filename = $label;
	$filename =~ s/[ *?&]/_/g;

	# check for existing pages

	my $n = undef;
	my @exist = $db->fetch_match("page",{section_id=>$site->{section_id},
					     filename=>$filename."$n.html"});
    
	while (@exist > 0) {
	    $n++;
	    @exist = $db->fetch_match("page",{section_id=>$site->{section_id},
					      filename=>$filename."$n.html"});
	}
	
	$page{filename} = $filename."$n.html";

	$pid = $db->insert("page",\%page);
	if (! $pid) {
	    push @err, ("failed to add new page: ".
			$db->show_diagnostics("error","html"));
	}
	else {
	    # now create a body
	    my %content = (
			   page_id => $pid,
			   name => "body",
			   type => "editorial",
			   );
	    my $id = $db->insert("content",\%content);
	    if (! $id) {
		push @err, ("page was created, but failed to add page body: ".
			    $db->show_diagnostics("error","html"));
	    }
	}
    }
    else {
	push @err, "You must provide a page label.";
    }
    
    my $out;
    if (@err > 0) {
	$out .= "<h1>Error:</h1>\n<p class=\"error\">".
	    join("<br>\n",@err)."</p>\n";
	$out .= "<p><form><input type=button value=\"&lt; Back\" onclick=\"history.back()\"></form></p>\n";
    }
    else {
	my $url = &relink(cmscmd=>undef,ecmd=>undef);
	$out .= "<h1>New Page Created</h1>\n<ul>
<li><a href=\"$url\">Back to My Site</a></li>\n";
	$url = &relink(cmscmd=>undef,ecmd=>undef);
	$out .= "<li><a href=\"$config{server}{CGIpath}/page.cgi?_id=$pid&_cmsmode=1\">Edit page \"$label\"</a></li>\n</ul>\n";
    }
    return $out;
}

sub easy_content_update {
    my $this = shift;
    my $db = $share{DB};
    my ($out, %content, @cdata, $selected);

    my $reply;
    if ($ENV{HTTP_REFERER} =~ /page\.cgi/) {
	$reply = $ENV{HTTP_REFERER};
    }
    else {
	$reply = "page.cgi?_id=$this->{page}";
    }
    my $action = &relink(cmscmd=>"doupdate1");
    $out .= "<form METHOD=POST action=\"$action\" onSubmit=\"validate()\">\n";

    $out .= $db->input_html(type=>"hidden",name=>"content_data#_2#data#html");
    $out .= $db->input_html(type=>"hidden",name=>"action",value=>"insert");
    $out .= $db->input_html(type=>"hidden",name=>"reply",value=>$reply);
    $out .= $db->input_html(type=>"hidden",name=>"content_data#_2#mime_type",value=>"text/html");

    # get the content id from the query data
    my $id = $this->{content};

    $out .= $db->input_html(type=>"hidden",name=>"content_data#_2#content_id", value=>$id);

    if ($id =~ /\d+/) {
	%content = $db->fetch("content",$id);
	@cdata = $db->fetch_child("content_data","content",$id,"content_data_id");
    }
    if (@cdata == 0) {
	# set an assumed mime type, depending on content type
	if ($content{type} eq "design") {
	    # also check for content#_1#allow_mime#_set_N = "image"
	    $cdata[0] = { content_data_id=>"_2", mime_type => "image/gif" };
	}
	else {
	    $cdata[0] = { content_data_id=>"_2", mime_type => "text/html" };
	}
    }
    my $cid = $cdata[-1]{content_data_id};
    my $name = $content{name};
    $out .= "<input type=hidden name=\"content_data_id\" value=\"$cid\">\n"; 

    my $update_method;
    if ($cdata[-1]->{mime_type} =~ /^text\/html$/o) {
	$update_method = "html";
    }
    else {
	$update_method = "upload";
    }

    my $textdata;
    if ($update_method =~ /html$/ && $db->use_wysiwyg) {
	$out .= $db->input_htmleditor(name=>"content_data#_2#data",
				      table=>"content_data",
				      record=>$cid,
				      page=>$this->{page},
				      section=>$this->{section},
				      column=>"data",
				      value=>"$textdata",
				      editor=>{
					  op_image=>0,
					  op_link=>1,
					  op_table=>1,
					  op_dyncontent=>0,
				      });
	$out .= $db->editor_javascript;
    }
    elsif ($update_method =~ /^upload$/) {
	$out .= "<p>Click on browse to select a file to upload:</p>\n";
	$out .= "<p><input type=file size=60 name=\"content_data#_2#fdata#filename\"></p>\n";
	$this->{db}->{enctype} = "multipart/form-data";
	######################## need a multipart-encoded form!
	########## should be done by Form
    }
    else {
	# default to text area
	$out .= "<p>You may enter your content changes directly into this text editing area:</p>\n";
	$out .= "<p><textarea cols=80 rows=20 wrap=virtual name=\"content_data#_2#data\">\n";
	$out .= $textdata;
	$out .= "</textarea></p>\n";
    }

    # 3) revision control comment

    my $comment = "updated by ".$db->my_name." on ".scalar(localtime)."\n";
    $out .= $db->input_html(type=>"hidden",name=>"content_data#_2#note",value=>$comment);

    $out .= "</p>\n<INPUT TYPE=submit NAME=submit_action VALUE=\"Submit\">&nbsp;&nbsp;\n<INPUT TYPE=submit NAME=submit_action VALUE=\"Cancel\">&nbsp;&nbsp;\n";

    return $out;
}

sub do_easy_content_update {
    my $this = shift;
    my $form = $share{DB};
    $form->get_data;

    # patch up those mangled page links
    $form->{form}->{content_data}->{_2}->{data} =~ s/%7B/{/g;
    $form->{form}->{content_data}->{_2}->{data} =~ s/%7D/}/g;

    $form->do_db("insert");
    my $reply = $this->{input}->{reply} || "$config{server}{CGIpath}/page.cgi?_id=$this->{page}";
    if ($share{Page}) {
	$share{Page}->redirect($reply);
    }
    elsif ($share{PageBuf}) {
	$share{PageBuf}->location($reply);
    }
}

# put up a link to edit content right in the page, instead of jumping out
# to the CMS control panel
#
# WARNING: same name as ExSite::Page::edit_content_link, because it
# essentially duplicates that function

sub edit_content_link {
    my ($this,$path) = @_;
    my $url = &relink(cmscmd=>"update1",path=>$path);
    return "<a href=\"$url\"><img src=\"$config{server}{HTMLpath}/_ExSite/images/plus16.gif\" height=16 width=16 border=0 title=\"update this content\"></a>";
}

# create a form with a list of available templates

sub changetemplate{
    my $this = shift;
    my ($type) = @_;
    my ($sel,$out,$reply,$templid,$path,$itab,$iid,$icol,$selid);
    my $db = $share{DB};
    $reply = $type ? $this->goto(@{$this->{path}}[1]) : $this->goto(@{$this->{path}}[1..2]);
    $path = "/".$this->{section};
    $path .= $type ? "" : "/$this->{page}";
    $sel = $type ? $db->fetch('section',$this->{section}) : $db->fetch('page',$this->{page});
    $itab = $type ? 'section' : 'page';
    $iid = $type ? $this->{section} : $this->{page};
    $icol = $type ? 'section_id' : 'page_id';
    $selid = $type ? "" : $sel->{template_id};
    my $url = &relink(cmscmd=>"updatetemplate$type",path=>$path);
    $out .= "
	<form method=\"post\" action=\"$url\">\n
	<input type=\"hidden\" name=\"$itab#$iid#$icol\" value=\"$iid\">\n
	<input type=\"hidden\" name=\"$itab#$iid#title\" value=\"$sel->{title}\">";
    $out .= &list_template($this,"$itab#$iid#template_id",$selid,1);
    $out .= "<input type=\"hidden\" name=\"reply\" value=\"$reply\">\n";
    $out .= "<p><input type=\"submit\" value=\" Submit \">\n"; 
    $out .= "<input type=\"button\" value=\" Cancel \" onClick=\"history.back()\">\n</p>"; 
    $out .= "</form>";
    return $out;
}

# updates the database with the new body template

sub updatetemplate{
    my $this = shift;
    my ($type) = @_;
    my $db = $share{DB};
    my ($out,$do_out,$reply,$title,$itab,$iid,$icol,$allpages,$tempid);
    $title = "page template";
    $title .= $type ? "for all pages" : "";
    $itab = $type ? 'section' : 'page';
    $iid = $type ? $this->{section} : $this->{page};
    $icol = $type ? 'section_id' : 'page_id';
    $reply = $type ? $this->goto(@{$this->{path}}[1]) : $this->goto(@{$this->{path}}[1..2]);
    $tempid = $this->{formdata}{"$itab#$iid#template_id"};
    # now, perform all the necessary DB mods
    $db->get_data;
    if (!$tempid)
    {
	$out .=  "<p>Error: You need to select a $title</p>\n";
	$out .= "<p><form><input type=\"button\" value=\"Back\" onClick=\"history.back()\"></form></p>\n";
	return $out;
    }
    if ($type){$do_out = $db->update('page',{template_id=>$tempid},{section_id=>$iid,type=>"page"});}
    else
    {
	$db->{action} = "insert";
	$do_out = $db->do;
    }
    if ($db->errorcheck > 0){return ($do_out);}
    $this->redirect($reply);
    return;
}

# creates a form with a list of available body templates

sub changebody{
    my $this = shift;
    my ($type,$id,$out,$reply);
    $reply = $this->goto(@{$this->{path}}[1..2]);
    $type = "body content";
    my $db = $share{DB};
    $id = $db->fetch_match('content',{'page_id'=>$this->{page},'name'=>'body'},{'content_id'=>'desc'});
    $id = $id->[0];
    $id = $id->{content_id};
    $out .= "<h3>Change Page Body</h3>
<p>The <b>body</b> is the primary editable section of the page.  This 
tool lets you swap in pre-configured page bodies, if any are available 
for your website to use.</p>
<p>Note that if you simply want to edit or customize your page body, 
you should use the update function.</p>
<p>Also note that some pages may have structures that are too complicated
to be contained in a simple body;  with these pages, changing the body
may only affect part of the editable content, or may even have no effect 
at all if the page doesn't use a body element.</p>
<p>The following pre-configured page bodies are available:</p>\n<p>";
    my $url = &relink(cmscmd=>"updatebody",path=>"/$this->{section}/$this->{page}");
    $out .= "
	<form method=\"post\" action=\"$url\">\n
	<input type=\"hidden\" name=\"content_data#_1#content_id\" value=\"$id\">\n";
    my (@liblist,$libs,$templates,$sql);
    $libs = $db->fetch_match('page',{type=>"library",section_id=>$this->{section}});
    foreach my $l (@$libs){push(@liblist, $l->{page_id});}
    $sql = "select * from content where name='body' and page_id in (".join(',',@liblist).")";
    $templates = $db->custom_query($sql);
    if (@$templates < 1)
    {
	$out = "<p>There are no body templates available.<br>";
	$out .= "<form><input type=\"button\" value=\"Back\" onClick=\"history.back()\"></form></p>";
	return $out;
    }
    foreach my $t (@$templates)
    {
	$out .= "<input type=\"radio\" name=\"body_id\" value=\"$t->{content_id}\"";
	$out .= ">$t->{description}<br>\n";
    }
    $out .= "<input type=\"hidden\" name=\"content_data#_1#mime_type\" value=\"text/html\">\n";
    $out .= "<input type=\"hidden\" name=\"reply\" value=\"$reply\">\n";
    $out .= "<p><input type=\"submit\" value=\" Submit \">\n"; 
    $out .= "<input type=\"button\" value=\" Cancel \" onClick=\"history.back()\">\n</p>"; 
    $out .= "</form>";
    return $out;
}

# updates the database with the new body template

sub updatebody{
    my $this = shift;
    my ($out,$do_out,$reply);
    $reply = $this->goto(@{$this->{path}}[1..2]);
    # now, perform all the necessary DB mods
    my $db = $share{DB};
    $db->get_data;
    if (!$this->{formdata}{body_id})
    {
	$out .=  "<p>Error: You need to select a body template</p>\n";
	$out .= "<p><form><input type=\"button\" value=\"Back\" onClick=\"history.back()\"></form></p>\n";
	return $out;
    }
    if ($this->{formdata}{body_id})
    {
	# substitute the selected body into the form data
	my $c_data = $db->fetch_child("content_data","content",$this->{formdata}{"body_id"},"content_data_id");
	$c_data = $c_data->[-1];
	$db->{'form'}->{'content_data'}->{'_1'}->{'content_id'} = $this->{formdata}{"content_data#_1#content_id"};
	foreach my $c (keys %$c_data)
	{
	    next if $c =~ /^(content_data_id|content_id|ctime)$/;
	    $db->{'form'}->{'content_data'}->{'_1'}->{"$c"} = $c_data->{"$c"};
	}
    }
    $db->{action} = "insert";
    $do_out = $db->do;
    if ($db->errorcheck > 0){return ($do_out);}
    $this->redirect($reply);
    return;
}

# sub list_template: create a list of available templates

sub list_template {
    my $this = shift; # unused: this routine cheats and is tied to both the
                      # CMS and Wizard classes.
    # $input_name: name of the input field
    # $id: template id
    # $select: 0 => select all templates
    #          1 => select templates useable by this site
    #          2 => select templates useable by children of this site
    my ($input_name,$id,$select) = @_;
    my ($out,@ptempl,@options);
    $out .= "<p>Select a template:</p>\n<blockquote>\n";

    # a little javascript to control the target window

    $out .= "<script LANGUAGE=\"javascript\">
<!--
function NewWindow(target){
	var NewWin = window.open(target,'','width=800,height=500,scrollbars,resizable')
	}
-->
</script>\n";

    my $db = $share{DB};

    # select from the templates we are allowed to use
    # FIXME: would probably be faster to fetch all templates, but selectively show them

    if ($select) {
	# fetch public templates
	@ptempl = $db->fetch_match("page",{type=>"template",access=>"public"});
	# ... plus the member-only templates from the current section
	push @ptempl, $db->fetch_match("page",{type=>"template",access=>"members",section_id=>$cms->{section}});
	if ($select == 1) {
	    my $sec = $db->fetch("section",$cms->{section});
	    if ($sec && $sec->{parent_id}) {
		# ... plus the member-only templates from the parent section
		push @ptempl, $db->fetch_match("page",{type=>"template",access=>"members",section_id=>$sec->{parent_id}});
	    }
	    # ...plus the private templates from this section
	    push @ptempl, $db->fetch_match("page",{type=>"template",access=>"administrators",section_id=>$cms->{section}});
	}
    }
    else {
	# fetch all templates in the system
	@ptempl = $db->fetch_match("page",{type=>"template"});
    }

   if (@ptempl > 0) {
	foreach my $pt (@ptempl) {
	    $out .= "<p><dl><dt><input type=\"radio\" name=\"$input_name\" value=\"$pt->{page_id}\"";
	    $out .= ($pt->{page_id} == $id) ? " checked" : "";
	    $out .= "> <b>$pt->{title}</b></dt>\n";
	    #my $preview_url = "$config{server}{CGIpath}/page.cgi?_id=$pt->{page_id}')";
	    my $preview_url = "javascript:NewWindow('$config{server}{CGIpath}/page.cgi?_id=$pt->{page_id}')";
	    $out .= "<dd>$pt->{description}  [<A HREF=\"$preview_url\">preview</A>]</dd></dl>\n";
	    #$out .= "TARGET=\"preview\">preview</A>]</dd></dl>\n";
	    # push @options, [$pt->{page_id}, $pt->{title}]; # used for input_html function
	}
    }
    else {
	$out .= "<p>...hmm, there don't seem to be any templates to choose from.  This is not a problem;  just proceed normally.</p>\n"; 
    }
    $out .= "</blockquote>\n";
    return $out;
}

# bulk_import : upload multiple files at once

sub bulk_import {
    my $this = shift;
    my $db = $share{DB};
    my $url = &relink(cmscmd=>"do_import");
    my $page = $db->fetch("page",$this->{page});
    my $type;
    if ($page->{type} eq "template") {
	# default to design if this is a template
	$type = "<select name=\"type\">
<option>editorial</option>
<option selected>design</option>
</select>\n";
    }
    else {
	# default to editorial if this is a page or library
	$type = "<select name=\"type\">
<option selected>editorial</option>
<option>design</option>
</select>\n";
    }
    my $out = "<h1>Bulk Import</h1>
<p>Bulk importing allows you to upload multiple files at once into a single
library, template, or page.  Every file will upload into a new content
object named the same as the file, although the details of each content 
object can be reconfigured later.</p>
<form method=\"post\" enctype=\"multipart/form-data\" action=\"$url\">
<p>Import all files as 
$type
content.</p>
<p>You can upload up to 10 files at a time:</p><p>
<input type=\"file\" name=\"file0\"><br>
<input type=\"file\" name=\"file1\"><br>
<input type=\"file\" name=\"file2\"><br>
<input type=\"file\" name=\"file3\"><br>
<input type=\"file\" name=\"file4\"><br>
<input type=\"file\" name=\"file5\"><br>
<input type=\"file\" name=\"file6\"><br>
<input type=\"file\" name=\"file7\"><br>
<input type=\"file\" name=\"file8\"><br>
<input type=\"file\" name=\"file9\"><br>
</p>
<p><input type=\"checkbox\" name=\"thumb\"> Check here to thumbnail all files.</p>
<p><input type=\"submit\" value=\"Submit\"></p>
</form>
\n";
    return $out;
}

#<form method=\"post\" enctype=\"multipart/form-data\" action=\"/cgi/exsiteonline/debug2.cgi\">
#

sub do_import {
    my $this = shift;
    if (! $this->{page}) {
	return "<p>No page to import into!</p>\n";
    }
    my $db = $share{DB};
    my $data = $this->{input};
    my $out = "<h1>Bulk Import</h1>\n";
    #$out .= &ShowHash($data);
    foreach my $num (0..9) {
	my $name = "file$num";
	$out .= "$name: ";
	my $val = $data->{$name};
	my $fdata = undef;
	my $size = 0;
	my $mimedata = "";
	# read file contents
	my $continue = 1;
	while ($continue) {
	    $continue = read($val,$fdata,1024,$size);
	    $size += $continue;
	}
	# MIME-encode the file contents
	$mimedata = encode_base64($fdata,"");
	my @path = split /[:\\\/]/, $val;
	my $filename;
	if ($path[-1]) { $filename = &url_escape($path[-1]); }
	my $img = new ExSite::Image($filename,$fdata);
	$mimedata = $img->encode;
	if ($size) {
	    # create new content object
	    ### fixme - do not create the cobj if it exists.  This will
	    ### allow for bulk updates :-)
	    my %content = (page_id=>$this->{page},
			   name=>$filename,
			   type=>($data->{type} || "editorial"));
	    my $cid = $db->insert("content",\%content);
	    if ($cid) {
		my $mime = &MimeType($filename);
		my %cdata = (content_id=>$cid,
			     mime_type=>$mime,
			     fdata=>$mimedata,
			     );
		my ($attr, $th_attr, $thumbimg);
		if ($mime =~ /^image/) {
		    # determine image attributes
		    my ($x,$y) = $img->dim;
		    if ($x && $y) {
			$cdata{attributes} = "width=\"$x\" height=\"$y\"";
		    }
		    if ($data->{thumb}) {
			# make a thumbnail
			if ($img->thumb) {
			    ($x,$y) = $img->dim;  # get new dimensions
			    if ($x && $y) { 
				$cdata{thumb_attributes} = "width=\"$x\" height=\"$y\"";
				$cdata{thumb} = $img->encode;
			    }
			}
		    }
		}
		my $cdid = $db->insert("content_data",\%cdata);
		if (! $cdid) {
		    $out .= "failed to install $filename";
		}
		else {
		    $out .= "installed $filename ($size bytes) as $mime";
		}
	    }
	    else {
		$out .= "failed to configure $filename";
	    }
	}
	else {
	    $out .= "no data";
	}
	$out .= "<br>\n";
    }
    my $url = &relink(cmscmd=>undef);
    $out .= "<p><a href=\"$url\">Back to control panel</a></p>\n";
    return $out;
}

sub show_top_menu {
    my $this = shift;
    my $out;
    # page.cgi stores the page record in %share for our use
    my $page_id = $share{Page}->id;
    my $section_id = $share{Page}->id("section");
    my $JS = "$config{server}{HTMLpath}/_ExSite/js";
    my $CGI = "$config{server}{CGIpath}";
    my $HTML = "$config{server}{HTMLpath}";

    # add CBE menu stylesheet
    $out .= "<LINK REL=\"stylesheet\" TYPE=\"text/css\" HREF=\"$JS/menu.css\">\n";

    # add JavaScript libraries
    $out .= "<script type='text/javascript' src='$JS/cbe_core.js'></script>
<script type='text/javascript' src='$JS/cbe_event.js'></script>
<script type='text/javascript' src='$JS/cbe_util.js'></script><!-- for debugging -->
<script type='text/javascript' src='$JS/cbe_debug.js'></script><!-- for debugging -->
<script type='text/javascript' src='$JS/menu_exsite.js'></script>

<script type='text/javascript'><!--
var mark1, menu1, menu2, menu3;
function windowOnload() {
  mark1 = cbeGetElementById('menuMarker1').cbe;
  menu1 = new cbeMenu(
    true,                       // horizontal,
//    (document.cbe.width()-(3*75))/2, mark1.pageY(), 3, // centered
    mark1.pageX(), mark1.pageY(), 3, // menuX, menuY, menuZ (left-justified)
    [75,75,75], 20, 8,          // lblWidthsArray, lblHeight, lblspacing
    [140,170,155],              // boxWidthsArray,
    '#000099', '#000000',       // activeColor, inactiveColor,
    '#FFFFFF', '#C0C0C0',       // activeBkgnd, inactiveBkgnd
    '#C0C0C0'                   // boxBkgnd
  );
  window.cbe.addEventListener(\"resize\", resizeListener, false);
}

--></script>\n";

    # now build the HTML for the menubar

    $out .= "<a name=\"topofpage\"></a>
 <TABLE WIDTH=100% CELLSPACING=0 CELLPADDING=0 BORDER=0> 
  <TR>
   <TD BGCOLOR=#DDDDDD COLSPAN=4><IMG SRC=null.gif HEIGHT=1 WIDTH=1></TD>
  </TR>
   <TR>
   <TD BGCOLOR=#DDDDDD><IMG SRC=null.gif HEIGHT=1 WIDTH=1></TD>
   <TD BGCOLOR=#C0C0C0>
     &nbsp;&nbsp;
   </TD>
   <TD BGCOLOR=#C0C0C0 WIDTH=100%>
    <div id='menuMarker1' class='menuMarker'>&nbsp;</div>
   </TD>
   <TD BGCOLOR=#666666><IMG SRC=null.gif HEIGHT=1 WIDTH=1></TD>
  </TR>
  <TR>
   <TD BGCOLOR=#666666 COLSPAN=4><IMG SRC=null.gif HEIGHT=1 WIDTH=1></TD>
  </TR>
 </TABLE>\n";

    # now create the menus...

    # menu 1 : website manager

    $out .= "<div id='label11' class='mLabel'>Website</div>
<div id='box11' class='mBox'>
  <a class=\"m\" href=\"".&relink(cmscmd=>"map2page",path=>"/$section_id")."\">Site Map</a>
  <a class=\"m\" href=\"".&relink(cmscmd=>undef,path=>"/$section_id")."\">Advanced Tools</a>
  <a class=\"m\" href=\"$CGI/publish.cgi?section=$section_id\" target=publish>Publish Site</a>
</div>\n";

    # menu 2 : page manager

    ########### view published link will fail for non-home sections

    $out .= "<div id='label12' class='mLabel'>Page</div>
<div id='box12' class='mBox'>
  <a class=\"m\" href=\"$HTML/$share{Page}->{page}->{filename}\">View Published</a>
  <a class=\"m\" href=\"$CGI/page.cgi?_id=$page_id\">View Current</a>
  <a class=\"m\" href=\"$CGI/page.cgi?_id=$page_id&_hot=1&_cms=1\">View Editable</a>
  <hr>
  <a class=\"m\" href=\"".&relink(cmscmd=>"changebody",path=>"/$section_id/$page_id")."\">Change Body</a>
  <a class=\"m\" href=\"".&relink(cmscmd=>"editpage",path=>"/$section_id/$page_id")."\">Page Settings</a>
  <a class=\"m\" href=\"".&relink(cmscmd=>undef,path=>"/$section_id/$page_id")."\">Advanced Tools</a>
  <hr>
  <a class=\"m\" href=\"$CGI/publish.cgi?page=$page_id\" target=publish>Publish Page</a>
  <a class=\"m\" href=\"".&relink(cmscmd=>"newpage",path=>"/$section_id/$page_id")."\">New Page</a>
  <a class=\"m\" href=\"".&relink(cmscmd=>"deletepage",path=>"/$section_id/$page_id")."\">Delete Page</a>
</div>\n";

    # menu 3 : help

    $out .= "<div id='label13' class='mLabel'>Help</div>
<div id='box13' class='mBox'>
  <a class=\"m\" href=\"javascript:alert('Help not enabled.')\">Help</a>
  <a class=\"m\" href=\"javascript:alert('About not enabled.')\">About</a>
  <hr>
  <a class=\"m\" target=\"log\" href=\"$CGI/exsite_log.cgi\">View System Logs</a>
</div>\n";

    # menu 4 : exit

    $out .= "<div id='label14' class='mLabel'>Exit</div>
<div id='box14' class='mBox'>
  <a class=\"m\" href=\"$CGI/home.cgi\">Close</a>
  <a class=\"m\" href=\"$CGI/page.cgi?_id=$page_id&_cms=0\">Turn off CMS - this page</a>
  <a class=\"m\" href=\"$CGI/page.cgi?_id=$page_id&_cmsmode=0\">Turn off CMS - all pages</a>
  <a class=\"m\" href=\"$CGI/logout.cgi\">Log out</a>
</div>\n";

    return $out;
}

sub redirect {
    my ($this,$reply) = @_;
    if ($share{Page}) {
	$share{Page}->redirect($reply);
    }
    elsif ($share{PageBuf}) {
	$share{PageBuf}->location($reply);
    }
    else {
	print "Location: $reply\n\n";
    }
}

#=======================================================================
# handlers
#=======================================================================

sub select_foreign_key_handler {
    my ($this,$tab,$rec,$col,$ftab) = @_;
    my $db = $share{DB};
    if ($tab eq "page" && $ftab eq "page") {
	if ($col eq "parent_id") {
	    my $p = $db->fetch("page",$rec);
	    my $pages = $share{DB}->fetch_match("page",{type=>"page",section_id=>$p->{section_id}});
	    #### should probably drop the current page from this list to
	    #### avoid circular page references
	    return $pages;
	}
	elsif ($col eq "template_id") {
	    my $pages = $share{DB}->fetch_match("page",{type=>"template"});
	    return $pages;
	}
    }
    elsif ($tab eq "content" && $ftab eq "page") {
	my $c = $db->fetch("content",$rec);
	my $p = $db->fetch("page",$c->{page_id});
	my $pages = $share{DB}->fetch_match("page",{section_id=>$p->{section_id}});
	return $pages;
    }
}


#=======================================================================
# wizards
#
# Wizards are multi-screen forms that save their inputs in hidden fields
# until the last screen is submitted.
#=======================================================================
#
#                          *** WARNING ***
#
# The method &wizard belongs to the current (CMS) class.  All remaining
# wizard methods are dynamically scoped to the Wizard class.  Therefore
# private CMS variables will not be available via $this.  If you need
# to access data/methods in the CMS object, use the global $cms instead.
#
#=======================================================================

sub wizard {
    my $this = shift;
    my ($type,$title,$init) = @_;

    my $action = &relink();
    my $wiz = new ExSite::Wizard(title=>$title,(action=>$action));

    if ($type eq "content_data") {
	#$wiz->add_step("Update Method",\&wiz_cdata_page1,\&do_wiz_cdata_page1);
	$wiz->add_step("Enter New Content",\&wiz_cdata_page2,\&do_wiz_cdata_page2);
	$wiz->add_step("Content Type and Comment",\&wiz_cdata_page3,\&do_wiz_cdata_page3);
    }
    elsif ($type eq "content") {
	$wiz->add_step("Content Parameters",\&wiz_content_page1,\&do_wiz_content_page1);
	#$wiz->add_step("Update Method",\&wiz_cdata_page1,\&do_wiz_cdata_page1);
	$wiz->add_step("Enter New Content",\&wiz_cdata_page2,\&do_wiz_cdata_page2);
	$wiz->add_step("Content Type and Comment",\&wiz_cdata_page3,\&do_wiz_cdata_page3);
    }
    elsif ($type eq "page") {
	$wiz->add_step("Details",\&wiz_page_page1,\&do_wiz_page_page1);
	$wiz->add_step("Template",\&wiz_page_page2,\&do_wiz_page_page2);
	$wiz->add_step("Body",\&wiz_page_page3,\&do_wiz_page_page3);
    }
    elsif ($type eq "template") {
	$wiz->add_step("Details",\&wiz_template_page1,\&do_wiz_template_page1);
	$wiz->add_step("Parent Template",\&wiz_template_page2,\&do_wiz_template_page2);
    }
    elsif ($type eq "import_template") {
	$wiz->add_step("Details",\&wiz_import_template_page1,\&do_wiz_import_template_page1);
	$wiz->add_step("Upload Graphics and Other Files",\&wiz_import_template_page2,\&do_wiz_import_template_page2);
    }
    elsif ($type eq "website_template_page") {
	$wiz->add_step("Details",\&wiz_page_page1,\&do_wiz_page_page1);
	# no template selection in this case
	$wiz->add_step("Body",\&wiz_templatepage_page2,\&do_wiz_templatepage_page2);
    }
    elsif ($type eq "library") {
	$wiz->add_step("Details",\&wiz_library_page1,\&do_wiz_library_page1);
    }
    elsif ($type eq "section") {
	$wiz->add_step("Details",\&wiz_section_page1,\&do_wiz_section_page1);
	$wiz->add_step("Location",\&wiz_section_page2,\&do_wiz_section_page2);
	$wiz->add_step("Pages",\&wiz_section_page3,\&do_wiz_section_page3);
	$wiz->add_step("Webmaster",\&wiz_section_page4,\&do_wiz_section_page4);
    }
    elsif ($type eq "easysection") {
	$wiz->add_step("Identification",\&wiz_easy_section_page1,\&do_wiz_easy_section_page1);
	$wiz->add_step("Website Type",\&wiz_easy_section_page2,\&do_wiz_easy_section_page2);
	$wiz->add_step("Pick a Design",\&wiz_easy_section_page3,\&do_wiz_easy_section_page3);
	$wiz->add_step("Instructions",\&wiz_easy_section_page4,\&do_wiz_easy_section_page4);
    }

    return $wiz->go($init);

}

########################
# content_data wizard
########################

sub wiz_cdata_page1 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{content_id};
    my ($out, %content, @cdata, $selected);
    my $db = $share{DB};
    # set reply page on first screen
    if (! exists $in->{reply}) {
	my $reply = $ENV{HTTP_REFERER} || &relink(cmscmd=>undef);
	$out .= $db->input_html(type=>"hidden",name=>"reply",value=>$reply);
    }
    if ($id =~ /\d+/) {
	%content = $db->fetch("content",$id);
	@cdata = $db->fetch_child("content_data","content",$id,"content_data_id");
    }
    if (@cdata == 0) {
	$cdata[0] = { content_data_id=>"_2", mime_type => "text/html" };  # default mime type
    }
    my $cid = $cdata[-1]{content_data_id};
    my $cid = $cdata[-1]{content_data_id};
    $out .= $db->input_html(type=>"hidden",name=>"content_data#_2#content_id", value=>$id);
    my $name = $content{name} || $in->{"content#${id}#name"};
    $out .= "<p>This wizard will guide you through the steps of updating \"$name\".</p>\n";
    if (! exists $in->{content_id}) {
	# we've not come from the content wizard, so we'll need to include
	# a reply screen
	$out .= "<input type=hidden name=\"reply\" value=\"$in->{reply}\">\n";
    }
    $out .= "<input type=hidden name=\"content_data_id\" value=\"$cid\">\n"; 
#    if ($id && @cdata > 0) {
#	######### should in-line the content into the same screen
#	$out .= "<p><a href=javascript:popup_link(\"$CGI/viewcontent.cgi?id=$cid\")>View the latest version of \"$content{name}\" here</a>.</p>\n";
#	$out .= "<script LANGUAGE=javascript>\nfunction popup_link(url) {\n  window.open(url,\"\",\"width=400,height=400,resizable,scrollbars=1,menubar=0\")\n}\n</SCRIPT>\n";
#    }
    $out .= "<p>Please select how you will be updating this content:</p>\n";
    if ($cdata[-1]->{mime_type} =~ /^text\/html$/o) {
	$out .= "<p><input type=radio name=\"update_method\" value=\"html\" checked> Edit using WYSIWYG HTML editor</p>\n";
	$selected = 1;
    }
    else {
	$out .= "<p><input type=radio name=\"update_method\" value=\"newhtml\"> Replace with HTML, using WYSIWYG editor</p>\n";
    }
    if ($cdata[-1]->{mime_type} =~ /^text/o) {
	$out .= "<p><input type=radio name=\"update_method\" value=\"edit\"";
	if (! $selected) { $out .= " checked"; $selected = 1; }
	$out .= "> Edit using a simple text editor</p>\n";
    }
    else {
	$out .= "<p><input type=radio name=\"update_method\" value=\"newedit\"> Replace with text, using a simple text editor</p>\n";
    }
    $out .= "<p><input type=radio name=\"update_method\" value=\"upload\"";
    if (! $selected) { $out .= " checked"; $selected = 1; }
    $out .= "> Upload new version in a file</p>\n";
    $out .= "<p><input type=radio name=\"update_method\" value=\"file\"> Point to a file already in place on the server</p>\n";
    $out .= "<p><input type=radio name=\"update_method\" value=\"url\"> Point to a 3rd-party resource from another website</p>\n";
    return $out;
}

sub wiz_cdata_page2 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my $id;
    my ($out, %content, @cdata, $selected);
    
    # set reply page on first screen
    if (! exists $in->{reply}) {
	my $reply;
	if ($ENV{HTTP_REFERER} =~ /page\.cgi/) {
	    $reply = $ENV{HTTP_REFERER};
	}
	else {
	    $reply = $cms->goto(@{$cms->{path}}[1..2]);
	}
	$out .= $db->input_html(type=>"hidden",name=>"reply",value=>$reply);
    }

    if (exists $in->{'content_data#_2#content_id'}) {
	#$out .= "resubmit! removing content_data values<br>\n";
	# this is a resubmit of screen 1 - delete all record data so the form
	# can be regenerated fresh (with a new update method, though)
#### warning: this code causes segmentation faults (!?) 
#	foreach my $key (%$in) {
#	    if ($key =~ /$config{form}{sepchar}/) {
#		delete $in->{$key};
#	    }
#	}
#### we'll use the brute-force method instead...
	$id = $in->{'content_data#_2#content_id'};
	delete $in->{'content_data#_2#content_id'};
	delete $in->{'content_data#_2#data#html'};
	delete $in->{'content_data#_2#fdata#filename'};
    }
    else {
	#$out .= "no resubmit - adding content_id values<br>\n";
	# get the content id from the init data
	$id = $in->{content_id};
    }
    $out .= $db->input_html(type=>"hidden",name=>"content_data#_2#content_id", value=>$id);

    if ($id =~ /\d+/) {
	%content = $db->fetch("content",$id);
	@cdata = $db->fetch_child("content_data","content",$id,"content_data_id");
    }
    if (@cdata == 0) {
	# set an assumed mime type, depending on content type
	if ($in->{'content#_1#type'} eq "design") {
	    # also check for content#_1#allow_mime#_set_N = "image"
	    $cdata[0] = { content_data_id=>"_2", mime_type => "image/gif" };
	}
	else {
	    $cdata[0] = { content_data_id=>"_2", mime_type => "text/html" };
	}
    }
    my $cid = $cdata[-1]{content_data_id};
    my $name = $content{name} || $in->{"content#${id}#name"};
    $out .= "<p>This wizard will guide you through the steps of updating \"$name\".</p>\n";
#    if (! exists $in->{content_id}) {
#	########## will always exist due to line 3
#	# we've not come from the content wizard, so we'll need to include
#	# a reply screen
#	$out .= "<input type=hidden name=\"reply\" value=\"$in->{reply}\">\n";
#    }
    if (! exists $in->{content_data_id}) {
	$out .= "<input type=hidden name=\"content_data_id\" value=\"$cid\">\n"; 
    }

    my $update_method = $in->{update_method};
    if ($update_method =~ /$config{form}{multi_sep}/) {
	# we now have multiple update methods (due to reselecting the 
	# update method several times).  Delete all but the most recent.
	my @update_methods = split /$config{form}{multi_sep}/,$update_method;
	$update_method = @update_methods[-1];
	#$in->{update_method} = $update_method;
    }

    if (! $update_method) {
	if ($cdata[-1]->{mime_type} =~ /^text\/html$/o) {
	    $update_method = "html";
	}
	elsif ($cdata[-1]->{mime_type} =~ /^text/o) {
	    $update_method = "edit";
	}
	else {
	    $update_method = "upload";
	}
    }

    $out .= "<script>
function formsubmit(){
    document.forms[0].screen.value--;
    document.forms[0].submit();
}
</script>\n";
    $out .= "<p>Update method: \n<select name=\"update_method\" onchange=\"formsubmit()\">\n";
    $out .= "<option value=\"html\"";
    if ($update_method =~ /html/) { $out .= " selected"; }
    $out .= ">Edit using WYSIWYG HTML editor</option>\n";
    $out .= "<option value=\"edit\"";
    if ($update_method =~ /edit/) { $out .= " selected"; }
    $out .= ">Edit using a simple text editor</option>\n";
    $out .= "<option value=\"upload\"";
    if ($update_method eq "upload" || ! $update_method) { 
	$out .= " selected"; 
    }
    $out .= ">Upload new version in a file</option>\n";
    $out .= "<option value=\"file\"";
    if ($update_method eq "file") { $out .= " selected"; }
    $out .= ">Point to file on server</option>\n";
    $out .= "<option value=\"url\"";
    if ($update_method eq "url") { $out .= " selected"; }
    $out .= ">Point to URL</option>\n";
    $out .= "</select>\n";
    $out .= "<br><img src=\"$config{server}{HTMLpath}/_ExSite/images/caution20.gif\" height=\"20\" width=\"20\" border=\"0\" align=\"middle\"> Changing the update method will revert to the most recent saved content.\n";

    # step 2
    my $textdata;
    if ($update_method !~ /^new/ && $cid !~ /^_/) {
	my $c = new ExSite::ContentData(content_data=>$cdata[-1]);
	if ($c->is_text) {
	    $textdata = $c->get_raw;
	}
    }
    if ($update_method =~ /html$/ && $db->use_wysiwyg) {
	$out .= "<p>You may enter your content changes directly into this WYSIWYG editor:</p>\n";
	$out .= $db->input_htmleditor(name=>"content_data#_2#data",
				      table=>"content_data",
				      record=>$cid,
				      page=>$cms->{page},
				      section=>$cms->{section},
				      column=>"data",
				      value=>"$textdata",
				      editor=>{
					  op_image=>1,
					  op_link=>1,
					  op_table=>1,
					  op_dyncontent=>1,
				      });
	$out .= "<p>Note: unresolved content is represented using the missing graphic image(<img src=$config{server}{HTMLpath}/_ExSite/images/checker.png HEIGHT=10 WIDTH=20 border=1>).  This missing content can only be shown when this content is displayed as part of a complete page.</p>\n";
    }
    elsif ($update_method =~ /^upload$/) {
	$out .= "<p>Click on browse to select a file to upload:</p>\n";
	$out .= "<p><input type=file size=60 name=\"content_data#_2#fdata#filename\"></p>\n";
	$this->{db}->{enctype} = "multipart/form-data";
	######################## need a multipart-encoded form!
	########## should be done by Form
    }
    elsif ($update_method =~ /^file$/) {
	$out .= "<p>Enter the full path to the file on the server:</p>\n";
	$out .= "<p><input type=text size=80 name=\"content_data#_2#file\"></p>\n";
	$out .= "<p>WARNING: content in external files is outside of ExSite's revision control system.  The file may be deleted or modified without ExSite's knowledge, which could affect the correct rendering of your page.  To bring this file into ExSite's revision control, use the \"Upload new version in a file\" option, instead. (You may first have to copy the file from the server to your computer.)</p>\n";
    }
    elsif ($update_method =~ /^url$/) {
	$out .= "<p>Enter the full URL to the web-based content:</p>\n";
	$out .= "<p><input type=text size=80 name=\"content_data#_2#url\"></p>\n";
	$out .= "<p>WARNING: content on remote websites is beyond the control of ExSite.  If the remote webmaster changes or removes this content, your page may not render correctly.</p>\n";
	$out .= "<p>Also note that remote content may be covered by copyright.  Although technically you are not copying this content, you may end up giving the appearance of having done so.  Please ensure that you will not be offending the owners of this content by linking to it in this way.</p>\n";
    }
    else {
	# default to text area
	$out .= "<p>You may enter your content changes directly into this text editing area:</p>\n";
	$out .= "<p><textarea cols=80 rows=20 wrap=virtual name=\"content_data#_2#data\">\n";
	$out .= $textdata;
	$out .= "</textarea></p>\n";
    }
    return $out;
}

sub wiz_cdata_page3 {
    my $this = shift;
    # screen's html in a hidden input
    my $in = shift;
    my $id = $in->{content_data_id};
    my $out;
    my $db = $share{DB};
    
    # 1) mime-type

    my $mimetype;
    if ($in->{update_method} =~ /html/) {
	$mimetype = "text/html";
    }
    elsif ($in->{update_method} =~ /edit/) {
	# could be html, plain text, or css
	#$out .= &ShowHash($in);
	if ($in->{"content_data#_2#data"} =~ /<(!--|[\/a-zA-Z])/) {
	    $mimetype = "text/html";
	}
	elsif ($in->{"content_data#_2#data"} =~ /\/\*|[\w-]+:\s*[\w\#]+\;|^\w*\.\w+\s*{/) {
	    $mimetype = "text/css";
	}
	else {
	    $mimetype = "text/plain";
	}
    }
    else {
	my $file;
	if ($in->{update_method} =~ /upload/) {
	    ($file,undef) = split(/\#/,$in->{"content_data#_2#fdata"});
	}
	else {
	    $file = $in->{"content_data#${id}#file"} || $in->{"content_data#_2#url"};
	}
	# guess based on file name
	$mimetype = &MimeType($file);
    }

    # above mimetype is just our best guess.  Show options, so that it
    # can be over-ridden.

    $out .= "<ul>\n";
    $out .= "<li><p>ExSite thinks this content is of type ";
    $out .= $db->input_exsite(datatype=>"list:mime", name=>"content_data#_2#mime_type", 
			      value=>$mimetype);
    $out .= "</p>\n<p>(If ExSite has guessed incorrectly, you can set the content type manually.)</p></li>\n";

    # 2) image attributes

    if ($mimetype =~ /^image/) {
	my $img = new ExSite::Image($in->{"content_data#_2#fdata"});
	my $info = $img->info;
	my $attributes;
	if (exists $info->{width}) {
	    $attributes = "width=&#34;$info->{width}&#34; height=&#34;$info->{height}&#34; border=&#34;0&#34;";
	}
	$out .= "<li><p>Specify any image attributes you would like applied to this image (optional):</p>\n";
	$out .= $db->input_exsite(datatype=>"string", size=>60, name=>"content_data#_2#attributes", value=>$attributes);
	$out .= "<p><input type=\"checkbox\" name=\"thumb\"> check here to thumbnail this image</p>\n";
	$out .= "</li>\n";
    }

    # 3) revision control comment

    my $comment = "updated by ".$db->my_name." on ".scalar(localtime)."\n";
    $out .= "<li><p>For the purposes of auditing, please include a note describing the changes you made:</p>\n";
    $out .= $db->input_exsite(datatype=>"text",size=>100,name=>"content_data#_2#note",value=>$comment);
    $out .= "</li></ul>\n";

    $out .= "<p>Note that content updates may require approval before becoming visible on public pages.</p>\n";
	
    if ($mimetype =~ /^text/) {
	$out .= "<p><img src=\"$config{server}{HTMLpath}/_ExSite/images/caution20.gif\" height=\"20\" width=\"20\" border=\"0\" align=\"middle\"> Hitting \"Back\" may revert to the last saved version of your content.</p>\n";
    }

    return $out;
}

sub do_wiz_cdata_page1 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my @err;
    if (! $in->{update_method}) {
	push @err, "Please specify which update method you want to use.";
    }
    return ();
}

sub do_wiz_cdata_page2 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my @err = &do_wiz_cdata_page1($this,$in);
    if (! ($in->{"content_data#_2#data"} || $in->{"content_data#_2#fdata"} || 
	   $in->{"content_data#_2#file"} || $in->{"content_data#_2#url"})) {
	push @err, "Please provide some content!";
    }
    # set up file data
    if ($in->{"content_data#_2#fdata#filename"}) {
	$in->{"content_data#_2#fdata"} = $this->{db}{form}{content_data}{_2}{fdata};
	delete $in->{"content_data#_2#fdata#filename"};
    }
    return ();
}

sub do_wiz_cdata_page3 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    $this->unescape_input($in);
    my $id = $in->{content_data_id};
    my @err;
    if ($id) {
	@err = &do_wiz_cdata_page2($this,$in);
	if (! $in->{"content_data#_2#mime_type"}) {
	    push @err, "Please ensure that a content type is selected.";
	}
	if (! $in->{"content_data#_2#note"}) {
	    push @err, "Please provide a note explaining this content update.";
	}
    }
    return @err if (@err > 0);

    # fix reply value (KLUDGE!)  ############

    if ($db->{data}{reply} =~ /;/) {
	my @replies = split /;/, $db->{data}{reply};
	$db->{data}{reply} = $replies[0];
	push @err, "reply = $db->{data}{reply}";
    }
    return @err if (@err > 0);

    if ($in->{thumb}) {
	# request to thumbnail the image
	if ($in->{"content_data#_2#mime_type"} =~ /^image/) {
	    my $img = new ExSite::Image($in->{"content_data#_2#fdata"});
	    if ($img->thumb) {
		my ($x,$y) = $img->dim;  # get new dimensions
		$db->{form}->{content_data}->{_2}->{thumb} = 
		    "th_".$img->encode;
		$db->{form}->{content_data}->{_2}->{thumb_attributes} = 
		    "width=&#34;$x&#34; height=&#34;$y&#34;";
	    }
	}
    }

    # if we get here, we're good to go

    $db->{action} = "insert";

    # note to self: why do we have to do this??
    $db->{form}->{content_data}->{_2}->{data} =~ s/%7B/{/g;
    $db->{form}->{content_data}->{_2}->{data} =~ s/%7D/}/g;

    my $do_out = $db->do;
    if ($db->errorcheck) {
	return ($do_out);
    }
    if (! $db->{last_insert_id}) {
	return ("Content update failed!");
    }
    return ();
}

########################
# content wizard
########################

sub wiz_content_page1 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{content_id};
    my $db = $share{DB};
    my %content;
    if ($id) {
	%content = $db->fetch("content",$id);
	$db->{action} = "update";
    }
    else {
	$id = "_1";
	%content = (content_id => $id );
	$db->{action} = "insert";
    }
    my $out = "<p>A content item is like a variable that can contain website content.  
It has a name, which used to reference the content, and 
values (or versions), which can change over time.</p>
<p>On this first screen, you define the content's name, category, and the 
types of content it can legally hold.</p>\n";

    # set reply page on first screen
    if (! exists $in->{reply}) {
	my $reply = $ENV{HTTP_REFERER} || &relink(cmscmd=>undef);
	$out .= $db->input_html(type=>"hidden",name=>"reply",value=>$reply);
    }

    $out .= "<p>Name of content item:<br>\n";
    $out .= $db->input_column(table=>"content",column=>"name",size=>20,
			      record=>$id,data=>\%content);
    $out .= "</p>\n<p>A brief description of this content item (optional):<br>\n";
    $out .= $db->input_column(table=>"content",column=>"description",
			      record=>$id,data=>\%content);
    $out .= "</p>\n<p>Content category (determines who is allowed to update this content item):<br>\n";
    $out .= $db->input_column(table=>"content",column=>"type",
			      record=>$id,data=>\%content);
    $out .= "</p>\n<p>The allowed types of content for this content item (uncheck everything to allow all content types):<br>\n";
    $out .= $db->input_column(table=>"content",column=>"allow_mime",
			      record=>$id,data=>\%content);

    $out .= "\n<br><br><P>".$db->input_html(type=>"checkbox",name=>"skip_update")." Skip remaining content-update steps.  Select this if not actually creating/modifying this content at this time.</p>\n";
    $out .= $db->input_html(type=>"hidden", name=>"reply", value=>$in->{reply});
    $out .= $db->input_html(type=>"hidden", name=>"content_id", value=>$id);
    $out .= $db->input_html(type=>"hidden", name=>"content#${id}#page_id", value=>($content{page_id}||$cms->{page}));
			    
    return $out;
}

sub do_wiz_content_page1 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{content_id};
    my $db = $share{DB};
    my @err;
    if (! $in->{"content#${id}#name"}) {
	push @err, "Please specify a name for this content item.";
    }
    elsif ($in->{"content#${id}#name"} !~ /^[-\w]+\.?\w+$/) {
	push @err, "The content name can only contain alphanumeric characters, and an optional file suffix - no spaces!";
    }
    if (! $in->{"content#${id}#type"}) {
	push @err, "Please specify a category for this content item.";
    }
    return @err if (@err > 0);
    if ($in->{skip_update}) {
	# manually reset to last step
	$this->{current} = 3;

	$db->{action} = "insert";
	my $do_out = $db->do;
	if ($db->errorcheck) {
	    return ($do_out);
	}
	if (! $db->{last_insert_id}) {
	    return ("Content update failed!");
	}
    }
    return ();
}

########################
# page wizard
########################

sub wiz_page_page1 {
    my $this = shift;
    my $in = shift;
    my %page;
    my $db = $share{DB};
    my $id = "_1";
    $db->{action} = "insert";
    my $out;
    $out .= $db->input_html(name=>"page#${id}#type",
			    value=>"page",
			    type=>"hidden");
    if ($in->{reply}) {
	$out .= $db->input_html(name=>"reply",
				type=>"hidden",
				value=>"$in->{reply}");
    }
    
    $out .= "<h4>Page Description</h4>\n<blockquote>\n";
    ###### next line sets primary key to "_1"
    $out .= $db->input_html(name=>"page_id",type=>"hidden",value=>$id); #######
    $out .= $db->input_html(name=>"page#${id}#section_id",
			    type=>"hidden",
			    value=>$cms->{section});
    $out .= "<p>Title:<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"title",
			      record=>$id);
    $out .= "<p>Description:<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"description",
			      record=>$id);
    $out .= "<p>Status:<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"status",
			      record=>$id,
			      value=>"active");
    $out .= "</p>\n</blockquote>\n";
    $out .= "<h4>Page Viewing Parameters</h4>\n<blockquote>\n";
    $out .= "<p>Accessibility:<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"access",
			      record=>$id,
			      value=>"public");
    $out .= "</p>\n<p>Is this page static or dynamic?<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"publish_method",
			      record=>$id,
			      value=>"static");
    $out .= "</p>\n</blockquote>\n";

    $out .= "<h4>Page Positioning</h4>\n<blockquote>\n";
    $out .= "<p>Is this page visible in menus and site maps?<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"visibility",
			      record=>$id,
			      value=>"visible");
    $out .= "</p>\n<p>Specify the parent page (sets the position in menus and site maps):<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"parent_id",
			      record=>$id, 
			      value=>$in->{parent_id},
			      fmatch=>{section_id=>$cms->{section}, type=>"page"});
    $out .= "</p>\n<p>Menu Rank (typically 0-9):<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"rank",
			      record=>$id,
			      value=>1);
    $out .= "</p>\n<p>Menu label for this page:<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"label",
			      record=>$id);
    $out .= "</p>\n</blockquote>\n";

#    $out .= "<h4>Other Page Settings</h4>\n<blockquote>\n";
#    $out .= $db->input_column(table=>"page",
#			      column=>"flags",
#			      record=>$id);
#    $out .= "<input type=\"hidden\" name=\"page#_1#flags#_set_0\" value=\"\">
#<input type=\"checkbox\" name=\"page#_1#flags#_set_1\" value=\"not searchable\"> This page should NOT be indexed by search tools.<br>
#<input type=\"checkbox\" name=\"page#_1#flags#_set_2\" value=\"fast\"> Precompile dynamic pages for better performance.<br>\n";

    $out .= "\n</blockquote>\n";

    $out .= "<h4>Meta-information</h4>\n<blockquote>\n";
    $out .= "</p>\n<p>Keywords:<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"keywords",
			      record=>$id);
    $out .= "</p>\n";

    $out .= "</p>\n</blockquote>\n<h4>Publishing info</h4>\n<blockquote>\n<p>File name (must end in \".htm\", \".html\", \".php\", or \".asp\"):<br>\n";
    $out .= $db->input_column(table=>"page",
			      column=>"filename",
			      record=>$id);
    $out .= "</p>\n</blockquote>\n";
    return $out;
}

sub do_wiz_page_page1 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{page_id};
    my $db = $share{DB};
    my @err;
    if (! $in->{"page#${id}#type"}) {
	push @err, "No page type specified!";
    }
    if (! $in->{"page#${id}#title"}) {
	push @err, "Please specify a page title.";
    }
    if (! $in->{"page#${id}#status"}) {
	push @err, "Please specify a page status.";
    }
    if (! $in->{"page#${id}#access"}) {
	push @err, "Please specify the page accessibility.";
    }
    if ($in->{"page#${id}#access"} ne "public" && 
	$in->{"page#${id}#publish_method"} eq "static") {
	push @err, "Only public pages can be static;  please select a different accessibility or static/dynamic setting.";
    }
    if (! $in->{"page#${id}#visibility"}) {
	push @err, "Please specify the page visibility.";
    }
    if ($in->{"page#${id}#visibility"} ne "hidden" && ! $in->{"page#${id}#label"}) {
	push @err, "Please specify a menu label.";
    }
    if ($in->{"page#${id}#filename"} !~ /^[\w-]+\.(html?|php|asp)$/) {
	push @err, "Invalid filename -- please use the characters A-Z, a-z, 0-9, -, and _ only.  The filename must end in \".htm\", \".html\", \".php\", or \".asp\".";
    }
    return @err;
}

sub wiz_page_page2 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my $out;
    my %page;
    my $id = $in->{page_id};

    $out .= "<p>Select the template that will define this page's layout, 
graphics, and style.  (Click on the preview links to view the templates.)</p>\n";

    $out .= &list_template($this,"page#_1#template_id#key",undef,1);

    $out .= "</p>\n<p>If you copy the layout (the \"page\" object) from 
the template, you will be able to make layout customizations for this 
page only.  However, you will no longer see updates to the layout of the 
template.  Only select this option if you are certain that this page
requires customizations to the layout.</p>
<blockquote><p>\n".
        $db->input_html(type=>"checkbox",name=>"copy_layout",selected=>0).
        " Copy Page Layout</p></blockquote>\n";

    return $out;
}

sub do_wiz_page_page2 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{page_id};
    my $db = $share{DB};
    my @err = &do_wiz_page_page1($this,$in);
    if ($in->{skip_template} || ! $in->{"page#${id}#template_id#key"}) {
	push @err, "Please select a template.";
    }
    return @err;
}

sub wiz_page_page3 {
    my $this = shift;
    my $in = shift;
    my $out;
    my %page;
    my $pdata;
    my $db = $share{DB};
    my $id = $in->{page_id};

    ########### compare to custom version, if it exists

    %page = $db->fetch("page",$in->{"page#${id}#template_id#key"});
    my $p = new ExSite::Page(page=>\%page);
    $p->set_context;
    my $pdata = $p->get_html;

    if ($pdata =~ /\<\!--\&?content\(body\)--\>/) {
	# this page has a conventional body
	# list all body templates available
	my (@liblist,$libs,$templates,$sql);
        $libs = $db->fetch_match('page',{type=>"library",section_id=>$cms->{section}});
	$out .= "<p>This page appears to have a conventional layout, with a \"body\" area for holding page-specific content.</p>\n";
	$out .= "<input type=hidden name=content#_2#name value=\"body\">
<input type=hidden name=content#_2#type value=\"editorial\">
<input type=hidden name=content#_2#page_id value=\"$id\">\n";
	if (@$libs > 0) {
	    $out .= "<p>You can optionally select a pre-defined body for this 
page, from the following list of available body templates.</p>\n<p>";
	    foreach my $l (@$libs) {
		push(@liblist, $l->{page_id});
	    }
	    ######### FIXME: re-write this to avoid a custom query
	    $sql = "select * from content where name='body' and page_id in (".join(',',@liblist).")";
	    $templates = $db->custom_query($sql);
	    #$out .= "<input type=\"hidden\" name=\"content_data#_3#content_id\" value=\"_2\">\n";
	    $out .= "<input type=\"radio\" name=\"body_id\" value=\"0\" checked> normal body, fully editable (will be blank to start with)<br>\n";
	    foreach my $t (@$templates) {
		$out .= "<input type=\"radio\" name=\"body_id\" value=\"$t->{content_id}\">$t->{description}<br>\n";
	    }
	}
    }
    elsif (! $pdata) {
	# this page has no framework 
	$out .= "<IMG SRC=$config{server}->{HTMLpath}/_ExSite/images/caution.png HEIGHT=44 WIDTH=53 ALIGN=left>
<p>This page has no layout defined at this time.  There is no further
pre-configuration ExSite can do on this page -- please proceed
by clicking on FINISH.</p>
<p>Later, you will need to manually add the basic layout framework for this page.
To do this, create a content item named \"page\", and fill it with an
ExSite template (HTML with embedded ExSite templating tags).</p>\n";
    }
    else {
	# this page has an irregular layout, with no body
	$out .= "<IMG SRC=$config{server}->{HTMLpath}/_ExSite/images/caution.png HEIGHT=44 WIDTH=53 ALIGN=left>
<p>This page appears to have an irregular layout, with no explicit 
\"body\" area for holding page-specific content.  This is not an error,
since some pages have special layout requirements that cannot be met by
a conventional layout.  However, there is
no further pre-configuration ExSite can do on this page -- please proceed
by clicking on FINISH. Customization of this page can be performed
using the regular content-management tools, once the page is created.</p>\n";
    }

    $out .= "<br><br><br>\n";

    return $out;
}

sub do_wiz_page_page3 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{page_id};
    my $db = $share{DB};

    my @err = &do_wiz_page_page2($this,$in);
    if (@err > 0) { return @err; }
    
    # now, perform all the necessary DB mods

    # copy layout

    if ($in->{copy_layout}) {  ## && $in->{'page#_1#template_id#key'}
	&wiz_copy_page_layout($this,$in);
    }
    
    # copy selected page body

    if ($in->{body_id}) {
	# substitute the selected body into the form data
	my $c_data = $db->fetch_child("content_data","content",$in->{"body_id"},"content_data_id");
	$c_data = $c_data->[-1];
	$c_data->{note} = "copied from content_data $c_data->{content_data_id}";
	$c_data->{content_id} = "_2";
	delete $c_data->{content_data_id};
	delete $c_data->{ctime};
	$db->{'form'}{'content_data'} = { '_3' => $c_data };
	push @{$db->{recordlist}},"_3:content_data";
    }

    $db->{action} = "insert";

    my $do_out = $db->do;

    if ($db->errorcheck > 0) {
	return ($do_out);
    }
    return ();
    
}

########################
# template page wizard
########################

# A template page is a page for a template website.
# It uses same screen 1 as a regular page, but has different 
# logic for selecting the body.

sub wiz_templatepage_page2 {
    my $this = shift;
    my $in = shift;
    my $out;
    my %page;
    my $pdata;
    my $id = $in->{page_id};
    my $db = $share{DB};

    # always choose a body

    # list all body templates available
    my (@liblist,$libs,$templates,$sql);
    $libs = $db->fetch_match('page',{type=>"library",section_id=>$cms->{section}});
    $out .= "<p>This page appears to have a conventional layout, with a \"body\" area for holding page-specific content.</p>\n";
    $out .= "<input type=hidden name=content#_2#name value=\"body\">
<input type=hidden name=content#_2#type value=\"editorial\">
<input type=hidden name=content#_2#page_id value=\"$id\">\n";
    $out .= "<p>You can optionally select a pre-defined body for this 
page, from the following list of available body templates.</p>\n<p>";
    foreach my $l (@$libs) {
	push(@liblist, $l->{page_id});
    }
    ######### FIXME: re-write this to avoid a custom query
    if (@$libs > 0) {
	$sql = "select * from content where name='body' and page_id in (".join(',',@liblist).")";
	$templates = $db->custom_query($sql);
    }
    else {
	$templates = [];
    }
    $out .= "<input type=\"radio\" name=\"body_id\" value=\"0\" checked> normal body, fully editable (will be blank to start with)<br>\n";
    foreach my $t (@$templates) {
	$out .= "<input type=\"radio\" name=\"body_id\" value=\"$t->{content_id}\">$t->{description}<br>\n";
    }

    $out .= "<br><br><br>\n";

    return $out;
}

sub do_wiz_templatepage_page2 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{page_id};
    my $db = $share{DB};

    my @err = &do_wiz_page_page1($this,$in);
    if (@err > 0) { return @err; }
    
    # now, perform all the necessary DB mods

    # copy selected page body

    if ($in->{body_id}) {
	# substitute the selected body into the form data
	my $c_data = $db->fetch_child("content_data","content",$in->{"body_id"},"content_data_id");
	$c_data = $c_data->[-1];
	$c_data->{note} = "copied from content_data $c_data->{content_data_id}";
	$c_data->{content_id} = "_2";
	delete $c_data->{content_data_id};
	delete $c_data->{ctime};
	$db->{'form'}{'content_data'} = { '_3' => $c_data };
	push @{$db->{recordlist}},"_3:content_data";
    }

    $db->{action} = "insert";

    my $do_out = $db->do;

    if ($db->errorcheck > 0) {
	return ($do_out);
    }
    return ();
    
}

########################
# template wizard
########################

sub wiz_template_page1 {
    my $this = shift;
    my $in = shift;
    my $out;
    my $id = "_1";
    my $type = "template";
    my $db = $share{DB};

    if (! exists $in->{reply}) {
	my $reply = $ENV{HTTP_REFERER} || "$config{server}{CGIpath}/CMS.cgi$ENV{PATH_INFO}";
	$out .= $db->input_html(type=>"hidden",name=>"reply",value=>$reply);
    }
    $out .= $db->input_html(name=>"page_id",type=>"hidden",value=>$id);
    $out .= $db->input_html(name=>"template#${id}#section_id",type=>"hidden",value=>$cms->{section});
    $out .= $db->input_html(name=>"template#${id}#type",type=>"hidden",value=>$type);
    $out .= $db->input_html(name=>"template#${id}#status",type=>"hidden",value=>"active");
    $out .= "<p>Name:<br>\n";
    $out .= $db->input_column(table=>"template",column=>"title",record=>$id);
    $out .= "<p>Description (optional):<br>\n";
    $out .= $db->input_column(table=>"template",column=>"description",record=>$id);
    $out .= "<p>Template subdirectory name (alphanumeric characters only; no spaces or periods):<br>\n";
    $out .= $db->input_column(table=>"template",column=>"filename",record=>$id);
    $out .= "</p>\n";
    $out .= "<p>Template publishing:<br>\n";
    $out .= $db->input_column(table=>"template",column=>"publish_method",record=>$id, value=>"static");
    $out .= "</p>\n";
    $out .= "<p>Template sharing:<br>
<select name=\"template#${id}#access\">\n";
    if ($db->authorize >= 3) {
	$out .= "<option value=\"public\">public</option>
<option value=\"members\">semi-private</option>
<option value=\"administrators\">private</option>\n";
    }
    else {
	# lvl 2 users cannot make public changes
	$out .= "<option value=\"members\">semi-private</option>
<option value=\"administrators\">private</option>\n";
    }
    $out .= "</select>
<a HREF=javascript:popup_link(\"$config{server}{CGIpath}/help.cgi/dbmap/help/template/access\")><img SRC=\"$config{server}{HTMLpath}/_ExSite/images/qmark.gif\" WIDTH=\"20\" HEIGHT=\"20\" BORDER=\"0\" TITLE=\"click for help\" ALIGN=\"middle\"></a></p>\n";
    return $out;
}

sub do_wiz_template_page1 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{page_id};
    my $db = $share{DB};
    my @err;
    if (! $in->{"template#${id}#type"}) {
	push @err, "No page type specified!";
    }
    if (! $in->{"template#${id}#title"}) {
	push @err, "Please specify a title.";
    }
    if (! $in->{"template#${id}#filename"}) {
	push @err, "Please specify a publication directory.";
    }
    if (! $in->{"template#${id}#status"}) {
	push @err, "No status specified!";
    }
    return @err;
}

sub wiz_template_page2 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my $out = "<p>A template can itself have a template.  All content in
the parent template will be inherited by the new template.  This allows you
to create numerous variations on a template, without re-importing all the
graphics and stylesheets.  Templates can inherit from other templates to
any depth.</p>\n";
    my %page;
    my $id = $in->{page_id};
    my $type = $in->{"template#${id}#type"};
    
    $out .= "<p>\n";
    # in this case, we want to include private templates to inherit from
    $out .= &list_template($this,"template#_1#template_id#key",undef,1);
    if ($out !~ /\.\.\.hmm,/) {
	# kludge: can't see the no-template error message, so continue
	# prompting for template input
	$out .= "</p>\n<p>If you copy the layout (the \"page\" object) from 
the parent template, you will be able to update/customize the layout for 
this template only.  However, updates to the layout of the parent
template will no longer be inherited by the new template.</p>
<blockquote><p>\n".
           $db->input_html(type=>"checkbox",name=>"copy_layout",selected=>1).
           " Copy Page Layout</p></blockquote>\n";
    }
    return $out;
}

sub do_wiz_template_page2 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{page_id};
    my $db = $share{DB};

    # no special validation for this step
    my @err = &do_wiz_template_page1($this,$in);
    if (@err > 0) { return @err; }

    if ($in->{copy_layout} && $in->{'template#_1#template_id#key'}) {
	&wiz_copy_page_layout($this,$in);
    }
    
    $db->{action} = "insert";

    my $do_out = $db->do;

    if ($db->errorcheck > 0) {
	return ($do_out);
    }
    return ();
    
}

# copy the page layout into the form data

sub wiz_copy_page_layout {
    my ($this,$in) = @_;
    my $db = $share{DB};
    my $key = exists $in->{'page#_1#template_id#key'} ?
	$in->{'page#_1#template_id#key'} : 
	$in->{'template#_1#template_id#key'};
    if ($key) {
	my $p = new ExSite::Page(id=>$key);
	my $pobj = $p->find("page");
	if ($pobj) {
	    # copy the content object
	    my $pobj_data = $pobj->get("content");
	    delete $pobj_data->{content_id}; # force an insert
	    $pobj_data->{page_id} = '_1';
		
	    # copy the content data
	    my $p_data = $pobj->get("content_data");
	    $p_data->{note} = "copied from content_data $p_data->{content_data_id}";
	    delete $p_data->{content_data_id}; # force an insert
	    delete $p_data->{ctime}; 
	    $p_data->{content_id} = '_2';

	    # hack the form data to retroactively insert page object info
	    $db->{form}->{content} = {'_2' => $pobj_data };
	    $db->{form}->{content_data} = {'_3' => $p_data };
	    push @{$db->{recordlist}},"_2:content","_3:content_data";
	}
	else {
	    $this->error("Failed to copy page layout: no page object found!");
	    # log it in the db log, too, since that one is checked more often
	    $db->error("Failed to copy page layout: no page object found!");
	}
    }
}

########################
# import template wizard
########################

sub wiz_import_template_page1 {
    my $this = shift;
    my $in = shift;
    my $out;
    my $id = "_1";
    my $type = "template";
    my $db = $share{DB};

    if (! exists $in->{reply}) {
	my $reply = $ENV{HTTP_REFERER} || "$config{server}{CGIpath}/CMS.cgi$ENV{PATH_INFO}";
	$out .= $db->input_html(type=>"hidden",name=>"reply",value=>$reply);
    }
    $out .= $db->input_html(name=>"page_id",type=>"hidden",value=>$id);
    $out .= $db->input_html(name=>"template#${id}#section_id",type=>"hidden",value=>$cms->{section});
    $out .= $db->input_html(name=>"template#${id}#type",type=>"hidden",value=>$type);
    $out .= $db->input_html(name=>"template#${id}#status",type=>"hidden",value=>"active");
    $out .= "<p>Name:<br>\n";
    $out .= $db->input_column(table=>"template",column=>"title",record=>$id);
    $out .= "<p>Description (optional):<br>\n";
    $out .= $db->input_column(table=>"template",column=>"description",record=>$id);
    $out .= "<p>Template subdirectory name (alphanumeric characters only; no spaces or periods):<br>\n";
    $out .= $db->input_column(table=>"template",column=>"filename",record=>$id);
    $out .= "</p>\n";
    $out .= "<p>Template publishing:<br>\n";
    $out .= $db->input_column(table=>"template",column=>"publish_method",record=>$id, value=>"static");
    $out .= "</p>\n";
    $out .= "<p>Upload Template HTML file:<br>\n";
    $out .= $db->input_exsite(datatype=>"file",name=>"template#${id}#file");
    $out .= "</p>\n";
    return $out;
}

sub do_wiz_import_template_page1 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{page_id};
    my $db = $share{DB};
    my @err;
    if (! $in->{"template#${id}#type"}) {
	push @err, "No page type specified!";
    }
    if (! $in->{"template#${id}#title"}) {
	push @err, "Please specify a title.";
    }
    if (! $in->{"template#${id}#filename"}) {
	push @err, "Please specify a publication directory.";
    }
    if (! $in->{"template#${id}#status"}) {
	push @err, "No status specified!";
    }
    if (! $in->{"template#${id}#file#filename"}) {
	push @err, "No template file uploaded!";
    }
    else {
	my ($fname,$fdata) = 
	  split /$config{form}{sepchar}/,$this->{db}{form}{template}{$id}{file};
	my $data = &decode_base64($fdata);
	$in->{"template#${id}#file"} = $data;
	delete $in->{"template##{id}#file#filename"};
    }
    return @err;
}

sub wiz_import_template_page2 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my $out = "<p>A scan of the uploaded template HTML file shows that it
requires the following external files:</p>\n";
    my $id = $in->{page_id};

    my $html = $in->{"template#${id}#file"};
    $html =~ s/&#34;/\"/g;
    $html =~ s/&#60;/\</g;
    $html =~ s/&#62;/\>/g;

    $out .= "<table><tr><th>File</th><th>Upload</th></tr>\n";
    while ($html =~ /src=([^>\s]+)/i) {
	my $file = $1;
	$file =~ s/[\"\']//g;
	my @file = split /\//,$file;
	$html =~ s/src=$1/src%3D"[[$1]]"/i;  # note extra space to foil RE
	$out .= "<tr><td>$file[-1]</td><td><input type=\"file\" name=\"template#${id}#$file[-1]#filename\"></td></tr>\n";
    }
    while ($html =~ /background=([^>\s]+)/i) {
	my $file = $1;
	$file =~ s/[\"\']//g;
	my @file = split /\//,$file;
	$html =~ s/background=$1/background%3D"[[$1]]"/i;  # note extra space to foil RE
	$out .= "<tr><td>$file[-1]</td><td><input type=\"file\" name=\"template#${id}#$file[-1]#filename\"></td></tr>\n";
    }
    while ($html =~ /<link [^>]*href=([^>\s]+)/i) {
	my $file = $1;
	$file =~ s/[\"\']//g;
	my @file = split /\//,$file;
	$html =~ s/href=$1/href%3D"[[$1]]"/i;  # note extra space to foil RE
	$out .= "<tr><td>$file[-1]</td><td><input type=\"file\" name=\"template#${id}#$file[-1]#filename\"></td></tr>\n";
    }
    $out .= "</table>\n<p>";

    $in->{"template#${id}#file"} = $html;

    return $out;
}

sub do_wiz_import_template_page2 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{page_id};
    my $db = $share{DB};
    $db->{enctype} = "multipart/form-data";
    $this->unescape_input($in);

    # no special validation for this step
    #my @err = &do_wiz_template_page1($this,$in);
    #if (@err > 0) { return @err; }
    my @err;

    $db->{action} = "insert";

    my $template = {
		    section_id => $in->{"template#${id}#section_id"},
		    type => $in->{"template#${id}#type"},
		    status => $in->{"template#${id}#status"},
		    title => $in->{"template#${id}#title"},
		    description => $in->{"template#${id}#template"},
		    filename => $in->{"template#${id}#filename"},
		    publish_method => $in->{"template#${id}#publish_method"},
		    };

    my $tid;
    #my $tid = $db->insert("template",$template);
    push @err, "insert template ".&ShowHash($template);

    #if ($tid) {
	# insert page element
	my $content = {
	    name => "page",
	    type => "design",
	    page_id => $tid,
	};
	my $cid;
	#my $cid = $db->insert("content",$content);
	push @err, "insert content ".&ShowHash($content);

	#if ($cid) {
	    #my $cdid = $db->insert("content_data",{
	push @err, "insert content_data ".&ShowHash({
		content_id => $cid,
		mime_type => "text/html",
		data => $in->{"template#${id}#file"},
	    });
	    ### check result?
	#}
    delete $in->{"template#${id}#file"};

	foreach my $key (keys %$in) {
	    if ($key =~ /^template\#${id}\#(.+?)\#filename$/) {
		next if ($1 eq "file");
		# insert content object
		my $content = {
		    name => $1,
		    type => "design",
		    page_id => $tid,
		};
		#my $cid = $db->insert("content",$content);
		push @err, "insert content ".&ShowHash($content);
		#if ($cid) {
		    #my $cdid = $db->insert("content_data",{
		push @err, "insert content_data $1 ".&ShowHash({
			content_id => $cid,
			mime_type => "text/html",
			fdata => $this->{db}{form}{template}{$id}{$1},
		    });
		#}
	    }
	}
    #}
    push @err, "FINAL: ".&ShowHash($this->{db}{form}{template});
    return @err;

#    if ($db->errorcheck > 0) {
#	return ($do_out);
#    }
#    return ();
    
}

########################
# library wizard
########################

sub wiz_library_page1 {
    my $this = shift;
    my $in = shift;
    my $out;
    my %page;
    my $id = "_1";
    my $type = "library";
    my $db = $share{DB};

    $out .= "<p>A library is a just a container to hold shareable web content.  
Any page on the site may reference a content item placed in any of that
site's libraries.</p>\n";

    if (! exists $in->{reply}) {
	my $reply = $ENV{HTTP_REFERER} || &relink(cmscmd=>undef);
	$out .= $db->input_html(type=>"hidden",name=>"reply",value=>$reply);
    }
    $out .= $db->input_html(name=>"page_id",type=>"hidden",value=>$id);
    $out .= $db->input_html(name=>"page#${id}#section_id",type=>"hidden",value=>$cms->{section});
    $out .= $db->input_html(name=>"page#${id}#type",type=>"hidden",value=>$type);
    $out .= $db->input_html(name=>"page#${id}#status",type=>"hidden",value=>"active");
    $out .= "<p>Library Name:<br>\n";
    $out .= $db->input_column(table=>"page",column=>"title",record=>$id, 
			      data=>\%page);
    $out .= "<p>Description (optional):<br>\n";
    $out .= $db->input_column(table=>"page",column=>"description",record=>$id, 
			      data=>\%page);
    $out .= "<p>Library subdirectory name (alphanumeric characters only; no spaces or periods):<br>\n";
    $out .= $db->input_column(table=>"page",column=>"filename",record=>$id, data=>\%page);
    $out .= "</p>\n";
    return $out;
}

sub do_wiz_library_page1 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{page_id};
    my $db = $share{DB};
    my @err;
    if (! $in->{"page#${id}#type"}) {
	push @err, "No page type specified!";
    }
    if (! $in->{"page#${id}#title"}) {
	push @err, "Please specify a title.";
    }
    if (! $in->{"page#${id}#filename"}) {
	push @err, "Please specify a publication directory.";
    }
    if (! $in->{"page#${id}#status"}) {
	push @err, "No status specified!";
    }
    return @err if (@err > 0);

    my $do_out = $db->do;

    if ($db->errorcheck > 0) {
	return ($do_out);
    }
    return ();
}

################################
# website wizard - insert only!
################################

sub wiz_section_page1 {
    my $this = shift;
    my $in = shift;
    my $id = "_1";  # virtual record id
    my $db = $share{DB};
    my %sec;
    my $out = "<p>This wizard can be used to configure entirely new websites, 
or add sub-sections to existing websites.</p>\n";

    # set reply page on first screen
    if (! exists $in->{reply}) {
	my $reply = $ENV{HTTP_REFERER} || &relink(cmscmd=>undef);
	$out .= $db->input_html(type=>"hidden",name=>"reply",value=>$reply);
    }

    $out .= "<p>Select the website type.<br>\n";

    $out .= $db->input_column(table=>"section",column=>"type",record=>$id,data=>\%sec);
    $out .= "</p>\n<p>Title:<br>\n";
    $out .= $db->input_column(table=>"section",column=>"title",record=>$id,data=>\%sec);
    $out .= "</p>\n<p>Description:<br>\n";
    $out .= $db->input_column(table=>"section",column=>"description",record=>$id,data=>\%sec);
    $out .= $db->input_html(type=>"hidden", name=>"section_id", value=>$id);
    $out .= "</p>\n";
    return $out;
}

sub do_wiz_section_page1 {
    my $this = shift;
    my $in = shift;
    my $id = "_1";  # virtual record id
    my $db = $share{DB};
    my @err;
    if (! ($in->{"section#${id}#type"} && $in->{"section#${id}#title"} && $in->{"section#${id}#description"})) {
	push @err, "Please provide values for all fields.";
    }
    if ($in->{"section#${id}#type"} =~ /template/) {
	# templates have no design, publication setup, or default pages
	$this->{current} += 3 ;
	return $db->do;  # submit form now
    }
    return @err;
}

sub wiz_section_page2 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my $out;
    my %sec;
    my $id = "_1";  # virtual record id
#    my $id = $in->{page_id};
    my $type = $in->{"section#${id}#type"};
#    if ($id) {
#	%sec = $db->fetch("section",$in->{page_id});
#    }

    $out .= "<p>If this website or sub-site is related to an existing site, specify the parent site:<br>\n";
    $out .= $db->input_column(table=>"section",column=>"parent_id",record=>$id, data=>\%sec);

    $out .= "</p>\n<p>Menu Label to use if this site is linked to from other sites:<br>";
    $out .= $db->input_column(table=>"section",column=>"label",record=>$id,data=>\%sec);

    $out .= "<p>Specify the publication location ";
    if ($type =~ /remote/) {
	$out .= "as a URL of the form \"ftp://hostname/path/...\":";
    }
    elsif ($type =~ /stand-alone/) {
	$out .= "as a complete disk path to the publication directory:";
    }
    else {
	$out .= "as a subdirectory name (alphanumeric characters and underscores, no spaces):";
    }
    $out .= "<br>\n";
    $out .= $db->input_column(table=>"section",column=>"publish_to",record=>$id, data=>\%sec);
    $out .= "</p>\n";
    if ($type =~ /stand-alone|remote/) {
	$out .= "<p>This site will be accessed publicly from the following domain/URL:<br>\n";
	$out .= $db->input_column(table=>"section",column=>"url",record=>$id, data=>\%sec);
	$out .= "</p>\n";
    }

    return $out;
}

sub do_wiz_section_page2 {
    my $this = shift;
    my $in = shift;
    my $id = "_1";  # virtual record id
#    my $id = $in->{section_id};
    my $db = $share{DB};
    my @err = &do_wiz_section_page1($this,$in);
    my $type = $in->{"section#${id}#type"};
    
    if ($type =~ /remote/ && 
	$in->{"section#${id}#publish_to"} !~ /^ftp:\/\//) {
	push @err, "Please specify a URL to an FTP location.";
    }
    elsif ($type =~ /stand-alone/ && 
	$in->{"section#${id}#publish_to"} !~ /^\/.+/) {
	push @err, "Please specify a complete file path to publish to.";
    }
    else {
	if (! $in->{"section#${id}#publish_to"}) {
	    # only one site may publish to the root directory
	    my @rootsites = $db->fetch_match("section",{publish_to=>""});
	    if (@rootsites > 0) {
		push @err, "Another website is already publishing to the root directory;  please specify a publication subdirectory using alphanumeric characters and underscores only.";
	    }
	}
	elsif ($in->{"section#${id}#publish_to"} !~ /\w+/) {
	    push @err, "Please specify a publication subdirectory using alphanumeric characters and underscores only.";
	}
    }

    return @err;
}

sub wiz_section_page3{
    my $this = shift;
    my $in = shift;
    my $out;
    my %sec;
    my $pdata;
    my $id = "_1";  # virtual record id
#    my $id = $in->{section_id};
    my $db = $share{DB};

    %sec = $db->fetch("section",$id);

    $out .= "<p>You can select a pre-configured set of pages to populate
this website with.  If you do not select a set of pages, the website
will contain no pages to begin with, and you will have to create each
page manually.</p>\n<blockquote>\n";

    my @wtempl = $db->fetch_match("section","type=\"website template\" and (parent_id=0 or parent_id is NULL)");

    if (@wtempl > 0) {
	$out .= "<p><dl>";
	foreach my $wt (@wtempl) {
	    $out .= "<dt>".
		$db->input_html(type=>"radio",name=>"website_template",
				value=>$wt->{section_id},
				label=>"<b>$wt->{title}</b></dt>\n<dd>$wt->{description}</dd>\n");
	}
	$out .= "</dl></p>\n";
    }
    $out .= "</blockquote>\n";
    $out .= &list_template($this,"page_template",undef,2);
    return $out;
}

sub do_wiz_section_page3 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    # no special validation for this step
    return &do_wiz_section_page2($this,$in);
}

sub wiz_section_page4 {
    my $this = shift;
    my $in = shift;
    my $out;
    my %sec;
    my $pdata;
    my $id = $in->{section_id};
    my $db = $share{DB};

    %sec = $db->fetch("section",$id);

    $out .= "<p>You may optionally define a webmaster for this site.</p> 
<p>If you leave the webmaster undefined, this site will use the
webmaster of its parent site.  If you define a new webmaster, this site 
will function semi-independently of the parent site.</p>\n";

    $out .= "<table cellspacing=20><tr><td width=50% valign=top>
<H4>Method 1: Define a New Webmaster</h4>\n";

    $out .= "<p>\n";
    $db->{action} = "insert";  # dummy, to get input_record to work
    $out .= $db->input_record({table=>"member",record=>"_2",data=>{access=>3},hide=>["access"]});
    #$out .= "<p>Priveleges<br>";
    $out .= $db->input_record({table=>"member_site_link",record=>"_3",data=>{section_id=>"_1",member_id=>"_2",access=>3},hide=>["section_id","member_id"]});
#    $out .= $db->input_exsite(name=>"member_site_link#_1#priveleges",datatype=>"set:member_priv");
    $out .= "</p>\n";

    $out .= "</td><td width=50% valign=top>
<H4>Method 2: Use an Existing Webmaster</h4>\n";

    $out .= $db->input_exsite(name=>"webmaster",datatype=>"key:member",fmatch=>{access=>3});

    #$out .= "<p>Priveleges<br>";
    $out .= $db->input_record({table=>"member_site_link",record=>"_4",data=>{section_id=>"_1",member_id=>"_2",access=>3},hide=>["section_id","member_id"]});
    #$out .= $db->input_exsite(name=>"member_site_link#_2#priveleges",datatype=>"set:member_priv");

    $out .= "<p>This will not affect this webmaster's responsibilities for other web sites.</p>\n";

    $out .= "\n</td></tr></table>\n";

    $out .= "<p>NOTE: if you use both of the above methods, both webmasters will be created.</p>\n";

    return $out;
}

sub do_wiz_section_page4 {
    my $this = shift;
    my $in = shift;
    my $id = $in->{section_id};
    my @err;
    my $new_webmaster;
    my $db = $share{DB};

    # validate all previous inputs

    my @err = &do_wiz_section_page3($this,$in);

    # check new webmaster fields

    if ($in->{"member#_2#login"} || $in->{"member#_2#password"} || $in->{"member#_2#name"} || $in->{"member#_2#email"}) {
	if (! ($in->{"member#_2#login"} && $in->{"member#_2#password"} && $in->{"member#_2#name"} && $in->{"member#_2#email"})) {
	    push @err, "Missing fields for new webmaster.  If not creating a new webmaster, then leave all fields blank.";
	}
	elsif ($in->{"member#_2#email"} !~ /^[\w](\.?[\w\-]+)*@([\w\-]+\.)+[A-Za-z]{2,4}$/) {
	    push @err, "Invalid email address format for new webmaster.";
	}
	else {
	    $new_webmaster = 1;  # the new webmaster form has been filled out
	    my @wmdata = $db->fetch_match("member",{login=>$in->{"member#_2#login"}});
	    if (@wmdata > 0) {
		push @err,"Sorry, a user with that login already exists;  please choose another.";
	    }
	}
    }

    if (@err > 0) { return @err; }
    
    # Looks like we're good to go...
    
    # 1) load up the form data
    $db->get_data;

    # 2) delete any webmaster forms that aren't used
    #print "Content-type: text/plain\n\n";
    if (! $new_webmaster) {
	$db->delete_record("member","_2") or push @err, "failed to ignore new webmaster form";
	$db->delete_record("member_site_link","_3") or push @err, "failed to ignore new webmaster priveleges form";
    }
    if (! $in->{'webmaster#key'}) {
	$db->delete_record("member_site_link","_4") or push @err, "failed to ignore old webmaster priveleges form";
    }
    else {
	# add the existing webmaster to the member_site_link
	$db->{form}{member_site_link}{_4}{member_id} = $in->{'webmaster#key'};
    }

    # delete the webmaster#key form input
    # KLUDGE - use delete_record

    $db->delete_record("webmaster","key") or push @err, "failed to remove old webmaster from form";

    if (@err > 0) { return @err; }

    # insert the basic website records
    my $stat = $db->do;
    if ($stat && $stat !~ /Done!/) {  
	# yuck!  need better error management for doform
	push @err, $stat;
	return @err;
    }

    # now create the default pages
    if ($in->{website_template}) {
	push @err, "using website $in->{website_template}";
	my @pages = $db->fetch_match("page",{section_id=>$in->{website_template}});
	if (@pages > 0) {
	    foreach my $page (@pages) {
		push @err, "making page $page->{page_id}";
		my %newpage = %$page;
		delete $newpage{page_id};  # force an insert
		$newpage{template_id} = $in->{page_template};
		$newpage{section_id} = $db->{form}{section}{_1}{section_id};
		my $id = $db->insert("page",\%newpage);
		if (! $id) {
		    push @err, "failed to add page \"$page->{title}\"";
		}
		else {
		    # make copies of all editorial content objects
		    my @content = $db->fetch_match("content",{page_id=>$page->{page_id},type=>"editorial"});
		    foreach my $cont (@content) {
			delete $cont->{content_id};  # force an insert
			$cont->{page_id} = $id;
			my $cid = $db->insert("content",$cont);
			if (! $cid) {
			    push @err, "failed to add \"$cont->{name}\" content to page \"$page->{title}\"";
			}
		    }
		}
	    }
	}
    }
    return @err;
}

# -------- easy site creation wizard
#
# assumes: native site, new webmaster

sub wiz_easy_section_page1 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my $out;

    # set reply page on first screen
    if (! exists $in->{reply}) {
	my $reply = $ENV{HTTP_REFERER} || &relink(cmscmd=>undef);
	$out .= $db->input_html(type=>"hidden",name=>"reply",value=>$reply);
    }

    $out = "<p>First, provide us with a valid e-mail address.  This is needed
to communicate with you on the status of your website, send password and login
instructions, and so on.  It is kept strictly confidential.</p>
<p>Email address:<br>\n";
    $out .= $db->input_html(type=>"text", name=>"email");

    $out .= "<p>Next, give us the full name or title of your site.  For
example: <b>Mainland Trucking Company</b>
<p>Full Website Name:<br>\n";
    $out .= $db->input_html(type=>"text", name=>"title", size=>60);

    $out .= "<p>You can optionally provide a longer website description.
Some graphic designs may display this information as a subtitle, while 
others will simply use it as a hidden description for search engines. 
For example: <b>Serving your transportation needs since 1975</b>
<p>Description/subtitle (optional):<br>\n";
    $out .= $db->input_html(type=>"text", name=>"description", size=>60);

    $out .= "<p>You also need to define a short version of the website
name.  This must be a short text string with no spaces or punctuation.
It will be used in your
website URL, and also as a login ID for you to access your site and
edit it.  You may use up to 32 letters, numerals, underscores, and
hyphens.  For example: <b>MainlandTrucking</b></p>
<p>Short Website ID:<br>\n";
    $out .= $db->input_html(type=>"text", name=>"sitename");
    $out .= "</p>\n";
    return $out;
}

sub do_wiz_easy_section_page1 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my @err;
    if ($in->{"email"} !~ /^[\w](\.?[\w\-]+)*@([\w\-]+\.)+[A-Za-z]{2,4}$/) {
	push @err, "Invalid e-mail address.";
    }
    if ($in->{"sitename"} !~ /^[A-Za-z0-9_\-]{3,32}$/) {
	push @err, "Invalid website ID.  The site ID must be between
3 and 32 characters long, containing only letters, numerals, underscores, 
and hyphens.";
    }
    if (! $in->{"title"}) {
	# not an error
	$in->{title} = $in->{sitename};
    }

    if ($in->{"sitename"}) {
	my @match = $db->fetch_match("section",{publish_to=>$in->{sitename}});
	if (@match > 0) {
	    push @err, "Sorry, a website with that ID already exists in the system.  Please try another.";
	}
	else {
	    @match = $db->fetch_match("member",{login=>$in->{sitename}});
	    if (@match > 0) {
		push @err, "Sorry, a website with that ID already exists in the system.  Please try another.";
	    }
	}
    }

    return @err;
}

sub wiz_easy_section_page2 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my $out;

    $out .= "<p>Choose your website type.  This determines the pages that
will be created for you automatically.</p>\n";

    my @wtempl = $db->fetch_match("section","type=\"website template\" and (parent_id=0 or parent_id is NULL)");

    if (@wtempl > 0) {
	$out .= "<p><dl>";
	foreach my $wt (@wtempl) {
	    $out .= "<dt>".
		$db->input_html(type=>"radio",name=>"website_template",
				value=>$wt->{section_id},
				label=>"<b>$wt->{title}</b></dt>\n<dd>$wt->{description}</dd>\n");
	}
	$out .= "</dl></p>\n";
    }
    
    return $out;
}

sub do_wiz_easy_section_page2 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my @err = &do_wiz_easy_section_page1($this,$in);
    if (! $in->{website_template}) {
	push @err, "Please select a website type!";
    }
    return @err;
}

sub wiz_easy_section_page3 {
    my $this = shift;
    my $in = shift;
    my $out;
    my $db = $share{DB};
    $out .= "<p>Now you can select the graphic design of your site, from the
options provided below.  Use the <b>preview</b> links to view samples of
each design.</p>\n";
    $out .= &list_template($this,"page_template",undef,2);
    return $out;
}

sub do_wiz_easy_section_page3 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my @err = &do_wiz_easy_section_page2($this,$in);
    if (! $in->{page_template}) {
	push @err, "Please select a graphic design template!";
    }
    return @err;
}

sub wiz_easy_section_page4 {
    my $this = shift;
    my $in = shift;
    return "<p>Thank You! We have collected all the information we need to create
your website.  When you click on \"FINISH\", below, your website will be
configured for you.  Instructions for how to log in to manage your 
website will be emailed to <font color=#0000FF><tt>$in->{email}</tt></font>.
</p><p>
After you receive this email, click on the \"My Site\" tab at the top.
You can enter your login information there to access your website
administration tools.
</p>\n";
}

sub do_wiz_easy_section_page4 {
    my $this = shift;
    my $in = shift;
    my $db = $share{DB};
    my @err = &do_wiz_easy_section_page3($this,$in);
    return @err if (@err > 0);

    # if we get here, we're ready to go

    # we need to create:
    # 1) a user
    # 2) a site (section)
    # 3) pages and content objects
    # 4) then mail the login info to the user

    my $pwd = &randtext();
    my %user = (login=>$in->{sitename},
		password=>$pwd,
		name=>$in->{sitename},
		email=>$in->{email},
		access=>2);

    my %section = (title=>$in->{title},
		   type=>"native website",
		   description=>$in->{description},
		   publish_to=>$in->{sitename},
		   template_id=>$in->{page_template});

    # add user
    my $uid = $db->insert("member",\%user);
    if (! $uid) {
	return ("failed to create new user account!");
    }

    # add website
    my $sid = $db->insert("section",\%section);
    if (! $sid) {
	# undo member insert
	$db->delete("member",$uid);
	return ("failed to create new website!");
    }

    # add administrator ticket
    my $tid = $db->insert("member_site_link",{section_id=>$sid,
					      member_id=>$uid,
					      priveleges=>"editor; administrator"});
    if (! $tid) {
	return ("failed to create new administrator ticket!");
    }

    # now create the default pages
    if ($in->{website_template}) {
	my @pages = $db->fetch_match("page",{section_id=>$in->{website_template}});
	if (@pages > 0) {
	    foreach my $page (@pages) {
		my %newpage = %$page;
		delete $newpage{page_id};  # force an insert
		$newpage{template_id} = $in->{page_template};
		$newpage{section_id} = $sid;
		my $pid = $db->insert("page",\%newpage);
		if (! $pid) {
		    push @err, "failed to add page \"$page->{title}\"";
		}
		else {
		    # make copies of all editorial content objects
		    my @content = $db->fetch_match("content",{page_id=>$page->{page_id},type=>"editorial"});
		    foreach my $cont (@content) {
			delete $cont->{content_id};  # force an insert
			$cont->{page_id} = $pid;
			my $cid = $db->insert("content",$cont);
			if (! $cid) {
			    push @err, "failed to add \"$cont->{name}\" content to page \"$page->{title}\"";
			}
			else {
			    # copy current content version

			    my @cdata = $db->fetch_match("content_data",{content_id=>$cid},"content_data_id");
			    if (@cdata > 0) {
				my $cdata = $cdata[-1];
				delete $cdata->{content_data_id};  # force an insert
				$cdata->{content_id} = $cid;
				my $cdid = $db->insert("content_data",$cdata);
				if (! $cdid) {
				    push @err, "failed to add \"$cont->{name}\" content data";
				}
			    }
			}
		    }
		}
	    }
	}
    }
	
    if (! @err) {
	# we made it!
	# send the user their login info

	require ExSite::Mail;

	my $msg = "This message has been sent to you because someone created
a new website at $config{server}{server}$config{server}{HTMLpath} and provided 
this email address for contact information.

To access the administrator functions for this website, direct your
browser to $config{server}{server}$config{server}{HTMLpath} and click on the 
My Site tab.

Your login information:

Login ID: $in->{sitename}
Password: $pwd

Enjoy your new website!
";

	&ExSite::Mail::SendMail($in->{email},$config{site}{administrator},
		  "Your Exsite login information",
		  $msg);
    }
			  
    return @err;
}


1;
