package Modules::SimpleMenu;

#----------------------------------------------------------------------------
#
#   Copyright 2001-2004 Exware Solutions, Inc.  http://www.exware.com
#
#   This file is part of ExSite WebWare (ExSite, for short).
#
#   ExSite is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   ExSite is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with ExSite; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Users requiring warranty coverage and/or support may arrange alternate
#   commercial licensing for ExSite, by contacting Exware Solutions 
#   via the website noted above.
#
#-----------------------------------------------------------------------
# Modules::SimpleMenu
# Author: Morgan Burke
# (c) 2003 Exware Solutions
#-----------------------------------------------------------------------
# generic menu DCM
# possible extensions to SimpleMenu
# - graphical menu buttons
# - spacer HTML/graphics inserted between menu items
# - publish menu snippets, for faster lookup on dynamic pages
# - control panel for easy menu layout/configuration

=pod

=head1 Modules::SimpleMenu - generic menu-builder for ExSite

SimpleMenu is a multi-purpose menu-builder for ExSite websites.  It
will create context-specific menus on any web page in a site.
Context-specific means that the site navigation links will reflect the
current location in the site.

SimpleMenu can also serve as a code template for a customized menu
builder, since it contains a general purpose menu-construnction
algorithm based on the ExSite page placement parameters.

SimpleMenu can build four types of menus:

=over 4

=item * B<vertical> (default) - navigation links are stacked in a
vertical list, with submenus exploded within this list.

=item * B<horizontal> - navigation links are listed left-to-right,
with each link separated by a "|" (bar) character.  Only the top-level
links in the site are shown, and no submenus are displayed.

=item * B<path> - the path from the site home page to the current page
is displayed in the "cookie crumb" format, ie. page1 > page2 > page3

=item * B<sitemap> - the full site structure is shown, with child
pages indented beneath their parents.

=back

When inserting a web application tag into a block of HTML, any of the
above types may be provided as a parameter.

When manually coding a SimpleMenu dynamic content tag, the format is:

C<E<lt>!--&SimpleMenu(format)--E<gt>>

where "format" is one of the above menu types.

=cut

use strict;
use ExSite::Config;
use ExSite::Misc;
use ExSite::Util;
use ExSite::Page;
use ExSite::Input;
# inherit from BaseDCD class
use Modules::BaseDCD;
use vars qw(@ISA);
@ISA = qw(Modules::BaseDCD);

sub write {
    my ($this, $opt) = @_;
    if ($opt =~ /horizontal/) { return $this->hmenu($opt); }
    if ($opt =~ /path/) { return $this->pmenu($opt); }
    if ($opt =~ /sitemap/) { return $this->sitemap($opt); }
    else { return $this->vmenu($opt); }
}

sub ioctl {
    my $this = shift;
    $_ = shift;
    if (/ModuleInfo/) {
	return "SimpleMenu is a multi-purpose menu-building system.
It automatically assembles site menus and site maps by inspecting the
relationships between your webpages.  It can build vertically-formatted
menus with exploded sub-menus, horizontally-formatted menus, and 
path menus (also known as \"cookie-crumb\" menus) that show you where
you are in the site.  It has options for customizing the formatting and
layout of menus so that they can be incorporated into most graphic designs.
SimpleMenu does not do dynamic pop-up menus, or menus with graphical 
buttons.  These options will require specialized menu plug-ins.";
    }
    return undef;
}

# hmenu - return a horizontal menu (global or child pages)

sub hmenu {
    my $this = shift;
    my $opt = shift;

    my $CGIpath = $config{server}->{CGIpath};
    my $HTMLpath = $config{server}->{HTMLpath};
    my $IMG = "$config{server}->{HTMLpath}/_ExSite/images";
    my $db = $share{DB};
    my $Page = $share{Page};
    my $this_page = $Page->id;
    my $menu_mode = $this->menu_mode;
    my $out = "<!-- begin menu -->\n<SPAN CLASS=menu>\n";

    # fetch all menu-able pages from this section
    my @pages;
    if ($opt =~ /child/) {
	@pages = $this->get_pages({
	    type=>"page",
	    visibility=>"visible",
	    status=>"active",
	    parent_id=>$Page->id(),
	    section_id=>$Page->id("section"),
	},
				  "rank",
				  );
    }
    else {
	@pages = $this->get_pages({
	    type=>"page",
	    visibility=>"visible",
	    status=>"active",
	    section_id=>$Page->id("section"),
	},
				  "rank",
				  );
    }
    my $pages_by_id = &keywise("page_id",\@pages);
    my $access = $db->authorize;
    my @menu;
    foreach my $page (@pages) {
	# global pages only
	next if ($page->{parent_id} && $opt !~ /child/);
	# skip access-controlled pages, if necessary
	next if (($page->{access} eq "members" && ! $access) ||
		 ($page->{access} eq "administrator" && ($access >= 3))
		 );
	my $url = $page->{url};
	if (!$url) {
	    my $p = new ExSite::Page(page=>$page);
	    $url = $p->get_url($menu_mode);
	}
	push @menu, "<a href=\"$url\" class=\"menu\">$page->{label}</a>";
    }

    # optional links for members only
    if ($access) {
	if (! $Page->{publish} && ! $Page->is_template) {
	    push @menu, "<a href=\"$CGIpath/logout.cgi\" class=\"menu\">Logout</a>";
	}
    }

    $out .= join("&nbsp;|&nbsp;", @menu);

    $out .= "</SPAN>\n<!-- end menu -->\n";
}

# pmenu - return a path menu
#
# Note that for pmenu we do not put the content into the menu class,
# because most commonly a pmenu is placed into the page body, not the
# menu bar.

sub pmenu {
    my $this = shift;
    my $opt = shift;

    my $CGIpath = $config{server}->{CGIpath};
    my $HTMLpath = $config{server}->{HTMLpath};
    my $IMG = "$config{server}->{HTMLpath}/_ExSite/images";
    my $db = $share{DB};
    my $Page = $share{Page};
    my $this_page = $Page->id;
    my $pdata = $Page->get;
    my $menu_mode = $this->menu_mode;
    my $purl = $Page->get_url($menu_mode);
    my @path = "<a href=\"$purl\">$pdata->{label}</a>";
    my $out = "<!-- begin menu -->\n";

    while ($pdata->{parent_id}) {
	$Page = new ExSite::Page(id=>$pdata->{parent_id});
	$purl = $Page->get_url($menu_mode);
	$pdata = $Page->get;
	unshift @path, "<a href=\"$purl\">$pdata->{label}</a>";
    }
    if ($pdata->{filename} !~ /^index\./) {
	# also include the index page in the path
	my @p = $this->get_pages({section_id=>$pdata->{section_id},
				  filename=>"index.%",
				  status=>"active",
				  visibility=>"visible"});
	if (@p > 0) {
	    $Page = new ExSite::Page(page=>$p[-1]);
	    $purl = $Page->get_url($menu_mode);
	    $pdata = $Page->get;
	    unshift @path, "<a href=\"$purl\">$pdata->{label}</a>";
	}
    }

    $out .= join "&nbsp;&gt;&nbsp;",@path;
    $out .= "<!-- end menu -->\n";

    return $out;
}

sub vmenu {
    my $this = shift;
    my $opt = shift;

    my $CGIpath = $config{server}->{CGIpath};
    my $HTMLpath = $config{server}->{HTMLpath};
    my $IMG = "$config{server}{server}$config{server}{HTMLpath}/_ExSite/images";
    my $db = $share{DB};
    my $Page = $share{Page};
    my $this_page = $Page->id;
    my $pdata = $Page->get;
    my $menu_mode = $this->menu_mode;

    my $out;
    my ($menutop, $menumid, $menubot);
    $menutop = $Page->find("SimpleMenuTop")->get_html();
    if ($menutop) {
	$menumid = $Page->find("SimpleMenuMid")->get_html();
	$menubot = $Page->find("SimpleMenuBot")->get_html();
    }
    else {
	$menutop = "<!-- begin menu -->\n<span class=\"menu\">\n";
	$menumid = "<br>";
	$menubot = "</span>\n<!-- end menu -->\n";
    }

    # fetch all pages
    my @pages = $this->get_pages({
				     type=>"page",
				     visibility=>"visible",
				     status=>"active",
				     section_id=>$Page->id("section"),
				 },
				 "rank",
				 );
    my $pages_by_id = &keywise("page_id",\@pages);
    if (! exists $pages_by_id->{$this_page}) {
	# we're on a hidden page
	$pages_by_id->{$this_page} = $db->fetch("page",$this_page);
    }
    my $access = $db->authorize;

    my @items;
    
    foreach my $page (@pages) {
	# do not display child pages in top-level of menu
	next if ($page->{parent_id});
	# skip pages that should not be menu'd
	next if ($page->{type} ne "page" ||
#--- show links to member-only pages, just don't show the pages---
#		 ($page->{access} eq "members" && ! $access) ||
		 ($page->{access} eq "administrator" && ($access >= 3))
		 );
	# skip pages not in this section
	next if ($page->{section_id} != $pages_by_id->{$this_page}->{section_id});
	my $url = $page->{url};
	if (!$url) {
	    my $p = new ExSite::Page(page=>$page);
	    $url = $p->get_url($menu_mode);
	}
	# child menu
	if ($Page && $pdata->{page_id}) {
	    # we seem to be on a real page - make a submenu
	    my @submenu = ();
	    foreach my $child (@pages) {
		next if (! $child->{parent_id});
		next if ($page->{visibility} eq "hidden");
		# if this is a child of the current page
		if (($page->{page_id} == $this_page && $child->{parent_id} == $this_page) ||
		    (($child->{parent_id} == $pages_by_id->{$this_page}->{parent_id}) &&
		    ($page->{page_id} == $pages_by_id->{$this_page}->{parent_id}))) {
		    my $url = $child->{url};
		    if (!$url) {
			my $cp = new ExSite::Page(page=>$child);
			$url = $cp->get_url($menu_mode);
		    }
		    push @submenu, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href=\"$url\" class=\"submenu\">$child->{label}</a>";
		}
	    }
	    if (@submenu > 0) {
		push @items, "<a href=\"$url\" class=\"menu\">$page->{label}</a><br>\n<span class=\"submenu\">\n".join("<br>\n",@submenu)."\n</span>\n";
	    }
	    else {
		push @items, "<a href=\"$url\" class=\"menu\">$page->{label}</a>";
	    }
	}
	#$out .= "<IMG SRC=$IMG/menuborder.gif HEIGHT=10 WIDTH=150><br>\n";
	#$out .= "<!--&content(menu-spacer)--><br>\n";
    }

    # optional links for members only
    if ($db) {
	if ($Page) {
	    if (! $Page->{publish}) {
		my $level = $db->authorize;
		if ($level) {
		    push @items, "<a href=\"$CGIpath/logout.cgi\" class=\"menu\">Logout</A>";
		}
	    }
	}
    }

    return "$menutop\n".join("$menumid\n",@items)."\n$menubot";
}

sub menu_mode {
    my $this = shift;
    # always do static menus when publishing
    return 1 if ($share{Page}->{publish});
    # otherwise, use static menus only when not in CMS mode
    return &get_cookie("ExSiteCMS") ? 0 : 1;
}

sub sitemap {
    my $this = shift;
    my $out;
    my $db = $share{DB};
    my $page = $share{Page}->get;
    my @pages = $db->fetch_match("page",{section_id=>$page->{section_id},
					 type=>"page",
					 parent_id=>0,
					 visibility=>"visible"},"rank");
    $out .= "<P><UL CLASS=\"sitemap\">\n";
    foreach my $page (@pages) {
	$out .= $this->sitemap_r($page);
    }
    $out .= "</UL></P>\n";
    return $out;
}

sub sitemap_r {
    # recursive mapping routine
    my ($this,$page) = @_;
    my $db = $share{DB};
    my $pid = $page->{page_id};
    my $p = new ExSite::Page(id=>$pid);
    my $url = $p->get_url;
    my $out = "<LI><A HREF=\"$url\">$page->{label}</A>";
    if ($page->{access} ne "public") { $out .= " (restricted access)"; }
    #elsif ($page->{access} eq "public") { $out .= " ($page->{filename})"; }
    $out .= "</LI>\n";
    if ($page->{access} ne "public") { return $out; }
    my $subpage = $db->fetch_match("page",{parent_id=>$pid,
					   type=>"page",
					   visibility=>"visible"},"rank");
    if (@$subpage > 0) {
	$out .= "<UL CLASS=\"sitemap\">\n";
	foreach my $page (@$subpage) { $out .= $this->sitemap_r($page); }
	$out .= "</UL>\n";
    }
    return $out;
}

sub get_pages {
    my ($this,$match,$rank) = @_;
    if ($share{Page}) {
	if ($share{Page}->is_template) {
	    my $site = $share{Page}->id("section");
	    return ( 
		     { page_id=>1, label=>"Home", type=>"page", access=>0, url=>"#", section_id=>$site },
		     { page_id=>1, label=>"About", type=>"page", access=>0, url=>"#", section_id=>$site },
		     { page_id=>1, label=>"Contact", type=>"page", access=>0, url=>"#", section_id=>$site },
		     { page_id=>1, label=>"Links", type=>"page", access=>0, url=>"#", section_id=>$site },
		     );
	}
    }
    return $share{DB}->fetch_match("page",$match,$rank);
}

1;


